/*
 * Copyright 1998-2008 VIA Technologies, Inc. All Rights Reserved.
 * Copyright 2001-2008 S3 Graphics, Inc. All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHOR(S) OR COPYRIGHT HOLDER(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

/*
 * INCLUDES 
 */
#include "via.h"
#include "capture.h"
#include "ddmpeg.h"
#include "regrec.h"
#include "ddover.h"
#include "Vinline.h"            /* Universal wait function */
#include "VidInit.h"
#include "VidStruc.h"           /* VIDDATA, video top level structure */
#include "debug.h"              /* for DBG_DD */
#include "engmgr.h"             /* ENGINEMANAGER */
#include "vdefCME.h"
#include "overlay.h"
#include "drm.h"
#include "mm.h"

#ifdef XFree86Server
#include "xf86_ansic.h"         /* needed by compiler.h */
#include "compiler.h"           /* for inb/outb */
#include "xf86drm.h"            /* for drmOpen, etc. */
#endif

/* DEFINES */
#define IGA1                1
#define IGA2                2
#define IGA1_IGA2           3
#define SCREEN_INDEX_0      0
#define SCREEN_INDEX_1      1

/* EXTERN GLOBALS */
extern CLIENT VidClient[CLIENT_PORT_MAX];
extern unsigned long uAreadyInit;

/* EXTERN FUNCTIONS */
extern void vfCMHQVW(unsigned long dwIndex, unsigned long dwData, PVIDDATA pVidData);
extern unsigned long vfCMHQVR(unsigned long dwIndex, PVIDDATA pVidData);
extern void vfCM(unsigned long dwIndex, unsigned long dwData, PVIDDATA pVidData);

/* GLOBAL VARIABLES */
int sammVideoOnScrnNum = 0;
unsigned int VideoInitial;

/* FUNCTIONS */
void InitEngineManager(void)
{
    int i;

    DBG_DD(("\n"));
    /* Initial all the variables. */
    for (i = 0; i < CLIENT_PORT_MAX; i++)
    {
        VidClient[i].uClientHandle = CLIENT_INVALID_HANDLE;
    }
    EngMgr_Construct();
    uAreadyInit = 1;
}

unsigned long VIA_Engine_Allocate(PVIDDATA pVidData, unsigned long uPortId,
                                  unsigned long uEng_type, unsigned long uEng_id)
{
    EngMgr_AllocateEngine_Para allocate_para;
    unsigned long dwRet = 0;
    unsigned long uActual;

    DBG_DD(("\n"));
    if (uPortId > CLIENT_PORT_MAX)
        return ENGMGR_FAIL;
    /* alloate the request engine */
    allocate_para.uClient_handle = VidClient[uPortId].uClientHandle;
    allocate_para.uEng_type = uEng_type;
    allocate_para.uEng_id = uEng_id;
    allocate_para.pEng_id_actual = &uActual;

    dwRet = EngMgr_AllocateEngine(pVidData, &allocate_para);

    return dwRet;
}

unsigned long VIA_Engine_Swap(PVIDDATA pVidData, unsigned long uPortId,
                              unsigned long uEng_type, unsigned long uEng_id_src,
                              unsigned long uEng_id_dst)
{
    unsigned long dwRet = 0;
    EngMgr_SwapEngine_Para SwapEngine_Para;

    if (uPortId > CLIENT_PORT_MAX)
        return ENGMGR_FAIL;
    SwapEngine_Para.uClient_handle = VidClient[uPortId].uClientHandle;
    SwapEngine_Para.uEng_type = uEng_type;
    SwapEngine_Para.uEng_id_src = uEng_id_src;
    SwapEngine_Para.uEng_id_dst = uEng_id_dst;
    dwRet = EngMgr_SwapEngine(pVidData, &SwapEngine_Para);

    return dwRet;
}

unsigned long VIA_AdjustClientPriority(unsigned long uPortId, unsigned long uPriority)
{
    unsigned long dwRet = 0;
    EngMgr_AdjustClientPriority_Para AdjustClientPriority_Para;

    if (uPortId > CLIENT_PORT_MAX)
        return ENGMGR_FAIL;
    AdjustClientPriority_Para.uClient_handle = VidClient[uPortId].uClientHandle;
    AdjustClientPriority_Para.uPriority = uPriority;
    dwRet = EngMgr_AdjustClientPriority(&AdjustClientPriority_Para);

    return dwRet;
}

/* VIA_Engine_Free_All: free possessed engines & unregister */
void VIA_Engine_Free_All(PVIDDATA pVidData, unsigned long uPortId)
{
    EngMgr_UnRegClient_Para unregclient_para;

    DBG_DD(("\n"));
    if (uPortId > CLIENT_PORT_MAX)
        return;

    unregclient_para.uClient_handle = VidClient[uPortId].uClientHandle;
    DBG_DD((" unregclient_para.uClient_handle =%ld :\n", unregclient_para.uClient_handle));
    DBG_DD((" VidClient[%ld].uClientHandle =%ld :\n", uPortId,
             VidClient[uPortId].uClientHandle));
    EngMgr_UnRegClient(pVidData, &unregclient_para);
    VidClient[uPortId].uClientHandle = CLIENT_INVALID_HANDLE;
}

/* Lock Surface */
unsigned long LockSurface(PVIDDATA pVidData, LPDDLOCK lpLock)
{
    LPVIAGRAPHICINFO lpVIAGraphicInfo = &pVidData->VIAGraphicInfo;
    unsigned long handle = pVidData->VMI_Info.handle;

    switch (lpLock->dwFourCC)
    {
        case FOURCC_SUBP:
            memcpy(&lpLock->SubDev[handle], &pVidData->ddLock.SubDev[handle], sizeof(SUBDEVICE));
            lpLock->dwPhysicalBase = lpVIAGraphicInfo->ScreenPhysAddr;
            break;

        case FOURCC_ALPHA:
            memcpy(&lpLock->ALPDevice, &pVidData->ddLock.ALPDevice, sizeof(ALPHADEVICE));
            lpLock->dwPhysicalBase = lpVIAGraphicInfo->ScreenPhysAddr;
            break;

        case FOURCC_YUY2:
        case FOURCC_YV12:
        case FOURCC_NV12:      /* for support NV12 overlay mode test-4 */
            memcpy(&lpLock->SWDevice, &pVidData->SWDevice[handle], sizeof(SWDEVICE));
            lpLock->dwPhysicalBase = lpVIAGraphicInfo->ScreenPhysAddr;
            break;

        default:
            break;
    }
    return VMI_OK;

}

void VIAAlphaWin(PVIDDATA pVidData, LPALPHACTRL lpParam)
{
    unsigned long dwControl = 0;

    unsigned char *pjVideo = pVidData->Vhwinfo.pjVideo;

    unsigned long dwCompose = VIDInD(V_COMPOSE_MODE);
    LPVIAGRAPHICINFO lpVIAGraphicInfo = &pVidData->VIAGraphicInfo;

    if ((!lpParam->AlphaEnable) ||
        ((lpVIAGraphicInfo->duoview & DUOVIEW_ENABLED) && (lpParam->type != ALPHA_CONSTANT)
         && (lpParam->type != ALPHA_DISABLE)))
        return;
    else
    {
        switch (lpParam->type)
        {
            case ALPHA_CONSTANT:
                dwControl = 0x00000000;

                switch (lpVIAGraphicInfo->dwDeviceID)
                {
                    case VIA_DEVICE_VT3205:
                        dwControl |= (lpParam->ConstantFactor & 0x0F) << 12;
                        dwCompose = VIDInD(V_COMPOSE_MODE) & ~SELECT_VIDEO_IF_COLOR_KEY;
                        break;
                    case VIA_DEVICE_VT3204:
                    case VIA_DEVICE_VT3314:
                    case VIA_DEVICE_VT3327:
                    case VIA_DEVICE_VT3336:
                    case VIA_DEVICE_VT3364:

                        dwControl |= (lpParam->ConstantFactor & 0xFF) << 8;
                        dwCompose = VIDInD(V_COMPOSE_MODE) & ~SELECT_VIDEO_IF_COLOR_KEY;
                        break;
                    case VIA_DEVICE_CLE1:
                    case VIA_DEVICE_CLE2:
                        switch (DispatchVGARevisionID(pVidData))
                        {
                            case VIA_REVISION_CLECX:
                                dwControl |= (lpParam->ConstantFactor & 0x0F) << 12;
                                dwCompose = VIDInD(V_COMPOSE_MODE) & ~(SELECT_VIDEO_IF_COLOR_KEY |
                                                                SELECT_VIDEO3_IF_COLOR_KEY);
                                break;
                            default:
                                dwControl |= (lpParam->ConstantFactor & 0x0F) << 12;
                                dwCompose = VIDInD(V_COMPOSE_MODE) & ~SELECT_VIDEO_IF_COLOR_KEY;
                                break;
                        }
                        break;
                    case VIA_DEVICE_VT3259:
                    case VIA_DEVICE_VT3324:
                    case VIA_DEVICE_VT3293:
                    case VIA_DEVICE_VT3353:
                        dwControl |= (lpParam->ConstantFactor & 0xFF) << 8;
                        vfCM(V_COLOR_KEY,
                              dMMRD(pjVideo, V_COLOR_KEY) & (~VAL_VIDEO_COLOR_KEY_ENABLE_CME),
                              pVidData);
                        vfCM(V3_COLOR_KEY,
                              dMMRD(pjVideo, V3_COLOR_KEY) & (~VAL_VIDEO_COLOR_KEY_ENABLE_CME),
                              pVidData);
                        break;

                    default:
                        DBG_DD(("Unknown DeviceID\n"));
                        break;
                }

                pVidData->VidStatus.dwAlphaEnabled = 1;
                break;

            case ALPHA_STREAM:

                switch (lpVIAGraphicInfo->dwDeviceID)
                {
                    case VIA_DEVICE_VT3205:
                        dwControl = 0x00000001 | 0x0f << 12;
                        dwCompose = VIDInD(V_COMPOSE_MODE) & ~SELECT_VIDEO_IF_COLOR_KEY;
                        break;
                    case VIA_DEVICE_VT3204:
                    case VIA_DEVICE_VT3314:
                    case VIA_DEVICE_VT3327:
                    case VIA_DEVICE_VT3336:
                    case VIA_DEVICE_VT3364:
                        dwControl = 0x00000001 | 0xff << 8;
                        dwCompose = VIDInD(V_COMPOSE_MODE) & ~SELECT_VIDEO_IF_COLOR_KEY;
                        break;
                    case VIA_DEVICE_CLE1:
                    case VIA_DEVICE_CLE2:
                        switch (DispatchVGARevisionID(pVidData))
                        {
                            case VIA_REVISION_CLECX:
                                dwControl = 0x00000001 | 0x0f << 12;
                                dwCompose = VIDInD(V_COMPOSE_MODE) & 
                                            ~(SELECT_VIDEO_IF_COLOR_KEY |
                                            SELECT_VIDEO3_IF_COLOR_KEY);
                                break;
                            default:
                                dwControl = 0x00000001 | 0x0f << 12;
                                /* Overlay source format for V3 */
                                dwCompose = VIDInD(V_COMPOSE_MODE) & ~SELECT_VIDEO_IF_COLOR_KEY;
                                break;
                        }
                        break;
                    case VIA_DEVICE_VT3259:
                    case VIA_DEVICE_VT3324:
                    case VIA_DEVICE_VT3293:
                    case VIA_DEVICE_VT3353:
                        dwControl = 0x00000001 | 0xff << 8;
                        vfCM(V_COLOR_KEY,
                              dMMRD(pjVideo, V_COLOR_KEY) & (~VAL_VIDEO_COLOR_KEY_ENABLE_CME),
                              pVidData);
                        vfCM(V3_COLOR_KEY,
                              dMMRD(pjVideo, V3_COLOR_KEY) & (~VAL_VIDEO_COLOR_KEY_ENABLE_CME),
                              pVidData);
                        break;

                    default:
                        DBG_DD(("Unknown DeviceID\n"));
                        break;
                }

                pVidData->VidStatus.dwAlphaEnabled = 1;
                break;

            case ALPHA_GRAPHIC:
                dwControl = 0x00000002;
                break;

            case ALPHA_COMBINE:

                switch (lpVIAGraphicInfo->dwDeviceID)
                {
                    case VIA_DEVICE_VT3205:
                        dwControl = 0x00000001 | (lpParam->ConstantFactor & 0x0F) << 12;
                        dwCompose = VIDInD(V_COMPOSE_MODE) & ~SELECT_VIDEO_IF_COLOR_KEY;
                        break;
                    case VIA_DEVICE_VT3204:
                    case VIA_DEVICE_VT3314:
                    case VIA_DEVICE_VT3327:
                    case VIA_DEVICE_VT3336:
                    case VIA_DEVICE_VT3364:
                        dwControl = 0x00000001 | (lpParam->ConstantFactor & 0xFF) << 8;
                        dwCompose = VIDInD(V_COMPOSE_MODE) & ~SELECT_VIDEO_IF_COLOR_KEY;
                        break;
                    case VIA_DEVICE_CLE1:
                    case VIA_DEVICE_CLE2:
                        switch (DispatchVGARevisionID(pVidData))
                        {
                            case VIA_REVISION_CLECX:
                                dwControl = 0x00000001 | (lpParam->ConstantFactor & 0x0F) << 12;
                                dwCompose = VIDInD(V_COMPOSE_MODE) & 
                                            ~(SELECT_VIDEO_IF_COLOR_KEY |
                                            SELECT_VIDEO3_IF_COLOR_KEY);
                                break;
                            default:
                                dwControl = 0x00000001 | (lpParam->ConstantFactor & 0x0F) << 12;
                                /* Overlay source format for V3 */
                                dwCompose = VIDInD(V_COMPOSE_MODE) & ~SELECT_VIDEO_IF_COLOR_KEY;
                                break;
                        }
                        break;
                    case VIA_DEVICE_VT3259:
                    case VIA_DEVICE_VT3324:
                    case VIA_DEVICE_VT3293:
                    case VIA_DEVICE_VT3353:
                        dwControl = 0x00000001 | (lpParam->ConstantFactor & 0xFF) << 8;
                        vfCM(V_COLOR_KEY,
                              dMMRD(pjVideo, V_COLOR_KEY) & (~VAL_VIDEO_COLOR_KEY_ENABLE_CME),
                              pVidData);
                        vfCM(V3_COLOR_KEY,
                              dMMRD(pjVideo, V3_COLOR_KEY) & (~VAL_VIDEO_COLOR_KEY_ENABLE_CME),
                              pVidData);
                        break;

                    default:
                        DBG_DD(("Unknown DeviceID\n"));
                        break;
                }

                pVidData->VidStatus.dwAlphaEnabled = 1;
                break;

            case ALPHA_DISABLE:
                /* H/W has no disable blending bit(s), so use Constant Alpha */
                dwControl = 0x00000000;

                switch (lpVIAGraphicInfo->dwDeviceID)
                {
                    case VIA_DEVICE_VT3205:
                        dwControl |= (0x0f << 12);
                        dwCompose = VIDInD(V_COMPOSE_MODE) | SELECT_VIDEO_IF_COLOR_KEY;
                        break;
                    case VIA_DEVICE_VT3204:
                    case VIA_DEVICE_VT3314:
                    case VIA_DEVICE_VT3327:
                    case VIA_DEVICE_VT3336:
                    case VIA_DEVICE_VT3364:
                        dwControl |= (0xff << 8);
                        dwCompose = VIDInD(V_COMPOSE_MODE) | SELECT_VIDEO_IF_COLOR_KEY;
                        break;
                    case VIA_DEVICE_CLE1:
                    case VIA_DEVICE_CLE2:
                        switch (DispatchVGARevisionID(pVidData))
                        {
                            case VIA_REVISION_CLECX:
                                dwControl |= (0x0f << 12);
                                dwCompose = VIDInD(V_COMPOSE_MODE) | SELECT_VIDEO_IF_COLOR_KEY |
                                            SELECT_VIDEO3_IF_COLOR_KEY;
                                break;
                            default:
                                dwControl |= (0x0f << 12);
                                /* Overlay source format for V3 */
                                dwCompose = VIDInD(V_COMPOSE_MODE) | SELECT_VIDEO_IF_COLOR_KEY;
                                break;
                        }
                        break;
                    case VIA_DEVICE_VT3259:
                    case VIA_DEVICE_VT3324:
                    case VIA_DEVICE_VT3293:
                    case VIA_DEVICE_VT3353:
                        dwControl |= (0xff << 8);
                        vfCM(V_COLOR_KEY,
                              dMMRD(pjVideo, V_COLOR_KEY) | VAL_VIDEO_COLOR_KEY_ENABLE_CME,
                              pVidData);
                        vfCM(V3_COLOR_KEY,
                              dMMRD(pjVideo, V3_COLOR_KEY) | VAL_VIDEO_COLOR_KEY_ENABLE_CME,
                              pVidData);
                        break;
                    default:
                        DBG_DD(("Unknown DeviceID\n"));
                        break;
                }

                pVidData->VidStatus.dwAlphaEnabled = 0;
                break;
        }

        dwControl |= 0x00080000;        /* Expire number ;divided by 4 */

        VIDOutD(V_ALPHA_CONTROL, dwControl);
        VIDOutD(V_COMPOSE_MODE, dwCompose);
        DBG_DD(("00000210 %08lx\n", VIDInD(V_ALPHA_CONTROL)));
        DBG_DD(("00000220 %08lx\n", VIDInD(V_COMPOSE_MODE)));

    }

}

/************************************************************************************************
*  Function :   To determine V1 is on Screen 0 (V1onScreen1 = 0) or on screen 1 (V1onScreen1 = 1)
*               and V3 is on screen 0 (V3offScreen0 = 0) or on screen 1 (V3offScreen0 = 1)      
*               For SAMM mode
*************************************************************************************************/
void ScreenSwitchManager(PVIDDATA pVidData, LPDDUPDATEOVERLAY lpUpdate)
{
    LPVIAGRAPHICINFO lpVIAGraphicInfo = &pVidData->VIAGraphicInfo;
    unsigned long handle = pVidData->VMI_Info.handle;

    DBG_DD(("\n"));
    DBG_DD(("lpVIAGraphicInfo->SAMM=%ld\n", (unsigned long) lpVIAGraphicInfo->SAMM));
    DBG_DD(("lpUpdate->rDest.left=%ld\n", lpUpdate->rDest.left));
    DBG_DD(("lpVIAGraphicInfo->Screen[0].dwSaveWidth=%ld\n",
             lpVIAGraphicInfo->Screen[0].dwSaveWidth));
    DBG_DD(("lpVIAGraphicInfo->Screen1IsLeft=%ld\n",
             (unsigned long) lpVIAGraphicInfo->Screen1IsLeft));

    DBG_DD(("lpUpdate->rDest.top=%ld\n", lpUpdate->rDest.top));
    DBG_DD(("lpVIAGraphicInfo->Screen[0].dwSaveHeight=%ld\n",
             lpVIAGraphicInfo->Screen[0].dwSaveHeight));
    DBG_DD(("lpVIAGraphicInfo->Screen1IsAbove=%ld\n",
             (unsigned long) lpVIAGraphicInfo->Screen1IsAbove));

    if ((lpVIAGraphicInfo->SAMM)
        && (lpUpdate->rDest.left >= lpVIAGraphicInfo->Screen[0].dwSaveWidth)
        && (lpUpdate->rDest.left < 65535) && !(lpVIAGraphicInfo->Screen1IsLeft))
    {
        /* Screen1 RightOf Screen0, V1 or V3 on Screen1 */
        lpUpdate->rDest.left = lpUpdate->rDest.left - lpVIAGraphicInfo->Screen[0].dwSaveWidth;
        lpUpdate->rDest.right = lpUpdate->rDest.right - lpVIAGraphicInfo->Screen[0].dwSaveWidth;

        switch (lpUpdate->dwFourcc)
        {
            case FOURCC_NV12:
            case FOURCC_YV12:
            case FOURCC_YUY2:
                pVidData->VidStatus.dwVideoFlagSW[handle] |= SW_ON_SND;
                break;
            default:
                break;
        }
        sammVideoOnScrnNum = 1;
        DBG_DD(("Case1:  Screen1 RightOf Screen0, V1 or V3 on Screen1 \n"));
    }

    else if ((lpVIAGraphicInfo->SAMM)
             && (lpUpdate->rDest.top >= lpVIAGraphicInfo->Screen[0].dwSaveHeight)
             && (lpUpdate->rDest.top < 65535) && !(lpVIAGraphicInfo->Screen1IsAbove))
    {
        /* Screen1 Below Screen0, V1 or V3 on Screen1 */
        lpUpdate->rDest.top = lpUpdate->rDest.top - lpVIAGraphicInfo->Screen[0].dwSaveHeight;
        lpUpdate->rDest.bottom = lpUpdate->rDest.bottom - lpVIAGraphicInfo->Screen[0].dwSaveHeight;

        switch (lpUpdate->dwFourcc)
        {
            case FOURCC_NV12:
            case FOURCC_YV12:
            case FOURCC_YUY2:
                pVidData->VidStatus.dwVideoFlagSW[handle] |= SW_ON_SND;
                break;
            default:
                break;
        }
        sammVideoOnScrnNum = 1;
        DBG_DD(("Case2 : V1orV3onScreen1_Below\n"));
    }

    else if ((lpVIAGraphicInfo->Screen1IsLeft) &&
             ((lpUpdate->rDest.left < lpVIAGraphicInfo->Screen[1].dwSaveWidth)
              || (lpUpdate->rDest.left > 65535)))
    {
        /* Screen1 LeftOf Screen0, V1 or V3 on Screen1 */
        switch (lpUpdate->dwFourcc)
        {
            case FOURCC_NV12:
            case FOURCC_YV12:
            case FOURCC_YUY2:
                pVidData->VidStatus.dwVideoFlagSW[handle] |= SW_ON_SND;
                break;
            default:
                break;
        }
        sammVideoOnScrnNum = 1;
        DBG_DD(("Case3: Screen1 LeftOf Screen0, V1 or V3 on Screen1\n"));
    }

    else if ((lpVIAGraphicInfo->Screen1IsLeft)
             && (lpUpdate->rDest.left >= lpVIAGraphicInfo->Screen[1].dwSaveWidth)
             && (lpUpdate->rDest.left < 65535))
    {
        /* Screen1 LeftOf Screen0, V1 or V3 on Screen0 */
        lpUpdate->rDest.left = lpUpdate->rDest.left - lpVIAGraphicInfo->Screen[1].dwSaveWidth;
        lpUpdate->rDest.right = lpUpdate->rDest.right - lpVIAGraphicInfo->Screen[1].dwSaveWidth;

        switch (lpUpdate->dwFourcc)
        {
            case FOURCC_NV12:
            case FOURCC_YV12:
            case FOURCC_YUY2:
                pVidData->VidStatus.dwVideoFlagSW[handle] &= ~SW_ON_SND;
                break;
            default:
                break;
        }
        sammVideoOnScrnNum = 0;
        DBG_DD(("Case4: Screen1 LeftOf Screen0, V1 or V3 on Screen0\n"));
    }

    else if ((lpVIAGraphicInfo->Screen1IsAbove) &&
             ((lpUpdate->rDest.top < lpVIAGraphicInfo->Screen[1].dwSaveHeight)
              || (lpUpdate->rDest.top > 65535)))
    {
        /* Screen1 Above Screen0, V1 or V3 on Screen1 */
        switch (lpUpdate->dwFourcc)
        {
            case FOURCC_NV12:
            case FOURCC_YV12:
            case FOURCC_YUY2:
                pVidData->VidStatus.dwVideoFlagSW[handle] |= SW_ON_SND;
                break;
            default:
                break;
        }
        sammVideoOnScrnNum = 1;
        DBG_DD(("Case5: Screen1 Above Screen0, V1 or V3 on Screen1\n"));
    }
    else if ((lpVIAGraphicInfo->Screen1IsAbove)
             && (lpUpdate->rDest.top >= lpVIAGraphicInfo->Screen[1].dwSaveHeight)
             && (lpUpdate->rDest.top < 65535))
    {
        /* Screen1 Above Screen0, V1 or V3 on Screen0 */
        lpUpdate->rDest.top = lpUpdate->rDest.top - lpVIAGraphicInfo->Screen[1].dwSaveHeight;
        lpUpdate->rDest.bottom = lpUpdate->rDest.bottom - lpVIAGraphicInfo->Screen[1].dwSaveHeight;

        switch (lpUpdate->dwFourcc)
        {
            case FOURCC_NV12:
            case FOURCC_YV12:
            case FOURCC_YUY2:
                pVidData->VidStatus.dwVideoFlagSW[handle] &= ~SW_ON_SND;
                break;
            default:
                break;
        }
        sammVideoOnScrnNum = 0;
        DBG_DD(("Case6: Screen1 Above Screen0, V1 or V3 on Screen0\n"));
    }
    else
    {
        /* V1 and V3 on Screen0 */
        switch (lpUpdate->dwFourcc)
        {
            case FOURCC_NV12:
            case FOURCC_YV12:
            case FOURCC_YUY2:
                pVidData->VidStatus.dwVideoFlagSW[handle] &= ~SW_ON_SND;
                break;
            default:
                break;
        }
        sammVideoOnScrnNum = 0;
        DBG_DD(("Case7: V1 and V3 on Screen0\n"));
    }
}

unsigned long UpdateOverlay_CheckVideoStatus(PVIDDATA pVidData, LPDDUPDATEOVERLAY lpUpdate)
{
    unsigned long dwRet = VMI_OK;
    unsigned long handle = pVidData->VMI_Info.handle;

    DBG_DD((" UpdateOverlay_CheckVideoStatus !! \n"));

    switch (lpUpdate->dwFourcc)
    {
        /*********************************** UpdateOverlay_CheckVideoStatus for SWOV ***********************************/
        case FOURCC_YV12:
        case FOURCC_NV12:
        case FOURCC_YUY2:
            {
                DBG_DD((" Upd_Video case: \n"));

                pVidData->VidStatus.dwVideoFlagSW[handle] |= VIDEO_SHOW;
                VIA_AdjustClientPriority(CLIENT_SW, URGENT_PRIORITY);

                if (pVidData->VidStatus.dwVideoFlagSW[handle] & SW_ON_SND)
                {
                    if (pVidData->VideoHWDifference.dwOnlyOneVideoOverlay)
                    {
                        DBG_DD((" only one video engine in device !! \n"));
                    }
                    else
                    {
                        if (!(pVidData->VidStatus.dwVideoFlagSW[handle] & VIDEO_1_INUSE))
                        {
                            if (VIA_Engine_Swap(pVidData, CLIENT_SW, VID_ENGINE, 1, 0))
                            {
                                pVidData->VidStatus.dwVideoFlagSW[handle] &= ~VIDEO_3_INUSE;
                                pVidData->VidStatus.dwVideoFlagSW[handle] |= VIDEO_1_INUSE;
                                DBG_DD((" Switch V3 to V1 !! \n"));
                            }
                            else
                            {
                                pVidData->VidStatus.dwVideoFlagSW[handle] &= ~VIDEO_SHOW;
                                DBG_DD((" SW Hide !!==> coz no V1 for SW to show on SND screen!  \n"));
                            }
                        }
                    }
                }
                dwRet = fpUpd_Video(pVidData);
                if (dwRet == VMI_ERR)
                {
                    return VMI_ERR;
                }
                pVidData->VidStatus.dwVideoFlagSW[handle] |= VIDEO_ON;
                VIA_AdjustClientPriority(CLIENT_SW, NORMAL_PRIORITY);
                return dwRet;
            }
            break;
        default:
            break;
    }

    return dwRet;
}


void scaleOverlayDestWindow(LPDDUPDATEOVERLAY lpUpdate,
                            int OrigHActive, int OrigVActive,
                            int RealHActive, int RealVActive,
                            long XSIZEOffset, long YSIZEOffset,
                            long XPOSITIONOffset, long YPOSITIONOffset)
{
    int tmpLeft, tmpRight, tmpTop, tmpBottom;
    int screenX, screenY;

    DBG_DD(("scaleOverlayDestWindow: OriginalActiveSize=%dx%d \n", OrigHActive, OrigVActive));
    DBG_DD(("scaleOverlayDestWindow: RealActiveSize=%dx%d \n", RealHActive, RealVActive));
    DBG_DD(("scaleOverlayDestWindow: Offset XSize=%d, YSize=%d, XPosition=%d, YPosition=%d \n",
             XSIZEOffset, YSIZEOffset, XPOSITIONOffset, YPOSITIONOffset));
    DBG_DD(("scaleOverlayDestWindow: Old rDest >> left=%d, right=%d, top=%d, bottom=%d \n",
             lpUpdate->rDest.left, lpUpdate->rDest.right, lpUpdate->rDest.top,
             lpUpdate->rDest.bottom));

    tmpLeft = ((int) lpUpdate->rDest.left * RealHActive) / OrigHActive;
    tmpRight = ((int) lpUpdate->rDest.right * RealHActive) / OrigHActive;
    tmpTop = ((int) lpUpdate->rDest.top * RealVActive) / OrigVActive;
    tmpBottom = ((int) lpUpdate->rDest.bottom * RealVActive) / OrigVActive;

    screenX = RealHActive - XSIZEOffset * 2;
    screenY = RealVActive - YSIZEOffset * 2;

    tmpLeft = (screenX / 2) - (RealHActive / 2 - tmpLeft) * screenX / RealHActive;
    tmpRight = (screenX / 2) - (RealHActive / 2 - tmpRight) * screenX / RealHActive;
    tmpTop = (screenY / 2) - (RealVActive / 2 - tmpTop) * screenY / RealVActive;
    tmpBottom = (screenY / 2) - (RealVActive / 2 - tmpBottom) * screenY / RealVActive;

    /* software shrink adjust windows size */
    tmpLeft += XSIZEOffset + XPOSITIONOffset;
    tmpRight += XSIZEOffset + XPOSITIONOffset;
    tmpTop += YSIZEOffset - YPOSITIONOffset;
    tmpBottom += YSIZEOffset - YPOSITIONOffset;

    lpUpdate->rDest.left = (unsigned long) tmpLeft;
    lpUpdate->rDest.right = (unsigned long) tmpRight;
    lpUpdate->rDest.top = (unsigned long) tmpTop;
    lpUpdate->rDest.bottom = (unsigned long) tmpBottom;

    DBG_DD(("scaleOverlayDestWindow: new rDest >> left=%d, right=%d, top=%d, bottom=%d \n",
             lpUpdate->rDest.left, lpUpdate->rDest.right, lpUpdate->rDest.top,
             lpUpdate->rDest.bottom));
}

unsigned long UpdateOverlay_Internal(PVIDDATA pVidData, LPDDUPDATEOVERLAY lpUpdate, int ScrnIndex)
{
    unsigned long handle = pVidData->VMI_Info.handle;
    unsigned long dwFlags = lpUpdate->dwFlags;

    int nDstTop, nDstBottom, nDstLeft, nDstRight, nTopBak = 0, nDstWidth, nDstHeight;

    LPMPGDEVICE lpMPGDevice = &(pVidData->MPGDevice[handle]);
    LPSWDEVICE lpSWDevice = &(pVidData->SWDevice[handle]);
    LPVIAGRAPHICINFO lpVIAGraphicInfo = &pVidData->VIAGraphicInfo;

    LPOVERLAYRECORD lpOverlayRecordMPEG = &pVidData->OverlayRecordSet.OverlayRecordMPEG[handle];
    LPOVERLAYRECORD lpOverlayRecordSW = &pVidData->OverlayRecordSet.OverlayRecordSW[handle];

    LPDDPIXELFORMAT lpDPFsrc = &pVidData->DPFsrc[handle];
    LPVIDHWDIFFERENCE lpVideoHWDifference = &(pVidData->VideoHWDifference);
    VIDYUVBYPASS vidYUVByPass;

#ifdef XFree86Server
    unsigned char *pjVideo = pVidData->Vhwinfo.pjVideo;
#endif

    /* for SAMM panning Mode */
    if (lpVIAGraphicInfo->Screen[ScrnIndex].dwPanning)
    {
        DBG_DD(("Panning!!\n"));
        lpUpdate->rDest.left = lpUpdate->rDest.left - (pVidData->panning_x[ScrnIndex] -
                               pVidData->panning_old_x[ScrnIndex]);
        lpUpdate->rDest.top = lpUpdate->rDest.top - (pVidData->panning_y[ScrnIndex] -
                              pVidData->panning_old_y[ScrnIndex]);
        lpUpdate->rDest.right = lpUpdate->rDest.right - (pVidData->panning_x[ScrnIndex] -
                                pVidData->panning_old_x[ScrnIndex]);
        lpUpdate->rDest.bottom = lpUpdate->rDest.bottom - (pVidData->panning_y[ScrnIndex] -
                                 pVidData->panning_old_y[ScrnIndex]);
    }

    /* Recalculate new destination position when down scaling is enabled */
    if (lpVIAGraphicInfo->Screen[ScrnIndex].IsDownScaling)
    {
        scaleOverlayDestWindow(lpUpdate, lpVIAGraphicInfo->Screen[ScrnIndex].dwCurWidth,
                                lpVIAGraphicInfo->Screen[ScrnIndex].dwCurHeight,
                                lpVIAGraphicInfo->Screen[ScrnIndex].dwWidth,
                                lpVIAGraphicInfo->Screen[ScrnIndex].dwHeight, 0, 0, 0, 0);
        if ((lpUpdate->rDest.right - lpUpdate->rDest.left) <
             (lpUpdate->rSrc.right - lpUpdate->rSrc.left))
        {
            pVidData->isVideoWidthDownScaling = TRUE;
        }

        if ((lpUpdate->rDest.bottom - lpUpdate->rDest.top) <
            (lpUpdate->rSrc.bottom - lpUpdate->rSrc.top))
        {
            pVidData->isVideoHeightDownScaling = TRUE;
        }
    }

    DBG_DD(("Raw rSrc  X (%ld,%ld) Y (%ld,%ld)\n",
             lpUpdate->rSrc.left, lpUpdate->rSrc.right, lpUpdate->rSrc.top, lpUpdate->rSrc.bottom));
    DBG_DD(("Raw rDest  X (%ld,%ld) Y (%ld,%ld)\n",
             lpUpdate->rDest.left, lpUpdate->rDest.right, lpUpdate->rDest.top,
             lpUpdate->rDest.bottom));

    dwFlags |= DDOVER_INTERLEAVED;

    /* For Alpha windows setting */
    if (pVidData->VidStatus.dwAlphaEnabled)
    {
        dwFlags &= ~DDOVER_KEYDEST;
    }

    Macro_VidREGRec(VIDREGREC_RESET_COUNTER, 0, 0);

    if (dwFlags & DDOVER_HIDE)
    {
        DBG_DD(("DDOVER_HIDE \n"));

        switch (lpUpdate->dwFourcc)
        {
            case FOURCC_NV12:
            case FOURCC_YV12:
            case FOURCC_YUY2:
                pVidData->VidStatus.dwVideoFlagSW[handle] &= ~VIDEO_SHOW;
                if (fpUpd_Video(pVidData) == VMI_ERR)
                {
                    return VMI_ERR;
                }

                pVidData->VidStatus.dwVideoFlagSW[handle] &= ~VIDEO_ON;
                pVidData->UpdateOverlaySWBackup.rDest.left = 0;
                pVidData->UpdateOverlaySWBackup.rDest.top = 0;
                pVidData->UpdateOverlaySWBackup.rDest.right = 0;
                pVidData->UpdateOverlaySWBackup.rDest.bottom = 0;
                break;
            default:
                break;
        }

        if (pVidData->dwUseExtendedFIFO)
        {
            /* Restore Display fifo */
#ifdef XFree86Server
            OUTB (0x16, 0x3C4);
            OUTB (pVidData->Save_3C4_16, 0x3C5);
            DBG_DD (("Restore 3c4.16 : %08x \n", INB (0x3C5)));
            OUTB (0x17, 0x3C4);
            OUTB (pVidData->Save_3C4_17, 0x3C5);
            DBG_DD (("        3c4.17 : %08x \n", INB (0x3C5)));
            OUTB (0x18, 0x3C4);
            OUTB (pVidData->Save_3C4_18, 0x3C5);
            DBG_DD (("        3c4.18 : %08x \n", INB (0x3C5)));
#endif

            pVidData->dwUseExtendedFIFO = 0;
        }

        return VMI_OK;
    }

    /* Assigning one video on top of another in two stream's case */
    if (dwFlags & DDOVER_ON_TOP)
    {
        DBG_DD(("DDOVER_ON_TOP \n"));
        switch (lpUpdate->dwFourcc)
        {
            case FOURCC_NV12:
            case FOURCC_YV12:
            case FOURCC_YUY2:
                lpOverlayRecordSW->dwVideoFlag |= DDOVER_ON_TOP;
                pVidData->VidStatus.dwVideoFlagSW[handle] |= VIDEO_ON_TOP;
                break;
            default:
                break;
        }
        if (UpdateOverlay_CheckVideoStatus(pVidData, lpUpdate) == VMI_ERR)
        {
            return VMI_ERR;
        }
        else
        {
            /* after on_top action over,clean the flag */
            switch (lpUpdate->dwFourcc)
            {
                case FOURCC_NV12:
                case FOURCC_YV12:
                case FOURCC_YUY2:
                    lpOverlayRecordSW->dwVideoFlag &= ~DDOVER_ON_TOP;
                    pVidData->VidStatus.dwVideoFlagSW[handle] &= ~VIDEO_ON_TOP;
                    break;
                default:
                    break;
            }
            return VMI_OK;
        }
    }
    if (dwFlags & DDOVER_KEYDEST)
    {
        DBG_DD(("DDOVER_KEYDEST \n"));

        if (lpVIAGraphicInfo->Screen[ScrnIndex].dwBPP == 8)
        {
            lpUpdate->dwColorSpaceLowValue &= 0xFF;
        }
        else if (lpVIAGraphicInfo->Screen[ScrnIndex].dwBPP == 16)
        {
            lpUpdate->dwColorSpaceLowValue &= 0xFFFF;
        }
        else if (lpVIAGraphicInfo->Screen[ScrnIndex].dwBPP == 32)
        {
            lpUpdate->dwColorSpaceLowValue &= 0x00FFFFFF;
        }

        switch (lpUpdate->dwFourcc)
        {
            case FOURCC_NV12:
            case FOURCC_YV12:
            case FOURCC_YUY2:
                lpOverlayRecordSW->dwColorKeyOn = 1;
                lpOverlayRecordSW->dwKeyLow = lpUpdate->dwColorSpaceLowValue;
                break;
            default:
                break;
        }
    }
    else
    {
        switch (lpUpdate->dwFourcc)
        {
            case FOURCC_NV12:
            case FOURCC_YV12:
            case FOURCC_YUY2:
                lpOverlayRecordSW->dwColorKeyOn = 0;
                break;
            default:
                break;
        }
    }

    if (dwFlags & DDOVER_SHOW)
    {
        unsigned long dwScnWidth, dwScnHeight;
        unsigned long dwRawSrcWidth, dwRawSrcHeight;

        DBG_DD(("DDOVER_SHOW \n"));

        if (dwFlags & DDOVER_INTERLEAVED)
        {
            DBG_DD(("DDOVER_INTERLEAVED\n"));

            switch (lpUpdate->dwFourcc)
            {
                case FOURCC_NV12:
                case FOURCC_YV12:
                case FOURCC_YUY2:
                    lpOverlayRecordSW->dwDeinterlaceMode |= DDOVER_INTERLEAVED;
                    break;
                default:
                    break;
            }
        }

        if (dwFlags & DDOVER_BOB)
        {
            DBG_DD(("DDOVER_BOB : bob-deinterlace(for HW mpeg path only)\n"));

            switch (lpUpdate->dwFourcc)
            {
                case FOURCC_NV12:
                case FOURCC_YV12:
                case FOURCC_YUY2:
                    lpOverlayRecordSW->dwDeinterlaceMode |= DDOVER_BOB;
                    break;
                default:
                    break;
            }
        }
        else
        {
            DBG_DD(("No DDOVER_BOB : weave-deinterlace(for HW mpeg path only)\n"));

            switch (lpUpdate->dwFourcc)
            {
                case FOURCC_NV12:
                case FOURCC_YV12:
                case FOURCC_YUY2:
                    lpOverlayRecordSW->dwDeinterlaceMode &= ~DDOVER_BOB;
                    break;
                default:
                    break;
            }
        }
        if (dwFlags & DDOVER_STAY_OLD)
        {
            DBG_DD(("DDOVER_STAY_OLD \n"));
            switch (lpUpdate->dwFourcc)
            {
                case FOURCC_NV12:
                case FOURCC_YV12:
                case FOURCC_YUY2:
                    lpOverlayRecordSW->dwVideoFlag |= DDOVER_STAY_OLD;
                    pVidData->VidStatus.dwVideoFlagSW[handle] |= VIDEO_STAY_OLD;
                    break;
                default:
                    break;
            }
        }
        switch (lpVIAGraphicInfo->dwDeviceID)
        {
            case VIA_DEVICE_CLE1:
            case VIA_DEVICE_CLE2:
                switch (DispatchVGARevisionID(pVidData))
                {
                    case VIA_REVISION_CLEAX:
                        if ((lpVIAGraphicInfo->Screen[0].dwWidth > 1024)
                            || (lpVIAGraphicInfo->SAMM))
                        {
                            DBG_DD(("UseExtendedFIFO\n"));
                            pVidData->dwUseExtendedFIFO = 1;
                        }
                        else
                        {
                            pVidData->dwUseExtendedFIFO = 0;
                        }
                        break;
                    case VIA_REVISION_CLECX:
                    default:
                        pVidData->dwUseExtendedFIFO = 0;
                        break;
                }
                break;
            case VIA_DEVICE_VT3204:
            case VIA_DEVICE_VT3205:
            case VIA_DEVICE_VT3259:
            case VIA_DEVICE_VT3314:
            case VIA_DEVICE_VT3324:
            case VIA_DEVICE_VT3353:
            case VIA_DEVICE_VT3293:
                pVidData->dwUseExtendedFIFO = 0;
                break;
            default:
                pVidData->dwUseExtendedFIFO = 0;
                DBG_DD(("Unknown DeviceID\n"));
                break;
        }

        /* Chage the value of width and height to the same as screen size when full screen. */
        if ((lpVIAGraphicInfo->rotate == 1) || (lpVIAGraphicInfo->rotate == -1))
        {
            if (((lpUpdate->rDest.right - lpUpdate->rDest.left) ==
                 lpVIAGraphicInfo->Screen[ScrnIndex].dwWidth)
                || ((lpUpdate->rDest.bottom - lpUpdate->rDest.top) ==
                    lpVIAGraphicInfo->Screen[ScrnIndex].dwHeight))
            {
                lpUpdate->rDest.left = 0;
                lpUpdate->rDest.top = 0;
                lpUpdate->rDest.right = lpVIAGraphicInfo->Screen[ScrnIndex].dwHeight;
                lpUpdate->rDest.bottom = lpVIAGraphicInfo->Screen[ScrnIndex].dwWidth;
            }
        }

        nDstLeft = lpUpdate->rDest.left;
        nDstTop = lpUpdate->rDest.top;
        nDstRight = lpUpdate->rDest.right;
        nDstBottom = lpUpdate->rDest.bottom;

        /* Screen width & height */
        if ((lpVIAGraphicInfo->rotate == 1) || (lpVIAGraphicInfo->rotate == -1))
        {
            dwScnWidth = lpVIAGraphicInfo->Screen[ScrnIndex].dwHeight;
            dwScnHeight = lpVIAGraphicInfo->Screen[ScrnIndex].dwWidth;
        }
        else
        {
            dwScnWidth = lpVIAGraphicInfo->Screen[ScrnIndex].dwWidth;
            dwScnHeight = lpVIAGraphicInfo->Screen[ScrnIndex].dwHeight;
        }

        /* the original width & height AP sent */
        dwRawSrcWidth = lpUpdate->rSrc.right - lpUpdate->rSrc.left;
        dwRawSrcHeight = lpUpdate->rSrc.bottom - lpUpdate->rSrc.top;

        if (0 == nDstRight - nDstLeft)
        {
            nDstWidth = 1;
        }
        else
        {
            nDstWidth = nDstRight - nDstLeft;
        }
        if (0 == nDstBottom - nDstTop)
        {
            nDstHeight = 1;
        }
        else
        {
            nDstHeight = nDstBottom - nDstTop;
        }

        switch (lpUpdate->dwFourcc)
        {
            case FOURCC_YV12:
            case FOURCC_YUY2:
            case FOURCC_NV12:
                if (nDstLeft < 0)
                {
                    lpUpdate->rSrc.left += (((-nDstLeft) * (dwRawSrcWidth)) +
                                                  ((nDstRight - nDstLeft) >> 1)) / nDstWidth;
                }

                if (nDstRight > dwScnWidth)
                {
                    lpUpdate->rSrc.right -= (((nDstRight - dwScnWidth) * (dwRawSrcWidth)) +
                                                  ((nDstRight - nDstLeft) >> 1)) / nDstWidth;
                }

                if (nDstTop < 0)
                {
                    lpUpdate->rSrc.top += (((-nDstTop) * (dwRawSrcHeight)) +
                                                 ((nDstBottom - nDstTop) >> 1)) / nDstHeight;
                    nTopBak = (-nDstTop);
                }

                if (nDstBottom > dwScnHeight)
                {
                    lpUpdate->rSrc.bottom -= (((nDstBottom - dwScnHeight) * (dwRawSrcHeight)) +
                                                 ((nDstBottom - nDstTop) >> 1)) / nDstHeight;
                }

                break;
            default:
                break;
        }

        /* save modified src & original dest rectangle param. */
        switch (lpUpdate->dwFourcc)
        {
            case FOURCC_NV12:
            case FOURCC_YV12:
            case FOURCC_YUY2:
                if (lpVIAGraphicInfo->Screen[ScrnIndex].dwPanning)
                {
                    lpSWDevice->gdwSWDstLeft = lpUpdate->rDest.left + 
                                               (pVidData->panning_x[ScrnIndex] -
                                               pVidData->panning_old_x[ScrnIndex]);
                    lpSWDevice->gdwSWDstTop = lpUpdate->rDest.top + 
                                              (pVidData->panning_y[ScrnIndex] -
                                              pVidData->panning_old_y[ScrnIndex]);
                }
                else
                {
                    lpSWDevice->gdwSWDstLeft = lpUpdate->rDest.left;
                    lpSWDevice->gdwSWDstTop = lpUpdate->rDest.top;
                }
                lpSWDevice->gdwSWDstWidth = lpUpdate->rDest.right - lpUpdate->rDest.left;
                lpSWDevice->gdwSWDstHeight = lpUpdate->rDest.bottom - lpUpdate->rDest.top;
                lpSWDevice->gdwSWSrcWidth = lpUpdate->rSrc.right - lpUpdate->rSrc.left;
                lpSWDevice->gdwSWSrcHeight = lpUpdate->rSrc.bottom - lpUpdate->rSrc.top;
                break;
            default:
                break;
        }

        /* Figure out actual rDest rectangle */
        lpUpdate->rDest.left = nDstLeft < 0 ? 0 : nDstLeft;
        lpUpdate->rDest.top = nDstTop < 0 ? 0 : nDstTop;
        if (lpUpdate->rDest.top >= dwScnHeight)
            lpUpdate->rDest.top = dwScnHeight - 1;
        lpUpdate->rDest.right = nDstRight > dwScnWidth ? dwScnWidth : nDstRight;
        lpUpdate->rDest.bottom = nDstBottom > dwScnHeight ? dwScnHeight : nDstBottom;

        DBG_DD((" screen index is  %d \n", ScrnIndex));

        /* Save overlay rSrc & rDest info to OverlayRecord struct */
        switch (lpUpdate->dwFourcc)
        {
            case FOURCC_NV12:
            case FOURCC_YV12:
            case FOURCC_YUY2:
                lpOverlayRecordSW->dwOverlaySrcRect = lpUpdate->rSrc;
                lpOverlayRecordSW->dwOverlayDestRect = lpOverlayRecordSW->dwExpandDestRect = lpUpdate->rDest;
                lpOverlayRecordSW->dwOverlayDestWidth = lpUpdate->rDest.right - lpUpdate->rDest.left;
                lpOverlayRecordSW->dwOverlayDestHeight = lpUpdate->rDest.bottom - lpUpdate->rDest.top;

                if (lpVIAGraphicInfo->Screen[ScrnIndex].dwExpand)
                {
                    DBG_DD(("===dwExpand!! \n"));
                    lpOverlayRecordSW->dwExpandDestRect.top = (lpUpdate->rDest.top) *
                                            (lpVIAGraphicInfo->Screen[ScrnIndex].dwPanelHeight) /
                                            (lpVIAGraphicInfo->Screen[ScrnIndex].dwHeight);
                    lpOverlayRecordSW->dwExpandDestRect.bottom = (lpUpdate->rDest.bottom) *
                                            (lpVIAGraphicInfo->Screen[ScrnIndex].dwPanelHeight) /
                                            (lpVIAGraphicInfo->Screen[ScrnIndex].dwHeight);

                    /* This will be used for rotation: */
                    lpOverlayRecordSW->dwExpandDestRect.left = (lpUpdate->rDest.left) *
                                            (lpVIAGraphicInfo->Screen[ScrnIndex].dwPanelWidth) /
                                            (lpVIAGraphicInfo->Screen[ScrnIndex].dwWidth);
                    lpOverlayRecordSW->dwExpandDestRect.right = (lpUpdate->rDest.right) *
                                            (lpVIAGraphicInfo->Screen[ScrnIndex].dwPanelWidth) /
                                            (lpVIAGraphicInfo->Screen[ScrnIndex].dwWidth);
                }
                break;
            default:
                break;
        }

        /* check which update func. (upd_MPEG, upd_video, upd_capture) to call */
        if (UpdateOverlay_CheckVideoStatus(pVidData, lpUpdate) == VMI_ERR)
        {
            return VMI_ERR;
        }
        else
        {
            /* after stay_old action over,clean the flag,otherwise the next updateovelay may have error */
            switch (lpUpdate->dwFourcc)
            {
                case FOURCC_NV12:
                case FOURCC_YV12:
                case FOURCC_YUY2:
                    lpOverlayRecordSW->dwVideoFlag &= ~DDOVER_STAY_OLD;
                    pVidData->VidStatus.dwVideoFlagSW[handle] &= ~VIDEO_STAY_OLD;
                    break;
                default:
                    break;
            }
            return VMI_OK;
        }

    }
    pVidData->panning_old_x[ScrnIndex] = pVidData->panning_x[ScrnIndex];
    pVidData->panning_old_y[ScrnIndex] = pVidData->panning_y[ScrnIndex];

    return VMI_OK;

}


unsigned long UpdateOverlay(PVIDDATA pVidData, LPDDUPDATEOVERLAY lpUpdate)
{
    unsigned long handle = pVidData->VMI_Info.handle;
    unsigned long dwFlags = lpUpdate->dwFlags;
    unsigned long dwRet = VMI_OK;
    int ScrnIndex = 0;

    LPVIAGRAPHICINFO lpVIAGraphicInfo = &pVidData->VIAGraphicInfo;
    LPDDPIXELFORMAT lpDPFsrcTV = &pVidData->DPFsrcTV;
    LPOVERLAYRECORD lpOverlayRecordMPEG = &pVidData->OverlayRecordSet.OverlayRecordMPEG[handle];
    LPOVERLAYRECORD lpOverlayRecordSW = &pVidData->OverlayRecordSet.OverlayRecordSW[handle];

    DDUPDATEOVERLAY UpdateBakup;

    /* Which IGA on which Screen, for SAMM use, and DuoView is fixed to {IGA1, IGA2, -1} */
    unsigned long *Screen_IGA_Map = lpVIAGraphicInfo->Screen_IGA_Map;

    unsigned long *IGA_Video_Map = lpVIAGraphicInfo->IGA_Video_Map;

    DBG_DD(("lpVIAGraphicInfo->duoview=%p SAMM=%p dwActiveDevice=%p dwVideoFlag=%p \n",
             lpVIAGraphicInfo->duoview, lpVIAGraphicInfo->SAMM, lpVIAGraphicInfo->dwActiveDevice,
             pVidData->VidStatus.dwVideoFlagMPEG[handle]));

    DBG_DD(("into viaUpdateOverlay : lpUpdate.rDest  >> left=%d, right=%d, top = %d, bottom=%d \n",
             lpUpdate->rDest.left, lpUpdate->rDest.right, lpUpdate->rDest.top,
             lpUpdate->rDest.bottom));

    /* Save unmodified overlay rSrc & rDest info to OverlayRecord struct */
    switch (lpUpdate->dwFourcc)
    {
        case FOURCC_NV12:
        case FOURCC_YV12:
        case FOURCC_YUY2:
            DBG_DD(("UpdateOverlay: SW\n"));

            /* Added for two h/w overlays */
            if (pVidData->VidStatus.dwVideoFlagSW[handle] == 0)
            {
                return VMI_OK;
            }

            lpOverlayRecordSW->dwOriSrcRect = lpUpdate->rSrc;
            lpOverlayRecordSW->dwOriDestRect = lpUpdate->rDest;

            /* Save updateoverlay info to global VidData struct */
            pVidData->UpdateOverlaySWBackup = *(LPDDUPDATEOVERLAY) lpUpdate;    /*Need to check */

            break;
        default:
            break;
    }

    if (lpVIAGraphicInfo->SAMM)
    {
        ScreenSwitchManager(pVidData, lpUpdate);
        pVidData->VDevInfo.dwVDevFlag = CTL_UseDualView;
    }

    /*1. Single case*/
    if (!((lpVIAGraphicInfo->duoview & DUOVIEW_ENABLED) || (lpVIAGraphicInfo->SAMM)))
    {
        lpVIAGraphicInfo->bVideoEngineUsedOut = 0;
        /* Single device that use IGA2 */
        if (Screen_IGA_Map[SCREEN_INDEX_0] == IGA2)
        {
            switch (lpUpdate->dwFourcc)
            {
                case FOURCC_NV12:
                case FOURCC_YV12:
                case FOURCC_YUY2:
                    pVidData->VidStatus.dwVideoFlagSW[handle] |= SW_ON_SND;
                    DBG_DD(("LCD Only: SW_ON_SND\n"));
                    break;
                default:
                    break;
            }
            pVidData->VDevInfo.dwVDevFlag |= CTL_UseIGA2Path;
        }
        else
        {
            switch (lpUpdate->dwFourcc)
            {
                case FOURCC_NV12:
                case FOURCC_YV12:
                case FOURCC_YUY2:
                    pVidData->VidStatus.dwVideoFlagSW[handle] &= ~SW_ON_SND;
                    break;
                default:
                    break;
            }
            pVidData->VDevInfo.dwVDevFlag &= ~CTL_UseIGA2Path;
        }
    }

    /* 2. SAMM case */
    if (lpVIAGraphicInfo->SAMM)
    {
        lpVIAGraphicInfo->bVideoEngineUsedOut = 0;
        if (dwFlags & DDOVER_ON_SND)
        {
            ScrnIndex = 1;

            switch (lpUpdate->dwFourcc)
            {
                case FOURCC_NV12:
                case FOURCC_YV12:
                case FOURCC_YUY2:
                    pVidData->VidStatus.dwVideoFlagSW[handle] |= SW_ON_SND;
                    DBG_DD(("SW_ON_SND\n"));
                    break;
                default:
                    break;
            }
        }
        else
        {
            switch (lpUpdate->dwFourcc)
            {
                case FOURCC_NV12:
                case FOURCC_YV12:
                case FOURCC_YUY2:
                    if (pVidData->VidStatus.dwVideoFlagSW[handle] & SW_ON_SND)
                    {
                        DBG_DD(("SW_ON_SND\n"));
                        ScrnIndex = 1;
                        dwFlags |= DDOVER_ON_SND;
                    }
                    break;
                default:
                    break;
            }
        }

        DBG_DD(("Screen device[0:1]=%d ,%d \n", lpVIAGraphicInfo->Screen[0].dwActiveDevice,
                 lpVIAGraphicInfo->Screen[1].dwActiveDevice));
        DBG_DD(("video on %d screen , its device is %d\n", sammVideoOnScrnNum,
                 lpVIAGraphicInfo->Screen[sammVideoOnScrnNum].dwActiveDevice));

        if (lpVIAGraphicInfo->Screen_IGA_Map[SCREEN_INDEX_0] == IGA2)
        {
            if (dwFlags & DDOVER_ON_SND)
                dwFlags &= ~DDOVER_ON_SND;
            else
                dwFlags |= DDOVER_ON_SND;

            switch (lpUpdate->dwFourcc)
            {
                case FOURCC_NV12:
                case FOURCC_YV12:
                case FOURCC_YUY2:
                    if (pVidData->VidStatus.dwVideoFlagSW[handle] & SW_ON_SND)
                        pVidData->VidStatus.dwVideoFlagSW[handle] &= ~SW_ON_SND;
                    else
                        pVidData->VidStatus.dwVideoFlagSW[handle] |= SW_ON_SND;
                    DBG_DD(("SW_ON_SND\n"));
                    break;
                default:
                    break;
            }
        }

    }

    /* 3. duoview  case, only one video engine chip need set xxx_ON_SND flag. */
    if (lpVIAGraphicInfo->duoview & DUOVIEW_ENABLED)
    {
        lpVIAGraphicInfo->bVideoEngineUsedOut = 1;
        /* switch in duoview mode */
        /* DDOVER_ON_SND in duoveiw is only use for one-video chips */
        if (lpVIAGraphicInfo->duoview & DUOVIEW_V3_ON_IGA2)
            dwFlags |= DDOVER_ON_SND;
        else
            dwFlags &= ~DDOVER_ON_SND;

        switch (lpVIAGraphicInfo->dwDeviceID)
        {
            case VIA_DEVICE_VT3205:
            case VIA_DEVICE_VT3204:
            case VIA_DEVICE_VT3314:
            case VIA_DEVICE_VT3336:
            case VIA_DEVICE_VT3327:
            case VIA_DEVICE_VT3364:

                if (dwFlags & DDOVER_ON_SND)
                {
                    pVidData->VDevInfo.dwVDevFlag |= CTL_UseIGA2Path;
                    switch (lpUpdate->dwFourcc)
                    {
                        case FOURCC_NV12:
                        case FOURCC_YV12:
                        case FOURCC_YUY2:
                            pVidData->VidStatus.dwVideoFlagSW[handle] |= SW_ON_SND;
                            DBG_DD(("DuoView: SW_ON_SND\n"));
                            break;
                         default:
                            break;
                    }

                }
                else
                {
                    pVidData->VDevInfo.dwVDevFlag &= ~CTL_UseIGA2Path;
                    switch (lpUpdate->dwFourcc)
                    {
                        case FOURCC_NV12:
                        case FOURCC_YV12:
                        case FOURCC_YUY2:
                            pVidData->VidStatus.dwVideoFlagSW[handle] &= ~SW_ON_SND;
                            DBG_DD(("DuoView: SW_ON_PRI\n"));
                            break;
                        default:
                            break;
                    }
                }
                pVidData->VidStatus.dwVideoFlagSW[handle] &= ~VIDEO_1_ON_SND;
                pVidData->VidStatus.dwVideoFlagMPEG[handle] &= ~VIDEO_1_ON_SND;
                pVidData->VidStatus.dwVideoFlagTV0 &= ~VIDEO_1_ON_SND;
                pVidData->VidStatus.dwVideoFlagTV1 &= ~VIDEO_1_ON_SND;

                break;
            case VIA_DEVICE_CLE1:
            case VIA_DEVICE_CLE2:
            case VIA_DEVICE_VT3259:
            case VIA_DEVICE_VT3324:
            case VIA_DEVICE_VT3293:
            case VIA_DEVICE_VT3353:

                dwFlags &= ~DDOVER_ON_SND;
                pVidData->VDevInfo.dwVDevFlag |= CTL_UseIGA2Path;
                switch (lpUpdate->dwFourcc)
                {
                    case FOURCC_YV12:
                    case FOURCC_NV12:
                    case FOURCC_YUY2:
                        pVidData->VidStatus.dwVideoFlagSW[handle] |= VIDEO_1_ON_SND;
                        break;
                    default:
                        break;
                }
                pVidData->VidStatus.dwVideoFlagSW[handle] &= ~SW_ON_SND;
                pVidData->VidStatus.dwVideoFlagMPEG[handle] &= ~MPEG_ON_SND;
                pVidData->VidStatus.dwVideoFlagTV0 &= ~CAP0_ON_SND;
                pVidData->VidStatus.dwVideoFlagTV1 &= ~CAP1_ON_SND;
                break;
            default:
                break;

        }
    }

    DBG_DD(("UpdateOverlay: dwFlags = %08lx\n", dwFlags));

    if ((pVidData->VidStatus.dwVideoFlagMPEG[handle] & MPEG_ON_SND) ||
        (pVidData->VidStatus.dwVideoFlagSW[handle] & SW_ON_SND) ||
        (pVidData->VidStatus.dwVideoFlagTV0 & CAP0_ON_SND) ||
        (pVidData->VidStatus.dwVideoFlagTV1 & CAP1_ON_SND) ||
        ((lpVIAGraphicInfo->duoview & DUOVIEW_ENABLED)
         && (pVidData->VDevInfo.dwVDevFlag & CTL_UseIGA2Path)))
    {
        switch (lpVIAGraphicInfo->dwDeviceID)
        {
            case VIA_DEVICE_CLE1:
            case VIA_DEVICE_CLE2:
            case VIA_DEVICE_VT3259:
            case VIA_DEVICE_VT3324:
            case VIA_DEVICE_VT3293:
            case VIA_DEVICE_VT3353:
                IGA_Video_Map[IGA1] = VIDEO_3;  /* V3 fixed to IGA1 */
                IGA_Video_Map[Screen_IGA_Map[SCREEN_INDEX_1]] = VIDEO_1;
                break;
            case VIA_DEVICE_VT3205:
            case VIA_DEVICE_VT3204:
            case VIA_DEVICE_VT3314:
                /* Screen0 have not Video */
                IGA_Video_Map[Screen_IGA_Map[SCREEN_INDEX_0]] = 0;
                /* V3 on Screen1 */
                IGA_Video_Map[Screen_IGA_Map[SCREEN_INDEX_1]] = VIDEO_3;
                break;
            case VIA_DEVICE_VT3327:
            case VIA_DEVICE_VT3336:
            case VIA_DEVICE_VT3364:
                IGA_Video_Map[Screen_IGA_Map[SCREEN_INDEX_0]] = 0;
                IGA_Video_Map[Screen_IGA_Map[SCREEN_INDEX_1]] = VIDEO_1;
                break;
            default:
                IGA_Video_Map[IGA1] = VIDEO_3;
                IGA_Video_Map[IGA2] = VIDEO_1;
                break;
        }
    }
    else
    {
        switch (lpVIAGraphicInfo->dwDeviceID)
        {
            case VIA_DEVICE_CLE1:
            case VIA_DEVICE_CLE2:
            case VIA_DEVICE_VT3259:
            case VIA_DEVICE_VT3324:
            case VIA_DEVICE_VT3293:
            case VIA_DEVICE_VT3353:
                IGA_Video_Map[IGA1] = VIDEO_1 | VIDEO_3;
                IGA_Video_Map[IGA2] = 0;
                break;
            case VIA_DEVICE_VT3205:
            case VIA_DEVICE_VT3204:
            case VIA_DEVICE_VT3314:
                IGA_Video_Map[IGA1] = VIDEO_3;
                IGA_Video_Map[IGA2] = 0;
                break;
            case VIA_DEVICE_VT3327:
            case VIA_DEVICE_VT3336:
            case VIA_DEVICE_VT3364:
                IGA_Video_Map[IGA1] = VIDEO_1;
                IGA_Video_Map[IGA2] = 0;
                break;
            default:
                IGA_Video_Map[IGA1] = VIDEO_1 | VIDEO_3;
                IGA_Video_Map[IGA2] = 0;
                break;
        }
    }

    if (lpVIAGraphicInfo->duoview & DUOVIEW_ENABLED)
    {
        UpdateBakup = *lpUpdate;
        pVidData->panning_y[1] = pVidData->panning_y[0];
        pVidData->panning_x[1] = pVidData->panning_x[0];
        pVidData->panning_old_y[1] = pVidData->panning_old_y[0];
        pVidData->panning_old_x[1] = pVidData->panning_old_x[0];
        switch (lpVIAGraphicInfo->IGAPanningStatus)
        {
            /* HQV+V1 should be updated first,otherwise HQV's dst buffer is not set
             * there will be garbage on the screen */
            case 0x01:
                /* IGA1 panning, IGA2 not panning */
                if (IGA_Video_Map[IGA1] == VIDEO_1)
                {
                    /* IGA1 has V1 and panning */
                    lpVIAGraphicInfo->OverlayToUpdate = IGA_Video_Map[IGA1];
                    dwRet = UpdateOverlay_Internal(pVidData, lpUpdate, ScrnIndex = 0);
                    if (IGA_Video_Map[IGA2])
                    {
                        lpVIAGraphicInfo->OverlayToUpdate = IGA_Video_Map[IGA2];
                        dwRet = UpdateOverlay_Internal(pVidData, &UpdateBakup, ScrnIndex = 1);
                    }
                }
                else if (IGA_Video_Map[IGA2] == VIDEO_1)
                {
                    /* IGA2 has V1 */
                    lpVIAGraphicInfo->OverlayToUpdate = IGA_Video_Map[IGA2];
                    dwRet = UpdateOverlay_Internal(pVidData, &UpdateBakup, ScrnIndex = 1);
                    if (IGA_Video_Map[IGA1])
                    {
                        lpVIAGraphicInfo->OverlayToUpdate = IGA_Video_Map[IGA1];
                        dwRet = UpdateOverlay_Internal(pVidData, lpUpdate, ScrnIndex = 0);
                    }
                }
                else
                {
                    /* vt3314 only has V3, keep its update behavior as formerly */
                    if (IGA_Video_Map[IGA1])
                    {
                        lpVIAGraphicInfo->OverlayToUpdate = IGA_Video_Map[IGA1];
                        dwRet = UpdateOverlay_Internal(pVidData, lpUpdate, ScrnIndex = 0);
                    }
                    if (IGA_Video_Map[IGA2])
                    {
                        lpVIAGraphicInfo->OverlayToUpdate = IGA_Video_Map[IGA2];
                        dwRet = UpdateOverlay_Internal(pVidData, &UpdateBakup, ScrnIndex = 1);
                    }
                }
                break;

            case 0x2:
                /* IGA1 not panning, IGA2 panning */
                if (IGA_Video_Map[IGA1] == VIDEO_1)
                {
                    /* IGA1 has V1,IGA2 panning */
                    lpVIAGraphicInfo->OverlayToUpdate = IGA_Video_Map[IGA1];
                    dwRet = UpdateOverlay_Internal(pVidData, &UpdateBakup, ScrnIndex = 0);
                    if (IGA_Video_Map[IGA2])
                    {
                        lpVIAGraphicInfo->OverlayToUpdate = IGA_Video_Map[IGA2];
                        dwRet = UpdateOverlay_Internal(pVidData, lpUpdate, ScrnIndex = 1);
                    }
                }
                else if (IGA_Video_Map[IGA2] == VIDEO_1)
                {
                    /* IGA2 has V1 and panning */
                    lpVIAGraphicInfo->OverlayToUpdate = IGA_Video_Map[IGA2];
                    dwRet = UpdateOverlay_Internal(pVidData, lpUpdate, ScrnIndex = 1);
                    if (IGA_Video_Map[IGA1])
                    {
                        lpVIAGraphicInfo->OverlayToUpdate = IGA_Video_Map[IGA1];
                        dwRet = UpdateOverlay_Internal(pVidData, &UpdateBakup, ScrnIndex = 0);
                    }
                }
                else
                {
                    if (IGA_Video_Map[IGA1])
                    {
                        lpVIAGraphicInfo->OverlayToUpdate = IGA_Video_Map[IGA1];
                        dwRet = UpdateOverlay_Internal(pVidData, lpUpdate, ScrnIndex = 0);
                    }
                    if (IGA_Video_Map[IGA2])
                    {
                        lpVIAGraphicInfo->OverlayToUpdate = IGA_Video_Map[IGA2];
                        dwRet = UpdateOverlay_Internal(pVidData, &UpdateBakup, ScrnIndex = 1);
                    }
                }
                break;

            default:
                if (IGA_Video_Map[IGA1] == VIDEO_1)
                {
                    /* IGA1 has V1 */
                    lpVIAGraphicInfo->OverlayToUpdate = IGA_Video_Map[IGA1];
                    dwRet = UpdateOverlay_Internal(pVidData, lpUpdate, ScrnIndex);
                    lpVIAGraphicInfo->OverlayToUpdate = IGA_Video_Map[IGA2];
                    dwRet = UpdateOverlay_Internal(pVidData, &UpdateBakup, ScrnIndex);
                }
                else if (IGA_Video_Map[IGA2] == VIDEO_1)
                {
                    /* IGA2 has V1 */
                    pVidData->NeedToCorrectVideoSize = FALSE;
                    pVidData->isVideoWidthDownScaling = FALSE;
                    pVidData->isVideoHeightDownScaling = FALSE;

                    lpVIAGraphicInfo->OverlayToUpdate = IGA_Video_Map[IGA2];
                    dwRet = UpdateOverlay_Internal(pVidData, lpUpdate, ScrnIndex = 1);

                    if (pVidData->isVideoWidthDownScaling || pVidData->isVideoHeightDownScaling)
                    {
                        pVidData->DownScalingDestRect.left = lpUpdate->rDest.left;
                        pVidData->DownScalingDestRect.right = lpUpdate->rDest.right;
                        pVidData->DownScalingDestRect.top = lpUpdate->rDest.top;
                        pVidData->DownScalingDestRect.bottom = lpUpdate->rDest.bottom;
                        pVidData->NeedToCorrectVideoSize = TRUE;
                    }

                    lpVIAGraphicInfo->OverlayToUpdate = IGA_Video_Map[IGA1];
                    dwRet = UpdateOverlay_Internal(pVidData, &UpdateBakup, ScrnIndex = 0);
                }
                else
                {
                    /* vt3314 only has V3, keep its update behavior as formerly */
                    if (IGA_Video_Map[IGA1])
                    {
                        lpVIAGraphicInfo->OverlayToUpdate = IGA_Video_Map[IGA1];
                        dwRet = UpdateOverlay_Internal(pVidData, lpUpdate, ScrnIndex = 0);
                    }
                    if (IGA_Video_Map[IGA2])
                    {
                        lpVIAGraphicInfo->OverlayToUpdate = IGA_Video_Map[IGA2];
                        dwRet = UpdateOverlay_Internal(pVidData, &UpdateBakup, ScrnIndex = 1);
                    }
                }
                break;

        }
        lpVIAGraphicInfo->OverlayToUpdate = VIDEO_1 | VIDEO_3;
    }
    else
    {
        lpVIAGraphicInfo->OverlayToUpdate = VIDEO_1 | VIDEO_3;
        dwRet = UpdateOverlay_Internal(pVidData, lpUpdate, ScrnIndex);
    }
    return dwRet;
}
