 /*
 * Copyright 1998-2008 VIA Technologies, Inc. All Rights Reserved.
 * Copyright 2001-2008 S3 Graphics, Inc. All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHOR(S) OR COPYRIGHT HOLDER(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */
 
           
/* I N C L U D E S ------------------------------------------------- */

#include "drm.h"
#include "mm.h"
#ifdef XFree86Server
#include "xf86.h"
#include "xf86_ansic.h" /* for Bool */
#include "compiler.h"   /* for inb/outb */
#include "xf86drm.h"    /* for drmOpen, etc. */
#ifdef HAVE_CONFIG_H            /* for define XF86DRI on X11R7(FC5)*/
#include "config_via.h"
#endif
#else
#include "via_types.h"  /* for Bool */
#endif

#include "via.h"
#include "ddmpeg.h"
#include "capture.h"
#include "VidInit.h"
#include "Vinline.h"    /* Universal wait function */

#include "VidStruc.h"   /* VIDDATA, video top level structure */
#include "debug.h"

#include "ovlfCME.h"
#include "vregCME.h"
#include "vdefCME.h"

#include <math.h>   /* for sin & cos function */

#include "overlay.h"

/* DEFINE */
#define PI 3.1415926535897932


/* EXTERN   FUNCTIONS  */

extern void vfHM(PVIDDATA pVidData,unsigned long dwFunSel);
extern void vfCM(unsigned long dwIndex,unsigned long dwData,PVIDDATA pVidData);
extern void vfCommandManager(PVIDDATA pVidData);
extern void vfCMHQVW(unsigned long dwIndex,unsigned long dwData,PVIDDATA pVidData);
extern unsigned long vfCMHQVR(unsigned long dwIndex,PVIDDATA pVidData);
extern int viaFreeVideoSurface(PVIDDATA pVidData,int wSurfaceID);
extern unsigned long viaAllocVideoSurface(PVIDDATA pVidData,\
            int wSurfaceID, unsigned long dwBufSize);
extern VIDEOMEMORY     VidMemory;          /* video memory struct */
#ifdef XF86DRI
extern int  drm_fd;
#endif


/* FUNCTION*/
/*
 * In :lpVIAGraphicInfo: graphic info structure pointer
 *     pjVideo :the maped register address,used for VIDoutD macro
 * Out:
 * Ret
 * Description:Setting video engine color space register.
 *             If graphicinfo's color space value is not changed,we have
 *             default values for them.Otherwise new values stored in the
 *             pointer are used.
 */ 
static void SetColorSpace(LPVIAGRAPHICINFO lpVIAGraphicInfo,unsigned char *pjVideo)
{
    if (!lpVIAGraphicInfo->ColorSpaceChanged)    
    {
        VIDOutD(ColorSpaceReg_1, ColorSpaceValue_1_3123C0);
        VIDOutD(ColorSpaceReg_2, ColorSpaceValue_2_3123C0);
        VIDOutD(V3_ColorSpaceReg_1, ColorSpaceValue_1_3123C0);
        VIDOutD(V3_ColorSpaceReg_2, ColorSpaceValue_2_3123C0);
    }
    else
    {
        VIDOutD(ColorSpaceReg_1, lpVIAGraphicInfo->ColorSpaceValue1);
        VIDOutD(ColorSpaceReg_2, lpVIAGraphicInfo->ColorSpaceValue2);
        VIDOutD(V3_ColorSpaceReg_1, lpVIAGraphicInfo->ColorSpaceValue1);
        VIDOutD(V3_ColorSpaceReg_2, lpVIAGraphicInfo->ColorSpaceValue2);                
   } 
}

/*
 * In :lpDDSurfaceDesc:surface descriptor pointer
 *     SurfaceType: surface ID
 *     dwSurfaceCount:the number of surface needed for a given surface ID
 * In/Out:dwWidth,dwHeight,dwPitch,dwAlignHeight,dwSingleSize   
 * Ret:the actual surface size
 * Desc:calculate actual surface size needed for a given surface type.
 *      For some surface type, the size is easy to figured out,unneeded to
 *      call this.
 */
static unsigned long CalSurfaceSize(LPDDSURFACEDESC lpDDSurfaceDesc,enum SurfaceID SurfaceType,
        unsigned long dwSurfaceCount, unsigned long *dwWidth,unsigned long *dwHeight,
        unsigned long *dwPitch, unsigned long *dwAlignHeight,unsigned long *dwSingleSize)
{
    *dwWidth=(*dwWidth)?(*dwWidth):lpDDSurfaceDesc->dwWidth;
    *dwHeight=(*dwHeight)?(*dwHeight):lpDDSurfaceDesc->dwHeight;
    unsigned long dwSize=0;         /*the sum size of surfaces */

    switch (SurfaceType)
    {
        case MPEG_REQ_SURFACE_ID:
        {
            *dwPitch  = (*dwPitch)?(*dwPitch):ALIGN_TO_64_BYTE(*dwWidth);/*maybe already has value*/
            *dwAlignHeight = ALIGN_TO_32(*dwHeight);
            *dwSingleSize=((((*dwPitch)*3)/2)*(*dwAlignHeight));    /* 1.5 bytes per pixel */
            DBG_DD(("MPEG   dwWidth: %ld\n", *dwWidth));
            DBG_DD(("MPEG   dwHeight: %ld\n", *dwHeight));
            DBG_DD(("MPEG   dwPitch: %ld\n", *dwPitch));
            DBG_DD(("MPEG   dwAlignHeight: %ld\n", *dwAlignHeight));
            break;
        }
        case HQV_REQ_SURFACE_ID:
        {    
        	/*HQV size differs according to different FOURCCCC*/
            switch (lpDDSurfaceDesc->dwFourCC)
            {
                case FOURCC_YUY2:    	
                case FOURCC_YV12:		        	
                case FOURCC_NV12:	       	
                    if ((*dwPitch)&&(*dwAlignHeight))
                        *dwSingleSize= (*dwPitch)*(*dwAlignHeight) * 2;
                    break;	            		            	
                default:
                	DBG_DD(("Invalid Fourcc"));
                    break;
            }
            DBG_DD(("HQV   dwWidth: %ld\n", *dwWidth));
            DBG_DD(("HQV   dwHeight: %ld\n", *dwHeight));
            DBG_DD(("HQV   dwPitch: %ld\n", *dwPitch));
            DBG_DD(("HQV   dwAlignHeight: %ld\n", *dwAlignHeight));
            break;

        }
        case SUBP_REQ_SURFACE_ID:
        {
            *dwPitch  = (*dwPitch)?(*dwPitch):ALIGN_TO_32_BYTE(*dwWidth);
            *dwAlignHeight=*dwPitch;
            *dwSingleSize=(*dwPitch)*(*dwHeight);     /* 4 bits per pixel,AI44 mode */
            DBG_DD(("SUBP   dwWidth: %ld\n", *dwWidth));
            DBG_DD(("SUBP   dwHeight: %ld\n", *dwHeight));
            DBG_DD(("SUBP   dwPitch: %ld\n", *dwPitch));
            DBG_DD(("SUBP   dwAlignHeight: %ld\n", *dwAlignHeight));
            break;
        }
        case SW_REQ_SURFACE_ID:
        {
            /*sw surface differs according to different FOURCC*/
            switch (lpDDSurfaceDesc->dwFourCC)
            {
                case FOURCC_YUY2:    
                {	   
                    *dwPitch  =(*dwPitch)?(*dwPitch): ALIGN_TO_32_BYTE(*dwWidth)*2;
                    *dwAlignHeight = ALIGN_TO_16(*dwHeight);
                    *dwSingleSize =  (*dwPitch)*(*dwAlignHeight);    /* YUYV */
                    break;
                }    

                case FOURCC_YV12:
                case FOURCC_NV12:			
                {
                    *dwPitch  =(*dwPitch)?(*dwPitch): ALIGN_TO_32_BYTE(*dwWidth);
                    *dwAlignHeight = *dwHeight;
                    *dwSingleSize = ((((*dwPitch)*3)/2)*(*dwAlignHeight));    /* 1.5 bytes per pixel */
                    break;
                }		

                default:
                    DBG_DD(("Invalid Fourcc"));
                    break;
	
            }
            DBG_DD(("SW   dwWidth: %ld\n", *dwWidth));
            DBG_DD(("SW   dwHeight: %ld\n", *dwHeight));
            DBG_DD(("SW   dwPitch: %ld\n", *dwPitch));
            DBG_DD(("SW   dwAlignHeight: %ld\n", *dwAlignHeight));
            break;
        }
        case ALPHA_REQ_SURFACE_ID:
        {
            *dwPitch=(*dwPitch)?(*dwPitch): ALIGN_TO_16_BYTE((*dwWidth)/2);
            *dwSingleSize=(*dwPitch)*(*dwHeight); 
            DBG_DD(("ALPHA   dwWidth: %ld\n", *dwWidth));
            DBG_DD(("ALPHA   dwHeight: %ld\n", *dwHeight));
            DBG_DD(("ALPHA   dwPitch: %ld\n", *dwPitch));
            break;
        }
        default:
            DBG_DD(("Invalid Surface Type"));
            break;
    }
    
    if (SurfaceType==MPEG_REQ_SURFACE_ID)
        dwSize=(*dwSingleSize)*(dwSurfaceCount+1)+((*dwPitch)*4);
    else
        dwSize = (*dwSingleSize)* dwSurfaceCount;
    DBG_DD(("SingleSurfaceSize:%ld\n",*dwSingleSize)); 
    DBG_DD(("Total Size Request:%ld\n",dwSize));
    return dwSize;
}

unsigned long Flip_CME(PVIDDATA pVidData, unsigned long DisplayBufferIndex)
{
    unsigned long handle=pVidData->VMI_Info.handle;
    LPSWDEVICE  lpSWDevice  = &(pVidData->SWDevice[handle]);
    unsigned char* pjVideo = pVidData->Vhwinfo.pjVideo;

    LPDDPIXELFORMAT lpDPFsrc    = &pVidData->DPFsrc[handle];
    LPVIAGRAPHICINFO lpVIAGraphicInfo = &pVidData->VIAGraphicInfo;    
    unsigned long V3_temp;

    DBG_DD(("handle =%d  pVidData->VidStatus.dwVideoFlagSW[0:1]=0x%lx 0x%lx\n",\
                pVidData->VMI_Info.handle,pVidData->VidStatus.dwVideoFlagSW[0],\
                pVidData->VidStatus.dwVideoFlagSW[1]));
    DBG_DD(("lpSWDevice->lpSWOverlaySurface[0:1]=%lx,%lx\n",\
               (unsigned long)lpSWDevice->lpSWOverlaySurface[0],\
               (unsigned long)lpSWDevice->lpSWOverlaySurface[1]));
    DBG_DD(("lpSWDevice->dwSWPhysicalAddr[0:1]=%lx %lx\n",\
               lpSWDevice->dwSWPhysicalAddr[0],lpSWDevice->dwSWPhysicalAddr[1]));
    DBG_DD(("lpSWDevice->dwHQVAddr[0:2]=%lx %lx %lx \n",\
             lpSWDevice->dwHQVAddr[0],lpSWDevice->dwHQVAddr[1],lpSWDevice->dwHQVAddr[2]));
    DBG_DD(("VidCtl.dwVideoFlag=0x%lx\n",pVidData->VidCtl.dwVideoFlag[handle]));
    switch(lpDPFsrc->dwFourCC)
    {
        case FOURCC_UYVY:
        case FOURCC_YUY2:
            vfHM(pVidData,HM_WAIT_HQV_DONE);
            vfCMHQVW(HQV_SRC_STARTADDR_Y, lpSWDevice->dwSWPhysicalAddr[DisplayBufferIndex],pVidData);
#ifdef CS_FLAG_HWROTATION 
        if(lpVIAGraphicInfo->rotate&& (handle == 0))        
        {
            /* Horace add hqv0 irq*/			        
            vfCMHQVW(HQV_CONTROL,(( vfCMHQVR(HQV_CONTROL,pVidData) &~HQV_FLIP_ODD) |HQV_SW_FLIP|HQV_FLIP_STATUS|0x00000080),pVidData);	
        }
	    else
#endif /* CS_FLAG_HWROTATION */		            
            vfCMHQVW(HQV_CONTROL,(( vfCMHQVR(HQV_CONTROL,pVidData) &~HQV_FLIP_ODD) |HQV_SW_FLIP|HQV_FLIP_STATUS),pVidData);

        /* In rotation case, we will update the frame buffer starting address
        * after the video is rotated. So there is no need to set these registers here in rotation case,
        * to do this can help to reduce garbage or flicker (rotation image overlaps unrotation image)
        */
            if(!lpVIAGraphicInfo->rotate && (pVidData->VidStatus.dwVideoFlagSW[handle]&VIDEO_1_ON_SND))
            {
                V3_temp = VIDInD(V3_CONTROL)&~V3_FLIP_HW_CAPTURE1;
                V3_temp = V3_temp|V3_SWAP_SW;
                /* For duoview case, we have different dwVSrcAddr for V1,V3 */
                vfCM(V3_STARTADDR_0, pVidData->Vhwinfo.dwV3SrcAddr[(VIDInD(HQV_CONTROL)>>1)&0x03],pVidData);
                vfCM(V_COMPOSE_MODE, (VIDInD(V_COMPOSE_MODE)|V3_COMMAND_FIRE),pVidData);
            }
            break;

        case FOURCC_NV12:/* for support NV12 overlay mode test-2*/
        case FOURCC_YV12:
        default:
            DBG_DD(("ovlCME.c:Flip_CME:lpDPFsrc->dwFourCC=%p \n",lpDPFsrc->dwFourCC));
            vfHM(pVidData,HM_WAIT_HQV_DONE);
            vfCMHQVW(HQV_SRC_STARTADDR_Y, lpSWDevice->dwSWPhysicalAddr[DisplayBufferIndex], pVidData);
            vfCMHQVW(HQV_SRC_STARTADDR_U, lpSWDevice->dwSWCbPhysicalAddr[DisplayBufferIndex], pVidData);
            vfCMHQVW(HQV_CONTROL,(( vfCMHQVR(HQV_CONTROL,pVidData) &~HQV_FLIP_ODD) |HQV_SW_FLIP|HQV_FLIP_STATUS),pVidData);

            /* In rotation case, we will update the frame buffer starting address
            * after the video is rotated. So there is no need to set these registers here in rotation case,
            * to do this can help to reduce garbage or flicker (rotation image overlaps unrotation image)
            */
            if(!lpVIAGraphicInfo->rotate && (pVidData->VidStatus.dwVideoFlagSW[handle]&VIDEO_1_ON_SND))
            {
                V3_temp = VIDInD(V3_CONTROL)&~V3_FLIP_HW_CAPTURE1;
                V3_temp = V3_temp|V3_SWAP_SW;
                 /* For duoview case, we have different dwVSrcAddr for V1,V3 */
                vfCM(V3_STARTADDR_0, pVidData->Vhwinfo.dwV3SrcAddr[(VIDInD(HQV_CONTROL)>>1)&0x03],pVidData);
                vfCM(V_COMPOSE_MODE, (VIDInD(V_COMPOSE_MODE)|V3_COMMAND_FIRE),pVidData);
            }
            break;
    }

    DBG_DD(("ovlCME.c: Flip_End \n\n"));
    return VMI_OK;
}


void UpdYInterpolationFlag
(
    LPVIAGRAPHICINFO lpVIAGraphicInfo, 
    unsigned long dwSrcWidth, 
    unsigned long dwVideoFlag, 
    unsigned long *pdwMiniCtl, 
    unsigned long *pdwCompose
)
{
    if (dwSrcWidth >= 800)
    {
        if ((lpVIAGraphicInfo->dwDeviceID == VIA_DEVICE_VT3327) ||
            (lpVIAGraphicInfo->dwDeviceID == VIA_DEVICE_VT3336) ||
            (lpVIAGraphicInfo->dwDeviceID == VIA_DEVICE_VT3364))
        {
            /* Disable vertical interpolation because the size of line buffer (limited to 800) is */
            /* too small to do interpolation. */
            *pdwMiniCtl &= ~(V1_Y_INTERPOLY);
        }
        else
        {
            if(lpVIAGraphicInfo->duoview & DUOVIEW_ENABLED)
            {
                *pdwMiniCtl &= ~(V1_Y_INTERPOLY | V3_Y_INTERPOLY);
            }
            else
            {
                *pdwMiniCtl |= (V1_Y_INTERPOLY | V1_YCBCR_INTERPOLY);

                if (dwVideoFlag & VIDEO_1_INUSE)
                {
                    *pdwCompose |= VAL_V1HDTV_CME;
                }
                else
                {
                    *pdwCompose |= VAL_V3HDTV_CME;
                }
            }
        }
    }
    else
    {
        *pdwMiniCtl |= (V1_Y_INTERPOLY | V1_YCBCR_INTERPOLY);
    }
}


/****************************************************************************
 *
 * Upd_Video_CME()
 *
 ***************************************************************************/
unsigned long Upd_Video_CME(PVIDDATA pVidData)
{
    unsigned long handle=pVidData->VMI_Info.handle;
    LPOVERLAYRECORD lpOverlayRecord = &pVidData->OverlayRecordSet.OverlayRecordSW[handle];
    LPVIDHWDIFFERENCE lpVideoHWDifference = &pVidData->VideoHWDifference;

    unsigned char* pjVideo              = pVidData->Vhwinfo.pjVideo;
    unsigned long dwCtlHQVflag          = pVidData->VidCtl.dwCtlHQVflag;
    unsigned long dwVideoFlag           = pVidData->VidCtl.dwVideoFlag[handle] = pVidData->VidStatus.dwVideoFlagSW[handle];
    unsigned long dwVDevFlag            = pVidData->VDevInfo.dwVDevFlag;
    LPVIAGRAPHICINFO lpVIAGraphicInfo = &pVidData->VIAGraphicInfo;

    unsigned long dwStartAddr           = 0;
    unsigned long dwSrcPitch            = lpOverlayRecord->dwOriSrcPitch;
    unsigned long dwOriSrcWidth         = lpOverlayRecord->dwOriSrcWidth;
    unsigned long dwOriSrcHeight        = lpOverlayRecord->dwOriSrcHeight;
    unsigned long dwDeinterlaceMode     = lpOverlayRecord->dwDeinterlaceMode;
    unsigned long dwColorKeyOn          = lpOverlayRecord->dwColorKeyOn;
    unsigned long dwChromaKeyOn         = lpOverlayRecord->dwChromaKeyOn;
    unsigned long dwKeyLow              = lpOverlayRecord->dwKeyLow;
    unsigned long dwChromaLow           = lpOverlayRecord->dwChromaLow;
    unsigned long dwChromaHigh          = lpOverlayRecord->dwChromaHigh;
    unsigned long dwDstWidth            = lpOverlayRecord->dwOverlayDestWidth;
    unsigned long dwDstHeight           = lpOverlayRecord->dwOverlayDestHeight;
    OV_RECTL rSrc                  = lpOverlayRecord->dwOverlaySrcRect;
    OV_RECTL rDest                 = lpOverlayRecord->dwOverlayDestRect;    

    LPDDPIXELFORMAT lpDPFsrc    = &pVidData->DPFsrc[handle];
    LPVIDEOHWINFO lphwinfo      = &pVidData->Vhwinfo;
    unsigned long* lpdwCtlCMFlag       = &pVidData->VidCtl.dwCtlCMFlag;

    unsigned long dwVidCtl      =0;
    unsigned long dwCompose     =0;
    unsigned long dwSrcWidth    =0;
    unsigned long dwSrcHeight   =0;
    unsigned long dwZoomCtl     =0;
    unsigned long dwMiniCtl     =0;
    unsigned long dwHQVCtl      =0;
    unsigned long dwHQVminiCtl  =0;
    /* 3293 will not use dwHQVminiCtl but dwHQVScaleCtlH, dwHQVScaleCtlV*/
    unsigned long dwHQVScaleCtlH=0;  
    unsigned long dwHQVScaleCtlV=0;  
    
    unsigned long dwHQVzoomflagH=0;
    unsigned long dwHQVzoomflagV=0;
    unsigned long dwHQVsrcFetch =0;
    unsigned long dwOffset      =0;
    unsigned long dwFetch       =0;
    unsigned long dwTmp         =0;
    unsigned long dwDisplayCountW=0;
    unsigned long dwTCGDRegVal=0;       
    unsigned long dwTmpHQVSrcFetch = 0;

    VIDSETTILEANDCSC vidSetTileAndCSC;
    VIDGETFORMAT vidGetFormat;
    VIDGETHQVSRCOFFSET vidGetHQVOffset;
    VIDGETSRCADDR vidGetSrcAddr;
    VIDGETFETCH vidGetFetch;
    VIDCALCZOOMW vidCalcZoomW;
    VIDGETDISPLAYCOUNT vidGetDisplayCount;
    VIDCALCZOOMH vidCalcZoomH;
    
    DBG_DD(("Upd_Video_CME:\n"));
    DBG_DD(("videoflag=%p\n",dwVideoFlag));

    lphwinfo->dwzoomCtlV1DuoView = 0;
    if (dwVideoFlag&VIDEO_ON_TOP)  
   {
        /* only update two video's relative postion
        which on top of another*/
        DBG_DD(("vesper:adjust on_top\n"));
        dwCompose=dMMRD(pjVideo,V_COMPOSE_MODE);
        dwCompose &= ~COMPOSE_V3_V1;
        if (dwVideoFlag & VIDEO_1_INUSE)
        {    
            dwCompose|=COMPOSE_V1_V3; /*redundant but keep the logic clear*/
            vfCM(V_COMPOSE_MODE,dwCompose|V1_COMMAND_FIRE,pVidData);
        }   
        else 
        {
            dwCompose|=COMPOSE_V3_V1;
            vfCM(V_COMPOSE_MODE,dwCompose|V3_COMMAND_FIRE,pVidData);
        }
        return VMI_OK;
    }
    if (dwVideoFlag & VIDEO_SHOW)    
    {        
        if ( dwVideoFlag & SW_ON_SND )
        {
            dwVDevFlag |= CTL_UseIGA2Path;
        }    
        /*add code here to set LCD expansion flag*/
        if((lpOverlayRecord->dwExpandDestRect.bottom !=0) &&
            (lpOverlayRecord->dwExpandDestRect.bottom != rDest.bottom))/*expansion*/
        {
            dwVDevFlag |= CTL_UseExpand;
        }           
       
        switch ( lpVIAGraphicInfo->dwDeviceID )
        {
        case VIA_DEVICE_VT3336:
        case VIA_DEVICE_VT3327:
        case VIA_DEVICE_VT3364:
            DBG_DD(("dwVDevFlag=0x%lx\n",dwVDevFlag));
            if(lpVIAGraphicInfo->duoview&DUOVIEW_ENABLED)/*duo view*/
            {
                if(!(lpVIAGraphicInfo->duoview&DUOVIEW_V3_ON_IGA2))
                {
                    pVidData->VidStatus.dwVideoFlagSW[handle] &= ~SW_ON_SND;
                    dwVideoFlag = pVidData->VidStatus.dwVideoFlagSW[handle];
                    dwVDevFlag &= ~CTL_UseIGA2Path;
                }
                else
                {
                    DBG_DD(("V3_ON_IGA2:dwVDevFlag=0x%lx\n",dwVDevFlag));
                    dwVDevFlag |= CTL_UseIGA2Path;
                    /*add code here to set LCD expansion flag*/
                    if((lpOverlayRecord->dwExpandDestRect.bottom !=0) &&
                       (lpOverlayRecord->dwExpandDestRect.bottom != rDest.bottom))/*expansion*/
                    {
                        dwVDevFlag |= CTL_UseExpand;
                    }           
                }
            }
            break;
        default:
            break;
        }

        if (dwVideoFlag & VIDEO_HQV_INUSE)
        {
            /* vfSetHQVTileModeAndCSCcontrol_CME */
            dwTCGDRegVal = vfCMHQVR(REG_HQV_TILE_CSC_CTL_CME,pVidData);
            lphwinfo->dwHQVDstAddr[0]   = lpOverlayRecord->dwHQVAddr[0];
            lphwinfo->dwHQVDstAddr[1]   = lpOverlayRecord->dwHQVAddr[1];
            lphwinfo->dwHQVDstAddr[2]   = lpOverlayRecord->dwHQVAddr[2];
            
            /* Tile mode can be used only with HW mpeg decoding */
            dwCtlHQVflag &= ~CTL_HQV_TILEMODE_ENABLE_CME;
            vidSetTileAndCSC.dwCtlHQVflag      = dwCtlHQVflag;
            vidSetTileAndCSC.dwTCGDRegVal      = dwTCGDRegVal;
            vidSetTileAndCSC.dwHQVDstAddr0     = lphwinfo->dwHQVDstAddr[0];
            vidSetTileAndCSC.dwHQVDstAddr1     = lphwinfo->dwHQVDstAddr[1];
            vidSetTileAndCSC.dwOriSrcWidth     = dwOriSrcWidth;
            vidSetTileAndCSC.dwOriSrcHeight    = dwOriSrcHeight;
            vidSetTileAndCSC.dwUseCSC          = (lpDPFsrc->dwFlags & DDPF_FOURCC)?1:0;
            vidSetTileAndCSC.dwFourCC          = lpDPFsrc->dwFourCC;
            vidSetTileAndCSC.lpdwHQVTileCSCVal = &lphwinfo->dwHQVTCGDCtl;
     
            vfSetHQVTileModeAndCSCcontrol_CME(&vidSetTileAndCSC);

            lphwinfo->dwHQVDeinterlaceCtl = 0;
        }

        dwSrcWidth = (unsigned long)rSrc.right - rSrc.left;
        dwSrcHeight = (unsigned long)rSrc.bottom - rSrc.top;

        if(lpVideoHWDifference->dwNeedV1Prefetch)
        {
            dwVidCtl = (VAL_VIDEO_ENABLE_CME|VAL_VIDEO_EXPIRE_NUM_16_CME);
        }
        else
        {
            dwVidCtl = (VAL_VIDEO_ENABLE_CME|VAL_VIDEO_EXPIRE_NUM_CME);
        }
        /* vfGetVideoFormat_CME */
        vidGetFormat.dwVideoFlag    = dwVideoFlag;
        vidGetFormat.dwCtlHQVflag   = dwCtlHQVflag;
        vidGetFormat.lpDPFsrc       = lpDPFsrc;
        vidGetFormat.dwDisplayPictStruct    = 0; /* For MPEG only */
        vidGetFormat.dwDeinterlaceMode      = 0; /* For MPEG only */
        vidGetFormat.dwMPEGProgressiveMode  = 0; /* For MPEG only */
        vidGetFormat.dwMPEGInTVFullScreen   = lpOverlayRecord->dwMPEGInTVFullScreen;/* For MPEG only */
        vidGetFormat.dwVDevFlag = dwVDevFlag;
        vidGetFormat.lpdwVidCtl = &dwVidCtl;
        vidGetFormat.lpdwHQVCtl = &dwHQVCtl;
        
        vfGetVideoFormat_CME(&vidGetFormat);
        
        /* Starting address of source and Source offset */
        if (dwCtlHQVflag&CTL_HQV_SCALING_UP_ENABLE_CME)
        {
            /* vfGetHQVSourceOffset_CME */
            vidGetHQVOffset.rSrc    = rSrc;

            if (pVidData->NeedToCorrectVideoSize)
            {
                vidGetHQVOffset.rDest = pVidData->DownScalingDestRect;
            }
            else
            {
                vidGetHQVOffset.rDest = rDest;
            }
            
            vidGetHQVOffset.lpDPFsrc= lpDPFsrc;
            vidGetHQVOffset.lpdwHQVOffset = &lphwinfo->dwHQVOffset;

            vfGetHQVSourceOffset_CME(&vidGetHQVOffset);        
        }
        else
        {
            /* vfGetSrcStartAddress_CME */
            vidGetSrcAddr.dwVideoFlag   = dwVideoFlag;
            vidGetSrcAddr.dwCtlHQVflag  = dwCtlHQVflag;
            vidGetSrcAddr.lpDPFsrc      = lpDPFsrc;
            vidGetSrcAddr.dwSrcPitch    = dwSrcPitch;
            vidGetSrcAddr.rSrc          = rSrc;

            if (pVidData->NeedToCorrectVideoSize)
            {
                vidGetSrcAddr.rDest = pVidData->DownScalingDestRect;
            }
            else
            {
                vidGetSrcAddr.rDest = rDest;
            }
            vidGetSrcAddr.pVDevInfo     = &pVidData->VDevInfo;
            vidGetSrcAddr.lpdwOffset    = &dwOffset;            
            
            vfGetSrcStartAddress_CME(&vidGetSrcAddr);
            lphwinfo->dwHQVOffset = 0;
        }

        lpOverlayRecord->dwOffset = dwOffset;
        lphwinfo->dwVOffset       = dwOffset;

        if (dwVideoFlag & VIDEO_HQV_INUSE)    
        {
            dwStartAddr = vfCMHQVR(HQV_SRC_STARTADDR_U,pVidData) & 0xFFFFFFF0;
            lphwinfo->dwHQVSrcAddr[1]   = dwStartAddr;
            dwStartAddr = vfCMHQVR(HQV_SRC_STARTADDR_Y,pVidData) & 0xFFFFFFF0;
            lphwinfo->dwHQVSrcAddr[0]   = dwStartAddr;
            lphwinfo->dwVSrcAddr[0] = lpOverlayRecord->dwHQVAddr[0]+dwOffset;
            lphwinfo->dwVSrcAddr[1] = lpOverlayRecord->dwHQVAddr[1]+dwOffset;
            lphwinfo->dwVSrcAddr[2] = lpOverlayRecord->dwHQVAddr[2]+dwOffset;
        }
        else
        {
            lphwinfo->dwVSrcAddr[0]     = dwStartAddr+dwOffset;
        }

		if(lpVIAGraphicInfo->dwDeviceID == VIA_DEVICE_VT3353)
		{
        	/* Get HQV Source Data Offset: */
        	vidGetHQVOffset.rSrc = rSrc;
        	vidGetHQVOffset.lpdwHQVSrcDataOffsetStart = &lphwinfo->dwHQVSrcDataOffsetStart;
        	vidGetHQVOffset.lpdwHQVSrcDataOffsetEnd = &lphwinfo->dwHQVSrcDataOffsetEnd;

        	vfGetHQVSrcDataOffset_CME(&vidGetHQVOffset);
		}


        /* vfGetFetch_CME */
        vidGetFetch.dwVideoFlag     = dwVideoFlag;
        vidGetFetch.dwCtlHQVflag    = dwCtlHQVflag;
        vidGetFetch.lpDPFsrc        = lpDPFsrc;
        vidGetFetch.dwSrcWidth      = dwSrcWidth;
        vidGetFetch.dwDstWidth      = dwDstWidth;
        vidGetFetch.dwOriSrcWidth   = dwOriSrcWidth;
        vidGetFetch.lpdwFetch       = &dwFetch;
        vidGetFetch.lpdwHQVsrcFetch = &dwHQVsrcFetch;

        vfGetFetch_CME(&vidGetFetch);

        if (dwVideoFlag & VIDEO_HQV_INUSE)
        {
            if ( !(dwDeinterlaceMode & DDOVER_INTERLEAVED) && 
                 (dwDeinterlaceMode & DDOVER_BOB ) )
            {                                   
                lphwinfo->dwHQVSrcCountWH = 
                ((dwHQVsrcFetch-1)<<16)|((dwOriSrcHeight<<1)-1);
            }
            else
            {                                
                lphwinfo->dwHQVSrcCountWH = 
                ((dwHQVsrcFetch-1)<<16)|(dwOriSrcHeight-1);
            }


            if (dwCtlHQVflag&CTL_HQV_SCALING_UP_ENABLE_CME)
            {
                lphwinfo->dwVPitch =          ALIGN_TO_64_BYTE(pVidData->VDevInfo.dwcxScreen*4);
                lphwinfo->dwHQVDstPitch = ALIGN_TO_64_BYTE(pVidData->VDevInfo.dwcxScreen*4);
                lphwinfo->dwHQVSrcPitch = dwSrcPitch;
            }
            else
            {
                lphwinfo->dwHQVSrcPitch = dwSrcPitch;
                if ((dwCtlHQVflag&CTL_HQV_CSC_ENABLE_CME)&&
                    (dwCtlHQVflag&(CTL_HQV_CSC_RGB32_CME|CTL_HQV_CSC_RGB101010_CME2))&&
                    (lpDPFsrc->dwFlags & DDPF_FOURCC))
                {
                    lphwinfo->dwVPitch = dwSrcPitch<<1;
                    lphwinfo->dwHQVDstPitch = dwSrcPitch<<1;                
                }
                else
                { 
                    
                    if (lpDPFsrc->dwFourCC==FOURCC_NV12) 
                    {     
                        lphwinfo->dwVPitch = dwSrcPitch<<1;
                        lphwinfo->dwHQVDstPitch = dwSrcPitch<<1;
                    }
                    else 
                    {     
                        lphwinfo->dwVPitch = dwSrcPitch;
                        lphwinfo->dwHQVDstPitch = dwSrcPitch;
                    }
                }
            }
            lphwinfo->dwHQVSrcPitch |= VAL_HQV_FLIP_FIFODEPTH_1_BACK_DROP_CME;
        }
        else
        {
            lphwinfo->dwVPitch = dwSrcPitch;
        }

        DBG_DD(("rSrc  X (%ld,%ld) Y (%ld,%ld)\n",
                rSrc.left, rSrc.right,rSrc.top, rSrc.bottom));
        DBG_DD(("rDest  X (%ld,%ld) Y (%ld,%ld)\n",
                rDest.left, rDest.right,rDest.top, rDest.bottom));

        lphwinfo->dwWinStartXY = ((rDest.left<<16) + rDest.top);
        lphwinfo->dwWinEndXY= (((rDest.right-1)<<16) + (rDest.bottom - 1));   

        if (lpVideoHWDifference->dwBypassVerExpandMagnify == VID_HWDIFF_TRUE)
        {
            if (dwVDevFlag&CTL_UseExpand &&((dwVDevFlag&CTL_UseSIMView)||(dwVDevFlag&CTL_UseIGA2Path)))
            {
                dwCompose |= VAL_DISABLE_LCD_VERTICAL_SCALING_CME;
            }
        }        

        /* Setup X zoom factor */
        lpOverlayRecord->dwFetchAlignment = 0;

        /* vfCalcZoomWidth_CME */
        vidCalcZoomW.dwCtlHQVflag       = dwCtlHQVflag;

        if (pVidData->NeedToCorrectVideoSize && pVidData->isVideoWidthDownScaling)
        {
            vidCalcZoomW.dwSrcWidth = pVidData->DownScalingDestRect.right - pVidData->DownScalingDestRect.left;
        }
        else
        {
            vidCalcZoomW.dwSrcWidth = dwSrcWidth;
        }
        
        vidCalcZoomW.dwDstWidth         = dwDstWidth;
        vidCalcZoomW.lpdwZoomCtl        = &dwZoomCtl;
        vidCalcZoomW.lpdwMiniCtl        = &dwMiniCtl;

        vidCalcZoomW.lpdwHQVzoomflagH   = &dwHQVzoomflagH;
        vidCalcZoomW.lpdwMiniHFactor    = &lpOverlayRecord->dwminifyH;
        vidCalcZoomW.lpdwFetchAlignment = &lpOverlayRecord->dwFetchAlignment;
        
        
 		if((lpVIAGraphicInfo->dwDeviceID == VIA_DEVICE_VT3293) ||
          (lpVIAGraphicInfo->dwDeviceID == VIA_DEVICE_VT3353))  
        {
            vidCalcZoomW.lpdwHQVScaleCtlH =&dwHQVScaleCtlH;
            vfCalcZoomWidth_CME_293(&vidCalcZoomW);
        }
        else
        {
            vidCalcZoomW.lpdwHQVminiCtl     = &dwHQVminiCtl;
            vfCalcZoomWidth_CME(&vidCalcZoomW);
        }   
        
        dwFetch <<= 20;

        if (dwVideoFlag & VIDEO_1_INUSE)    
        {
            if(lpVIAGraphicInfo->duoview&DUOVIEW_ENABLED)/*for duoview*/
            {
                dwFetch |=(dMMRD(pjVideo,V3_ALPHA_FETCH_COUNT)&(~V3_FETCH_COUNT));
                dwVDevFlag |= CTL_UseDuoView|CTL_UseIGA2Path;
            }
            lphwinfo->dwFetch = dwFetch;
        }
        else
        {
            dwFetch |=(dMMRD(pjVideo,V3_ALPHA_FETCH_COUNT)&(~V3_FETCH_COUNT));
            lphwinfo->dwFetch = dwFetch;
        }

        /* Setup Y zoom factor */

        /* For DCT450 test-BOB INTERLEAVE */
        if ((dwDeinterlaceMode & DDOVER_INTERLEAVED) &&
            (dwDeinterlaceMode & DDOVER_BOB))
        {
            if (!(dwVideoFlag & VIDEO_HQV_INUSE))
            {
                dwSrcHeight /=2;
                if (dwVideoFlag & VIDEO_1_INUSE)    
                {
                    dwVidCtl |= (V1_BOB_ENABLE | V1_FRAME_BASE);
                    if(dwVideoFlag & VIDEO_1_ON_SND)
                        dwVidCtl |= (V3_BOB_ENABLE | V3_FRAME_BASE); 
                }
                else
                {
                    dwVidCtl |= (V3_BOB_ENABLE | V3_FRAME_BASE);                    
                }
            }
            else
            {             
                dwHQVCtl |= HQV_FIELD_2_FRAME|
                            HQV_FRAME_2_FIELD|
                            HQV_DEINTERLACE;   
            }
        }
        else if (dwDeinterlaceMode & DDOVER_BOB )
        {
            if (dwVideoFlag & VIDEO_HQV_INUSE)
            {
                dwSrcHeight <<=1;
                dwHQVCtl |= HQV_FIELD_2_FRAME|HQV_DEINTERLACE;
            }
            else
            {
                if (dwVideoFlag & VIDEO_1_INUSE)    
                {
                    dwVidCtl |= V1_BOB_ENABLE;  
                    if(dwVideoFlag & VIDEO_1_ON_SND)/*add for DuoView */              
                        dwVidCtl |= V3_BOB_ENABLE;  
                }
                else
                {
                    dwVidCtl |= V3_BOB_ENABLE;                    
                }
            }
        }

        /*
         * vfGetDisplayCount_CME
         */
        vidGetDisplayCount.dwVideoFlag      = dwVideoFlag;
        vidGetDisplayCount.lpDPFsrc         = lpDPFsrc;
        vidGetDisplayCount.dwSrcWidth       = dwSrcWidth;
        vidGetDisplayCount.dwMiniHFactor    = lpOverlayRecord->dwminifyH;
        vidGetDisplayCount.lpdwDisplayCountW= &dwDisplayCountW;
        
        vfGetDisplayCount_CME(&vidGetDisplayCount);
        
        lphwinfo->dwVDisplayCount = (dwSrcHeight<<16)|dwDisplayCountW;

        /*
         * vfCalcZoomHeight_CME
         */
        vidCalcZoomH.dwCtlHQVflag       = dwCtlHQVflag;

        if (pVidData->NeedToCorrectVideoSize && pVidData->isVideoHeightDownScaling)
        {
            vidCalcZoomH.dwSrcHeight = pVidData->DownScalingDestRect.bottom - pVidData->DownScalingDestRect.top;
        }
        else
        {
            vidCalcZoomH.dwSrcHeight = dwSrcHeight;
        }
        
        vidCalcZoomH.dwDstHeight        = dwDstHeight;
        vidCalcZoomH.lpdwZoomCtl        = &dwZoomCtl;
        vidCalcZoomH.lpdwMiniCtl        = &dwMiniCtl;
  
        vidCalcZoomH.lpdwHQVzoomflagV   = &dwHQVzoomflagV;

        if((lpVIAGraphicInfo->dwDeviceID == VIA_DEVICE_VT3293) ||
          (lpVIAGraphicInfo->dwDeviceID == VIA_DEVICE_VT3353))
        {
            vidCalcZoomH.lpdwHQVScaleCtlV =&dwHQVScaleCtlV;
            vfCalcZoomHeight_CME_293(&vidCalcZoomH);
        }
        else
        {
            vidCalcZoomH.lpdwHQVminiCtl     = &dwHQVminiCtl;
            vfCalcZoomHeight_CME(&vidCalcZoomH);
        }  

        if (dwVideoFlag & VIDEO_1_INUSE) 
        {
            if(lpVideoHWDifference->dwNeedV1Prefetch)
            {
                lphwinfo->dwV1FIFO = V1_FIFO_DEPTH122 | V1_FIFO_RRETHRESHOLD113| V1_FIFO_THRESHOLD113;
            }
            else
            {
                lphwinfo->dwV1FIFO = V1_FIFO_DEPTH64 | V1_FIFO_PRETHRESHOLD61 | V1_FIFO_THRESHOLD61;
            }
            
            if(lpVIAGraphicInfo->duoview&DUOVIEW_ENABLED)
            {
                lphwinfo->dwV3FIFO = (dMMRD(pjVideo,ALPHA_V3_FIFO_CONTROL)& \
                                      ALPHA_FIFO_MASK)|                 \
                                      V3_FIFO_DEPTH64 | V3_FIFO_THRESHOLD61;
                /*3324,3293 has different fifomask */
                if ((lpVIAGraphicInfo->dwProjectIDRevision == UNICHROME_VT3324) \
                        ||(lpVIAGraphicInfo->dwProjectIDRevision == UNICHROME_VT3324M)\
                        ||(lpVIAGraphicInfo->dwProjectIDRevision == UNICHROME_VT3324M2)\
                        || (lpVIAGraphicInfo->dwProjectIDRevision == UNICHROME_VT3293) \
                        || (lpVIAGraphicInfo->dwProjectIDRevision == UNICHROME_VT3353))
                {
                    lphwinfo->dwV3PreFIFO = (V3_FIFO_THRESHOLD61>>8) |\
                        (dMMRD(pjVideo,ALPHA_V3_PREFIFO_CONTROL)&\
                         (~V3_FIFO_MASK_3324));
                }    
                else
                {
                    lphwinfo->dwV3PreFIFO = (V3_FIFO_THRESHOLD61>>8) |\
                        (dMMRD(pjVideo,ALPHA_V3_PREFIFO_CONTROL)&\
                         (~V3_FIFO_MASK));
                }    
                /*for duo view calculate the video1 vertical scaling 
                  factor when LCD expansion */
                if((lpOverlayRecord->dwExpandDestRect.bottom !=0) &&
                   (lpOverlayRecord->dwExpandDestRect.bottom != rDest.bottom)&&
                   (lpOverlayRecord->dwExpandDestRect.bottom!=lpOverlayRecord->dwExpandDestRect.top))/*expansion*/
                {
                    if((lpVIAGraphicInfo->duoview&DUOVIEW_V3_ON_IGA2)\
                            ||(lpVIAGraphicInfo->dwDeviceID!=VIA_DEVICE_VT3336\
                                && lpVIAGraphicInfo->dwDeviceID!=VIA_DEVICE_VT3327\
                                && lpVIAGraphicInfo->dwDeviceID!=VIA_DEVICE_VT3364 ))
                    {/*not 336, 364 and 327  or  V3 On IGA2*/
                        if(dwDstHeight >= dwSrcHeight)
                            lphwinfo->dwzoomCtlV1DuoView = 
                                ((dwSrcHeight*0x400)/    
                                 (lpOverlayRecord->dwExpandDestRect.bottom - 
                                  lpOverlayRecord->dwExpandDestRect.top));
                        else
                            lphwinfo->dwzoomCtlV1DuoView = 
                                ((dwDstHeight*0x400)/    
                                 (lpOverlayRecord->dwExpandDestRect.bottom - 
                                  lpOverlayRecord->dwExpandDestRect.top));
                        lphwinfo->dwzoomCtlV1DuoView |= V1_Y_ZOOM_ENABLE; 
                    }

                }
                    
            }
        }
        else
        {
            lphwinfo->dwV3FIFO = (dMMRD(pjVideo,ALPHA_V3_FIFO_CONTROL)&ALPHA_FIFO_MASK)|
                                 V3_FIFO_DEPTH64 | V3_FIFO_THRESHOLD61;
            /*3324,3293 has different fifomask */
            if ((lpVIAGraphicInfo->dwProjectIDRevision == UNICHROME_VT3324)\
                    ||(lpVIAGraphicInfo->dwProjectIDRevision == UNICHROME_VT3324M)\
                    ||(lpVIAGraphicInfo->dwProjectIDRevision == UNICHROME_VT3324M2)\
                    ||(lpVIAGraphicInfo->dwProjectIDRevision == UNICHROME_VT3293)\
                    ||(lpVIAGraphicInfo->dwProjectIDRevision == UNICHROME_VT3353))
            {
                lphwinfo->dwV3PreFIFO = (V3_FIFO_THRESHOLD61>>8) |
                        (dMMRD(pjVideo,ALPHA_V3_PREFIFO_CONTROL)& (~V3_FIFO_MASK_3324));
            }    
            else
            {
                lphwinfo->dwV3PreFIFO = (V3_FIFO_THRESHOLD61>>8) |
                            (dMMRD(pjVideo,ALPHA_V3_PREFIFO_CONTROL)& (~V3_FIFO_MASK));
            }   
        }

        if (dwVideoFlag & VIDEO_HQV_INUSE)
        {
            dwMiniCtl=0;
            if (dwHQVzoomflagH||dwHQVzoomflagV)
            {
                dwTmp = 0;
                if (dwHQVzoomflagH)
                {
                    dwMiniCtl = V1_X_INTERPOLY;
                    dwTmp = (dwZoomCtl&MASK_VIDEO_ZOOM_FACTOR_H);
                }
                
                if (dwHQVzoomflagV)
                {
                    UpdYInterpolationFlag(lpVIAGraphicInfo, dwSrcWidth, dwVideoFlag, &dwMiniCtl, &dwCompose);
                    dwTmp |= (dwZoomCtl&MASK_VIDEO_ZOOM_FACTOR_V);
                }                   

                lphwinfo->dwminiCtl = dwMiniCtl;
                lphwinfo->dwzoomCtl = dwTmp;
            }
            else
            {
                lphwinfo->dwminiCtl = 0;
                lphwinfo->dwzoomCtl = 0;
            }
          
            if ((lpVIAGraphicInfo->dwDeviceID == VIA_DEVICE_VT3293) ||
                (lpVIAGraphicInfo->dwDeviceID == VIA_DEVICE_VT3353))
            {
#ifdef CS_FLAG_HWROTATION
                /* In rotation mode: No need to control HQV Scaling factor. */
                /* Because zoom in will be done by Video engine, and zoom out will be done by 3D engine. */
                if (lpVIAGraphicInfo->rotate)
                {
                    lphwinfo->dwHQVScaleCtlH = 0;  
                    lphwinfo->dwHQVScaleCtlV = 0;
                }
                else
#endif /* CS_FLAG_HWROTATION */
                {
                    lphwinfo->dwHQVScaleCtlH = dwHQVScaleCtlH;  
                    lphwinfo->dwHQVScaleCtlV = dwHQVScaleCtlV;
                }
            }
            else
            {
#ifdef CS_FLAG_HWROTATION
                /* In rotation mode: No need to control HQV Scaling factor. */
                /* Because zoom in will be done by Video engine, and zoom out will be done by 3D engine. */
                if (lpVIAGraphicInfo->rotate)
                {
                    lphwinfo->dwHQVMiniZoomCtl = 0;
                }
                else
#endif /* CS_FLAG_HWROTATION */
                {
                    lphwinfo->dwHQVMiniZoomCtl = dwHQVminiCtl;
                }
            }
            
        }
        else
        {
            lphwinfo->dwminiCtl = dwMiniCtl;
            lphwinfo->dwzoomCtl = dwZoomCtl;
        }

        if(dwVideoFlag & SW_ON_SND)
        {
            if((lpOverlayRecord->dwExpandDestRect.bottom !=0) &&
               (lpOverlayRecord->dwExpandDestRect.bottom != rDest.bottom)&&
               (lpOverlayRecord->dwExpandDestRect.bottom!=lpOverlayRecord->dwExpandDestRect.top))
                {/*expansion*/
                    if(dwDstHeight >= dwSrcHeight)
                        lphwinfo->dwzoomCtlV1DuoView = 
                            ((dwSrcHeight*0x400)/    
                             (lpOverlayRecord->dwExpandDestRect.bottom - 
                              lpOverlayRecord->dwExpandDestRect.top));
                    else
                        lphwinfo->dwzoomCtlV1DuoView = 
                            ((dwDstHeight*0x400)/    
                             (lpOverlayRecord->dwExpandDestRect.bottom - 
                              lpOverlayRecord->dwExpandDestRect.top));
                    lphwinfo->dwzoomCtlV1DuoView |= V1_Y_ZOOM_ENABLE; 
                    lphwinfo->dwzoomCtl &= 0xffff0000;
                    lphwinfo->dwzoomCtl |= lphwinfo->dwzoomCtlV1DuoView;
                    lphwinfo->dwminiCtl = V1_X_INTERPOLY;
                    UpdYInterpolationFlag(lpVIAGraphicInfo, dwSrcWidth, dwVideoFlag, &(lphwinfo->dwminiCtl), &dwCompose);
                }
                
        }  
        /* Colorkey */
        lphwinfo->dwColorKey    &= ~VAL_VIDEO_COLOR_KEY_ENABLE_CME;
        lphwinfo->dwLCDColorKey &= ~VAL_VIDEO_COLOR_KEY_ENABLE_CME;
        lphwinfo->dwChromaKeyLow&= ~VAL_VIDEO_CHROMA_KEY_ENABLE_CME;
        if (dwColorKeyOn) 
        {
            DBG_DD(("Overlay colorkey=%08lx\n", dwKeyLow));

            dwKeyLow &= VAL_VIDEO_COLOR_KEY_MASK;

            if (lpVIAGraphicInfo->Screen[0].dwBPP == 8)
            {
                dwKeyLow &= 0xff;
            }

            lphwinfo->dwColorKey = dwKeyLow|VAL_VIDEO_COLOR_KEY_ENABLE_CME;

            if(dwVDevFlag&CTL_UseIGA2Path)
            {
                /* LCD, Display 2 */
                lphwinfo->dwLCDColorKey=dwKeyLow|VAL_VIDEO_COLOR_KEY_ENABLE_CME;
            }

            if (dwVideoFlag & VIDEO_1_INUSE)
            {
            
                if(dwVideoFlag & VIDEO_1_ON_SND)
                    dwCompose   |=  SECOND_DISPLAY_COLOR_KEY_ENABLE;
                else
                    dwCompose   |=  SELECT_VIDEO_IF_COLOR_KEY;
                switch(lpVIAGraphicInfo->dwDeviceID)
                {
                case VIA_DEVICE_VT3336:
                case VIA_DEVICE_VT3327:
                case VIA_DEVICE_VT3364:
                    dwCompose   |=  SECOND_DISPLAY_COLOR_KEY_ENABLE|0x01;   
                    break;
                default:
                    break;
                }
            }
            else /* Use V3 */
                dwCompose   |=  SELECT_VIDEO_IF_COLOR_KEY;
        }
        if (dwChromaKeyOn) 
        {
            DBG_DD(("Overlay Chromakey=%08lx %08lx\n",dwChromaLow,dwChromaHigh));

            dwChromaLow  &= CHROMA_KEY_LOW;
            dwChromaHigh &= CHROMA_KEY_HIGH;

            dwChromaLow  |= (dMMRD(pjVideo,V_CHROMAKEY_LOW)&(~CHROMA_KEY_LOW));
            dwChromaHigh |= (dMMRD(pjVideo,V_CHROMAKEY_HIGH)&(~CHROMA_KEY_HIGH));

            /* for Chroma Key */
            if (lpDPFsrc->dwFlags & DDPF_FOURCC)
            {
                switch (lpDPFsrc->dwFourCC) 
                {
                case FOURCC_YV12:
                    /* to be continued... */
                    break;
                case FOURCC_YUY2:
                case FOURCC_NV12:
                    /* to be continued... */
                    break;
                default:
                    break;
                }
            }
            else if (lpDPFsrc->dwFlags & DDPF_RGB)
            {
                unsigned long dwtmpLowR;
                unsigned long dwtmpLowG;
                unsigned long dwtmpLowB;
                unsigned long dwtmpChromaLow;
                unsigned long dwtmpHighR;
                unsigned long dwtmpHighG;
                unsigned long dwtmpHighB;
                unsigned long dwtmpChromaHigh;

                switch (lpDPFsrc->dwRGBBitCount) 
                {
                case 16:
                    if (lpDPFsrc->dwGBitMask==0x07E0) /* RGB16(5:6:5) */
                    {
                        dwtmpLowR = (((dwChromaLow >> 11) << 3) | ((dwChromaLow >> 13) & 0x07)) & 0xFF;
                        dwtmpLowG = (((dwChromaLow >> 5) << 2) | ((dwChromaLow >> 9) & 0x03)) & 0xFF;

                        dwtmpHighR = (((dwChromaHigh >> 11) << 3) | ((dwChromaHigh >> 13) & 0x07)) & 0xFF;
                        dwtmpHighG = (((dwChromaHigh >> 5) << 2) | ((dwChromaHigh >> 9) & 0x03)) & 0xFF;
                    }
                    else /* RGB15(5:5:5) */
                    {
                        dwtmpLowR = (((dwChromaLow >> 10) << 3) | ((dwChromaLow >> 12) & 0x07)) & 0xFF;
                        dwtmpLowG = (((dwChromaLow >> 5) << 3) | ((dwChromaLow >> 7) & 0x07)) & 0xFF;

                        dwtmpHighR = (((dwChromaHigh >> 10) << 3) | ((dwChromaHigh >> 12) & 0x07)) & 0xFF;
                        dwtmpHighG = (((dwChromaHigh >> 5) << 3) | ((dwChromaHigh >> 7) & 0x07)) & 0xFF;
                    }
                    dwtmpLowB = (((dwChromaLow << 3) | (dwChromaLow >> 2)) & 0x07) & 0xFF;
                    dwtmpChromaLow = (dwtmpLowG << 16) | (dwtmpLowR << 8) | dwtmpLowB;
                    dwChromaLow = ((dwChromaLow >> 24) << 24) | dwtmpChromaLow;

                    dwtmpHighB = (((dwChromaHigh << 3) | (dwChromaHigh >> 2)) & 0x07) & 0xFF;
                    dwtmpChromaHigh = (dwtmpHighG << 16) | (dwtmpHighR << 8) | dwtmpHighB;
                    dwChromaHigh = ((dwChromaHigh >> 24) << 24) | dwtmpChromaHigh;
                    break;

                case 32: /* 32 bit RGB */
                    dwtmpLowR = (dwChromaLow >> 16) & 0xFF;
                    dwtmpLowG = (dwChromaLow >> 8) & 0xFF;
                    dwtmpLowB = dwChromaLow & 0xFF;
                    dwtmpChromaLow = (dwtmpLowG << 16) | (dwtmpLowR << 8) | dwtmpLowB;
                    dwChromaLow = ((dwChromaLow >> 24) << 24) | dwtmpChromaLow;

                    dwtmpHighR = (dwChromaHigh >> 16) & 0xFF;
                    dwtmpHighG = (dwChromaHigh >> 8) & 0xFF;
                    dwtmpHighB = dwChromaHigh & 0xFF;
                    dwtmpChromaHigh = (dwtmpHighG << 16) | (dwtmpHighR << 8) | dwtmpHighB;
                    dwChromaHigh = ((dwChromaHigh >> 24) << 24) | dwtmpChromaHigh;
                    break;

                default:
                    break;
                }
            }/* End of DDPF_FOURCC */

            lphwinfo->dwChromaKeyHigh = dwChromaHigh;
            if (dwVideoFlag & VIDEO_1_INUSE)
            {
                lphwinfo->dwChromaKeyLow = 
                (dwChromaLow&(~V_CHROMAKEY_V3))|VAL_VIDEO_CHROMA_KEY_ENABLE_CME;
                lphwinfo->dwminiCtl = dwMiniCtl & 0xFFFFFFF8;
            }
            else
            {
                lphwinfo->dwChromaKeyLow = 
                dwChromaLow|V_CHROMAKEY_V3|VAL_VIDEO_CHROMA_KEY_ENABLE_CME;
                lphwinfo->dwminiCtl = dwMiniCtl & 0xFFFFFFF8;
            }
        }

        if (dwVideoFlag & VIDEO_HQV_INUSE)
        {
            dwTmpHQVSrcFetch = ((lphwinfo->dwHQVSrcCountWH & MASK_HQV_SRC_FETCH)>>16);
            if(dwTmpHQVSrcFetch<31)
            {
                lphwinfo->dwHQVSrcCountWH = 
                (lphwinfo->dwHQVSrcCountWH&(~MASK_HQV_SRC_FETCH))|VAL_HQV_SRC_FETCH_31;
            }
        }
        /* Setup video control VIDEO */
        *lpdwCtlCMFlag = CTL_CM_UPDATEOVERLAY_CME;
        vfCommandManager(pVidData);
        if(lpVideoHWDifference->dwNeedV1Prefetch)
        {
             dwVidCtl |= V1_PREFETCH_ON_3336;
        }        
        /*maintain the old relative position to another video*/
        if (!(dwVideoFlag & VIDEO_STAY_OLD)) 
        {
            if (dwVideoFlag & VIDEO_1_INUSE)
            {
                dwVidCtl |= (dMMRD(pjVideo,V1_CONTROL)&VAL_VIDEO_GAMMA_ENABLE_CME);
		            dwCompose |=COMPOSE_V1_V3;     
                if((dwVideoFlag & VIDEO_1_ON_SND)||(dwVideoFlag & SW_ON_SND))
                {
                    dwVidCtl    |= (dMMRD(pjVideo,V3_CONTROL)&\
                                    VAL_VIDEO_GAMMA_ENABLE_CME);
                    dwCompose   |= COMPOSE_V3_V1; 
                 }
            }
            else
            {
                dwVidCtl    |= (dMMRD(pjVideo,V3_CONTROL)&VAL_VIDEO_GAMMA_ENABLE_CME);
	              dwCompose   |= COMPOSE_V3_V1;            
            }        	    
        }
        else
        {
             DBG_DD(("Video stays old position\n"));
             dwCompose=dMMRD(pjVideo,V_COMPOSE_MODE);
             if (dwVideoFlag &VIDEO_1_INUSE)
                 vfCM(V_COMPOSE_MODE,dwCompose|V1_COMMAND_FIRE,pVidData);
             else
                 vfCM(V_COMPOSE_MODE,dwCompose|V3_COMMAND_FIRE,pVidData);
             if ((dwVideoFlag&VIDEO_1_ON_SND)||(dwVideoFlag &MPEG_ON_SND))
                 vfCM(V_COMPOSE_MODE,dwCompose|V3_COMMAND_FIRE,pVidData);	
        }	
        lphwinfo->dwVidCtl  = dwVidCtl;
        lphwinfo->dwCompose = dwCompose;
#ifdef CS_FLAG_HWROTATION
        if( lpVIAGraphicInfo->rotate)  
        {
            /*rotate, enable hqv trigger V3*/
            dwVidCtl &= 0xf9ffffff;     /* When rotate, v1 use sw flip */
        }
#endif /* CS_FLAG_HWROTATION */
        if (!lpVIAGraphicInfo->IsHWCursorEnabled)
        {
            if ( dwVideoFlag & SW_ON_SND )
            {
                vfCM(HI_CONTROL, dMMRD(pjVideo,HI_CONTROL)|V4_DISPLAY_PATH,pVidData); 
            }
            else
            {
                vfCM(HI_CONTROL, dMMRD(pjVideo,HI_CONTROL)&(~V4_DISPLAY_PATH),pVidData);
            }
        }

        if (dwVideoFlag & VIDEO_HQV_INUSE)
        {
#ifdef CS_FLAG_HWROTATION
            if (lpVIAGraphicInfo->rotate) 
            {
                dwHQVCtl = (dwHQVCtl | HQV_INT_ENABLE);   /* Enable HQV interrupt for rotate. */
            }
#endif /* CS_FLAG_HWROTATION */
            if (lpOverlayRecord->dwHQVon != HQV_USE_ENABLE)
            {
/*for s3 sleep xv resume case, if flip with no real context in hqv buffer after waking, 
 * video scrambles and in worst case,system hang
 */
#ifdef XFree86Server 
                dwHQVCtl = dwHQVCtl | HQV_ENABLE; 
                vfCMHQVW(HQV_CONTROL, dwHQVCtl, pVidData);
#else
                vfCMHQVW(HQV_CONTROL, dwHQVCtl|HQV_FLIP_STATUS, pVidData);
                vfCMHQVW(HQV_CONTROL, dwHQVCtl|HQV_SW_FLIP, pVidData);
#endif

#ifdef CS_FLAG_HWROTATION
                /* In VT3353, we have to use CR to wait HQV SW Flip. */
                if((dwVideoFlag & VIDEO_VQ_INUSE) && (lpVIAGraphicInfo->dwDeviceID == VIA_DEVICE_VT3353))
                {
                    vfHM(pVidData, HM_WAIT_HQV_DONE_VQ); 
                }
                else
                {
                    vfHM(pVidData, HW_WAIT_HQV_FINISH);
                }
                
#endif /* CS_FLAG_HWROTATION */

                if (dwVideoFlag & VIDEO_1_INUSE)
                {
                    if(dwVideoFlag & VIDEO_1_ON_SND)
                        vfCM(V1_CONTROL, dwVidCtl|V1_ON_SND_DISPLAY,pVidData);
                    else
                        vfCM(V1_CONTROL, dwVidCtl,pVidData);
                    vfCM(V_COMPOSE_MODE, (dwCompose|V1_COMMAND_FIRE),pVidData);
                    if(dwVideoFlag & VIDEO_1_ON_SND)
                    {
                        vfCM(V3_CONTROL, dwVidCtl,pVidData);
                        vfCM(V_COMPOSE_MODE, (dwCompose|V3_COMMAND_FIRE ),pVidData);
                    }
                }
                else
                {
                    /* Gamma */
                    vfCM(V3_CONTROL, dwVidCtl,pVidData);
                    /* V4 Pre-Fetch Enable */
                    vfCM(V_COMPOSE_MODE, (dwCompose|V3_COMMAND_FIRE ),pVidData);
                }           
                lpOverlayRecord->dwHQVon = HQV_USE_ENABLE;
                lpOverlayRecord->dwHQVInUseType = HQV_IN_VIDEO;                
            }
            else
            {
                if (dwVideoFlag & VIDEO_1_INUSE)
                {
                    if(dwVideoFlag & VIDEO_1_ON_SND)
                        vfCM(V1_CONTROL, dwVidCtl|V1_ON_SND_DISPLAY,pVidData);
                    /*In duo view, V1 shoud be put on IGA2, cause V3 is fixed on IGA1 */
                    else
                        vfCM(V1_CONTROL, dwVidCtl,pVidData);
                    vfCM(V_COMPOSE_MODE, (dwCompose|V1_COMMAND_FIRE ),pVidData);
                    if(dwVideoFlag & VIDEO_1_ON_SND)
                    {
                        vfCM(V3_CONTROL, dwVidCtl,pVidData);
                        vfCM(V_COMPOSE_MODE, (dwCompose|V3_COMMAND_FIRE ),pVidData);
                    }
                }
                else
                {
                    vfCM(V3_CONTROL, dwVidCtl,pVidData);
                    vfCM(V_COMPOSE_MODE, (dwCompose|V3_COMMAND_FIRE ),pVidData);
                }
                
                vfCMHQVW(HQV_CONTROL, dwHQVCtl|HQV_SUB_FLIP, pVidData);

            }

            lphwinfo->dwHQVCtl = dwHQVCtl;
        }
        else /* Without HQV */
        {
            if (dwVideoFlag & VIDEO_1_INUSE)
            {
                vfCM(V1_CONTROL, dwVidCtl,pVidData);
                vfCM(V_COMPOSE_MODE, (dwCompose|V1_COMMAND_FIRE ),pVidData);
            }
            else
            {
                vfCM(V3_CONTROL, dwVidCtl,pVidData);
                vfCM(V_COMPOSE_MODE, (dwCompose|V3_COMMAND_FIRE ),pVidData);
            }

            lpOverlayRecord->dwVideoOn = 1;
            lpOverlayRecord->dwOverlayInUseType = Overlay_IN_VIDEO;
        }
    }
    else
    {
        /* Hide overlay */
        if(dwVideoFlag&VIDEO_HQV_INUSE)
        {
            if(vfCMHQVR(HQV_CONTROL,pVidData) & HQV_ENABLE)
            {
                DBG_DD(("Before wait, Reg HQV_CONTROL is %x \n",  vfCMHQVR(HQV_CONTROL,pVidData)));             
                vfHM(pVidData,HM_WAIT_HQV_DONE|HM_WAIT_VIRTUAL_QUEUE);
                vfHM(pVidData,HM_WAIT_SUBFLIP_FIRE|HM_WAIT_VIRTUAL_QUEUE);
                vfHM(pVidData, HW_WAIT_HQV_FINISH|HM_WAIT_VIRTUAL_QUEUE);
                vfHM(pVidData, HM_WAIT_FIRE|HM_WAIT_VIRTUAL_QUEUE);
                DBG_DD(("After wait, Reg HQV_CONTROL is %x \n",  vfCMHQVR(HQV_CONTROL,pVidData)));             
            }
        }

        if (dwVideoFlag&VIDEO_HQV_INUSE)
        {
            lpOverlayRecord->dwHQVon = HQV_USE_DISABLE;
            lpOverlayRecord->dwHQVInUseType &= (~HQV_IN_VIDEO);
            lphwinfo->dwHQVCtl = vfCMHQVR(HQV_CONTROL,pVidData) & (~HQV_ENABLE);
            vfCMHQVW(HQV_CONTROL, lphwinfo->dwHQVCtl,pVidData);
        }
        else
        {
            lpOverlayRecord->dwVideoOn = 0;
            lpOverlayRecord->dwOverlayInUseType &= (~Overlay_IN_VIDEO);
        }

        if (dwVideoFlag&VIDEO_1_INUSE)
        {
            lphwinfo->dwVidCtl=dMMRD(pjVideo,V1_CONTROL)&(~VAL_VIDEO_ENABLE_CME);
            lphwinfo->dwCompose=dMMRD(pjVideo,V_COMPOSE_MODE);
            dwCompose=lphwinfo->dwCompose|V1_COMMAND_FIRE;

            vfCM(V1_CONTROL, lphwinfo->dwVidCtl,pVidData);
            vfCM(V_COMPOSE_MODE, dwCompose,pVidData);
			
	        if(dwVideoFlag & VIDEO_1_ON_SND)
            {
                dwVidCtl=dMMRD(pjVideo,V3_CONTROL)&(~VAL_VIDEO_ENABLE_CME);
                dwCompose = dMMRD(pjVideo,V_COMPOSE_MODE);
                dwCompose |= V3_COMMAND_FIRE;

                vfCM(V3_CONTROL, dwVidCtl,pVidData);
                vfCM(V_COMPOSE_MODE, dwCompose,pVidData);
            }		
		 
        }
        else
        {
            lphwinfo->dwVidCtl=dMMRD(pjVideo,V3_CONTROL)&(~VAL_VIDEO_ENABLE_CME);
            lphwinfo->dwCompose = dMMRD(pjVideo,V_COMPOSE_MODE);
            dwCompose=lphwinfo->dwCompose|V3_COMMAND_FIRE;
            vfHM(pVidData, HM_WAIT_FIRE|HM_WAIT_VIRTUAL_QUEUE);
 #ifdef CS_FLAG_HWROTATION           
            vfCM(V3_CONTROL, lphwinfo->dwVidCtl,pVidData);
            vfCM(V_COMPOSE_MODE, dwCompose,pVidData);
 #else /* CS_FLAG_HWROTATION */          
            VIDOutD(V3_CONTROL,lphwinfo->dwVidCtl);
            VIDOutD(V_COMPOSE_MODE,dwCompose);
/* CS_FLAG_HWROTATION */
 #endif 
/* CS_FLAG_HWROTATION */
        }
    }

    return VMI_OK;    
}

/*
 *  Create overlay surface depend on FOURCC
 */
unsigned long CreateSurface_CME(PVIDDATA pVidData, LPDDSURFACEDESC lpDDSurfaceDesc)
{
    unsigned long handle=pVidData->VMI_Info.handle;
    unsigned long   dwWidth, dwHeight, dwPitch=0, dwAlignHeight=0;
    unsigned long   dwBackBufferCount;
    unsigned long   dwRet=VMI_OK;
    unsigned long   dwAddr = 0;
    unsigned long   MPEGFBSIZE  = 0;    
    unsigned long   SUBPFBSIZE  = 0;
    unsigned long   HQVFBSIZE   = 0;
    unsigned long   TVFBSIZE    = 0;
    unsigned long   ALPHAFBSIZE = 0;
    unsigned long   SWFBSIZE    = 0;
    unsigned long   dwSize = 0;
    int     i, iCount;       /* iCount for clean HQV FB use */
    unsigned char  * lpTmpAddr;    /* for clean HQV FB use */
   
    LPVIDHWDIFFERENCE lpVideoHWDifference = &pVidData->VideoHWDifference;
    LPMPGDEVICE lpMPGDevice = NULL;
    LPSUBDEVICE lpSubDevice = NULL;
    LPSWDEVICE  lpSWDevice  = &(pVidData->SWDevice[handle]);
    LPALPHADEVICE lpAlpDevice = &(pVidData->ddLock.ALPDevice);
    LPVIAGRAPHICINFO lpVIAGraphicInfo = &pVidData->VIAGraphicInfo;
    LPVIAVIDCTRL  lpVideoControl = &pVidData->VideoControl;
    LPDDPIXELFORMAT lpDPFsrc = &pVidData->DPFsrc[handle];    
    LPOVERLAYRECORD lpOverlayRecordMPEG = NULL;
    LPOVERLAYRECORD lpOverlayRecordSW = &pVidData->OverlayRecordSet.OverlayRecordSW[handle];
    
    unsigned char* pjVideo = pVidData->Vhwinfo.pjVideo;
    unsigned char* pjScreen = pVidData->Vhwinfo.pjScreen;

    int ScrnIndex = lpVIAGraphicInfo->scrnIndex;
    unsigned long dwVideoFlag = 0;   
    
    DBG_DD(("\n"));
    DBG_DD(("dwVideoFlag=%p\n",pVidData->VidCtl.dwVideoFlag[handle]));
    if ( lpDDSurfaceDesc == NULL )
        return VMI_ERR;

    if(lpVIAGraphicInfo->duoview)  
    {
    	switch (lpDDSurfaceDesc->dwFourCC)
    	{
        case FOURCC_NV12 :
		case FOURCC_YUY2:
		case FOURCC_YV12:
            lpVIAGraphicInfo->bVideoEngineUsedOut=1;
			break;
		default:
			break;
    	}
    }	

    switch (lpDDSurfaceDesc->dwFourCC)
    {
        case FOURCC_SUBP :
            DBG_DD(("    Create subpicture Surface: \n"));
            DBG_DD((" handle=%lx \n",pVidData->VMI_Info.handle));

            dwVideoFlag= pVidData->VidStatus.dwVideoFlagMPEG[handle];
            lpSubDevice= &(pVidData->ddLock.SubDev[handle]);
            lpMPGDevice = &(pVidData->MPGDevice[handle]);
            lpOverlayRecordMPEG = &pVidData->OverlayRecordSet.OverlayRecordMPEG[handle];
      
            dwBackBufferCount = lpDDSurfaceDesc-> dwBackBufferCount;
            dwWidth=dwHeight=dwPitch=dwAlignHeight=0;
            dwSize=CalSurfaceSize(lpDDSurfaceDesc,SUBP_REQ_SURFACE_ID,2,
                    &dwWidth,&dwHeight,&dwPitch,&dwAlignHeight,&SUBPFBSIZE);
            dwAddr = viaAllocVideoSurface(pVidData, SUBP_REQ_SURFACE_ID, dwSize);

            if (dwAddr==0)
            {
                DBG_DD(("Memory allocation for SUBPICTURE fails\n"));
                return VMI_ERR_CANNOT_CREATE_SURFACE;
            }
            lpSubDevice->dwSUBPhysicalAddr[0] = dwAddr;

            lpSubDevice->dwSUBPhysicalAddr[1] = lpSubDevice->dwSUBPhysicalAddr[0] + SUBPFBSIZE;

            DBG_DD(("SP_Buffer[0]: 0x%x\n", lpSubDevice->dwSUBPhysicalAddr[0]));
            DBG_DD(("SP_Buffer[1]: 0x%x\n", lpSubDevice->dwSUBPhysicalAddr[1]));

	    /* add the subdevice's overalay surface	*/
            lpSubDevice->lpSUBOverlaySurface[0] = pjScreen + lpSubDevice->dwSUBPhysicalAddr[0];
	        lpSubDevice->lpSUBOverlaySurface[1]	= pjScreen + lpSubDevice->dwSUBPhysicalAddr[1];
		 
            lpSubDevice->gdwSUBSrcWidth = dwWidth;
            lpSubDevice->gdwSUBSrcHeight= dwHeight;
            lpSubDevice->dwPitch        = dwPitch;
            lpOverlayRecordMPEG->dwSPPitch  = dwPitch;

            /* Subpicture Frame Buffer Stride */
            vfCMHQVW(SUBP_CONTROL_STRIDE, SUBP_AI44|dwPitch,pVidData); 

            lpMPGDevice->dwSUBPPitch  = dwPitch;
            lpMPGDevice->byDeviceType = DEV_SUBP;
            lpMPGDevice->gdwSUBP_NotVisible = 1;

            break;

        case FOURCC_ALPHA :
            DBG_DD(("    Create ALPHA Surface: \n"));
            switch ( lpVIAGraphicInfo->dwDeviceID )
            {
                case VIA_DEVICE_VT3259:
                case VIA_DEVICE_VT3324:
                case VIA_DEVICE_VT3293:  
                case VIA_DEVICE_VT3327: 
                case VIA_DEVICE_VT3336:     
                case VIA_DEVICE_VT3364:    
				case VIA_DEVICE_VT3353:
                    dwPitch  = ALIGN_TO_16_BYTE(lpDDSurfaceDesc->dwWidth);
                    break;
                default:
                    dwPitch  = ALIGN_TO_16_BYTE(lpDDSurfaceDesc->dwWidth/2);
                    break;
            }


            dwWidth=dwHeight=dwAlignHeight=0;
            dwSize=CalSurfaceSize(lpDDSurfaceDesc,ALPHA_REQ_SURFACE_ID,1,
                    &dwWidth,&dwHeight,&dwPitch,&dwAlignHeight,&ALPHAFBSIZE);
            dwAddr = viaAllocVideoSurface(pVidData, ALPHA_REQ_SURFACE_ID, dwSize);

            if (dwAddr==0)
            {
                DBG_DD(("Memory allocation for Alpha fails\n"));
                return VMI_ERR_CANNOT_CREATE_SURFACE;
            }
            lpAlpDevice->dwALPPhysicalAddr   = dwAddr;
            lpAlpDevice->lpALPOverlaySurface = pjScreen+dwAddr;
            lpAlpDevice->dwPitch = dwPitch;
            lpAlpDevice->gdwALPDstHeight = dwHeight;
            lpAlpDevice->gdwALPDstWidth  = dwWidth ;
            lpAlpDevice->gdwALPDstLeft   = 0;
            lpAlpDevice->gdwALPDstTop    = 0;

            break;

       case FOURCC_YUY2 :
            /* init Video status flag*/

            if (!lpVIAGraphicInfo->IsSurfaceUsedForRotate)
            {
                if (((pVidData->VidStatus.dwVideoFlagMPEG[0] != 0) &&(pVidData->VidStatus.dwVideoFlagMPEG[0]& VIDEO_1_INUSE)) ||
                    ((pVidData->VidStatus.dwVideoFlagMPEG[1] != 0) &&(pVidData->VidStatus.dwVideoFlagMPEG[1]& VIDEO_1_INUSE)) ||
                    ((pVidData->VidStatus.dwVideoFlagSW[0] != 0) && (pVidData->VidStatus.dwVideoFlagSW[0] & VIDEO_1_INUSE)) ||
                    ((pVidData->VidStatus.dwVideoFlagSW[1] != 0) && (pVidData->VidStatus.dwVideoFlagSW[1] & VIDEO_1_INUSE)) ||
                    ((pVidData->VidStatus.dwVideoFlagTV0 != 0) && (pVidData->VidStatus.dwVideoFlagTV0 & VIDEO_1_INUSE)) || 
                    ((pVidData->VidStatus.dwVideoFlagTV1 != 0) && (pVidData->VidStatus.dwVideoFlagTV1 & VIDEO_1_INUSE)))
                {
                    if (((pVidData->VidStatus.dwVideoFlagMPEG[0] != 0) && !(pVidData->VidStatus.dwVideoFlagMPEG[0]& VIDEO_1_INUSE)) ||
                        ((pVidData->VidStatus.dwVideoFlagMPEG[1] != 0) && !(pVidData->VidStatus.dwVideoFlagMPEG[1]& VIDEO_1_INUSE)) ||
                        ((pVidData->VidStatus.dwVideoFlagSW[0] != 0) && !(pVidData->VidStatus.dwVideoFlagSW[0] & VIDEO_1_INUSE)) || 
                        ((pVidData->VidStatus.dwVideoFlagSW[1] != 0) && !(pVidData->VidStatus.dwVideoFlagSW[1] & VIDEO_1_INUSE)) ||
                        ((pVidData->VidStatus.dwVideoFlagTV0 != 0) && !(pVidData->VidStatus.dwVideoFlagTV0 & VIDEO_1_INUSE)) || 
                        ((pVidData->VidStatus.dwVideoFlagTV1 != 0) && !(pVidData->VidStatus.dwVideoFlagTV1 & VIDEO_1_INUSE)))
                    {
                        DBG_DD((" No more Video Engine for YUY2!! \n"));                             
                        return VMI_ERR;
                    }
                    else
                    {
                        pVidData->VidStatus.dwVideoFlagSW[handle] = VIDEO_ACTIVE | VIDEO_SHOW | VIDEO_3_INUSE;
                        DBG_DD(("    Create YUY2 Surface: VIDEO_3_INUSE !!!!\n"));    
                        pVidData->VidStatus.dwVideoFlagSW[handle] |= VIDEO_HQV_INUSE | SW_USE_HQV;
                        lpVideoControl->dwHighQVDO = VW_HIGHQVDO_SW;
                        DBG_DD(("    Create YUY2 Surface: HQV_INUSE !!!!\n"));    
                        if(lpVideoHWDifference->dwSupportVideoVQ == VID_HWDIFF_TRUE)
                        {
                            pVidData->VidStatus.dwVideoFlagSW[handle] |= VIDEO_VQ_INUSE;
                            DBG_DD(("    Create YUY2 Surface: VIDEO use VQ !!!!\n"));
                        }
                    }
                }
                else
                {
                    pVidData->VidStatus.dwVideoFlagSW[handle] = VIDEO_ACTIVE | VIDEO_SHOW | VIDEO_1_INUSE;                
                    DBG_DD(("    Create YUY2 Surface: VIDEO_1_INUSE !!!!\n")); 
                    pVidData->VidStatus.dwVideoFlagSW[handle] |= VIDEO_HQV_INUSE | SW_USE_HQV;

                    if(lpVIAGraphicInfo->duoview &DUOVIEW_ENABLED)
                    {
                        switch ( lpVIAGraphicInfo->dwDeviceID )
                        {
                            case VIA_DEVICE_VT3336:
                            case VIA_DEVICE_VT3327:
                            case VIA_DEVICE_VT3364:
                                pVidData->VidStatus.dwVideoFlagSW[handle] &= ~VIDEO_1_ON_SND;
                                break;
                            default:
                                pVidData->VidStatus.dwVideoFlagSW[handle] |= VIDEO_1_ON_SND;
                                break;
                        }
                        
                        if(lpVIAGraphicInfo->Screen[ScrnIndex].dwExpand)
                        {
                            pVidData->VDevInfo.dwVDevFlag |= CTL_UseExpand|CTL_UseIGA2Path;
                        }
                    }

                    lpVideoControl->dwHighQVDO = VW_HIGHQVDO_SW;
                    DBG_DD(("    Create YUY2 Surface: HQV_INUSE !!!!\n"));    
                    if(lpVideoHWDifference->dwSupportVideoVQ == VID_HWDIFF_TRUE)
                    {
                        pVidData->VidStatus.dwVideoFlagSW[handle] |= VIDEO_VQ_INUSE;
                        DBG_DD(("    Create YUY2 Surface: VIDEO use VQ !!!!\n"));
                    }
                }
                /*synv Vidctl and VidStatus stucture*/
                pVidData->VidCtl.dwVideoFlag[handle]=pVidData->VidStatus.dwVideoFlagSW[handle];
            }
            
            /*write Color Space Conversion param.*/                          
            SetColorSpace(lpVIAGraphicInfo,pjVideo);

            dwBackBufferCount = lpDDSurfaceDesc->dwBackBufferCount;

            dwWidth=dwHeight=dwPitch=dwAlignHeight=0;
            dwSize=CalSurfaceSize(lpDDSurfaceDesc,SW_REQ_SURFACE_ID,2,&dwWidth,
                   &dwHeight,&dwPitch,&dwAlignHeight,&SWFBSIZE); 
#ifdef XF86DRI
            if (lpVIAGraphicInfo->DRMEnabled)
            {
                if((drm_fd = open("/dev/dri/card0",O_RDWR)) < 0)
                {
                    DBG_DD(("DRM Device for via could not be opened.\n"));
                    return BadAccess;
                }
                DBG_DD((" FOURCC_YUY2: Open DRM Device! \n"));
            }	
#endif

#ifdef CS_FLAG_HWROTATION /* add for video rotation */
            if(lpVIAGraphicInfo->IsSurfaceUsedForRotate)
            {
                DBG_DD(("    Create SW YUY2 Surface: for rotate. \n"));
                lpDPFsrc = &pVidData->DPFsrc_rot;
                lpSWDevice = &(pVidData->SWDevice_rot[handle]);
                dwAddr = viaAllocVideoSurface(pVidData, ROT_SW_REQ_SURFACE_ID, dwSize);
            }
            else
#endif /* CS_FLAG_HWROTATION */
            {
                DBG_DD(("    Create SW YUY2 Surface: \n"));
                dwAddr = viaAllocVideoSurface(pVidData, SW_REQ_SURFACE_ID, dwSize);
            }        

            if (dwAddr==0)
            {
                DBG_DD(("Memory allocation for SW YUY2 fails\n"));
                return VMI_ERR_CANNOT_CREATE_SURFACE;
            }

            lpDPFsrc->dwFlags = DDPF_FOURCC;
            lpDPFsrc->dwFourCC = FOURCC_YUY2;
            /* fill in the SW buffer with 0x8000 (YUY2-black color) to clear FB buffer */
            lpTmpAddr = pjScreen + dwAddr;
            for(iCount=0;iCount<(SWFBSIZE*2);iCount++)
            {
                if((iCount%2) == 0)
                    *lpTmpAddr++=0x00;
                else
                {
                    *lpTmpAddr=0x80;
#ifdef CS_FLAG_HWROTATION /* for video rotation */
		            if(lpVIAGraphicInfo->rotate)
            		    *(lpTmpAddr)=0x00; /* fill the buffer with rgb blcak */
#endif /* CS_FLAG_HWROTATION */		
		            lpTmpAddr++;
                }
            }

            lpSWDevice->dwSWPhysicalAddr[0]   = dwAddr;
            lpSWDevice->lpSWOverlaySurface[0] = pjScreen+dwAddr;

            lpSWDevice->dwSWPhysicalAddr[1] = lpSWDevice->dwSWPhysicalAddr[0] + SWFBSIZE;
            lpSWDevice->lpSWOverlaySurface[1] = lpSWDevice->lpSWOverlaySurface[0] + SWFBSIZE;

            DBG_DD(("lpSWDevice->dwSWPhysicalAddr[0] %08lx\n", dwAddr));
            DBG_DD(("lpSWDevice->dwSWPhysicalAddr[1] %08lx\n", dwAddr + SWFBSIZE));

            lpSWDevice->gdwSWSrcWidth = lpSWDevice->dwWidth = dwWidth;
            lpSWDevice->gdwSWSrcHeight = lpSWDevice->dwHeight = dwHeight;
            lpSWDevice->dwPitch = dwPitch;

            if(!lpVIAGraphicInfo->IsSurfaceUsedForRotate)
            {
                /* Fill image data in overlay record */
                lpOverlayRecordSW->dwOriSrcWidth  = dwWidth;
                lpOverlayRecordSW->dwOriSrcHeight = dwHeight;
                lpOverlayRecordSW->dwOriSrcPitch  = dwPitch;            
                DBG_DD(("Create HQV_YUY2 Surface\n"));

                dwSize=CalSurfaceSize(lpDDSurfaceDesc,HQV_REQ_SURFACE_ID,3,&dwWidth,\
                        &dwHeight,&dwPitch,&dwAlignHeight,&HQVFBSIZE); 
                if ( pVidData->VidStatus.dwVideoFlagSW[handle] & VIDEO_1_INUSE)
                {
                    dwAddr = viaAllocVideoSurface(pVidData, HQV_REQ_SURFACE_ID, dwSize);
                }
                else
                {
                    dwAddr = viaAllocVideoSurface(pVidData, HQV2_REQ_SURFACE_ID, dwSize);
                }
                
                if (dwAddr ==0)
                {
                    DBG_DD((" Memory allocation for SW YUY2 HQV fails .\n"));
                    return VMI_ERR_CANNOT_CREATE_SURFACE;
                }

    	        lpSWDevice->dwHQVAddr[0] = dwAddr;
                lpSWDevice->dwHQVAddr[1] = dwAddr + HQVFBSIZE;
                lpSWDevice->dwHQVAddr[2] = lpSWDevice->dwHQVAddr[1] + HQVFBSIZE;
      
                lpOverlayRecordSW->dwHQVAddr[0] = dwAddr;
                lpOverlayRecordSW->dwHQVAddr[1] = dwAddr + HQVFBSIZE;
                lpOverlayRecordSW->dwHQVAddr[2] = lpOverlayRecordSW->dwHQVAddr[1] + HQVFBSIZE;
                    
                if (lpOverlayRecordSW->dwHQVAddr[2] + HQVFBSIZE >=
                                            (unsigned long)lpVIAGraphicInfo->Screen[0].VideoHeapEnd)
                {
                    DBG_DD(("Memory not enough for MPEG with HQV\n"));
                    return VMI_ERR_CANNOT_CREATE_SURFACE;
                }

                /*fill in the HQV buffer with 0x8000 (YUY2-black color) to clear HQV buffers*/
                for(iCount=0;iCount<HQVFBSIZE*3;iCount++)
                {
                    lpTmpAddr = pjScreen + dwAddr + iCount;
                    if((iCount%2) == 0)
                        *(lpTmpAddr)=0x00;
                    else
                        *(lpTmpAddr)=0x80;
#ifdef CS_FLAG_HWROTATION /* for video rotation */
    		        if(lpVIAGraphicInfo->rotate)
                		*(lpTmpAddr)=0x00; /* fill the buffer with rgb blcak */
#endif /* CS_FLAG_HWROTATION */					
                }
                
                pVidData->VidCtl.dwVideoFlag[handle] = pVidData->VidStatus.dwVideoFlagSW[handle];

                DBG_DD(("shut down SUBP\n"));
                VIDOutD(SUBP_CONTROL_STRIDE,vfCMHQVR(SUBP_CONTROL_STRIDE,pVidData)&~SUBP_HQV_ENABLE);
                if (pVidData->VidCtl.dwVideoFlag[handle] & VIDEO_VQ_INUSE)
                {
                    vfEnableVQ_CME(pVidData);
                }
            }
            break;

        case FOURCC_YV12 :
            DBG_DD(("    Create SW YV12 Surface: \n"));
            lpDPFsrc->dwFlags = DDPF_FOURCC;
            lpDPFsrc->dwFourCC = FOURCC_YV12;

            /* init Video status flag*/
       	    if ( ( (pVidData->VidStatus.dwVideoFlagMPEG[0] != 0) &&(pVidData->VidStatus.dwVideoFlagMPEG[0]& VIDEO_1_INUSE) )||
                 ( (pVidData->VidStatus.dwVideoFlagMPEG[1] != 0) &&(pVidData->VidStatus.dwVideoFlagMPEG[1]& VIDEO_1_INUSE) ) ||                		
                 ( (pVidData->VidStatus.dwVideoFlagTV0 != 0) && (pVidData->VidStatus.dwVideoFlagTV0 & VIDEO_1_INUSE) ) || 
                 ( (pVidData->VidStatus.dwVideoFlagTV1 != 0) && (pVidData->VidStatus.dwVideoFlagTV1 & VIDEO_1_INUSE) ) )
            {
                if ( ( (pVidData->VidStatus.dwVideoFlagMPEG[0] != 0) && !(pVidData->VidStatus.dwVideoFlagMPEG[0]& VIDEO_1_INUSE) )||
                     ( (pVidData->VidStatus.dwVideoFlagMPEG[1] != 0) && !(pVidData->VidStatus.dwVideoFlagMPEG[1]& VIDEO_1_INUSE) ) ||                		
                     ( (pVidData->VidStatus.dwVideoFlagTV0 != 0) && !(pVidData->VidStatus.dwVideoFlagTV0 & VIDEO_1_INUSE) ) || 
                     ( (pVidData->VidStatus.dwVideoFlagTV1 != 0) && !(pVidData->VidStatus.dwVideoFlagTV1 & VIDEO_1_INUSE) ) )
                {
                    DBG_DD((" No more Video Engine for YV12!! \n")); 
                    return VMI_ERR;
                }
                else
                {
                    pVidData->VidStatus.dwVideoFlagSW[handle] = VIDEO_ACTIVE | VIDEO_SHOW | VIDEO_3_INUSE;
                    DBG_DD(("    Create YV12 Surface: VIDEO_3_INUSE !!!!\n"));    
                    pVidData->VidStatus.dwVideoFlagSW[handle] |= VIDEO_HQV_INUSE | SW_USE_HQV;                            
                    lpVideoControl->dwHighQVDO = VW_HIGHQVDO_SW;
                    DBG_DD(("    Create YV12 Surface: HQV_INUSE !!!!\n"));                                
  
                }
            }
            else
            {
                pVidData->VidStatus.dwVideoFlagSW[handle] = VIDEO_ACTIVE | VIDEO_SHOW | VIDEO_1_INUSE;                
                DBG_DD(("    Create YV12 Surface: VIDEO_1_INUSE !!!!\n")); 
                pVidData->VidStatus.dwVideoFlagSW[handle] |= VIDEO_HQV_INUSE | SW_USE_HQV;                            
                lpVideoControl->dwHighQVDO = VW_HIGHQVDO_SW;
                DBG_DD(("    Create YV12 Surface: HQV_INUSE !!!!\n"));                                
  
            }
                                        
            /* sync VidStatus and VidCtl structure */
            pVidData->VidCtl.dwVideoFlag[handle]=pVidData->VidStatus.dwVideoFlagSW[handle];                
            /* write Color Space Conversion param. */        
            SetColorSpace(lpVIAGraphicInfo,pjVideo);
            dwBackBufferCount = lpDDSurfaceDesc->dwBackBufferCount;
            
#ifdef XF86DRI
            if (lpVIAGraphicInfo->DRMEnabled)
            {
                if((drm_fd = open("/dev/dri/card0",O_RDWR)) < 0)
                {
                    DBG_DD(("DRM Device for via could not be opened.\n"));
                    return BadAccess;
                }
                DBG_DD((" FOURCC_YV12: Open DRM Device! \n"));
            }
#endif
            dwWidth=dwHeight=dwPitch=dwAlignHeight=0;
            dwSize=CalSurfaceSize(lpDDSurfaceDesc,SW_REQ_SURFACE_ID,2,&dwWidth,     
                    &dwHeight,&dwPitch,&dwAlignHeight,&SWFBSIZE);
            dwAddr = viaAllocVideoSurface(pVidData, SW_REQ_SURFACE_ID, dwSize);

            if (dwAddr ==0)
            {
                DBG_DD((" Memory allocation  for SW YV12 fails.\n"));
                return VMI_ERR_CANNOT_CREATE_SURFACE;
            }

            /* fill in the SW buffer with 0x8000 (YUY2-black color) to clear FB buffer */
            lpTmpAddr = pjScreen + dwAddr;
            for(iCount=0;iCount<(SWFBSIZE*2);iCount++)
            {
                if((iCount%2) == 0)
                    *lpTmpAddr++=0x00;
                else
                    *lpTmpAddr++=0x80;
            }

            lpSWDevice->dwSWPhysicalAddr[0]   = dwAddr;
            lpSWDevice->dwSWCrPhysicalAddr[0] = lpSWDevice->dwSWPhysicalAddr[0]+ (dwPitch*dwAlignHeight);
            lpSWDevice->dwSWCbPhysicalAddr[0] = lpSWDevice->dwSWCrPhysicalAddr[0]+ ((dwPitch>>1)*(dwAlignHeight>>1));
            lpSWDevice->lpSWOverlaySurface[0] = pjScreen+dwAddr;

            lpSWDevice->dwSWPhysicalAddr[1] = dwAddr + SWFBSIZE;
            lpSWDevice->dwSWCrPhysicalAddr[1] = lpSWDevice->dwSWPhysicalAddr[1] + (dwPitch*dwAlignHeight);
            lpSWDevice->dwSWCbPhysicalAddr[1] = lpSWDevice->dwSWCrPhysicalAddr[1]+ ((dwPitch>>1)*(dwAlignHeight>>1));
            lpSWDevice->lpSWOverlaySurface[1] = lpSWDevice->lpSWOverlaySurface[0] + SWFBSIZE;

            DBG_DD(("lpSWDevice->dwSWPhysicalAddr[0] %08lx\n", dwAddr));
            DBG_DD(("lpSWDevice->dwSWPhysicalAddr[1] %08lx\n", dwAddr + SWFBSIZE));

            lpSWDevice->gdwSWSrcWidth = dwWidth;
            lpSWDevice->gdwSWSrcHeight= dwHeight;
            lpSWDevice->dwPitch = dwPitch;

            /* Fill image data in overlay record */
            lpOverlayRecordSW->dwOriSrcWidth  = dwWidth;
            lpOverlayRecordSW->dwOriSrcHeight = dwHeight;
            lpOverlayRecordSW->dwOriSrcPitch  = dwPitch;

            /* 
             *  if sw video use HQV dwpitch should changed
             */ 
            DBG_DD(("Create HQV_YV12 Surface\n"));
            dwPitch  = lpSWDevice->dwPitch * 2; 

            dwSize=CalSurfaceSize(lpDDSurfaceDesc,HQV_REQ_SURFACE_ID,3,&dwWidth, 
                    &dwHeight,&dwPitch,&dwAlignHeight,&HQVFBSIZE);
            if ( pVidData->VidStatus.dwVideoFlagSW[handle] &VIDEO_1_INUSE)
            {
                dwAddr = viaAllocVideoSurface(pVidData, HQV_REQ_SURFACE_ID, dwSize);
            }
            else
            {
                dwAddr = viaAllocVideoSurface(pVidData, HQV2_REQ_SURFACE_ID, dwSize);
            }

            if (dwAddr ==0)
            {
                DBG_DD((" Memory allocation  for SW YV12 HQV fails.\n"));
                return VMI_ERR_CANNOT_CREATE_SURFACE;
            }

            lpOverlayRecordSW->dwHQVAddr[0] = dwAddr;
            lpOverlayRecordSW->dwHQVAddr[1] = dwAddr + HQVFBSIZE;            

            lpOverlayRecordSW->dwHQVAddr[2] = lpOverlayRecordSW->dwHQVAddr[1] + HQVFBSIZE;
                
            if (lpOverlayRecordSW->dwHQVAddr[2] + HQVFBSIZE >=
                        (unsigned long)lpVIAGraphicInfo->Screen[0].VideoHeapEnd)
            {
                DBG_DD(("Memory not enough for MPEG with HQV\n"));
                return VMI_ERR_CANNOT_CREATE_SURFACE;
            }

            /*fill in the HQV buffer with 0x8000 (YUY2-black color) to clear HQV buffers*/
            for(iCount=0;iCount<HQVFBSIZE*2;iCount++)
            {
                lpTmpAddr = pjScreen + dwAddr + iCount;
                if((iCount%2) == 0)
                    *(lpTmpAddr)=0x00;
                else
                    *(lpTmpAddr)=0x80;
            }
            
            pVidData->VidCtl.dwVideoFlag[handle] = pVidData->VidStatus.dwVideoFlagSW[handle];
            break;
       case FOURCC_NV12 :
            DBG_DD(("    Create SW NV12 Surface: \n"));
            lpDPFsrc->dwFlags = DDPF_FOURCC;
            lpDPFsrc->dwFourCC = FOURCC_NV12;/*for support NV12 overlay mode test-1*/

            lpSWDevice  = &(pVidData->SWDevice[handle]);
            lpOverlayRecordSW = &pVidData->OverlayRecordSet.OverlayRecordSW[handle];

            /* init Video status flag*/
            if ( ( (pVidData->VidStatus.dwVideoFlagSW[0] != 0) && (pVidData->VidStatus.dwVideoFlagSW[0] & VIDEO_1_INUSE) ) ||
                ( (pVidData->VidStatus.dwVideoFlagSW[1] != 0) && (pVidData->VidStatus.dwVideoFlagSW[1] & VIDEO_1_INUSE) ) ||
                ( (pVidData->VidStatus.dwVideoFlagTV0 != 0) && (pVidData->VidStatus.dwVideoFlagTV0 & VIDEO_1_INUSE) ) ||
                ( (pVidData->VidStatus.dwVideoFlagTV1 != 0) &&(pVidData->VidStatus.dwVideoFlagTV1 & VIDEO_1_INUSE) ) ||
                ( (pVidData->VidStatus.dwVideoFlagMPEG[0] != 0) &&(pVidData->VidStatus.dwVideoFlagMPEG[0]& VIDEO_1_INUSE) )||
               ( (pVidData->VidStatus.dwVideoFlagMPEG[1] != 0) &&(pVidData->VidStatus.dwVideoFlagMPEG[1]& VIDEO_1_INUSE) ) )
             {
                 if ( ( (pVidData->VidStatus.dwVideoFlagSW[0] != 0) && !(pVidData->VidStatus.dwVideoFlagSW[0] & VIDEO_1_INUSE) ) || 
                     ( (pVidData->VidStatus.dwVideoFlagSW[1] != 0) && !(pVidData->VidStatus.dwVideoFlagSW[1] & VIDEO_1_INUSE) ) || 
                     ( (pVidData->VidStatus.dwVideoFlagTV0 != 0) && !(pVidData->VidStatus.dwVideoFlagTV0 & VIDEO_1_INUSE) ) || 
                    ( (pVidData->VidStatus.dwVideoFlagTV1 != 0) && !(pVidData->VidStatus.dwVideoFlagTV1 & VIDEO_1_INUSE) )|| 
                    ( (pVidData->VidStatus.dwVideoFlagMPEG[0] != 0) && !(pVidData->VidStatus.dwVideoFlagMPEG[0]& VIDEO_1_INUSE) )||
                    ( (pVidData->VidStatus.dwVideoFlagMPEG[1] != 0) && !(pVidData->VidStatus.dwVideoFlagMPEG[1]& VIDEO_1_INUSE) ) )
                    {
                        DBG_DD((" No more Video Engine for NV12!! \n")); 
                        return VMI_ERR;
                    }
                    else
                    {
                        pVidData->VidStatus.dwVideoFlagSW[handle] = VIDEO_ACTIVE | VIDEO_SHOW | VIDEO_3_INUSE| VIDEO_HQV_INUSE | SW_USE_HQV;     
                        lpVideoControl->dwHighQVDO = VW_HIGHQVDO_SW;
                        DBG_DD(("    Create NV12 Surface: VIDEO_3_INUSE !!!!\n"));     
                        DBG_DD(("    Create NV12 Surface: HQV_INUSE !!!!\n"));                                
                        if(lpVideoHWDifference->dwSupportVideoVQ == VID_HWDIFF_TRUE)
                        {
                            pVidData->VidStatus.dwVideoFlagSW[handle] |= VIDEO_VQ_INUSE;
                            DBG_DD(("    Create NV12 Surface: VIDEO use VQ !!!!\n"));
                        }
                    }
             }
             else
             {
                pVidData->VidStatus.dwVideoFlagSW[handle] = VIDEO_ACTIVE | VIDEO_SHOW | VIDEO_1_INUSE|VIDEO_HQV_INUSE | SW_USE_HQV;               
                lpVideoControl->dwHighQVDO = VW_HIGHQVDO_SW;
                DBG_DD(("    Create NV12 Surface: VIDEO_1_INUSE !!!!\n")); 
                DBG_DD(("    Create NV12 Surface: HQV_INUSE !!!!\n"));                                
                if(lpVideoHWDifference->dwSupportVideoVQ == VID_HWDIFF_TRUE)
                {
                    pVidData->VidStatus.dwVideoFlagSW[handle] |= VIDEO_VQ_INUSE;
                    DBG_DD(("    Create NV12 Surface: VIDEO use VQ !!!!\n"));
                }
            }
            /* sync Ctl and Status structure*/                                        
            pVidData->VidCtl.dwVideoFlag[handle]=pVidData->VidStatus.dwVideoFlagSW[handle];                         
            /* write Color Space Conversion param. */        
            SetColorSpace(lpVIAGraphicInfo,pjVideo);
            dwBackBufferCount = lpDDSurfaceDesc->dwBackBufferCount;

#ifdef XF86DRI
            if (lpVIAGraphicInfo->DRMEnabled)
            {
                if((drm_fd = open("/dev/dri/card0",O_RDWR)) < 0)
                {
                    DBG_DD(("DRM Device for via could not be opened.\n"));
                    return BadAccess;
                }
                DBG_DD((" FOURCC_NV12: Open DRM Device! \n"));
            }
#endif

            dwWidth=dwHeight=dwPitch=dwAlignHeight=0;
            dwSize=CalSurfaceSize(lpDDSurfaceDesc,SW_REQ_SURFACE_ID,2,
                    &dwWidth,&dwHeight,&dwPitch,&dwAlignHeight,&SWFBSIZE);
            dwAddr = viaAllocVideoSurface(pVidData, SW_REQ_SURFACE_ID, dwSize);

            if (dwAddr ==0)
            {
                DBG_DD((" Memory allocation for SW NV12 fails.\n"));
                return VMI_ERR_CANNOT_CREATE_SURFACE;
            }

            /* fill in the SW buffer with 0x8000 (YUY2-black color) to clear FB buffer */
            lpTmpAddr = pjScreen + dwAddr;
            for(iCount=0;iCount<(SWFBSIZE*2);iCount++)
            {
                if((iCount%2) == 0)
                    *lpTmpAddr++=0x00;
                else
                    *lpTmpAddr++=0x80;
            }

            lpSWDevice->dwSWPhysicalAddr[0]   = dwAddr;
            lpSWDevice->dwSWCbPhysicalAddr[0] = lpSWDevice->dwSWPhysicalAddr[0]   + (dwPitch*dwAlignHeight);
            lpSWDevice->dwSWCrPhysicalAddr[0] = lpSWDevice->dwSWCbPhysicalAddr[0]  + ((dwPitch>>1)*(dwAlignHeight>>1));
            lpSWDevice->lpSWOverlaySurface[0] = pjScreen+dwAddr;

            lpSWDevice->dwSWPhysicalAddr[1] = dwAddr + SWFBSIZE;
            lpSWDevice->dwSWCbPhysicalAddr[1] = lpSWDevice->dwSWPhysicalAddr[1] + (dwPitch*dwAlignHeight);
            lpSWDevice->dwSWCrPhysicalAddr[1] = lpSWDevice->dwSWCbPhysicalAddr[1] + ((dwPitch>>1)*(dwAlignHeight>>1));
            lpSWDevice->lpSWOverlaySurface[1] = lpSWDevice->lpSWOverlaySurface[0] + SWFBSIZE;

            DBG_DD(("lpSWDevice->dwSWPhysicalAddr[0]=%08lx\n", dwAddr));
            DBG_DD(("lpSWDevice->dwSWPhysicalAddr[1]=%08lx\n", dwAddr + SWFBSIZE));

            lpSWDevice->dwWidth  = dwWidth;
            lpSWDevice->dwHeight = dwHeight; 
            lpSWDevice->dwPitch  = dwPitch;  

            lpSWDevice->gdwSWSrcWidth = dwWidth;
            lpSWDevice->gdwSWSrcHeight= dwHeight;
            lpSWDevice->dwPitch = dwPitch;

            /* Fill image data in overlay record */
            lpOverlayRecordSW->dwOriSrcWidth  = dwWidth;
            lpOverlayRecordSW->dwOriSrcHeight = dwHeight;
            lpOverlayRecordSW->dwOriSrcPitch  = dwPitch;


            /* 
             *  if sw video use HQV dwpitch should changed
             */ 
            DBG_DD(("Create HQV_NV12 Surface\n"));
            dwPitch  = lpSWDevice->dwPitch*2; 

            dwSize=CalSurfaceSize(lpDDSurfaceDesc,HQV_REQ_SURFACE_ID,3,&dwWidth, 
                    &dwHeight,&dwPitch,&dwAlignHeight,&HQVFBSIZE);
            if ( pVidData->VidStatus.dwVideoFlagSW[handle] &VIDEO_1_INUSE)
            {
                dwAddr = viaAllocVideoSurface(pVidData, HQV_REQ_SURFACE_ID, dwSize);
            }
            else
            {
                dwAddr = viaAllocVideoSurface(pVidData, HQV2_REQ_SURFACE_ID, dwSize);
            }

            if (dwAddr ==0)
            {
                DBG_DD((" Memory allocation for SW NV12 HQV fails.\n"));
                return VMI_ERR_CANNOT_CREATE_SURFACE;
            }

            lpSWDevice->dwHQVAddr[0] = dwAddr;
            lpSWDevice->dwHQVAddr[1] = dwAddr + HQVFBSIZE;
            lpSWDevice->dwHQVAddr[2] = lpSWDevice->dwHQVAddr[1] + HQVFBSIZE;

            lpOverlayRecordSW->dwHQVAddr[0] = dwAddr;
            lpOverlayRecordSW->dwHQVAddr[1] = dwAddr + HQVFBSIZE;            
            lpOverlayRecordSW->dwHQVAddr[2] = lpOverlayRecordSW->dwHQVAddr[1] + HQVFBSIZE;
                
            if (lpOverlayRecordSW->dwHQVAddr[2] + HQVFBSIZE >=
                        (unsigned long)lpVIAGraphicInfo->Screen[0].VideoHeapEnd)
            {
                DBG_DD(("Memory not enough for MPEG with HQV\n"));
                return VMI_ERR_CANNOT_CREATE_SURFACE;
            }

            /*fill in the HQV buffer with 0x8000 (YUY2-black color) to clear HQV buffers*/
            for(iCount=0;iCount<HQVFBSIZE*2;iCount++)
            {
                lpTmpAddr = pjScreen + dwAddr + iCount;
                if((iCount%2) == 0)
                    *(lpTmpAddr)=0x00;
                else
                    *(lpTmpAddr)=0x80;
            }
            
            pVidData->VidCtl.dwVideoFlag[handle] = pVidData->VidStatus.dwVideoFlagSW[handle];
            DBG_DD(("for NV12 pVidData->VidCtl.dwVideoFlag[handle]=%p\n",pVidData->VidCtl.dwVideoFlag[handle]));

            if (pVidData->VidCtl.dwVideoFlag[handle] & VIDEO_VQ_INUSE)
            {
                vfEnableVQ_CME(pVidData);
            }
            break;

        default :
            break;
    }

    return dwRet;

} /* CreateSurface */




/*
 *  Destroy Surface
 */
unsigned long DestroySurface_CME(PVIDDATA pVidData, LPDDSURFACEDESC lpDDSurfaceDesc)
{   
    unsigned long handle=pVidData->VMI_Info.handle;
    LPVIDEOMEMORY lpVidMemory = &VidMemory;
    LPVIAGRAPHICINFO lpVIAGraphicInfo = &pVidData->VIAGraphicInfo;
    LPDDPIXELFORMAT lpDPFsrc = &pVidData->DPFsrc[handle];
    LPMPGDEVICE lpMPGDevice=&pVidData->MPGDevice[handle];
    LPSUBDEVICE lpSubDevice = &(pVidData->ddLock.SubDev[handle]);
    LPALPHADEVICE lpAlpDevice = &(pVidData->ddLock.ALPDevice);
    LPSWDEVICE  lpSWDevice = &(pVidData->SWDevice[handle]);
    
    DBG_DD(("\n"));

    if(lpVIAGraphicInfo->duoview)  
    {
    	switch (lpDDSurfaceDesc->dwFourCC)
    	{
        case FOURCC_NV12 :
		case FOURCC_YUY2:
		case FOURCC_YV12:
            lpVIAGraphicInfo->bVideoEngineUsedOut=0;
			break;
		default:
			break;
    	}
    }	


    switch (lpDDSurfaceDesc->dwFourCC)
    {
        case FOURCC_SUBP :
            DBG_DD((" Destroy SUBP Surface! \n"));
            DBG_DD(("Disable SUBP engin first\n"));
            vfCMHQVW(SUBP_CONTROL_STRIDE, vfCMHQVR(SUBP_CONTROL_STRIDE,pVidData)&~SUBP_HQV_ENABLE,pVidData);
            vfCMHQVR(SUBP_CONTROL_STRIDE,pVidData);

            if (lpSubDevice->dwSUBPhysicalAddr[0])
            {
                viaFreeVideoSurface(pVidData, SUBP_REQ_SURFACE_ID);
                lpSubDevice->dwSUBPhysicalAddr[0] = 0;
            }

            memset(&pVidData->ddLock.SubDev[handle],0,sizeof(SUBDEVICE));
            break;

       case FOURCC_ALPHA :
            DBG_DD((" Destroy ALPHA Surface! \n"));

            if (lpAlpDevice->dwALPPhysicalAddr)
            {
                viaFreeVideoSurface(pVidData, ALPHA_REQ_SURFACE_ID);
                lpAlpDevice->dwALPPhysicalAddr = 0;
            }

            break;

       case FOURCC_YUY2 :
#ifdef CS_FLAG_HWROTATION /* for video rotation */
           if (lpVIAGraphicInfo->IsSurfaceUsedForRotate)
            {
                lpDPFsrc = &pVidData->DPFsrc_rot;
                lpSWDevice = &(pVidData->SWDevice_rot[handle]);
                
                DBG_DD((" Destroy YUY2 Surface: for rotate! \n"));
                
                if (lpSWDevice->dwSWPhysicalAddr[0])
                {
                    viaFreeVideoSurface(pVidData, ROT_SW_REQ_SURFACE_ID);
                    lpSWDevice->dwSWPhysicalAddr[0] = 0;
                }
            }
            else
#endif /* CS_FLAG_HWROTATION */
            {
                DBG_DD((" Destroy YUY2 Surface! \n"));

                if (lpSWDevice->dwSWPhysicalAddr[0])
                {
                    viaFreeVideoSurface(pVidData, SW_REQ_SURFACE_ID);
                    lpSWDevice->dwSWPhysicalAddr[0] = 0;
                }
            }

            lpDPFsrc->dwFlags = 0;
            lpDPFsrc->dwFourCC = 0;
            
            pVidData->VidStatus.dwVideoFlagSW[handle] &= ~VIDEO_ACTIVE;

            if (lpSWDevice->dwHQVAddr[0])
            {
                if ( pVidData->VidStatus.dwVideoFlagSW[handle] &VIDEO_1_INUSE)
                {
                    viaFreeVideoSurface(pVidData, HQV_REQ_SURFACE_ID);
                }
                else
                {
                    viaFreeVideoSurface(pVidData, HQV2_REQ_SURFACE_ID);
                }

                lpSWDevice->dwHQVAddr[0] = 0;
            }

#ifdef XF86DRI
            if (lpVIAGraphicInfo->DRMEnabled)
            {
                drmClose(drm_fd);
                DBG_DD((" FOURCC_YUY2: Close DRM Device! \n"));
            }
#endif

            if (pVidData->VidStatus.dwVideoFlagSW[handle] & VIDEO_VQ_INUSE)
            {
                vfDisableVQ_CME(pVidData);
            }
            
            if (!(pVidData->VidStatus.dwVideoFlagSW[handle] & VIDEO_ACTIVE))
                pVidData->VidStatus.dwVideoFlagSW[handle] = 0;
            if(lpVIAGraphicInfo->duoview)
            {
                pVidData->VidStatus.dwVideoFlagSW[handle] &= ~VIDEO_1_ON_SND;
                if(lpVIAGraphicInfo->Screen[0].dwExpand)
                    pVidData->VDevInfo.dwVDevFlag &= ~(CTL_UseExpand|CTL_UseIGA2Path);
            }
            pVidData->VidStatus.dwVideoFlagSW[handle]=pVidData->VidCtl.dwVideoFlag[handle]=0;
            memset(lpSWDevice,0,sizeof(SWDEVICE));
	        memset(&pVidData->OverlayRecordSet.OverlayRecordSW[handle],0,sizeof(OVERLAYRECORD));
            break;

       case FOURCC_YV12 :
            DBG_DD((" Destroy YV12 Surface! \n"));            
            lpDPFsrc->dwFlags = 0;
            lpDPFsrc->dwFourCC = 0;

            if (lpSWDevice->dwSWPhysicalAddr[0])
            {
                viaFreeVideoSurface(pVidData, SW_REQ_SURFACE_ID);
                lpSWDevice->dwSWPhysicalAddr[0] = 0;
            }

            pVidData->VidStatus.dwVideoFlagSW[handle] &= ~VIDEO_ACTIVE;

            if (lpSWDevice->dwHQVAddr[0])
            {
                if ( pVidData->VidStatus.dwVideoFlagSW[handle] & VIDEO_1_INUSE)
                {
                    viaFreeVideoSurface(pVidData, HQV_REQ_SURFACE_ID);
                }
                else
                {
                    viaFreeVideoSurface(pVidData, HQV2_REQ_SURFACE_ID);
                }

                lpSWDevice->dwHQVAddr[0] = 0;
            }

#ifdef XF86DRI
            if (lpVIAGraphicInfo->DRMEnabled)
            {
                drmClose(drm_fd);
                DBG_DD((" FOURCC_YV12: Close DRM Device! \n"));
            }
#endif

            if (!(pVidData->VidStatus.dwVideoFlagSW[handle] & VIDEO_ACTIVE))
                pVidData->VidStatus.dwVideoFlagSW[handle] = 0;           
            
            pVidData->VidCtl.dwVideoFlag[handle]=0;
            memset(&pVidData->SWDevice[handle],0,sizeof(SWDEVICE));
            memset(&pVidData->OverlayRecordSet.OverlayRecordSW[handle],0,sizeof(OVERLAYRECORD));
            break;

       case FOURCC_NV12 :
            DBG_DD((" Destroy NV12 Surface! \n"));            
            lpDPFsrc->dwFlags = 0;
            lpDPFsrc->dwFourCC = 0;

            if (lpSWDevice->dwSWPhysicalAddr[0])
            {
                viaFreeVideoSurface(pVidData, SW_REQ_SURFACE_ID);
                lpSWDevice->dwSWPhysicalAddr[0] = 0;
            }
                
            pVidData->VidStatus.dwVideoFlagSW[handle] &= ~VIDEO_ACTIVE;

            if (lpSWDevice->dwHQVAddr[0])
            {
                if ( pVidData->VidStatus.dwVideoFlagSW[handle] & VIDEO_1_INUSE)
                {
                    viaFreeVideoSurface(pVidData, HQV_REQ_SURFACE_ID);
                }
                else
                {
                    viaFreeVideoSurface(pVidData, HQV2_REQ_SURFACE_ID);
                }
                lpSWDevice->dwHQVAddr[0] = 0;
            }

#ifdef XF86DRI
            if (lpVIAGraphicInfo->DRMEnabled)
            {
                drmClose(drm_fd);
                DBG_DD((" FOURCC_NV12: Close DRM Device! \n"));
            }
#endif            
            
            /*disable VQ should be done before clear the flags
             *otherwise when playing next video right after this one, the system hangs
             */
            if (pVidData->VidStatus.dwVideoFlagSW[handle] & VIDEO_VQ_INUSE)
            {
                vfDisableVQ_CME(pVidData);
            }
            if (!(pVidData->VidStatus.dwVideoFlagSW[handle] & VIDEO_ACTIVE))
                pVidData->VidStatus.dwVideoFlagSW[handle] = 0;           
           
            pVidData->VidCtl.dwVideoFlag[handle]=0;
            memset(&pVidData->SWDevice[handle],0,sizeof(SWDEVICE));
            memset(&pVidData->OverlayRecordSet.OverlayRecordSW[handle],0,sizeof(OVERLAYRECORD));

            break;
        default :
            break;
    }

    return VMI_OK;

} /* DestroySurface */




