/*
 * Copyright 1998-2008 VIA Technologies, Inc. All Rights Reserved.
 * Copyright 2001-2008 S3 Graphics, Inc. All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHOR(S) OR COPYRIGHT HOLDER(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#ifndef __VINLINE__H__
#define __VINLINE__H__

#include "via.h"
#include "vregCME.h"
#include "debug.h"

#define IN_HQV_FIRE     (*((unsigned long volatile *)(pjVideo+HQV_CONTROL))&HQV_IDLE)
#define IN_VIDEO_FIRE   (*((unsigned long volatile *)(pjVideo+V_COMPOSE_MODE))&V1_COMMAND_FIRE)
#define IN_HQV_FLIP     (*((unsigned long volatile *)(pjVideo+HQV_CONTROL))&HQV_FLIP_STATUS) 
#define IN_VIDEO_DISPLAY     (*((unsigned long volatile *)(pjVideo+V_FLAGS))&VBI_STATUS) 


/* I N L I N E -------------------------------------------------------------- */
static __inline void vfWaitLastMBDone(unsigned char* pjVideo)
{    
    unsigned long volatile *pdwState = (unsigned long volatile *) (pjVideo+0x0c04);
    
    while (!(*pdwState & 0x00800000) );
}

static __inline void vfWaitSUBFlip(unsigned char* pjVideo)
{    
    unsigned long volatile *pdwState = (unsigned long volatile *) (pjVideo+HQV_CONTROL);    
    
    while ((*pdwState & HQV_SUB_FLIP) );
}

static __inline void vfWaitHQVIdle(unsigned char* pjVideo)
{
    while (!IN_HQV_FIRE);
}

static __inline void WaitVideoCommandFire(unsigned char* pjVideo)
{
    unsigned long volatile *pdwState = (unsigned long volatile *) (pjVideo+V_COMPOSE_MODE);
    while ((*pdwState & V1_COMMAND_FIRE));
}

static __inline void WaitVideo3CommandFire(unsigned char* pjVideo)
{
    unsigned long volatile *pdwState = (unsigned long volatile *) (pjVideo+V_COMPOSE_MODE);
    while ((*pdwState & V3_COMMAND_FIRE));
}

static __inline void vfWaitHQVFlip(unsigned char* pjVideo)
{
    unsigned long volatile *pdwState = (unsigned long volatile *) pjVideo;
    pdwState = (unsigned long volatile *) (pjVideo+HQV_CONTROL);
    while (!(*pdwState & HQV_FLIP_STATUS) );
}

static __inline void vfWaitHQVFIFOEmpty(unsigned char* pjVideo)
{
    unsigned long volatile *pdwState = (unsigned long volatile *) (pjVideo+HQV_CONTROL); 
    
    while ((*pdwState & HQV_FIFO_STATUS) );
}

static __inline void WaitVBI(unsigned char* pjVideo)
{
    while (IN_VIDEO_DISPLAY);
}

static __inline void WaitHQVDone(unsigned char* pjVideo)
{
    unsigned long volatile *pdwState = (unsigned long volatile *) (pjVideo+HQV_CONTROL);

    while ((*pdwState & HQV_SW_FLIP) );
}

static __inline void UniversalWait(unsigned char* pjVideo, unsigned long dwReg, unsigned long dwFlags, Bool WaitBitSet)
{
    unsigned long volatile *pdwState = (unsigned long volatile *) (pjVideo+dwReg); 

    if (WaitBitSet)
    {
        while (!(*pdwState & dwFlags) );
    }
    else
    {
        while ((*pdwState & dwFlags) );
    }
}

static __inline void WaitHQVFlipClear(unsigned char* pjVideo, unsigned long dwData)
{
    unsigned long volatile *pdwState = (unsigned long volatile *) (pjVideo+HQV_CONTROL);
    *pdwState =dwData;

    while ((*pdwState & HQV_FLIP_STATUS) )
    {
        VIDOutD(HQV_CONTROL, *pdwState|HQV_FLIP_STATUS);
    }
}

/* for VT3259 */
static __inline void vfWaitSUB1Flip(unsigned char* pjVideo)
{    
    unsigned long volatile *pdwState = 
          (unsigned long volatile *) (pjVideo+HQV_CONTROL+REG_HQV1_INDEX);
    
    while ((*pdwState & HQV_SUB_FLIP));
}

static __inline void vfWaitHQV1Done(unsigned char* pjVideo)
{    
    unsigned long volatile *pdwState = 
    (unsigned long volatile *) (pjVideo+HQV_CONTROL+REG_HQV1_INDEX);    
    
    while ((*pdwState & HQV_SW_FLIP));
}

static __inline void vfWaitHQV1FIFOEmpty(unsigned char* pjVideo)
{    
    unsigned long volatile *pdwState = 
    (unsigned long volatile *) (pjVideo+HQV_CONTROL+REG_HQV1_INDEX); 
    
    while ((*pdwState & HQV_FIFO_STATUS));
}

static __inline void vfWaitHQVFinish(unsigned char* pjVideo)
{
    unsigned long volatile *pdwState = 
    (unsigned long volatile *) (pjVideo+HQV_CONTROL); 

    while (!(*pdwState & HQV_FLIP_STATUS) );
}

static __inline void vfWaitHQV1Finish(unsigned char* pjVideo)
{
    unsigned long volatile *pdwState = 
    (unsigned long volatile *) (pjVideo+HQV_CONTROL+REG_HQV1_INDEX); 

    while (!(*pdwState & HQV_FLIP_STATUS) );
}

static __inline void vfWait3DIdle(unsigned char* pjVideo)
{
    unsigned long volatile *pdwState = (unsigned long volatile *) (pjVideo+0x400);

    while ((*pdwState & 0x83) );
}

static __inline void vfWaitVirtualQueueEmpty(unsigned char* pjVideo)
{
    unsigned long volatile *pdwState = (unsigned long volatile *) (pjVideo+REG_VQ_STATUS_CME);

    while ((*pdwState & VAL_VQ_NOT_EMPTY) );
}

#endif /* __VINLINE__H__ */
