/*
 * Copyright 1998-2008 VIA Technologies, Inc. All Rights Reserved.
 * Copyright 2001-2008 S3 Graphics, Inc. All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHOR(S) OR COPYRIGHT HOLDER(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */
 
#ifndef _ENGMGR_H_
#define _ENGMGR_H_ 1

#define ENGMGR_SUCCESS              1
#define ENGMGR_FAIL                 0  

#define ENGMGR_YES                  1
#define ENGMGR_NO                   0   

#define ENGMGR_FREE_ENG_MSG         0
#define ENGMGR_ALLOCATE_ENG_MSG     1
#define ENGMGR_SWAP_ENG_MSG         2  

#define VID_ENGINE               0x00
#define HQV_ENGINE               0x01
#define CAP_ENGINE               0x02
#define MPG_ENGINE               0x03
#define ENGINETYPE               0x04
                                      
#define ENG_ID_MAX               0x02 /*id can be 0,1,2..ENG_ID_MAX-1*/
                                      

#define ENGINE_NOT_EXIST         0xff
#define ENGINE_MGR               0xfe
/*      owner client id  0x00~0xfd*/

#define ANY_ENGINE_ID            0xff
#define CLIENTMAX                0x04  /*Engine Manager*/

/* CLIENT NAME */
#define CLIENT_MPEG                0
#define CLIENT_SW                  1
#define CLIENT_PORT_MAX            2   

/* PRIORITY stage*/
#define NORMAL_PRIORITY            0
#define URGENT_PRIORITY            1

#define CLIENT_INVALID_HANDLE      0xffffffff

typedef struct
{
	unsigned long uMessage_Type;
	unsigned long uClient_handle;
	unsigned long uSrcClient_Priority;
	unsigned long uEng_Type;
	unsigned long uEng_Id0;
	unsigned long uEng_Id1;
} Call_Back_Para;

typedef unsigned long EngMgrCallBack(PVIDDATA pVidData, Call_Back_Para *pPara); 


typedef struct
{
	EngMgrCallBack *pFunc;
	unsigned long uPriority;
} CLIENTINFO;/*for Manager*/

typedef struct ENGINE_STRUCT
{
	unsigned long uType;
	unsigned long uId;
	struct ENGINE_STRUCT * next;
}ENGINE;/*for Client*/

typedef struct
{
	unsigned long uClientHandle;
	ENGINE   *pPossessed_Eng;
} CLIENT;/*for Client*/


typedef struct
{
    EngMgrCallBack *func;
    unsigned long uPriority;
    unsigned long *pClient_handle;    
} EngMgr_RegClient_Para;

typedef struct
{
	unsigned long uClient_handle;
} EngMgr_UnRegClient_Para;

typedef struct
{
	unsigned long uClient_handle;
	unsigned long uPriority;
} EngMgr_AdjustClientPriority_Para;

typedef struct
{
	 unsigned long uClient_handle;
	 unsigned long uEng_type;
	 unsigned long uEng_id;
	 unsigned long *pEng_id_actual;
} EngMgr_AllocateEngine_Para;
 
typedef struct
{
	 unsigned long uClient_handle;
	 unsigned long uEng_type;
	 unsigned long uEng_id;
} EngMgr_FreeEngine_Para;

typedef struct
{
	 unsigned long uClient_handle;
	 unsigned long uEng_type;
	 unsigned long uEng_id_src;
	 unsigned long uEng_id_dst;
} EngMgr_SwapEngine_Para;

typedef struct
{
	 unsigned long uEng_type;
	 unsigned long uEng_id;
} EngMgr_CheckEngine_Para;


void EngMgr_Construct(void);
unsigned long EngMgr_RegClient(EngMgr_RegClient_Para *pInput);
unsigned long EngMgr_UnRegClient(PVIDDATA pVidData, EngMgr_UnRegClient_Para *pInput);
unsigned long EngMgr_AdjustClientPriority(EngMgr_AdjustClientPriority_Para *pInput);
unsigned long EngMgr_AllocateEngine(PVIDDATA pVidData, EngMgr_AllocateEngine_Para *pInput);
unsigned long EngMgr_CheckEngine(EngMgr_CheckEngine_Para *pInput);
unsigned long EngMgr_FreeEngine(PVIDDATA pVidData, EngMgr_FreeEngine_Para *pInput);
unsigned long EngMgr_SwapEngine(PVIDDATA pVidData, EngMgr_SwapEngine_Para *pInput);
unsigned long EngCallback_MPEG(PVIDDATA pVidData, Call_Back_Para *para);
unsigned long EngCallback_SW(PVIDDATA pVidData, Call_Back_Para *para);

#endif  /* _ENGMGR_H_ */
