 /*
 * Copyright 1998-2008 VIA Technologies, Inc. All Rights Reserved.
 * Copyright 2001-2008 S3 Graphics, Inc. All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHOR(S) OR COPYRIGHT HOLDER(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

 
#include "xf86.h"
#include "xf86_OSproc.h"
#include "xf86_ansic.h"
#include "xf86Priv.h"

#include "xf86PciInfo.h"
#include "xf86Pci.h"

#include "GL/glxtokens.h"

#include "via_driver.h"
#include "via_dri.h"
#include "xf86drm.h"
#include "xf86drmVIA.h"
extern void GlxSetVisualConfigs(
    int nconfigs,
    __GLXvisualConfig *configs,
    void **configprivs
);

#define H2_AGP_TEXTURE_ENABLE 1
#define H5_AGP_TEXTURE_ENABLE 1
#define H6_PCIE_TEXTURE_ENABLE 1

unsigned int agpaddr;
#define VIDEO	0 
#define AGP		1
#define AGP_PAGE_SIZE 4096

#define AGP_PAGES 4096
#define AGP_TEXBUF_PAGES_H5 0x3000
#define AGP_TEX_SIZE_H5 (AGP_PAGE_SIZE * AGP_TEXBUF_PAGES_H5)/*AGP_TEX_SIZE_H5 is used by H5*/
#define AGP_TEXBUF_PAGES_H2 0x1000
#define AGP_TEX_SIZE_H2 (AGP_PAGE_SIZE * AGP_TEXBUF_PAGES_H2)/*AGP_TEX_SIZE_H2 is used by H2*/
#define AGP_CMDBUF_PAGES 4096
#define AGP_CMDBUF_SIZE (AGP_PAGE_SIZE * AGP_CMDBUF_PAGES)/*AGP_CMDBUF_SIZE is used by H2,H5&H6*/

#if H5_AGP_TEXTURE_ENABLE
#define AGP_SIZE_H5 (AGP_CMDBUF_SIZE + AGP_TEX_SIZE_H5)/*AGP_SIZE_H5 is used by H5*/
#else
#define AGP_SIZE_H5 (AGP_CMDBUF_SIZE)
#endif

#if H2_AGP_TEXTURE_ENABLE
#define AGP_SIZE_H2 (AGP_CMDBUF_SIZE + AGP_TEX_SIZE_H2)/*AGP_SIZE_H2 is used by H2*/
#else
#define AGP_SIZE_H2 (AGP_CMDBUF_SIZE)
#endif

#if H6_PCIE_TEXTURE_ENABLE
#define PCIE_TEX_SIZE (AGP_PAGE_SIZE * 0x3000)/*default 48M*/
#endif

#define VIAH2KernelDriverName "via"
#define VIAH2ClientDriverName  "via_unichrome"
#define VIAH5KernelDriverName "via_chrome9"
#define VIAH5ClientDriverName  "via_chrome9"

int test_alloc_FB(ScreenPtr pScreen, VIAPtr pVia, int Size);
int test_alloc_AGP(ScreenPtr pScreen, VIAPtr pVia, int Size);
int VIAMapAGP(ScreenPtr pScreen, VIAPtr pVia);
int VIAS3GFinishScreenInit(ScreenPtr pScreen);
static Bool VIAInitVisualConfigs(ScreenPtr pScreen);
static Bool VIADRIAgpInit(ScreenPtr pScreen, VIAPtr pVia);
static Bool VIADRIPciInit(ScreenPtr pScreen, VIAPtr pVia);

Bool VIADRIFBInit(ScreenPtr pScreen, VIAPtr pVia);
static Bool VIADRIKernelInit(ScreenPtr pScreen, VIAPtr pVia);
static Bool VIADRIMapInit(ScreenPtr pScreen, VIAPtr pVia);

static Bool VIACreateContext(ScreenPtr pScreen, VisualPtr visual, 
                   drmContext hwContext, void *pVisualConfigPriv,
                   DRIContextType contextStore);
static void VIADestroyContext(ScreenPtr pScreen, drmContext hwContext,
                   DRIContextType contextStore);
static void VIADRISwapContext(ScreenPtr pScreen, DRISyncType syncType, 
                   DRIContextType readContextType, 
                   void *readContextStore,
                   DRIContextType writeContextType, 
                   void *writeContextStore);
static void VIADRIInitBuffers(WindowPtr pWin, RegionPtr prgn, CARD32 index);
static void VIADRIMoveBuffers(WindowPtr pParent, DDXPointRec ptOldOrg, 
                   RegionPtr prgnSrc, CARD32 index);
/*for vt3336*/
static int VIAInitS3gConf(ScreenPtr );
static int VIAReserveFBBuffer(ScreenPtr );
static int VIAReserveFBBuffer(ScreenPtr );
static int S3GInitVisualConfigs(ScreenPtr);


static Bool VIADRIAgpInit(ScreenPtr pScreen, VIAPtr pVia)
{
    unsigned long  agp_phys;
    VIADRIPtr pVIADRI;
    DRIInfoPtr pDRIInfo;
	unsigned long AGPSize;
    pDRIInfo = pVia->pDRIInfo;
    pVIADRI = pDRIInfo->devPrivate;
    pVia->agpSize = 0;
    
#ifdef __x86_64__
    return FALSE;
#endif

    if (drmAgpAcquire(pVia->drmFD) < 0) {
        xf86DrvMsg(pScreen->myNum, X_ERROR, "[drm] drmAgpAcquire failed\n");
        return FALSE;
    }

    if (drmAgpEnable(pVia->drmFD, drmAgpGetMode(pVia->drmFD)&~0x0) < 0) {
         xf86DrvMsg(pScreen->myNum, X_ERROR, "[drm] drmAgpEnable failed\n");
        return FALSE;
    }
    
    xf86DrvMsg(pScreen->myNum, X_INFO, "[drm] drmAgpEnabled succeeded\n");

    if(H5_UMA_CHIPID)
        AGPSize = min(drmAgpSize(pVia->drmFD),AGP_SIZE_H5);
	else
		AGPSize = min(drmAgpSize(pVia->drmFD),AGP_SIZE_H2);
	
    if (drmAgpAlloc(pVia->drmFD, AGPSize, 0, &agp_phys, &pVia->agpHandle) < 0) {
        xf86DrvMsg(pScreen->myNum, X_ERROR,
                 "[drm] drmAgpAlloc failed\n");
        drmAgpRelease(pVia->drmFD);
        return FALSE;
    }
   
    if (drmAgpBind(pVia->drmFD, pVia->agpHandle, 0) < 0) {
        xf86DrvMsg(pScreen->myNum, X_ERROR,
                 "[drm] drmAgpBind failed\n");
        drmAgpFree(pVia->drmFD, pVia->agpHandle);
        drmAgpRelease(pVia->drmFD);

        return FALSE;
    }

    pVia->agpSize = AGPSize;
    pVia->agpAddr = drmAgpBase(pVia->drmFD);
    xf86DrvMsg(pScreen->myNum, X_INFO,
                 "[drm] agpAddr = 0x%08lx\n",pVia->agpAddr);
		 
    /*h5s1 needn't share agp handle, and addmap drm buf is in VIAMapAGP()*/
    if(!(H5_UMA_CHIPID||H6_UMA_CHIPID))
    {
        pVIADRI->agp.size = pVia->agpSize;
        if (drmAddMap(pVia->drmFD, (drmHandle)0,
                     pVIADRI->agp.size, DRM_AGP, 0, 
                     &pVIADRI->agp.handle) < 0) {
    	xf86DrvMsg(pScreen->myNum, X_ERROR,
    	    "[drm] Failed to map public agp area\n");
            pVIADRI->agp.size = 0;
            return FALSE;
        }  
        /* Map AGP from kernel to Xserver - Not really needed */
        drmMap(pVia->drmFD, pVIADRI->agp.handle,pVIADRI->agp.size,
    	(drmAddressPtr)&agpaddr);
    	
    	/*agp flush drm*/
    	pVia->agp_handle=pVIADRI->agp.handle;

    	if(pVIADRI->agp.size > AGP_CMDBUF_SIZE)
            drmVIAAgpInit(pVia->drmFD, AGP_CMDBUF_SIZE, pVIADRI->agp.size - AGP_CMDBUF_SIZE);
    	/*Only let drm to handle AGP texture buffer*/
    }
    xf86DrvMsg(pScreen->myNum, X_INFO, 
                "[drm] agpBase = 0x%08lx\n", (unsigned long)pVia->agpBase);
    xf86DrvMsg(pScreen->myNum, X_INFO, 
                "[drm] agpAddr = 0x%08lx\n", pVia->agpAddr);
    xf86DrvMsg(pScreen->myNum, X_INFO, 
                "[drm] agpSize = 0x%08x\n", pVia->agpSize);
    xf86DrvMsg(pScreen->myNum, X_INFO, 
                "[drm] agp physical addr = 0x%08lx\n", agp_phys);
    return TRUE;
  
}

Bool VIADRIFBInit(ScreenPtr pScreen, VIAPtr pVia)
{   
    int FBSize = pVia->FBFreeEnd-pVia->FBFreeStart;
    int FBOffset = pVia->FBFreeStart; 
    VIADRIPtr pVIADRI = pVia->pDRIInfo->devPrivate;
    pVIADRI->fbOffset = FBOffset;
    pVIADRI->fbSize = pVia->videoRambytes;
    if(H5_UMA_CHIPID||H6_UMA_CHIPID) /*for H5s1_UMA, go another path*/
    {
        if(!(VIAReserveFBBuffer(pScreen)))
    		return FALSE;
        return TRUE;
    }
    else
    {
        if (drmVIAFBInit(pVia->drmFD, FBOffset, FBSize) < 0) {
    	xf86DrvMsg(pScreen->myNum, X_ERROR,"[drm] failed to init frame buffer area\n");
    	return FALSE;		
        }
        else {
    	xf86DrvMsg(pScreen->myNum, X_INFO,"[drm] FBFreeStart= 0x%08x FBFreeEnd= 0x%08x FBSize= 0x%08x\n", pVia->FBFreeStart, pVia->FBFreeEnd, FBSize);
    	return TRUE;	
        }
    }
}
static Bool VIADRIPciInit(ScreenPtr pScreen, VIAPtr pVia)
{
    return TRUE;	
}

static Bool
VIAInitVisualConfigs(ScreenPtr pScreen)
{
    ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
    VIAPtr pVia = VIAPTR(pScrn);
    int numConfigs = 0;
    __GLXvisualConfig *pConfigs = 0;
    VIAConfigPrivPtr pVIAConfigs = 0;
    VIAConfigPrivPtr *pVIAConfigPtrs = 0;
    int i, db, stencil, accum;

    switch (pScrn->bitsPerPixel) {
	case 8:
	case 24:
	    break;
	case 16:
	    /* for visual config */
	    numConfigs = 16;
	    
	    if (!(pConfigs = (__GLXvisualConfig*)xcalloc(sizeof(__GLXvisualConfig),
						   numConfigs)))
		    return FALSE;
	    if (!(pVIAConfigs = (VIAConfigPrivPtr)xcalloc(sizeof(VIAConfigPrivRec),
						    numConfigs))) {
    		xfree(pConfigs);
    		return FALSE;
	    }
	    if (!(pVIAConfigPtrs = (VIAConfigPrivPtr*)xcalloc(sizeof(VIAConfigPrivPtr),
							  numConfigs))) {
    		xfree(pConfigs);
    		xfree(pVIAConfigs);
    		return FALSE;
	    }
	    for (i=0; i<numConfigs; i++) 
    		pVIAConfigPtrs[i] = &pVIAConfigs[i];

	    i = 0;
    	for (accum = 0; accum <= 1; accum++) {
    		/* for visual config */
    		for (stencil = 0; stencil <= 3; stencil++) {
        		for (db = 0; db <= 1; db++) {
            		pConfigs[i].vid = -1;
            		pConfigs[i].class = -1;
            		pConfigs[i].rgba = TRUE;
            		pConfigs[i].redSize = -1;
            		pConfigs[i].greenSize = -1;
            		pConfigs[i].blueSize = -1;
            		pConfigs[i].redMask = -1;
            		pConfigs[i].greenMask = -1;
            		pConfigs[i].blueMask = -1;
            		pConfigs[i].alphaMask = 0;
    			
        			if (accum) {
            		    pConfigs[i].accumRedSize = 16;
            		    pConfigs[i].accumGreenSize = 16;
            		    pConfigs[i].accumBlueSize = 16;
            		    pConfigs[i].accumAlphaSize = 16;
            		}
        			else {
            		    pConfigs[i].accumRedSize = 0;
            		    pConfigs[i].accumGreenSize = 0;
            		    pConfigs[i].accumBlueSize = 0;
            		    pConfigs[i].accumAlphaSize = 0;
            		}

            		if (!db)
            		    pConfigs[i].doubleBuffer = TRUE;
            		else
            		    pConfigs[i].doubleBuffer = FALSE;
                		
        			pConfigs[i].stereo = FALSE;
                	pConfigs[i].bufferSize = -1;
                		
        			switch (stencil) {
            		    case 0:
                			pConfigs[i].depthSize = 0;
                			pConfigs[i].stencilSize = 0;
                			break;
            		    case 1:
                			pConfigs[i].depthSize = 16;
                			pConfigs[i].stencilSize = 0;
                			break;
            		    case 2:
                			pConfigs[i].depthSize = 32;
                			pConfigs[i].stencilSize = 0;
                			break;
            		    case 3:
                			pConfigs[i].depthSize = 24;
                			pConfigs[i].stencilSize = 8;
                			break;
            		}
            		
        			pConfigs[i].auxBuffers = 0;
            		pConfigs[i].level = 0;
            		pConfigs[i].visualRating = GLX_NONE_EXT;
            		pConfigs[i].transparentPixel = 0;
            		pConfigs[i].transparentRed = 0;
            		pConfigs[i].transparentGreen = 0;
            		pConfigs[i].transparentBlue = 0;
                	pConfigs[i].transparentAlpha = 0;
            		pConfigs[i].transparentIndex = 0;
            		i++;
    		    }
    		}
    	}
	
	    if (i != numConfigs) {
    	    xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
    		"[dri] Incorrect initialization of visuals.  Disabling DRI.\n");
        		return FALSE;
    	}
	    break;
	    
	/* expose RGBA visual configs */
	case 32:
	    numConfigs = 16;
	    
	    if (!(pConfigs = (__GLXvisualConfig*)xcalloc(sizeof(__GLXvisualConfig),
						   numConfigs)))
		    return FALSE;
	    if (!(pVIAConfigs = (VIAConfigPrivPtr)xcalloc(sizeof(VIAConfigPrivRec),
						    numConfigs))) {
    		xfree(pConfigs);
    		return FALSE;
	    }
	    if (!(pVIAConfigPtrs = (VIAConfigPrivPtr*)xcalloc(sizeof(VIAConfigPrivPtr),
							  numConfigs))) {
    		xfree(pConfigs);
    		xfree(pVIAConfigs);
    		return FALSE;
	    }
	    for (i=0; i<numConfigs; i++) 
    		pVIAConfigPtrs[i] = &pVIAConfigs[i];

	    i = 0;
	    for (accum = 0; accum <= 1; accum++) {
		    for (stencil = 0; stencil <= 3; stencil++) {
    		    for (db = 0; db <= 1; db++) {
            		pConfigs[i].vid = -1;
            		pConfigs[i].class = -1;
            		pConfigs[i].rgba = TRUE;
            		pConfigs[i].redSize = 8;
            		pConfigs[i].greenSize = 8;
            		pConfigs[i].blueSize = 8;
    			    pConfigs[i].alphaSize = 8;
            		pConfigs[i].redMask = 0x00FF0000;
            		pConfigs[i].greenMask = 0x0000FF00;
            		pConfigs[i].blueMask = 0x000000FF;
            		pConfigs[i].alphaMask = 0xFF000000;
			
			        if (accum) {
            		    pConfigs[i].accumRedSize = 16;
            		    pConfigs[i].accumGreenSize = 16;
            		    pConfigs[i].accumBlueSize = 16;
            		    pConfigs[i].accumAlphaSize = 16;
            		}
        			else {
            		    pConfigs[i].accumRedSize = 0;
            		    pConfigs[i].accumGreenSize = 0;
            		    pConfigs[i].accumBlueSize = 0;
            		    pConfigs[i].accumAlphaSize = 0;
                	}

            		if (!db)
            		    pConfigs[i].doubleBuffer = TRUE;
            		else
            		    pConfigs[i].doubleBuffer = FALSE;
        		
			        pConfigs[i].stereo = FALSE;
        		    pConfigs[i].bufferSize = 32;
        		
			        switch (stencil) {
        		    case 0:
            			pConfigs[i].depthSize = 0;
            			pConfigs[i].stencilSize = 0;
            			break;
        		    case 1:
            			pConfigs[i].depthSize = 24;
            			pConfigs[i].stencilSize = 0;
            			break;
        		    case 2:
            			pConfigs[i].depthSize = 32;
            			pConfigs[i].stencilSize = 0;
            			break;
        		    case 3:
            			pConfigs[i].depthSize = 24;
            			pConfigs[i].stencilSize = 8;
            			break;
        		    }
        		
			        pConfigs[i].auxBuffers = 0;
            		pConfigs[i].level = 0;
            		pConfigs[i].visualRating = GLX_NONE_EXT;
            		pConfigs[i].transparentPixel = 0;
            		pConfigs[i].transparentRed = 0;
            		pConfigs[i].transparentGreen = 0;
            		pConfigs[i].transparentBlue = 0;
    	        	pConfigs[i].transparentAlpha = 0;
            		pConfigs[i].transparentIndex = 0;
            		i++;
		        }
		    }
	    }
	
	    if (i != numConfigs) {
    	    xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
    		        "[dri] Incorrect initialization of visuals.  Disabling DRI.\n");
            return FALSE;
	    }
	    break;
    }
    
    pVia->numVisualConfigs = numConfigs;
    pVia->pVisualConfigs = pConfigs;
    pVia->pVisualConfigsPriv = pVIAConfigs;

    GlxSetVisualConfigs(numConfigs, pConfigs, (void**)pVIAConfigPtrs);

    return TRUE;
}

/*initialize AGP ring buffer in 2D driver*/
void VIADRIRingBufferCleanup(ScrnInfoPtr pScrn)
{
    VIAPtr pVia = VIAPTR(pScrn);
	drm_via_dma_init_t ringBufInit;
	xf86DrvMsg(pScrn->scrnIndex, X_INFO, 
		   "[drm] Cleaning up DMA ring-buffer.\n");
	ringBufInit.func = VIA_CLEANUP_DMA;
	if (drmCommandWrite(pVia->drmFD, DRM_VIA_DMA_INIT, &ringBufInit,
			    sizeof(ringBufInit))) {
	    xf86DrvMsg(pScrn->scrnIndex, X_WARNING, 
		       "[drm] Failed to clean up DMA ring-buffer: %d\n", errno);
	}
}
Bool VIADRIRingBufferInit(ScrnInfoPtr pScrn)
{
    VIAPtr pVia = VIAPTR(pScrn);
    if (!pVia->IsPCI) {
	drm_via_dma_init_t ringBufInit,init;

   	memset(&init, 0, sizeof(init));
   	memset(&ringBufInit, 0, sizeof(ringBufInit));
 	init.func = VIA_DMA_INITIALIZED;

    if(0 == drmCommandWrite(pVia->drmFD, DRM_VIA_DMA_INIT,&init, sizeof(init)))
		return TRUE;

    ringBufInit.reg_pause_addr = 0x418;
	ringBufInit.offset = 0x0;
	ringBufInit.size = AGP_CMDBUF_SIZE;

	ringBufInit.func = VIA_INIT_DMA;
	if (drmCommandWrite(pVia->drmFD, DRM_VIA_DMA_INIT, &ringBufInit,
			    sizeof(ringBufInit))) {
	    xf86DrvMsg(pScrn->scrnIndex, X_ERROR, 
		       "[drm] Failed to initialize DMA ring-buffer: %d\n", errno);
	    return FALSE;
	}
	xf86DrvMsg(pScrn->scrnIndex, X_INFO, 
		   "[drm] Initialized AGP ring-buffer, size 0x%lx at AGP offset 0x%lx.\n",
		   ringBufInit.size, ringBufInit.offset);
        return TRUE;		   
    }
    else
    {
        return FALSE;
    }
}

Bool VIADRIScreenInit(ScreenPtr pScreen)
{
    ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
    VIAPtr pVia = VIAPTR(pScrn);
    DRIInfoPtr pDRIInfo;
    VIADRIPtr pVIADRI;
    pVia->IsPCI = TRUE;

    if(H5_UMA_CHIPID||H6_UMA_CHIPID)
    {
        switch(pVia->ChipId)
        {
        case PCI_CHIP_VT3336:
 	   	case PCI_CHIP_VT3364:
            xf86DrvMsg(pScreen->myNum, X_INFO,
                   "[drm] Detect H5s1 chipset: %04x  chipid: %04x\n",
                    pVia->Chipset, pVia->ChipId);
            break;
        case PCI_CHIP_VT3293:
            xf86DrvMsg(pScreen->myNum, X_INFO,
                    "[drm] Detect H5 chipset: %04x  chipid: %04x\n",
                    pVia->Chipset, pVia->ChipId);
            break;
        case PCI_CHIP_VT3353:
            xf86DrvMsg(pScreen->myNum, X_INFO,
			         "[drm] Detect H6 chipset: %04x  chipid: %04x\n",
	                pVia->Chipset, pVia->ChipId);
            break;
        default:
            break;
        }
    }

		
    /* Check that the GLX, DRI, and DRM modules have been loaded by testing
    * for canonical symbols in each module. */
    if (!xf86LoaderCheckSymbol("GlxSetVisualConfigs")) return FALSE;
    if (!xf86LoaderCheckSymbol("DRIScreenInit"))       return FALSE;
    if (!xf86LoaderCheckSymbol("drmAvailable"))        return FALSE;
    if (!xf86LoaderCheckSymbol("DRIQueryVersion")) {
	    xf86DrvMsg(pScreen->myNum, X_ERROR,
                 "[dri] VIADRIScreenInit failed (libdri.a too old)\n");
        return FALSE;
    }
    
    /* Check the DRI version */
    /*change for get the libDRI version 5.0.0, so remove it first.*/
    if(!(H5_UMA_CHIPID||H6_UMA_CHIPID))
    {
        int major, minor, patch;
        DRIQueryVersion(&major, &minor, &patch);
        if (major < 4 || minor < 0) {
            xf86DrvMsg(pScreen->myNum, X_ERROR,
                    "[dri] VIADRIScreenInit failed because of a version mismatch.\n"
                    "[dri] libDRI version is %d.%d.%d but version 4.0.x is needed.\n"
                    "[dri] Disabling DRI.\n",
                    major, minor, patch);
            return FALSE;
        }
    }
    else /*for support h5s1(vt3336/vt3364), now allocate additional record*/
    {
        ((VIARec *)(pScrn->driverPrivate))->pScrnConf = xnfcalloc(sizeof(ScrnConfRec), 1);
        ((VIARec *)(pScrn->driverPrivate))->pDriConf = xnfcalloc(sizeof(DriConfRec), 1);
        ((VIARec *)(pScrn->driverPrivate))->pServerInfo = xnfcalloc(sizeof(ServerInfoRec), 1);
        ((VIARec *)(pScrn->driverPrivate))->pAddrConf = xnfcalloc(sizeof(AddrConfRec), 1);
        ((VIARec *)(pScrn->driverPrivate))->pPciConf = xnfcalloc(sizeof(PciConfRec), 1);
		((VIARec *)(pScrn->driverPrivate))->pChipConf = xnfcalloc(sizeof(ChipConfRec), 1);
    }

    pDRIInfo = DRICreateInfoRec();
    
    if (!pDRIInfo) return FALSE;
    
    pVia->pDRIInfo = pDRIInfo;

    if(H5_UMA_CHIPID||H6_UMA_CHIPID) 
    {
        pDRIInfo->drmDriverName = VIAH5KernelDriverName;
        pDRIInfo->clientDriverName = VIAH5ClientDriverName;
    }
    else
    {
        pDRIInfo->drmDriverName = VIAH2KernelDriverName;
        pDRIInfo->clientDriverName = VIAH2ClientDriverName;
    }
    pDRIInfo->busIdString = xalloc(64);
    sprintf(pDRIInfo->busIdString, "PCI:%d:%d:%d",
        ((pciConfigPtr)pVia->PciInfo->thisCard)->busnum,
        ((pciConfigPtr)pVia->PciInfo->thisCard)->devnum,
        ((pciConfigPtr)pVia->PciInfo->thisCard)->funcnum);
    pDRIInfo->ddxDriverMajorVersion = VIA_VERSION_MAJOR;
    pDRIInfo->ddxDriverMinorVersion = VIA_VERSION_MINOR;
    pDRIInfo->ddxDriverPatchVersion = PATCHLEVEL;
    pDRIInfo->frameBufferPhysicalAddress = (void *)pVia->FrameBufferBase;
    pDRIInfo->frameBufferSize = pVia->videoRambytes;  
  
    pDRIInfo->frameBufferStride = (pScrn->displayWidth *
					    pScrn->bitsPerPixel / 8);
    pDRIInfo->ddxDrawableTableEntry = VIA_MAX_DRAWABLES;

    if (SAREA_MAX_DRAWABLES < VIA_MAX_DRAWABLES)
	    pDRIInfo->maxDrawableTableEntry = SAREA_MAX_DRAWABLES;
    else
	    pDRIInfo->maxDrawableTableEntry = VIA_MAX_DRAWABLES;

#ifdef NOT_DONE

    pDRIInfo->SAREASize = 
	((sizeof(XF86DRISAREARec) + 0xfff) & 0x1000); /* round to page */
    /* + shared memory device private rec */
#else
    /* For now the mapping works by using a fixed size defined
    * in the SAREA header
    */
    if (sizeof(XF86DRISAREARec)+sizeof(VIASAREAPriv) > SAREA_MAX) {
	xf86DrvMsg(pScrn->scrnIndex, X_ERROR,
			"Data does not fit in SAREA\n");
	    return FALSE;
    }
    pDRIInfo->SAREASize = SAREA_MAX;
#endif

    if (!(pVIADRI = (VIADRIPtr)xcalloc(sizeof(VIADRIRec),1))) {
    	DRIDestroyInfoRec(pVia->pDRIInfo);
    	pVia->pDRIInfo=0;
    	return FALSE;
    }
    /* for h5s1_UMA, copy the s3g rec ptr from VIARec*/
    if(H5_UMA_CHIPID||H6_UMA_CHIPID)
    {
        pVIADRI->pServerInfo = pVia->pServerInfo;
        pDRIInfo->devPrivate= pVIADRI;
        pDRIInfo->devPrivateSize = sizeof(VIADRIRec);
        pDRIInfo->contextSize = sizeof(VIADRIContextRec);
    }
    else
    {
        pDRIInfo->devPrivate= pVIADRI;
        pDRIInfo->devPrivateSize = sizeof(VIADRIRec);
        pDRIInfo->contextSize = sizeof(VIADRIContextRec);
    }

    pDRIInfo->CreateContext = VIACreateContext;
    pDRIInfo->DestroyContext = VIADestroyContext;
    pDRIInfo->SwapContext = VIADRISwapContext;
    pDRIInfo->InitBuffers = VIADRIInitBuffers;
    pDRIInfo->MoveBuffers = VIADRIMoveBuffers;
    pDRIInfo->bufferRequests = DRI_ALL_WINDOWS;

    if (!DRIScreenInit(pScreen, pDRIInfo, &pVia->drmFD)) {
    	xf86DrvMsg(pScreen->myNum, X_ERROR,
    	    "[dri] DRIScreenInit failed.  Disabling DRI.\n");
    	xfree(pDRIInfo->devPrivate);
    	pDRIInfo->devPrivate=0;
    	DRIDestroyInfoRec(pVia->pDRIInfo);
    	pVia->pDRIInfo=0;
    	pVia->drmFD = -1;
    	return FALSE;
    }

	if(H6_UMA_CHIPID)
		pVia->IsPCI = TRUE;/*We can only know whether PCIE is OK or not after init drm modules*/
	else
        pVia->IsPCI = !VIADRIAgpInit(pScreen, pVia);

    if(!H6_UMA_CHIPID)
    {
        if (pVia->IsPCI) {
    	    VIADRIPciInit(pScreen, pVia);
    	    xf86DrvMsg(pScrn->scrnIndex, X_INFO, "[dri] use pci.\n" );
        }
        else
	        xf86DrvMsg(pScrn->scrnIndex, X_INFO, "[dri] use agp.\n" );
    }
    
    xf86DrvMsg(pScrn->scrnIndex, X_INFO, "[dri] frame buffer initialized.\n" );
    if(!(H5_UMA_CHIPID||H6_UMA_CHIPID))
    {
        if (!(VIAInitVisualConfigs(pScreen))) 
        {
            VIADRICloseScreen(pScreen);
            return FALSE;
        }
    } /*for  h5s1 UMA*/
    else
    {
        if(!(S3GInitVisualConfigs(pScreen)))
        {
            VIADRICloseScreen(pScreen);
            return FALSE;
        }
    }
    xf86DrvMsg(pScrn->scrnIndex, X_INFO, "[dri] visual configs initialized.\n" );
  
    /* DRIScreenInit doesn't add all the common mappings.  Add additional mappings here. */
    if (!VIADRIMapInit(pScreen, pVia)) {
    	VIADRICloseScreen(pScreen);
    	return FALSE;
    }
    pVIADRI->regs.size = VIA_MMIO_REGSIZE;
    pVIADRI->regs.map = 0;
    pVIADRI->regs.handle = pVia->registerHandle;
    xf86DrvMsg(pScreen->myNum, X_INFO, "[drm] mmio Registers = 0x%08lx\n",
	pVIADRI->regs.handle);
    
    pVIADRI->drixinerama = pVia->drixinerama;
    /* Tuxracer & VQ */
    pVIADRI->VQEnable = pVia->VQEnable;

    xf86DrvMsg(pScrn->scrnIndex, X_INFO, "[dri] mmio maped.\n" );

    return TRUE;
}

void
VIADRICloseScreen(ScreenPtr pScreen)
{
    ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
    VIAPtr pVia = VIAPTR(pScrn);
    if(H5_UMA_CHIPID||H6_UMA_CHIPID)
    {
        drm_s3g_init_t init;
        /*call drm clean up ioctl before close screen*/
        DriConfPtr pDriConf = pVia->pDriConf;
        memset(&init, 0, sizeof(drm_s3g_init_t));
        init.func = S3G_CLEANUP;
        drmCommandWrite(pDriConf->drmFd, DRM_S3G_INIT, &init, sizeof(drm_s3g_init_t));
    }
    DRICloseScreen(pScreen);

    if (pVia->pDRIInfo) {
    	if (pVia->pDRIInfo->devPrivate) {
            xfree(pVia->pDRIInfo->devPrivate);
            pVia->pDRIInfo->devPrivate=0;
    	}
    	DRIDestroyInfoRec(pVia->pDRIInfo);
    	pVia->pDRIInfo=0;
    }
    
    if (pVia->pVisualConfigs) 
        xfree(pVia->pVisualConfigs);
    if (pVia->pVisualConfigsPriv) 
        xfree(pVia->pVisualConfigsPriv);
    if (pVia->agpSize) {
    	xf86DrvMsg(pScreen->myNum, X_INFO, "[drm] Freeing agp memory\n");
            drmAgpFree(pVia->drmFD, pVia->agpHandle);
    	xf86DrvMsg(pScreen->myNum, X_INFO, "[drm] Releasing agp module\n");
        	drmAgpRelease(pVia->drmFD);
    }
}

/* TODO: xserver receives driver's swapping event and do something
 *       according the data initialized in this function
 */
static Bool
VIACreateContext(ScreenPtr pScreen, VisualPtr visual, 
          drmContext hwContext, void *pVisualConfigPriv,
          DRIContextType contextStore)
{
    return TRUE;
}

static void
VIADestroyContext(ScreenPtr pScreen, drmContext hwContext, 
           DRIContextType contextStore)
{
}

Bool
VIADRIFinishScreenInit(ScreenPtr pScreen)
{
    ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
    VIAPtr pVia = VIAPTR(pScrn);
    VIADRIPtr pVIADRI;
    pVIADRI=(VIADRIPtr)pVia->pDRIInfo->devPrivate;

    pVia->pDRIInfo->driverSwapMethod = DRI_HIDE_X_CONTEXT;
    
    DRIFinishScreenInit(pScreen);
    /*for vt3336*/
    if(H5_UMA_CHIPID||H6_UMA_CHIPID)
    {
        VIAS3GFinishScreenInit(pScreen);
    }
    if (!VIADRIKernelInit(pScreen, pVia)) {
    	VIADRICloseScreen(pScreen);
    	return FALSE;
    }
    xf86DrvMsg(pScreen->myNum, X_INFO, "[dri] kernel data initilized.\n");

    /* set SAREA value */
	VIASAREAPriv *saPriv;

	saPriv=(VIASAREAPriv*)DRIGetSAREAPrivate(pScreen);
	assert(saPriv);
	memset(saPriv, 0, sizeof(*saPriv));
	saPriv->CtxOwner = -1;

    pVIADRI=(VIADRIPtr)pVia->pDRIInfo->devPrivate;
    pVIADRI->deviceID=pVia->Chipset;  
    pVIADRI->width=pScrn->virtualX;
    pVIADRI->height=pScrn->virtualY;
    pVIADRI->mem=pScrn->videoRam*1024;
    pVIADRI->bytesPerPixel= (pScrn->bitsPerPixel+7) / 8; 
    pVIADRI->sarea_priv_offset = sizeof(XF86DRISAREARec);
    /* TODO */
    pVIADRI->scrnX=pVIADRI->width;
    pVIADRI->scrnY=pVIADRI->height;

    /*for vt3336*/
    pVia->useAgp = 0;
    if(H5_UMA_CHIPID||H6_UMA_CHIPID)
    {
        pVia->pDRIInfo->devPrivate= pVia->pServerInfo;
        pVia->pDRIInfo->devPrivateSize = sizeof(ServerInfoRec);
        pVia->pDRIInfo->contextSize = 256;
    }
    else
    {
        /* initia AGP ring buffer in 2D driver*/
    	pVia->useAgp=VIADRIRingBufferInit(pScrn);

    }
    return TRUE;
}

static void
VIADRISwapContext(ScreenPtr pScreen, DRISyncType syncType, 
           DRIContextType oldContextType, void *oldContext,
           DRIContextType newContextType, void *newContext)
{
    return;
}

static void
VIADRIInitBuffers(WindowPtr pWin, RegionPtr prgn, CARD32 index)
{
    return;
}

static void
VIADRIMoveBuffers(WindowPtr pParent, DDXPointRec ptOldOrg, 
           RegionPtr prgnSrc, CARD32 index)
{
    return;
}

/* Initialize the kernel data structures. */
static Bool VIADRIKernelInit(ScreenPtr pScreen, VIAPtr pVia)
{
    drmVIAInit drmInfo;
    drmInfo.sarea_priv_offset   = sizeof(XF86DRISAREARec);
    drmInfo.fb_offset           = pVia->FrameBufferBase;
    drmInfo.mmio_offset         = pVia->registerHandle;
    if (pVia->IsPCI)
	drmInfo.agpAddr = (CARD32)NULL;
    else {
/* For AMD64 */
#ifndef __x86_64__
	drmInfo.agpAddr = (CARD32)pVia->agpAddr;
#else
	drmInfo.agpAddr = (CARD64)pVia->agpAddr;
#endif
    }
 	/*agp flush drm*/
	drmInfo.agp_offset=(CARD32)pVia->agp_handle;
    if(H5_UMA_CHIPID||H6_UMA_CHIPID)/*drm init for H5s1_UMA*/
    {
        DriConfPtr pDriConf = pVia->pDriConf;
        ChipConfPtr pChipConf = pVia->pChipConf;
        AddrConfPtr pAddrConf = pVia->pAddrConf;
        ServerInfoPtr pServerInfo = pVia->pServerInfo;
        
        drm_s3g_init_t init;
	   {
            init.func = S3G_INIT;
            init.chip_index = S3G_INV;
			switch(pVia->ChipId)
			{
            case PCI_CHIP_VT3336:
            case PCI_CHIP_VT3364:
			   	init.chip_sub_index = CHIP_H5S1;
			    break;

            case PCI_CHIP_VT3293:
				init.chip_sub_index = CHIP_H5;
				break;

            case PCI_CHIP_VT3353:
 				init.chip_sub_index = CHIP_H6S2;
 				break;

            default:
				break;			   			   	
			}
            init.garttable_size = 0x0;
            init.garttable_offset = 0x0;  /*because our h5s1 is AGP, but not PCIE, we don't need to create garttable*/
            init.dma_handle = pDriConf->dmaHandle;   /*for h5s1, dma buffer is in agp*/
       }

		if(!H6_UMA_CHIPID)
		{
            /*tell drm how to init dma-agp*/
            /*if agp init failed, then use pci*/
            if(pVia->IsPCI)
                init.agp_type = AGP_DISABLED;
            /*For vt3336 and vt3364, the default agp type is ring buffer mode*/
            else
                init.agp_type = AGP_RING_BUFFER;
		}
		else
 			init.agp_type = AGP_RING_BUFFER;/*default choice*/

        init.sarea_priv_offset = sizeof(XF86DRISAREARec);
        init.fb_cpp = pVia->Bpp;
        init.front_offset = 0;
        init.depth_offset = pDriConf->depthOffset;
        
        init.mmio_handle = pDriConf->mmioHandle;
        init.hostBlt_handle = (unsigned int)pVia->hostBltHandle;
        
        init.fb_handle = pDriConf->fbHandle;
        init.front_handle = pDriConf->frontHandle;
        init.back_handle = pDriConf->frontHandle;
        init.depth_handle = pDriConf->frontHandle;

        /* Set total available video memory for 3D apps */
        init.available_fb_size = pVia->FBFreeEnd - pVia->FBFreeStart;/*because the meaning of fbFree in VIA and s3g is contrary to via*/
        init.fb_base_address = pAddrConf->fbPhyAddr;    
        init.back_offset = pVia->FBFreeStart;  /*give the region follow on-screen to 3d*/
        
        init.fb_tex_offset = pDriConf->fbTexOffset;
        init.fb_tex_size = pDriConf->fbTexSize;
        init.agp_tex_size = pDriConf->agpTexSize;
        init.agp_tex_handle = pDriConf->agpTexHandle;

        init.chip_agp = pChipConf->chipType;
        init.chip_sub_index = pServerInfo->chipSubIndex;
        init.chip_index = pServerInfo->chipIndex;

        init.usec_timeout = 50000; 
        /*now shadow is in video memory, so if agp init failed, shadow is still available*/
        init.shadow_size = pDriConf->shadowSize;
        init.shadow_handle = pDriConf->shadowHandle;


        if(H6_UMA_CHIPID)
        { 
            init.garttable_offset = init.back_offset;/*reserved for gart table*/
   			init.garttable_offset = (init.garttable_offset + 4095)&(~4095);/*4k align*/
   			init.DMA_size = AGP_CMDBUF_SIZE;
#if H6_PCIE_TEXTURE_ENABLE
            pServerInfo->agpTexSize = pDriConf->agpTexSize = 
            init.agp_tex_size = PCIE_TEX_SIZE;
#endif
 			init.garttable_size = (init.DMA_size + init.agp_tex_size) >> 10;
   			init.garttable_size = (init.garttable_size + 4095)&(~4095);
   			init.available_fb_size -= (init.garttable_size + init.garttable_offset - init.back_offset);
   			init.back_offset = init.garttable_offset + init.garttable_size;

        }
		else
		{
		    if (pVia->IsPCI) 
            {
                init.DMA_size = 0x0;
            }
            else
            {
                init.DMA_phys_address = pDriConf->agpBase; ;  /*dma buffer is follow the shadow in AGP*/
                init.DMA_size = AGP_CMDBUF_SIZE;   /*the real size of ring buffer agp buffer size*/
		    	init.agp_tex_size = pDriConf->agpTexSize;
            }
		}
        
        int  ret = drmCommandWriteRead(pDriConf->drmFd, DRM_S3G_INIT, &init, sizeof(drm_s3g_init_t));
        if (ret < 0)  
        {
            xf86DrvMsg(pScreen->myNum, X_ERROR,"[drm] ERROR! S3GDRIKernelInit:failed to initialize kernel module! (%d)\n", ret );
            return FALSE;
        }
        if(H6_UMA_CHIPID)
   	    {
 			if((init.agp_type == AGP_RING_BUFFER)||(init.agp_type == AGP_DOUBLE_BUFFER)){
                pServerInfo->isAGP = pServerInfo->DriConf.dmaBufAGP = 
					pDriConf->dmaBufAGP = TRUE;
					pVia->IsPCI = FALSE;
   			    xf86DrvMsg(pScreen->myNum, X_INFO,"[drm] use pcie\n");
   		    }/*PCIE path enable*/
   			else{
                pServerInfo->isAGP = pServerInfo->DriConf.dmaBufAGP = 
					pDriConf->dmaBufAGP = FALSE;
					pVia->IsPCI = TRUE;
   			    xf86DrvMsg(pScreen->myNum, X_INFO,"[drm] use pci\n");
   		    }				
        }

        return TRUE;
    }
    else
    {
        if (drmVIAInitMAP(pVia->drmFD, &drmInfo) < 0) 
            return FALSE;
    }
    return TRUE;
}
/* Add a map for the MMIO registers */
static Bool VIADRIMapInit(ScreenPtr pScreen, VIAPtr pVia)
{
    int flags = 0;
    unsigned long handle = 0;

    if (drmAddMap(pVia->drmFD, pVia->MmioBase, VIA_MMIO_REGSIZE,
		  DRM_REGISTERS, flags, &pVia->registerHandle) < 0) {
	    return FALSE;
    }

    pVia->pVidData->VIAGraphicInfo.MMIOhandle = pVia->registerHandle;
    pVia->pVidData->VIAGraphicInfo.MMIObase = pVia->MmioBase;
    pVia->pVidData->VIAGraphicInfo.MMIOsize = VIA_MMIO_REGSIZE;

    if (drmAddMap(pVia->drmFD, pVia->FrameBufferBase, pVia->videoRambytes,
                  DRM_FRAME_BUFFER, flags, &handle) < 0) {
        return FALSE;
    }

    pVia->pVidData->VIAGraphicInfo.FBhandle = handle;
    pVia->pVidData->VIAGraphicInfo.FBbase = pVia->FrameBufferBase;
    pVia->pVidData->VIAGraphicInfo.FBsize = pVia->videoRambytes;

    
    if(H5_UMA_CHIPID||H6_UMA_CHIPID) /*for H5s1_UMA*/
    {
        if (drmAddMap(pVia->drmFD, pVia->MmioBase + VIA_MMIO_BLTBASE, VIA_MMIO_REGSIZE,
    		  DRM_REGISTERS, flags, &pVia->hostBltHandle) < 0) {
    	    return FALSE;
        }
         
    	VIAInitS3gConf(pScreen);
    	VIAMapAGP(pScreen, pVia);
    }
    xf86DrvMsg(pScreen->myNum, X_INFO,
	"[drm] register handle = 0x%08lx, HostBlt handle = 0x%08lx\n", pVia->registerHandle, pVia->hostBltHandle);

    return TRUE;
}

/*following functions are only used by H5s1_UMA(vt3336/vt3364)*/
int VIAMapAGP(ScreenPtr pScreen, VIAPtr pVia)
{
    ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
    DriConfPtr pDriConf = pVia->pDriConf;
    ChipConfPtr pChipConf = pVia->pChipConf;
    AddrConfPtr pAddrConf = pVia->pAddrConf;
    ScrnConfPtr pScrnConf = pVia->pScrnConf;
	/*get pDriConf agp para*/
    pDriConf->agpHandle = pVia->agpHandle;
    pDriConf->agpBase = (unsigned long)pVia->agpBase;
    pDriConf->agpSize = pVia->agpSize;
	
    /*at first, get mmio handle from pVia*/
    pDriConf->mmioHandle = pVia->registerHandle;

    /* get linear frame buffer handle */
    int dummy,*dummyptr;
    DRIGetDeviceInfo(pScreen, &pDriConf->fbHandle,
                     &dummy,&dummy,&dummy,&dummy,(void **)&dummyptr);

    /* alloc shadow region */
    pDriConf->shadowSize = 16 * 1024; /* 16K */
    pDriConf->shadowOffset = 0x0;

    /* now remove the shadow region to pci space*/
    #define _DRM_CONSISTENT 5
    if (drmAddMap(pVia->drmFD,
                  0,
                  pDriConf->shadowSize,/* size of shadow region */
                  _DRM_CONSISTENT, 0,
                  &pDriConf->shadowHandle) < 0) {
        xf86DrvMsg(pScreen->myNum, X_ERROR,
                   "[drm] Could not add shadow region mapping\n");
        return FALSE;
    }

    if (!pVia->IsPCI) /*AGP has been enabled*/
    {
        pDriConf->dmaBufAGP = TRUE;   /*tell user space DRI, agp ok*/
            
        /* add dma buffer */
        int dmaBufSize = AGP_CMDBUF_SIZE;
        if (drmAddMap(pVia->drmFD,
                      0x0,/* offset of AGP = 0x0, because shadow is in sys mem */
                      dmaBufSize,/* size of AGP */
                      DRM_AGP, 0,
                      &pDriConf->dmaHandle) < 0) {
            xf86DrvMsg(pScreen->myNum, X_ERROR,
                       "[drm] Could not add DMA buffers mapping\n");
            return FALSE;
        }
        pDriConf->dmaBufOffs = 0x0;

		int AGPTexSize = pVia->agpSize - dmaBufSize;
		if(AGPTexSize > 0)
		{
		    if (drmAddMap(pVia->drmFD,
                          dmaBufSize,/*Follow the AGP DMA buffer*/
                          AGPTexSize,/* size of AGP */
                          DRM_AGP, 0,
                          &pDriConf->agpTexHandle) < 0) {
                xf86DrvMsg(pScreen->myNum, X_ERROR,
                           "[drm] Could not add AGP tex buffers mapping\n");
                return FALSE;
            }

		    pDriConf->agpTexOffset = dmaBufSize;
		    pDriConf->agpTexSize = AGPTexSize;
		}
    }
    else  /*AGP is disable, use PCI*/
    {
        pDriConf->dmaBufAGP = FALSE;   /*tell user module, agp init fail and use PCI*/
    }

    pDriConf->agpMode   = drmAgpGetMode(pVia->drmFD);
    pDriConf->agpVendor = drmAgpVendorId(pVia->drmFD);
    pDriConf->agpDevice = drmAgpDeviceId(pVia->drmFD);

    pDriConf->frontPhyAddr = pVia->FrameBufferBase + pDriConf->frontOffset;
    pDriConf->backPhyAddr = pVia->FrameBufferBase + pDriConf->backOffset;
    pDriConf->depthPhyAddr = pVia->FrameBufferBase + pDriConf->depthOffset;

    /*fill in Rec*/
    strcpy(pChipConf->chipName, pScrn->chipset);
    pChipConf->maxIgas = 2;
    pChipConf->maxApts = 16;
    if(H5_UMA_CHIPID||H6_UMA_CHIPID)
    {
        pChipConf->pciID = pVia->ChipId;
        pChipConf->pciTag = pVia->PciTag;
        pChipConf->primary = 1;
		if(H6_UMA_CHIPID)
			pChipConf->chipType = CHIP_PCIE;
        else
            pChipConf->chipType = CHIP_AGP;
        pChipConf->samm = 0;
        pChipConf->mergefb = 0;
        pChipConf->bcipm = 1; 
        pChipConf->docking = 0;
        pChipConf->swizzleOff = 1;
    }

    pAddrConf->fbVirtAddr = (unsigned long)pVia->FBBase;
    pAddrConf->fbPhyAddr = pVia->FrameBufferBase;
    pAddrConf->fbSize = pScrnConf->fbSize;

    pAddrConf->fbFree = pVia->FBFreeEnd; /*give the available end of fb*/
    pAddrConf->mmioPhyAddr = pVia->MmioBase;
    pAddrConf->mmioVirtAddr = (unsigned long)pVia->MapBase;
    pAddrConf->mmioSize = VIA_MMIO_REGSIZE;
    pAddrConf->bciPhyAddr = pAddrConf->mmioPhyAddr + 0x10000;
    pAddrConf->bciVirtAddr = pAddrConf->mmioVirtAddr + 0x10000;
    pAddrConf->bciSize = 0x10000;
    pAddrConf->aptPhyAddr[0] = pScrnConf->fbPhyAddr;
    pAddrConf->aptPhyAddr[1] = pScrnConf->fbPhyAddr + pScrnConf->   fbOffset;
    pAddrConf->aptVirtAddr[0] = pAddrConf->fbVirtAddr;
    pAddrConf->aptVirtAddr[1] = pAddrConf->fbVirtAddr + pScrnConf->fbOffset;
    pAddrConf->aptPitch[0] = pScrnConf->scrnDelta;
    pAddrConf->aptUsing[0] = 1;
    return TRUE;
}

int VIAReserveFBBuffer(ScreenPtr pScreen)
{
	/*get via rec*/
    ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
	VIAPtr    pVia = VIAPTR(pScrn);
	/*get s3g rec*/
	ScrnConfPtr pScrnConf = VIAPTR(pScrn)->pScrnConf;
	DriConfPtr pDriConf = VIAPTR(pScrn)->pDriConf;
    /*although now 3d will dynamic alloc and free vidmem with drm, these setting seems to be unuseful*/
    /*but such steps will effect set modes in 3d dri*/
    /*when call s3gFillInModes() to setmodes, if (pDevInfo->backOffset != pDevInfo->depthOffset), DRI will think that there is no backbuf*/
    ulong bufSize = pScrnConf->scrnScanline * pScrnConf->scrnDelta;
    int totalBufs = pScrnConf->memScanline / pScrnConf->scrnScanline;
    int texScanline,XBuffers,needBufs;
	Bool flag;

    if (pScrnConf->Bpp == 2) /*in s3g, &&pScrnConf->z24, no use?*/
	{
       	needBufs = 4;
       	flag = TRUE;
    } 
	else 
	{
       	needBufs = 3;
       	flag = FALSE;
    }
    pScrnConf->xaaBufs = (totalBufs - needBufs) / 3;  /*reserve 1/3 of fb for off-screen*/

	XBuffers = pScrnConf->xaaBufs + 1;         
	pScrnConf->XScanline = XBuffers * pScrnConf->scrnScanline;
   	/* reserver memory for back/depth */
    pDriConf->frontOffset = pScrnConf->fbOffset;  
	/*frontOffset means on-screen start?*/
   	pDriConf->backOffset = pVia->FBFreeStart; /*+ XBuffers * bufSize;*/
	/*because we do not use s3g XAA driver, so XBuffers is unuseful?*/
	/*backOffset begins after XAA(offscreen)*/
   	pDriConf->depthOffset = pDriConf->backOffset + bufSize;
	/*depthOffset just after back buffer, backbuffer size = bufSize?so small?*/
  	pDriConf->backY = pDriConf->backOffset / pScrnConf->scrnDelta;
   	pDriConf->depthY = pDriConf->depthOffset / pScrnConf->scrnDelta;
	/*reserve tex buffer*/
    texScanline = pScrnConf->memScanline - ((needBufs-1) + XBuffers) * pScrnConf->scrnScanline;
    if (flag) 
        pDriConf->fbTexOffset = pDriConf->depthOffset + 2 * bufSize;
    else
        pDriConf->fbTexOffset = pDriConf->depthOffset + bufSize;
    pDriConf->fbTexY = pDriConf->fbTexOffset / pScrnConf->scrnDelta;
    pDriConf->fbTexSize = texScanline * pScrnConf->scrnDelta;
    return TRUE;
}
#define TILEDFBINUSE 0
/*collect VIA screen info to fill in that of s3g*/
int VIAInitS3gConf(ScreenPtr pScreen)
{
	/*via rec*/
    ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
	VIAPtr    pVia = VIAPTR(pScrn);
	/*s3g rec*/
	ScrnConfPtr pScrnConf = VIAPTR(pScrn)->pScrnConf;
	DriConfPtr pDriConf = VIAPTR(pScrn)->pDriConf;

    /*get the physical address*/
    pScrnConf->fbPhyAddr=pVia->FrameBufferBase;
    pVia->MmioBase = pVia->PciInfo->memBase[1];

	
	/*copy samm flag*/
	pScrnConf->samm=pVia->pBIOSInfo->SAMM;   
	/*now set the default single-screen*/
	pScrnConf->master=1;
	/*set screen*/
    pScrnConf->virtualX = pScrn->virtualX;
    pScrnConf->virtualY = pScrn->virtualY;
    pScrnConf->bpp = pScrn->bitsPerPixel;
    pScrnConf->Bpp = pScrn->bitsPerPixel >> 3;
    pScrnConf->scrnDelta = pScrn->virtualX * pScrnConf->Bpp;
#if TILEDFBINUSE
    if (pScrnConf->bpp == 8)
    {
#endif
        /* 32: Alignment for nontiled mode */
        pScrnConf->scrnDelta = (pScrnConf->scrnDelta + 31) & (~31);
        pScrnConf->scrnScanline = pScrn->virtualY;
        pScrnConf->scrnWit = pScrnConf->scrnDelta  >> 7;
#if TILEDFBINUSE
    } 
    else 
    {
        /* align to a tile width */
        pScrnConf->scrnDelta = (pScrnConf->scrnDelta + 127) & (~127);
        /* align to a tile height */
        /* 4k per tile */
        pScrnConf->scrnScanline = (pScrn->virtualY + 31) & (~31);
        pScrnConf->scrnWit = pScrnConf->scrnDelta  >> 7;
    }
#endif
    pScrnConf->scrnWip = pScrnConf->scrnDelta / pScrnConf->Bpp;

    /* set framebuffer parameter:size/offset/linear address */
    if (!pScrnConf->samm) 
    {
        /* non-samm case */
        pScrnConf->fbOffset = 0; 
        pScrnConf->fbSize = (unsigned long)pVia->videoRambytes;
        pScrnConf->scrnLinear = (unsigned long)pVia->FBBase;
    } 
    /* allocate offscreen scanline by scanline */
    pScrnConf->memDelta = pScrnConf->scrnDelta;
    /* usable scanline(subtrace the scrnScanline) */
    pScrnConf->memScanline = pScrnConf->fbSize / pScrnConf->memDelta;
    pScrnConf->XScanline = pScrnConf->memScanline;
    /*fb start and fb pitch*/
    pScrnConf->scrnStart = (unsigned long)pVia->FBBase;
    pScrnConf->scrnPitch = pScrnConf->scrnDelta;
    pScrnConf->scrnPhyAddr = pVia->FrameBufferBase;
    pScrnConf->fbPhyAddr = pVia->FrameBufferBase + pScrnConf->fbOffset;
#if TILEDFBINUSE
    pScrnConf->tiled = 1; /*fb is tiled, like clb*/
#else
    pScrnConf->tiled = 0;
#endif
    pScrnConf->driEnabled = 1;
    pScrnConf->accel = 1;
    pScrnConf->video = 1;
#if TILEDFBINUSE
    pScrnConf->scrnTile = pScrnConf->scrnLinear; /*cracked, exactly should be pAddrConf->aptVirtAddr[0]*/
#else
    pScrnConf->scrnTile = 0;
#endif
    pDriConf->drmFd = pVia->drmFD;
    pDriConf->pDRIInfo= pVia->pDRIInfo;

    return TRUE;
}

int VIAS3GFinishScreenInit(ScreenPtr pScreen)
{
    ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
    VIAPtr pVia = VIAPTR(pScrn);
    ScrnConfPtr pScrnConf=VIAPTR(pScrn)->pScrnConf;
    DriConfPtr  pDriConf=VIAPTR(pScrn)->pDriConf;
    ServerInfoPtr pServerInfo = VIAPTR(pScrn)->pServerInfo;

    /*get additional datas for vt3336*/
    memcpy(&pServerInfo->AddrConf, pVia->pAddrConf, sizeof(AddrConfRec));
    memcpy(&pServerInfo->ChipConf, pVia->pChipConf, sizeof(ChipConfRec));
    memcpy(&pServerInfo->ScrnConf, pVia->pScrnConf, sizeof(ScrnConfRec));
    memcpy(&pServerInfo->DriConf, pVia->pDriConf, sizeof(DriConfRec));
    memcpy(&pServerInfo->PciConf, pVia->pPciConf, sizeof(PciConfRec));

    if(H5_UMA_CHIPID||H6_UMA_CHIPID)
    {
        pServerInfo->chipIndex = S3G_INV;
		if(H6_UMA_CHIPID)
			pServerInfo->chipType = CHIP_PCIE;
		else
            pServerInfo->chipType = CHIP_AGP; /*CHIP_AGP*/
        pServerInfo->chipIndex = S3G_INV; /*S3G_CLB*/
		switch(pVia->ChipId)
		{
        case PCI_CHIP_VT3336:
        case PCI_CHIP_VT3364:
			 pServerInfo->chipSubIndex =CHIPID_H5S1;
		    break;

        case PCI_CHIP_VT3293:
			pServerInfo->chipSubIndex =CHIPID_H5;
			break;

        case PCI_CHIP_VT3353:
			pServerInfo->chipSubIndex =CHIPID_H6S2;
			break;

        default:
			break;			   			   	
		}
         /*CHIP_CLB*/
        pServerInfo->isAGP=!pVia->IsPCI;
    }
    
    pServerInfo->sareaPrivOffset = sizeof(XF86DRISAREARec);
    pServerInfo->dmaBufOffs = pDriConf->dmaBufOffs;
    
    pServerInfo->bufGBD = pScrnConf->gbd;
    pServerInfo->bufBpp = pScrnConf->Bpp;
    pServerInfo->bufSize = pScrnConf->scrnScanline * pScrnConf->scrnPitch;
    pServerInfo->bufPitch = pScrnConf->scrnPitch;
    pServerInfo->bufWit = pScrnConf->scrnWit;

    if ((pScrnConf->Bpp == 4)
        || (pScrnConf->Bpp == 2 && !pDriConf->z24)) {
        pServerInfo->zbufGBD = pScrnConf->gbd;
        pServerInfo->zbufBpp = pScrnConf->Bpp;
        pServerInfo->zbufSize = pScrnConf->scrnScanline * pScrnConf->scrnPitch;
        pServerInfo->zbufPitch = pScrnConf->scrnPitch;
        pServerInfo->zbufWit = pScrnConf->scrnWit;
    } else { /* 16bpp mode but 32bpp depth buffer */
        pServerInfo->zbufGBD = (2 * pScrnConf->scrnDelta) | (1<<16) | (1<<18);
        pServerInfo->zbufBpp = 2 * pScrnConf->Bpp;
        pServerInfo->zbufSize = pScrnConf->scrnScanline * (2 * pScrnConf->scrnPitch);
        pServerInfo->zbufPitch = 2 * pScrnConf->scrnPitch;
        pServerInfo->zbufWit = 2 * pScrnConf->scrnWit;
    }
    
    pServerInfo->frontOffset = pDriConf->frontOffset;
    pServerInfo->backOffset = pDriConf->backOffset;
    pServerInfo->depthOffset = pDriConf->depthOffset;
    pServerInfo->backY = pDriConf->backY;
    pServerInfo->depthY = pDriConf->depthY;
    
    pServerInfo->dmaHandle = pDriConf->dmaHandle;
    pServerInfo->frontHandle = pDriConf->frontHandle;
    pServerInfo->backHandle = pDriConf->backHandle;
    pServerInfo->depthHandle = pDriConf->depthHandle;
    pServerInfo->fbTexOffset = pDriConf->fbTexOffset;
    pServerInfo->fbTexSize = pDriConf->fbTexSize;

    pServerInfo->shadowSize = pDriConf->shadowSize;
    pServerInfo->shadowHandle = pDriConf->shadowHandle;
    
    pServerInfo->agpTexOffset = pDriConf->agpTexOffset;
    pServerInfo->agpTexSize = pDriConf->agpTexSize;
    pServerInfo->agpTexHandle = pDriConf->agpTexHandle;

    pServerInfo->mmioHandle = pVia->registerHandle;
    pServerInfo->hostBltHandle = pVia->hostBltHandle;

    pDriConf->allowPageFlip = 0;
    /*sarea init*/
    drm_s3g_sarea_t  *pSAREA = DRIGetSAREAPrivate(pScreen);
    pSAREA->page_flip=0;
    return TRUE;
}


static __GLXvisualConfig * S3GInitVisConf16Bpp(ScreenPtr pScreen,int *numConfigs)
{
    __GLXvisualConfig *pConfigs=NULL ;
    ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
    DriConfPtr  pDriConf=VIAPTR(pScrn)->pDriConf;
    int i, db, stencil, accum;

    *numConfigs = 8;

    pConfigs = (__GLXvisualConfig*)xcalloc(sizeof(__GLXvisualConfig),*numConfigs);
    if (!pConfigs) {
        return NULL;
    }
    
    i = 0;
    for (db = 0 ; db <= 1 ; db++) 
    { /* for accum buffer */
        for (accum = 0 ; accum <= 1 ; accum++) 
        { /* for stencil buffer */
            for (stencil = 0 ; stencil <= 1 ; stencil++) 
            { /* for depth buffer */
                pConfigs[i].vid         = -1;
                pConfigs[i].class       = -1;
                pConfigs[i].rgba        = TRUE;
                pConfigs[i].redSize     = 5;
                pConfigs[i].greenSize   = 6;
                pConfigs[i].blueSize    = 5;
                pConfigs[i].alphaSize   = 0;
                pConfigs[i].redMask     = 0x0000F800;
                pConfigs[i].greenMask   = 0x000007E0;
                pConfigs[i].blueMask    = 0x0000001F;
                pConfigs[i].alphaMask   = 0;
                if (accum) 
                {
                    pConfigs[i].accumRedSize    = 16;
                    pConfigs[i].accumGreenSize  = 16;
                    pConfigs[i].accumBlueSize   = 16;
                    pConfigs[i].accumAlphaSize  = 0;
                } 
                else 
                {
                    pConfigs[i].accumRedSize    = 0;
                    pConfigs[i].accumGreenSize  = 0;
                    pConfigs[i].accumBlueSize   = 0;
                    pConfigs[i].accumAlphaSize  = 0;
                }
                
                if (!db) 
                {
                    pConfigs[i].doubleBuffer    = TRUE;
                } 
                else 
                {
                    pConfigs[i].doubleBuffer    = FALSE;
                }
                pConfigs[i].stereo        = FALSE;
                pConfigs[i].bufferSize    = 16;             
                if (pDriConf->z24) 
                { 
                    pConfigs[i].depthSize = 24;
                } 
                else 
                {
                    pConfigs[i].depthSize = 16;
                }
                if (stencil) 
                {
                    pConfigs[i].stencilSize  = 8;
                } 
                else 
                {
                    pConfigs[i].stencilSize  = 0;
                }
                pConfigs[i].auxBuffers       = 0;
                pConfigs[i].level            = 0;
                
                if (accum) 
                {
                    pConfigs[i].visualRating    = GLX_SLOW_CONFIG;
                } 
                else 
                {
                    pConfigs[i].visualRating    = GLX_NONE;
                }
                pConfigs[i].transparentPixel    = GLX_NONE;
                pConfigs[i].transparentRed      = 0;
                pConfigs[i].transparentGreen    = 0;
                pConfigs[i].transparentBlue     = 0;
                pConfigs[i].transparentAlpha    = 0;
                pConfigs[i].transparentIndex    = 0;
                i++;
            }
        }
    }

    return pConfigs;
}


static __GLXvisualConfig * S3GInitVisConf32Bpp(ScreenPtr pScreen,int *numConfigs)
{
    __GLXvisualConfig *pConfigs=NULL ;
    int i, db, accum,stencil;

    *numConfigs = 8;
    pConfigs = (__GLXvisualConfig*)xcalloc(sizeof(__GLXvisualConfig),*numConfigs);
    if (!pConfigs) 
    {
        return NULL;
    }
    
    i = 0;
    for (db = 0 ; db <= 1 ; db++) 
    {
        for (accum = 0 ; accum <= 1 ; accum++) 
        { /* and stencil */
            for (stencil = 0 ; stencil <= 1 ; stencil++) 
            {
                pConfigs[i].vid         = -1;
                pConfigs[i].class       = -1;
                pConfigs[i].rgba        = TRUE;
                pConfigs[i].redSize     = 8;
                pConfigs[i].greenSize   = 8;
                pConfigs[i].blueSize    = 8;
                pConfigs[i].alphaSize   = 8;
                pConfigs[i].redMask     = 0x00FF0000;
                pConfigs[i].greenMask   = 0x0000FF00;
                pConfigs[i].blueMask    = 0x000000FF;
                pConfigs[i].alphaMask   = 0xFF000000;
                if (accum) 
                {
                    pConfigs[i].accumRedSize    = 16;
                    pConfigs[i].accumGreenSize  = 16;
                    pConfigs[i].accumBlueSize   = 16;
                    pConfigs[i].accumAlphaSize  = 16;
                }
                else 
                {
                    pConfigs[i].accumRedSize    = 0;
                    pConfigs[i].accumGreenSize  = 0;
                    pConfigs[i].accumBlueSize   = 0;
                    pConfigs[i].accumAlphaSize  = 0;
                }
                if (!db) 
                {
                    pConfigs[i].doubleBuffer    = TRUE;
                } else 
                {
                    pConfigs[i].doubleBuffer    = FALSE;
                }
                pConfigs[i].stereo            = FALSE;
                pConfigs[i].bufferSize        = 32;
                pConfigs[i].depthSize     = 24;
                if (stencil) {
                    pConfigs[i].stencilSize   = 8;
                }
                else 
                {
                    pConfigs[i].stencilSize   = 0;
                }
                pConfigs[i].auxBuffers   = 0;
                pConfigs[i].level        = 0;
                if (accum) 
                {
                    pConfigs[i].visualRating    = GLX_SLOW_CONFIG;
                } 
                else 
                {
                    pConfigs[i].visualRating    = GLX_NONE;
                }
                pConfigs[i].transparentPixel    = GLX_NONE;
                pConfigs[i].transparentRed      = 0;
                pConfigs[i].transparentGreen    = 0;
                pConfigs[i].transparentBlue     = 0;
                pConfigs[i].transparentAlpha    = 0;
                pConfigs[i].transparentIndex    = 0;
                i++;
            }
        }
    }
    return pConfigs;    
}

static int S3GInitVisualConfigs(ScreenPtr pScreen)
{
    ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
    DriConfPtr  pDriConf=VIAPTR(pScrn)->pDriConf;
    __GLXvisualConfig *pConfigs=NULL ;
    int numConfigs = 0,i;
    VIAConfigVisualPrivPtr pVIAConfigs=NULL;
  	VIAConfigVisualPrivPtr* pVIAConfigPtrs=NULL;

    if (pScrn->bitsPerPixel == 16) 
    {
        pConfigs = S3GInitVisConf16Bpp(pScreen,&numConfigs);
        if (!(pVIAConfigs = (VIAConfigVisualPrivPtr)xcalloc(sizeof(VIAConfigVisualPrivRec),
  						    numConfigs))) 
  		{
      		xfree(pConfigs);
      		return FALSE;
  	    }
  
  	    if (!(pVIAConfigPtrs = (VIAConfigVisualPrivPtr*)xcalloc(sizeof(VIAConfigVisualPrivPtr),
  							  numConfigs))) {
      		xfree(pConfigs);
      		xfree(pVIAConfigs);
      		return FALSE;
  	    }
  	    for (i=0; i<numConfigs; i++) 
      	{
      	    pVIAConfigs[i].class = -1;
  			pVIAConfigs[i].depth = 16;
      	    pVIAConfigPtrs[i] = &pVIAConfigs[i];
  		}
    } 
    else if (pScrn->bitsPerPixel == 32) 
    {
        pConfigs = S3GInitVisConf32Bpp(pScreen,&numConfigs); 
        if (!(pVIAConfigs = (VIAConfigVisualPrivPtr)xcalloc(sizeof(VIAConfigVisualPrivRec),
  						    numConfigs))) 
  		{
      		xfree(pConfigs);
      		return FALSE;
 	    }
  
  	    if (!(pVIAConfigPtrs = (VIAConfigVisualPrivPtr*)xcalloc(sizeof(VIAConfigVisualPrivPtr),
  							  numConfigs))) {
      		xfree(pConfigs);
      		xfree(pVIAConfigs);
      		return FALSE;
  	    }
  	    for (i=0; i<numConfigs; i++) 
      	{
      	    pVIAConfigs[i].class = -1;
  			pVIAConfigs[i].depth = 32;
      	    pVIAConfigPtrs[i] = &pVIAConfigs[i];
  		}
       
    } 

    pDriConf->pVisualConfigs = pConfigs;
    GlxSetVisualConfigs(numConfigs, pConfigs, (void**)pVIAConfigPtrs);

    return TRUE;

}

