 /*
 * Copyright 1998-2008 VIA Technologies, Inc. All Rights Reserved.
 * Copyright 2001-2008 S3 Graphics, Inc. All Rights Reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sub license,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice (including the
 * next paragraph) shall be included in all copies or substantial portions
 * of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHOR(S) OR COPYRIGHT HOLDER(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR
 * OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
 * ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

 

/* I N C L U D E S ---------------------------------------------------------- */
/*vesper zhao*/
#include "drm.h"
#include "mm.h"

#ifdef XFree86Server
#include "xf86.h"
#include "xf86_ansic.h" /* for Bool */
#include "compiler.h"   /* for inb/outb */
#include "xf86drm.h"    /* for drmOpen, etc. */
#ifdef HAVE_CONFIG_H            /* for define XF86DRI on X11R7(FC5)*/
#include "config_via.h"
#endif
#else
#include "via_types.h"  /* for Bool */
#endif

#include "via.h"
#include "ddmpeg.h"
#include "capture.h"
#include "VidInit.h"
#include "Vinline.h"    /* Universal wait function */
#include "regrec.h"
#include "ddover.h"
#include "VidStruc.h"   /* VIDDATA, video top level structure */
#include "overlay.h"
#include "debug.h"
#include "engmgr.h"     /*ENGINEMANAGER*/

#include <math.h>   /* for sin & cos function */


/* D E F I N E -------------------------------------------------------------- */
#define PI 3.1415926535897932

#define IGA1        1
#define IGA2        2
#define IGA1_IGA2   3

#define SCREEN_INDEX_0  0
#define SCREEN_INDEX_1  1

/* F U N C T I O N  D E C L A R A T I O N ----------------------------------- */

/* E X T E R N   G L O B A L S ---------------------------------------------- */
CLIENT VidClient[CLIENT_PORT_MAX];/*ENGINEMANAGER*/
unsigned long uAreadyInit=0;/*ENGINEMANAGER*/

/* E X T E R N   F U N C T I O N S ------------------------------------------ */

/* G L O B A L   V A R I A B L E S ------------------------------------------ */
VIDEOMEMORY     VidMemory;          /* video memory struct */
unsigned long OverlayToUpdate;
unsigned long dwV3StartAddr_0;
#ifdef XF86DRI
int  drm_fd;
unsigned int  ddmpeg_drm_ctx = 1;
#endif

/*= Added for rejectoverlaysupport =*/
/*DClk table that match values with VESA Timings Spec.*/
DCLK_TABLE    dclktable[]=
{
    /*XRes, YRes, Refresh rate, DClk*/
    {  640, 480,  60,  25},
    {  640, 480,  75,  32},
    {  640, 480,  85,  36},
    {  640, 480, 100,  43},
    {  640, 480, 120,  53},

    {  800, 600,  60,  40},
    {  800, 600,  75,  50},
    {  800, 600,  85,  56},
    {  800, 600, 100,  68},
    {  800, 600, 120,  84},

    { 1024, 768,  60,  65},
    { 1024, 768,  75,  79},
    { 1024, 768,  85,  95},
    { 1024, 768, 100, 113},
    { 1024, 768, 120, 139},

    { 1152, 864,  60,  82},
    { 1152, 864,  75, 108},
    { 1152, 864,  85, 120},
    { 1152, 864, 100, 143},

    { 1280, 768,  60,  80},
    { 1280, 768,  75, 103},
    { 1280, 768,  85, 119},
    { 1280, 768, 100, 142},

    { 1280, 960,  60, 108},
    { 1280, 960,  75, 130},
    { 1280, 960,  85, 149},
    { 1280, 960, 100, 179},

    { 1280,1024,  60, 108},
    { 1280,1024,  75, 135},
    { 1280,1024,  85, 158},
    { 1280,1024, 100, 191},

    { 1400,1050,  60, 123},
    { 1400,1050,  75, 156},
    { 1400,1050,  85, 179},
    { 1400,1050, 100, 214},

    { 1600,1200,  60, 162},
    { 1600,1200,  75, 203},
    { 1600,1200,  85, 230},
    { 1600,1200, 100, 281},

    { 1920,1440,  60, 235},
};

int sz_dclktable = sizeof(dclktable) / sizeof(dclktable[0]);

TV_DCLK_TABLE    TV_dclktable[]=
{
    /*XRes, YRes, DClk*/
    {  640, 480, 30},
    {  800, 600, 48},
    { 1024, 768, 66},
    {  848, 480, 42},
    {  720, 480, 30},

};

int TVsz_dclktable = sizeof(TV_dclktable) / sizeof(TV_dclktable[0]);
/*= End of adding for rejectoverlaysupport =*/

/* F U N C T I O N ----------------------------------------------------------*/

/*  AllocFB
 *  input : FB struct(VIA self define), FB struct(DRM), FB size
 *  return : FB offset
 */

unsigned long VIAAllocFB(PVIDDATA pVidData, ViaMMReq * lpMemRequest, unsigned long dwFBsize)
{
    lpMemRequest->checkID = VIA_MID;
    lpMemRequest->size    = dwFBsize;
    lpMemRequest->type    = VIDMGR_TYPE_VIDEO;
    if ( MEM_OK != viaMgr_vid_ioctl(pVidData, VIAMGR_ALLOC_VIDEO, (char *)(lpMemRequest) ) )
    {
        DBG_DD((" AllocFB fail.\n"));
        return 0;
    }
    return ALIGN_TO_32_BYTE(lpMemRequest->offset);
}

#if defined(XFree86Server)
FBLinearPtr XAllocFB(PVIDDATA pVidData, FBLinearPtr lpFBlinear, unsigned long dwFBsize)
{
    LPVIAGRAPHICINFO lpVIAGraphicInfo = &pVidData->VIAGraphicInfo;
    ScrnInfoPtr pScrn0 = xf86Screens[0];
    ScrnInfoPtr pScrn1 = xf86Screens[1];
  

    if(lpFBlinear)
    {
            xf86FreeOffscreenLinear(lpFBlinear);
            DBG_DD(("xfree86 Manager Free Init_SWOVLinear Offscreen Memory Success!!!! \n"));
    }

    /*Fixed for SAMM with different resolution*/
    if (lpVIAGraphicInfo->SAMM)
    {
        unsigned long Area0= lpVIAGraphicInfo->Screen[0].dwSaveWidth*lpVIAGraphicInfo->Screen[0].dwSaveHeight;
        unsigned long Area1= lpVIAGraphicInfo->Screen[1].dwSaveWidth*lpVIAGraphicInfo->Screen[1].dwSaveHeight;

        if ( Area0 >= Area1 )
        {
            if(!(lpFBlinear = xf86AllocateOffscreenLinear(pScrn0->pScreen, dwFBsize,
                                                  16, NULL, NULL, NULL)))
            {
                xf86PurgeUnlockedOffscreenAreas(pScrn0->pScreen);
                lpFBlinear = xf86AllocateOffscreenLinear(pScrn0->pScreen, dwFBsize,
                                                  16, NULL, NULL, NULL);
                if(!lpFBlinear)
                {
                    ErrorF("SAMM(scrn0):xfree86 Manager allocate Offscreen Memory Fail!!!! \n");
                    return 0;
                }
            }
        }
        else
        {
            if(!(lpFBlinear = xf86AllocateOffscreenLinear(pScrn1->pScreen, dwFBsize,
                                                  16, NULL, NULL, NULL)))
            {
                xf86PurgeUnlockedOffscreenAreas(pScrn1->pScreen);
                lpFBlinear = xf86AllocateOffscreenLinear(pScrn1->pScreen, dwFBsize,
                                                  16, NULL, NULL, NULL);
                if(!lpFBlinear)
                {
                    ErrorF("SAMM(scrn1):xfree86 Manager allocate Offscreen Memory Fail!!!! \n");
                    return 0;
                }
            }
        }
    }
    else
    {
        if(!(lpFBlinear = xf86AllocateOffscreenLinear(pScrn0->pScreen, dwFBsize,
                                                      16, NULL, NULL, NULL)))
        {
            xf86PurgeUnlockedOffscreenAreas(pScrn0->pScreen);

            lpFBlinear = xf86AllocateOffscreenLinear(pScrn0->pScreen, dwFBsize,
                                                     16, NULL, NULL, NULL);
            if (!lpFBlinear)
            {
                ErrorF("xfree86 Manager allocate Offscreen Memory Fail!!!! \n");
                return 0;
            }
        }
    }

    return lpFBlinear;
}
#endif

Bool RejectOverlay(PVIDDATA pVidData)
{
    #ifdef XFree86Server
    unsigned char* pjVideo = pVidData->Vhwinfo.pjVideo;
    #endif
	

    unsigned long DRAMType;
    unsigned long NeedBandwidth,TotalBandwidth;
    int cxScreen[2], cyScreen[2], cBitsPerPel[2], RefreshRate[2];
    int MemEfficiency,MClk,DClk,DClk2,i;
    Bool bTV, bTV2;
    LPVIAGRAPHICINFO lpVIAGraphicInfo = &pVidData->VIAGraphicInfo;

    DClk = 0;
    bTV = 0;

    pVidData->dwUseExtendedFIFO = 0;
    /*initial array */
     memset((void*)cxScreen,0,sizeof(int)*2);
     memset((void*)cyScreen,0,sizeof(int)*2);
     memset((void*)cBitsPerPel,0,sizeof(int)*2);
     memset((void*)RefreshRate,0,sizeof(int)*2);

    #ifdef XFree86Server
    OUTB(0x3D, 0x3D4);
    DRAMType = ((INB(0x3D5)& 0xF0) >> 4);
    #endif


    DBG_DD((" RejectOverlay!!\n"));

    /*SAMM*/
    if ( (lpVIAGraphicInfo->SAMM) && (!pVidData->dwUseExtendedFIFO) )
    {
        bTV2 = 0;
        DClk2 = 0;

        switch(DRAMType)
        {
            case DDR_200:
                 MemEfficiency = (int)SAMM100;
                 MClk = 100;
                 break;
            case DDR_266:
                 MemEfficiency = (int)SAMM133;
                 MClk = 133;
                 break;
            default:
                 /*Unknow DRAM Type .*/
                 DBG_DD(("Unknow DRAM Type "));
                 MemEfficiency = (int)SAMM133;
                 MClk = 133;
                 break;
        }

        if ( lpVIAGraphicInfo->dwActiveDevice & VIA_DEVICE_LCD )
        {
            DBG_DD((" SAMM=> LCD+ELSE !!\n"));

            if ( lpVIAGraphicInfo->Screen[1].dwActiveDevice == VIA_DEVICE_LCD )
            {/* LCD on Secondary */
                cxScreen[0] = lpVIAGraphicInfo->Screen[0].dwWidth;
                cyScreen[0] = lpVIAGraphicInfo->Screen[0].dwHeight;
                cBitsPerPel[0] = lpVIAGraphicInfo->Screen[0].dwBPP;
                RefreshRate[0] = lpVIAGraphicInfo->Screen[0].dwRefreshRate;

                cxScreen[1] = lpVIAGraphicInfo->Screen[1].dwPanelWidth;
                cyScreen[1] = lpVIAGraphicInfo->Screen[1].dwPanelHeight;
                /*DCLK of each mode in the VESA specification,
                  except when LCD panel is 1400x1050, using 1280x1024 instead.*/
                if ((cxScreen[1] == 1400) && (cyScreen[1] == 1050))
                {
                    cxScreen[1] = 1280;
                    cyScreen[1] = 1024;
                }
                cBitsPerPel[1] = lpVIAGraphicInfo->Screen[1].dwBPP;
                RefreshRate[1] = 60;
            }
            else
            {
                cxScreen[0] = lpVIAGraphicInfo->Screen[0].dwPanelWidth;
                cyScreen[0] = lpVIAGraphicInfo->Screen[0].dwPanelHeight;
                /*DCLK of each mode in the VESA specification,
                  except when LCD panel is 1400x1050, using 1280x1024 instead.*/
                if ((cxScreen[0] == 1400) && (cyScreen[0] == 1050))
                {
                    cxScreen[0] = 1280;
                    cyScreen[0] = 1024;
                }
                cBitsPerPel[0] = lpVIAGraphicInfo->Screen[0].dwBPP;
                RefreshRate[0] = 60;

                cxScreen[1] = lpVIAGraphicInfo->Screen[1].dwWidth;
                cyScreen[1] = lpVIAGraphicInfo->Screen[1].dwHeight;
                cBitsPerPel[1] = lpVIAGraphicInfo->Screen[1].dwBPP;
                RefreshRate[1] = lpVIAGraphicInfo->Screen[1].dwRefreshRate;
            }

        }
        else
        {
            DBG_DD((" SAMM=> ELSE!!\n"));
            cxScreen[0] = lpVIAGraphicInfo->Screen[0].dwWidth;
            cyScreen[0] = lpVIAGraphicInfo->Screen[0].dwHeight;
            cBitsPerPel[0] = lpVIAGraphicInfo->Screen[0].dwBPP;
            RefreshRate[0] = lpVIAGraphicInfo->Screen[0].dwRefreshRate;
            cxScreen[1] = lpVIAGraphicInfo->Screen[1].dwWidth;
            cyScreen[1] = lpVIAGraphicInfo->Screen[1].dwHeight;
            cBitsPerPel[1] = lpVIAGraphicInfo->Screen[1].dwBPP;
            RefreshRate[1] = lpVIAGraphicInfo->Screen[1].dwRefreshRate;
        }

        if (bTV)
        {
            for (i = 0; i < TVsz_dclktable; i ++)
            {
                if((TV_dclktable[i].Width  == cxScreen[0]) &&
                   (TV_dclktable[i].Height == cyScreen[0]))
                {
                    DClk = TV_dclktable[i].DClk;
                    break;
                }
            }
        }
        else
        {
            /*Find the DClk from table that has the value from VESA Timings spec.*/
            for (i = 0; i < sz_dclktable; i ++)
            {
                if((dclktable[i].Width  == cxScreen[0]) &&
                   (dclktable[i].Height == cyScreen[0]) &&
                   (dclktable[i].Refresh == RefreshRate[0]))
                {
                    DClk = dclktable[i].DClk;
                    DBG_DD((" DCLK= %d\n",DClk));
                    break;
                }
            }
        }

        if (bTV2)
        {
            for (i = 0; i < TVsz_dclktable; i ++)
            {
                if((TV_dclktable[i].Width  == cxScreen[1]) &&
                   (TV_dclktable[i].Height == cyScreen[1]))
                {
                    DClk2 = TV_dclktable[i].DClk;
                    break;
                }
            }
        }
        else
        {
            for (i = 0; i < sz_dclktable; i ++)
            {
                if((dclktable[i].Width  == cxScreen[1]) &&
                   (dclktable[i].Height == cyScreen[1]) &&
                   (dclktable[i].Refresh == RefreshRate[1]))
                {
                    DClk2 = dclktable[i].DClk;
                    break;
                }

            }
        }

        if (DClk && DClk2)
        {
            NeedBandwidth = (unsigned long)(cBitsPerPel[0]/8*DClk + cBitsPerPel[1]/8*DClk2 + Video_Bpp*max(DClk,DClk2));
            TotalBandwidth = (unsigned long)(MClk*16*MemEfficiency/1000);
            /*TotalBandwidth is divided by 1000 because MemEfficiency was multiplied by 1000*/
            DBG_DD((" SAMM!!\n"));
            DBG_DD((" Screen[0]= %ld x %ld \n",lpVIAGraphicInfo->Screen[0].dwSaveWidth,lpVIAGraphicInfo->Screen[0].dwSaveHeight));
            DBG_DD((" Screen[1]= %ld x %ld \n",lpVIAGraphicInfo->Screen[1].dwSaveWidth,lpVIAGraphicInfo->Screen[1].dwSaveHeight));
            DBG_DD((" cBitsPerPel[0]= %d \n",cBitsPerPel[0]));
            DBG_DD((" cBitsPerPel[1]= %d \n",cBitsPerPel[1]));
            DBG_DD((" Video_Bpp= %d \n",Video_Bpp));
            DBG_DD((" DClk= %d \n",DClk));
            DBG_DD((" DClk2= %d \n",DClk2));
            DBG_DD((" MClk= %d \n",MClk));
            DBG_DD((" MemEfficiency= %d \n",MemEfficiency));
            DBG_DD((" NeedBandwidth= %ld \n",NeedBandwidth));
            DBG_DD((" TotalBandwidth= %ld \n",TotalBandwidth));

            if (NeedBandwidth < TotalBandwidth)
                return FALSE;
            else
                return TRUE;
        }
    }
    else if ( (lpVIAGraphicInfo->SAMM) && (pVidData->dwUseExtendedFIFO) )
    {
        switch (lpVIAGraphicInfo->dwProjectIDRevision)
        {
            case UNICHROME_VT3123AX:
                return TRUE;
                break;
            case UNICHROME_VT3123CX:
            case UNICHROME_VT3205:
            case UNICHROME_VT3204:
            case UNICHROME_VT3259:
            case UNICHROME_VT3259Lite:
            case UNICHROME_VT3314:
            case UNICHROME_VT3327:
                return FALSE;
                break;
            default:
                DBG_DD(("overlay.c : RejectOverlay: Unknow Device ID "));
                return FALSE;
                break;
        }
    }
    else /*Single*/
    {
        cxScreen[0] = lpVIAGraphicInfo->Screen[0].dwWidth;
        cyScreen[0] = lpVIAGraphicInfo->Screen[0].dwHeight;
        cBitsPerPel[0] = lpVIAGraphicInfo->Screen[0].dwBPP;
        RefreshRate[0] = lpVIAGraphicInfo->Screen[0].dwRefreshRate;

        switch(DRAMType)
        {
            case SDR_100:
                MClk = 50; /*HW base on 128 bit*/
                break;
            case SDR_133:
                MClk = 66.5 ;
                break;
            case DDR_200:
                MClk = 100;
                break;
            case DDR_266:
                MClk = 133;
                break;
            default:
                /*Unknow DRAM Type*/
                DBG_DD(("Unknow DRAM Type!\n"));
                MClk = 133;
                break;
        }
        MemEfficiency = (int)SINGLE;
        /*Find the DClk from table that has the value from VESA Timings spec.*/

        if ( lpVIAGraphicInfo->Screen[0].dwActiveDevice == VIA_DEVICE_LCD )
        {
            cxScreen[0] = lpVIAGraphicInfo->Screen[0].dwPanelWidth;
            cyScreen[0] = lpVIAGraphicInfo->Screen[0].dwPanelHeight;
            /*DCLK of each mode in the VESA specification,
              except when LCD panel is 1400x1050, using 1280x1024 instead.*/
            if ((cxScreen[0] == 1400) && (cyScreen[0] == 1050))
            {
                cxScreen[0] = 1280;
                cyScreen[0] = 1024;
            }
            RefreshRate[0] = 60;
        }

        if (bTV)
        {
            for (i = 0; i < TVsz_dclktable; i ++)
            {
                if((TV_dclktable[i].Width  == cxScreen[0]) &&
                   (TV_dclktable[i].Height == cyScreen[0]))
                {
                    DClk = TV_dclktable[i].DClk;
                    break;
                }
            }
        }
        else
        {
            for (i = 0; i < sz_dclktable; i ++)
            {
                if((dclktable[i].Width  == cxScreen[0]) &&
                   (dclktable[i].Height == cyScreen[0]) &&
                   (dclktable[i].Refresh == RefreshRate[0]))
                {
                    DClk = dclktable[i].DClk;
                    break;
                }
            }
        }

        if (DClk)
        {
            NeedBandwidth = (unsigned long)((cBitsPerPel[0]/8 + Video_Bpp)*DClk);
            TotalBandwidth = (unsigned long)(MClk*16*MemEfficiency/1000);
            /*TotalBandwidth is divided by 1000 because MemEfficiency was multiplied by 1000*/
            DBG_DD((" Single!!\n"));
            DBG_DD((" Screen[0]= %d x %d \n",cxScreen[0],cyScreen[0]));
            DBG_DD((" cBitsPerPel= %d \n",cBitsPerPel[0]));
            DBG_DD((" Video_Bpp= %d \n",Video_Bpp));
            DBG_DD((" DClk= %d \n",DClk));
            DBG_DD((" MClk= %d \n",MClk));
            DBG_DD((" MemEfficiency= %d \n",MemEfficiency));
            DBG_DD((" NeedBandwidth= %ld \n",NeedBandwidth));
            DBG_DD((" TotalBandwidth= %ld \n",TotalBandwidth));
            if (NeedBandwidth < TotalBandwidth)
                return FALSE;
            else
                return TRUE;
        }
    }
    return FALSE;
}

Bool RejectOverlay_3123C0(PVIDDATA pVidData)
{
    #ifdef XFree86Server
    unsigned char* pjVideo = pVidData->Vhwinfo.pjVideo;
    #endif
	
    unsigned long DRAMType;
    unsigned long NeedBandwidth,TotalBandwidth;
    int cxScreen[2], cyScreen[2], cBitsPerPel[2], RefreshRate[2];
    int  MemEfficiency,MClk,DClk,DClk2,i;
    Bool bTV, bTV2;
    LPVIAGRAPHICINFO lpVIAGraphicInfo = &pVidData->VIAGraphicInfo;

    DClk = 0;
    bTV = 0;

    /*initial array */
     memset((void*)cxScreen,0,sizeof(int)*2);
     memset((void*)cyScreen,0,sizeof(int)*2);
     memset((void*)cBitsPerPel,0,sizeof(int)*2);
     memset((void*)RefreshRate,0,sizeof(int)*2);
     
    #ifdef XFree86Server
    OUTB(0x3D, 0x3D4);
    DRAMType = ((INB(0x3D5)& 0xF0) >> 4);
    #endif
   DBG_DD((" RejectOverlay_3123C0!!\n"));
    /*SAMM*/
    if (lpVIAGraphicInfo->SAMM)
    {
        bTV2 = 0;
        DClk2 = 0;
        if (((lpVIAGraphicInfo->dwActiveDevice & VIA_DEVICE_CRT1) && (lpVIAGraphicInfo->dwActiveDevice & VIA_DEVICE_LCD))||
            ((lpVIAGraphicInfo->dwActiveDevice & VIA_DEVICE_CRT1) && (lpVIAGraphicInfo->dwActiveDevice & VIA_DEVICE_DFP)))
        {
            /*In DDR_266, We don't support overlay when the pannel resolution is 1600x1200x32 bpp.*/
            /*In DDR_200, We don't support overlay when the pannel resolution is 1600x1200x16 and 1600x1200x32 bpp.*/
            switch(DRAMType)
            {
                case DDR_200:
                    cxScreen[0] = lpVIAGraphicInfo->Screen[0].dwWidth;
                    cyScreen[0] = lpVIAGraphicInfo->Screen[0].dwHeight;
                    cBitsPerPel[0] = lpVIAGraphicInfo->Screen[0].dwBPP;
                    if (cxScreen[0] >= 1600)
                        return TRUE;
                    break;
                case DDR_266:
                    cxScreen[0] = lpVIAGraphicInfo->Screen[0].dwWidth;
                    cyScreen[0] = lpVIAGraphicInfo->Screen[0].dwHeight;
                    cBitsPerPel[0] = lpVIAGraphicInfo->Screen[0].dwBPP;
                    if ((cxScreen[0] == 1600) && (cyScreen[0] == 1200) && (cBitsPerPel[0] == 32))
                        return TRUE;
                    break;
                default:
                    /*Unknow DRAM Type*/
                    DBG_DD(("Unknow DRAM Type!\n"));
                    return TRUE;
                    break;
            }
        }

        switch(DRAMType)
        {
            case DDR_200:
                MemEfficiency = (int)SAMM100_3123C0;
                MClk = 100;
                break;
            case DDR_266:
                MemEfficiency = (int)SAMM133_3123C0;
                MClk = 133;
                break;
            default:
                /*Unknow DRAM Type*/
                DBG_DD(("Unknow DRAM Type!\n"));
                MemEfficiency = (int)SAMM133_3123C0;
                MClk = 133;
                break;
        }

        if ( lpVIAGraphicInfo->dwActiveDevice & VIA_DEVICE_LCD )
        {
            DBG_DD((" SAMM=> LCD+ELSE !!\n"));

            if ( lpVIAGraphicInfo->Screen[1].dwActiveDevice == VIA_DEVICE_LCD )
            {/* LCD on Secondary */
                cxScreen[0] = lpVIAGraphicInfo->Screen[0].dwWidth;
                cyScreen[0] = lpVIAGraphicInfo->Screen[0].dwHeight;
                cBitsPerPel[0] = lpVIAGraphicInfo->Screen[0].dwBPP;
                RefreshRate[0] = lpVIAGraphicInfo->Screen[0].dwRefreshRate;

                cxScreen[1] = lpVIAGraphicInfo->Screen[1].dwPanelWidth;
                cyScreen[1] = lpVIAGraphicInfo->Screen[1].dwPanelHeight;
                /*DCLK of each mode in the VESA specification,
                  except when LCD panel is 1400x1050, using 1280x1024 instead.*/
                if ((cxScreen[1] == 1400) && (cyScreen[1] == 1050))
                {
                    cxScreen[1] = 1280;
                    cyScreen[1] = 1024;
                }
                cBitsPerPel[1] = lpVIAGraphicInfo->Screen[1].dwBPP;
                RefreshRate[1] = 60;
            }
            else
            {
                cxScreen[0] = lpVIAGraphicInfo->Screen[0].dwPanelWidth;
                cyScreen[0] = lpVIAGraphicInfo->Screen[0].dwPanelHeight;
                /*DCLK of each mode in the VESA specification,
                  except when LCD panel is 1400x1050, using 1280x1024 instead.*/
                if ((cxScreen[0] == 1400) && (cyScreen[0] == 1050))
                {
                    cxScreen[0] = 1280;
                    cyScreen[0] = 1024;
                }
                cBitsPerPel[0] = lpVIAGraphicInfo->Screen[0].dwBPP;
                RefreshRate[0] = 60;

                cxScreen[1] = lpVIAGraphicInfo->Screen[1].dwWidth;
                cyScreen[1] = lpVIAGraphicInfo->Screen[1].dwHeight;
                cBitsPerPel[1] = lpVIAGraphicInfo->Screen[1].dwBPP;
                RefreshRate[1] = lpVIAGraphicInfo->Screen[1].dwRefreshRate;
            }

        }
        else
        {
            DBG_DD((" SAMM=> ELSE!!\n"));
            cxScreen[0] = lpVIAGraphicInfo->Screen[0].dwWidth;
            cyScreen[0] = lpVIAGraphicInfo->Screen[0].dwHeight;
            cBitsPerPel[0] = lpVIAGraphicInfo->Screen[0].dwBPP;
            RefreshRate[0] = lpVIAGraphicInfo->Screen[0].dwRefreshRate;
            cxScreen[1] = lpVIAGraphicInfo->Screen[1].dwWidth;
            cyScreen[1] = lpVIAGraphicInfo->Screen[1].dwHeight;
            cBitsPerPel[1] = lpVIAGraphicInfo->Screen[1].dwBPP;
            RefreshRate[1] = lpVIAGraphicInfo->Screen[1].dwRefreshRate;
        }

        if (bTV)
        {
            for (i = 0; i < TVsz_dclktable; i ++)
            {
                if((TV_dclktable[i].Width  == cxScreen[0]) &&
                   (TV_dclktable[i].Height == cyScreen[0]))
                {
                    DClk = TV_dclktable[i].DClk;
                    break;
                }
            }
        }
        else
        {
            /*Find the DClk from table that has the value from VESA Timings spec.*/
            for (i = 0; i < sz_dclktable; i ++)
            {
                if((dclktable[i].Width  == cxScreen[0]) &&
                   (dclktable[i].Height == cyScreen[0]) &&
                   (dclktable[i].Refresh == RefreshRate[0]))
                {
                    DClk = dclktable[i].DClk;
                    DBG_DD((" DCLK= %d\n",DClk));
                    break;
                }
            }
        }

        if (bTV2)
        {
            for (i = 0; i < TVsz_dclktable; i ++)
            {
                if((TV_dclktable[i].Width  == cxScreen[1]) &&
                   (TV_dclktable[i].Height == cyScreen[1]))
                {
                    DClk2 = TV_dclktable[i].DClk;
                    break;
                }
            }
        }
        else
        {
            for (i = 0; i < sz_dclktable; i ++)
            {
                if((dclktable[i].Width  == cxScreen[1]) &&
                   (dclktable[i].Height == cyScreen[1]) &&
                   (dclktable[i].Refresh == RefreshRate[1]))
                {
                    DClk2 = dclktable[i].DClk;
                    break;
                }

            }
        }

        if (DClk && DClk2)
        {
            NeedBandwidth = (unsigned long)(cBitsPerPel[0]/8*DClk + cBitsPerPel[1]/8*DClk2 + Video_Bpp*max(DClk,DClk2));
            TotalBandwidth = (unsigned long)(MClk*16*MemEfficiency/1000);
            /*TotalBandwidth is divided by 1000 because MemEfficiency was multiplied by 1000*/
            DBG_DD((" SAMM!!\n"));
            DBG_DD((" Screen[0]= %ld x %ld \n",lpVIAGraphicInfo->Screen[0].dwSaveWidth,lpVIAGraphicInfo->Screen[0].dwSaveHeight));
            DBG_DD((" Screen[1]= %ld x %ld \n",lpVIAGraphicInfo->Screen[1].dwSaveWidth,lpVIAGraphicInfo->Screen[1].dwSaveHeight));
            DBG_DD((" cBitsPerPel[0]= %d \n",cBitsPerPel[0]));
            DBG_DD((" cBitsPerPel[1]= %d \n",cBitsPerPel[1]));
            DBG_DD((" Video_Bpp= %d \n",Video_Bpp));
            DBG_DD((" DClk= %d \n",DClk));
            DBG_DD((" DClk2= %d \n",DClk2));
            DBG_DD((" MClk= %d \n",MClk));
            DBG_DD((" MemEfficiency= %d \n",MemEfficiency));
            DBG_DD((" NeedBandwidth= %ld \n",NeedBandwidth));
            DBG_DD((" TotalBandwidth= %ld \n",TotalBandwidth));

            if (NeedBandwidth < TotalBandwidth)
                return FALSE;
            else
                return TRUE;
        }
    }
    else /*Single*/
    {
        cxScreen[0] = lpVIAGraphicInfo->Screen[0].dwWidth;
        cyScreen[0] = lpVIAGraphicInfo->Screen[0].dwHeight;
        cBitsPerPel[0] = lpVIAGraphicInfo->Screen[0].dwBPP;
        RefreshRate[0] = lpVIAGraphicInfo->Screen[0].dwRefreshRate;

        switch(DRAMType)
        {
            case SDR_100:
                MClk = 50; /*HW base on 128 bit*/
                break;
            case SDR_133:
                MClk = 66.5 ;
                break;
            case DDR_200:
                MClk = 100;
                break;
            case DDR_266:
                MClk = 133;
                break;
            default:
                /*Unknow DRAM Type*/
                DBG_DD(("Unknow DRAM Type!\n"));
                MClk = 133;
                break;
        }
        MemEfficiency = (int)SINGLE_3123C0;
        /*Find the DClk from table that has the value from VESA Timings spec.*/

        if ( lpVIAGraphicInfo->Screen[0].dwActiveDevice == VIA_DEVICE_LCD )
        {
            cxScreen[0] = lpVIAGraphicInfo->Screen[0].dwPanelWidth;
            cyScreen[0] = lpVIAGraphicInfo->Screen[0].dwPanelHeight;
            /*DCLK of each mode in the VESA specification,
              except when LCD panel is 1400x1050, using 1280x1024 instead.*/
            if ((cxScreen[0] == 1400) && (cyScreen[0] == 1050))
            {
                cxScreen[0] = 1280;
                cyScreen[0] = 1024;
            }
            RefreshRate[0] = 60;
        }

        if (bTV)
        {
            for (i = 0; i < TVsz_dclktable; i ++)
            {
                if((TV_dclktable[i].Width  == cxScreen[0]) &&
                   (TV_dclktable[i].Height == cyScreen[0]))
                {
                    DClk = TV_dclktable[i].DClk;
                    break;
                }
            }
        }
        else
        {
            for (i = 0; i < sz_dclktable; i ++)
            {
                if((dclktable[i].Width  == cxScreen[0]) &&
                   (dclktable[i].Height == cyScreen[0]) &&
                   (dclktable[i].Refresh == RefreshRate[0]))
                {
                    DClk = dclktable[i].DClk;
                    break;
                }
            }
        }

        if (DClk)
        {
            NeedBandwidth = (unsigned long)((cBitsPerPel[0]/8 + Video_Bpp)*DClk);
            TotalBandwidth = (unsigned long)(MClk*16*MemEfficiency/1000);
            /*TotalBandwidth is divided by 1000 because MemEfficiency was multiplied by 1000*/
            DBG_DD((" Single!!\n"));
            DBG_DD((" Screen[0]= %d x %d \n",cxScreen[0],cyScreen[0]));
            DBG_DD((" cBitsPerPel= %d \n",cBitsPerPel[0]));
            DBG_DD((" Video_Bpp= %d \n",Video_Bpp));
            DBG_DD((" DClk= %d \n",DClk));
            DBG_DD((" MClk= %d \n",MClk));
            DBG_DD((" MemEfficiency= %d \n",MemEfficiency));
            DBG_DD((" NeedBandwidth= %ld \n",NeedBandwidth));
            DBG_DD((" TotalBandwidth= %ld \n",TotalBandwidth));
            if (NeedBandwidth < TotalBandwidth)
                return FALSE;
            else
                return TRUE;
        }
    }
    return FALSE;
}

/*  VT3204 DDR200 and DDR266  the bandwidth is not enough to support HW overlay In SAMM or DuoView.
    and Any other modes HW overlay can support.
*/
Bool RejectOverlay_3204(PVIDDATA pVidData)
{
    #ifdef XFree86Server
    unsigned char* pjVideo = pVidData->Vhwinfo.pjVideo;
    #endif
	
    unsigned long DRAMType;
    int cxScreen[2], cyScreen[2], cBitsPerPel[2], RefreshRate[2];
    LPVIAGRAPHICINFO lpVIAGraphicInfo = &pVidData->VIAGraphicInfo;

    /*initial array */
     memset((void*)cxScreen,0,sizeof(int)*2);
     memset((void*)cyScreen,0,sizeof(int)*2);
     memset((void*)cBitsPerPel,0,sizeof(int)*2);
     memset((void*)RefreshRate,0,sizeof(int)*2);

    #ifdef XFree86Server
    OUTB(0x3D, 0x3D4);
    DRAMType = ((INB(0x3D5)& 0xF0) >> 4);
    #endif

	
    DBG_DD((" RejectOverlay_3204!!\n"));
    DBG_DD((" DRAMType =%ld !!\n",DRAMType));

    /*SAMM*/
    if (lpVIAGraphicInfo->SAMM)
    {
        cxScreen[0] = lpVIAGraphicInfo->Screen[0].dwWidth;
        cyScreen[0] = lpVIAGraphicInfo->Screen[0].dwHeight;
        cBitsPerPel[0] = lpVIAGraphicInfo->Screen[0].dwBPP;
        RefreshRate[0] = lpVIAGraphicInfo->Screen[0].dwRefreshRate;


        cxScreen[1] = lpVIAGraphicInfo->Screen[1].dwWidth;
        cyScreen[1] = lpVIAGraphicInfo->Screen[1].dwHeight;
        cBitsPerPel[1] = lpVIAGraphicInfo->Screen[1].dwBPP;
        RefreshRate[1] = lpVIAGraphicInfo->Screen[1].dwRefreshRate;
        
        DBG_DD((" SAMM!!\n"));
        DBG_DD((" Screen[0]= %ld x %ld \n",lpVIAGraphicInfo->Screen[0].dwSaveWidth,lpVIAGraphicInfo->Screen[0].dwSaveHeight));
        DBG_DD((" Screen[1]= %ld x %ld \n",lpVIAGraphicInfo->Screen[1].dwSaveWidth,lpVIAGraphicInfo->Screen[1].dwSaveHeight));
        DBG_DD((" cBitsPerPel[0]= %d \n",cBitsPerPel[0]));
        DBG_DD((" cBitsPerPel[1]= %d \n",cBitsPerPel[1]));
        DBG_DD((" BPP[0]= %d , BPP[1]=%d\n",cBitsPerPel[0],cBitsPerPel[1]));
        DBG_DD((" RefreshRate[0]= %d \n",RefreshRate[0]));
        DBG_DD((" RefreshRate[1]= %d \n",RefreshRate[1]));

	switch (DRAMType)
        {
            case DDR_200:
            case DDR_266:
                if( (cxScreen[0]>= 1600 && cyScreen[0]>= 1200 &&
                       cBitsPerPel[0]==32 && RefreshRate[0]>= 75)  ||
                     (cxScreen[1]>= 1600 && cyScreen[1]>=1200 &&
                       cBitsPerPel[1]==32 && RefreshRate[1]>= 75))
                {
                    return TRUE;
                }
                break;
        }
    }
    else /*Single*/
    {
        cxScreen[0] = lpVIAGraphicInfo->Screen[0].dwWidth;
        cyScreen[0] = lpVIAGraphicInfo->Screen[0].dwHeight;
        cBitsPerPel[0] = lpVIAGraphicInfo->Screen[0].dwBPP;
        RefreshRate[0] = lpVIAGraphicInfo->Screen[0].dwRefreshRate;


        DBG_DD((" Screen[0]= %ld x %ld \n",lpVIAGraphicInfo->Screen[0].dwSaveWidth,lpVIAGraphicInfo->Screen[0].dwSaveHeight));
        DBG_DD((" cBitsPerPel[0]= %d \n",cBitsPerPel[0]));
        DBG_DD((" RefreshRate[0]= %d \n",RefreshRate[0]));
			
	switch (DRAMType)
        {
            case DDR_200:
            case DDR_266:
                if( (cxScreen[0]>= 1600 && cyScreen[0]>= 1200 &&cBitsPerPel[1]==32 && RefreshRate[0]>= 75))
                {
                    return TRUE;
                }
                break;
        }
    }
    
    return FALSE;
}

Bool RejectOverlay_3205(PVIDDATA pVidData)
{
    #ifdef XFree86Server
    unsigned char* pjVideo = pVidData->Vhwinfo.pjVideo;
    #endif

    unsigned long DRAMType;
    unsigned long NeedBandwidth,TotalBandwidth;
    int cxScreen[2], cyScreen[2], cBitsPerPel[2], RefreshRate[2];
    int MemEfficiency,MClk,DClk,DClk2,i;
    Bool bTV, bTV2;
    LPVIAGRAPHICINFO lpVIAGraphicInfo = &pVidData->VIAGraphicInfo;
    LPVIDHWDIFFERENCE lpVideoHWDifference = &pVidData->VideoHWDifference;

    DClk = 0;
    bTV = 0;

    /*initial array */
     memset((void*)cxScreen,0,sizeof(int)*2);
     memset((void*)cyScreen,0,sizeof(int)*2);
     memset((void*)cBitsPerPel,0,sizeof(int)*2);
     memset((void*)RefreshRate,0,sizeof(int)*2);

    #ifdef XFree86Server
    OUTB(0x3D, 0x3D4);
    DRAMType = ((INB(0x3D5)& 0xF0) >> 4);
    #endif

    DBG_DD((" RejectOverlay_3205!!\n"));
    DBG_DD((" DRAMType =%ld !!\n",DRAMType));

    /*SAMM*/
    if (lpVIAGraphicInfo->SAMM)
    {
        bTV2 = 0;
        DClk2 = 0;
        if (((lpVIAGraphicInfo->dwActiveDevice & VIA_DEVICE_CRT1) && (lpVIAGraphicInfo->dwActiveDevice & VIA_DEVICE_LCD))||
            ((lpVIAGraphicInfo->dwActiveDevice & VIA_DEVICE_CRT1) && (lpVIAGraphicInfo->dwActiveDevice & VIA_DEVICE_DFP)))
        {
            /*In DDR_333, We don't support overlay when the pannel resolution is 1600x1200x32 bpp.*/
            /*In DDR_266 and DDR_200, We don't support overlay when the pannel resolution is
              1600x1200x16 and 1600x1200x32 bpp. */
            switch(DRAMType)
            {
                case DDR_200:
                case DDR_266:
                    cxScreen[0] = lpVIAGraphicInfo->Screen[0].dwWidth;
                    cyScreen[0] = lpVIAGraphicInfo->Screen[0].dwHeight;
                    cBitsPerPel[0] = lpVIAGraphicInfo->Screen[0].dwBPP;
                    /*  For 205Ax */
                    if ((cxScreen[0] >= 1600)&&(lpVideoHWDifference->dw205BFIFO64 != VID_HWDIFF_TRUE ))
                    {
                       return TRUE;
                    }
                    /*  For 205Bx, We don't support overlay when LCD Panel Size is 1600x1200x32bpp */
                    else if ( (cxScreen[0] >=1600) && (cBitsPerPel[0] >=32 )&&
                              (lpVideoHWDifference->dw205BFIFO64 == VID_HWDIFF_TRUE))
                    {
                        return TRUE;
                    }
                    break;
                case DDR_333:
                case DDR_400:
                    cxScreen[0] = lpVIAGraphicInfo->Screen[0].dwWidth;
                    cyScreen[0] = lpVIAGraphicInfo->Screen[0].dwHeight;
                    cBitsPerPel[0] = lpVIAGraphicInfo->Screen[0].dwRefreshRate;
                    if ((cxScreen[0] == 1600) && (cyScreen[0] == 1200) && (cBitsPerPel[0] == 32))
                        return TRUE;
                    break;
                default:
                    /*Unknow DRAM Type*/
                    cxScreen[0] = lpVIAGraphicInfo->Screen[0].dwWidth;
                    cyScreen[0] = lpVIAGraphicInfo->Screen[0].dwHeight;
                    cBitsPerPel[0] = lpVIAGraphicInfo->Screen[0].dwRefreshRate;
                    if ((cxScreen[0] == 1600) && (cyScreen[0] == 1200) && (cBitsPerPel[0] == 32))
                        return TRUE;
                    break;
            }
        }

        switch(DRAMType)
        {
            case DDR_200:
                MemEfficiency = (int)SAMM100_3205;
                MClk = 100;
                break;
            case DDR_266:
                /*  For 205B0 */
                if ( lpVideoHWDifference->dw205BFIFO64 == VID_HWDIFF_TRUE )
                {
                    MemEfficiency = (int)SAMM133_3205B;
                }
                else /* 205Ax */
                {
                    MemEfficiency = (int)SAMM133_3205;
                }
                MClk = 133;
                break;
            case DDR_333:
                /*  For 205B0 */
                if ( lpVideoHWDifference->dw205BFIFO64 == VID_HWDIFF_TRUE )
                {
                    MemEfficiency = (int)SAMM166_3205B;
                }
                else /* 205Ax */
                {
                    MemEfficiency = (int)SAMM166_3205;
                }
                MClk = 166;
                break;
            case DDR_400:
                if ( lpVideoHWDifference->dw205BFIFO64 == VID_HWDIFF_TRUE )
                {
                    MemEfficiency = (int)SAMM200_3205B;
                }
                else /* 205Ax */
                {
                    MemEfficiency = (int)SAMM166_3205;
                }

                MClk = 200;
                break;
            default:
                /*Unknow DRAM Type*/
                DBG_DD(("Unknow DRAM Type!\n"));
                MemEfficiency = (int)SAMM166_3205;
                MClk = 166;
                break;
        }

        if ( lpVIAGraphicInfo->dwActiveDevice & VIA_DEVICE_LCD )
        {
            DBG_DD((" SAMM=> LCD+ELSE !!\n"));

            if ( lpVIAGraphicInfo->Screen[1].dwActiveDevice == VIA_DEVICE_LCD )

            {/* LCD on Secondary */
                DBG_DD((" LCD on Secondary !!\n"));
                cxScreen[0] = lpVIAGraphicInfo->Screen[0].dwWidth;
                cyScreen[0] = lpVIAGraphicInfo->Screen[0].dwHeight;
                cBitsPerPel[0] = lpVIAGraphicInfo->Screen[0].dwBPP;
                RefreshRate[0] = lpVIAGraphicInfo->Screen[0].dwRefreshRate;

                cxScreen[1] = lpVIAGraphicInfo->Screen[1].dwPanelWidth;
                cyScreen[1] = lpVIAGraphicInfo->Screen[1].dwPanelHeight;
                /*DCLK of each mode in the VESA specification,
                  except when LCD panel is 1400x1050, using 1280x1024 instead.*/
                if ((cxScreen[1] == 1400) && (cyScreen[1] == 1050))
                {
                    cxScreen[1] = 1280;
                    cyScreen[1] = 1024;
                }
                cBitsPerPel[1] = lpVIAGraphicInfo->Screen[1].dwBPP;
                RefreshRate[1] = 60;
            }
            else
            {
                DBG_DD((" LCD on Primary !!\n"));
                cxScreen[0] = lpVIAGraphicInfo->Screen[0].dwPanelWidth;
                cyScreen[0] = lpVIAGraphicInfo->Screen[0].dwPanelHeight;
                /*DCLK of each mode in the VESA specification,
                  except when LCD panel is 1400x1050, using 1280x1024 instead.*/
                if ((cxScreen[0] == 1400) && (cyScreen[0] == 1050))
                {
                    cxScreen[0] = 1280;
                    cyScreen[0] = 1024;
                }
                cBitsPerPel[0] = lpVIAGraphicInfo->Screen[0].dwBPP;
                RefreshRate[0] = 60;

                cxScreen[1] = lpVIAGraphicInfo->Screen[1].dwWidth;
                cyScreen[1] = lpVIAGraphicInfo->Screen[1].dwHeight;
                cBitsPerPel[1] = lpVIAGraphicInfo->Screen[1].dwBPP;
                RefreshRate[1] = lpVIAGraphicInfo->Screen[1].dwRefreshRate;
            }

        }
        else
        {
            DBG_DD((" SAMM=> ELSE!!\n"));
            cxScreen[0] = lpVIAGraphicInfo->Screen[0].dwWidth;
            cyScreen[0] = lpVIAGraphicInfo->Screen[0].dwHeight;
            cBitsPerPel[0] = lpVIAGraphicInfo->Screen[0].dwBPP;
            RefreshRate[0] = lpVIAGraphicInfo->Screen[0].dwRefreshRate;
            cxScreen[1] = lpVIAGraphicInfo->Screen[1].dwWidth;
            cyScreen[1] = lpVIAGraphicInfo->Screen[1].dwHeight;
            cBitsPerPel[1] = lpVIAGraphicInfo->Screen[1].dwBPP;
            RefreshRate[1] = lpVIAGraphicInfo->Screen[1].dwRefreshRate;
        }

        if (bTV)
        {
            for (i = 0; i < TVsz_dclktable; i ++)
            {
                if((TV_dclktable[i].Width  == cxScreen[0]) &&
                   (TV_dclktable[i].Height == cyScreen[0]))
                {
                    DClk = TV_dclktable[i].DClk;
                    break;
                }
            }
        }
        else
        {
            /*Find the DClk from table that has the value from VESA Timings spec.*/
            for (i = 0; i < sz_dclktable; i ++)
            {
                if((dclktable[i].Width  == cxScreen[0]) &&
                   (dclktable[i].Height == cyScreen[0]) &&
                   (dclktable[i].Refresh == RefreshRate[0]))
                {
                    DClk = dclktable[i].DClk;
                    DBG_DD((" DCLK= %d\n",DClk));
                    break;
                }
            }
        }

        if (bTV2)
        {
            for (i = 0; i < TVsz_dclktable; i ++)
            {
                if((TV_dclktable[i].Width  == cxScreen[1]) &&
                   (TV_dclktable[i].Height == cyScreen[1]))
                {
                    DClk2 = TV_dclktable[i].DClk;
		      DBG_DD((" DCLK2= %d\n",DClk2));
                    break;
                }
            }
        }
        else
        {
            for (i = 0; i < sz_dclktable; i ++)
            {
                if((dclktable[i].Width  == cxScreen[1]) &&
                   (dclktable[i].Height == cyScreen[1]) &&
                   (dclktable[i].Refresh == RefreshRate[1]))
                {
                    DClk2 = dclktable[i].DClk;
                    break;
                }

            }
        }

        if (DClk && DClk2)
        {
            NeedBandwidth = (unsigned long)(cBitsPerPel[0]/8*DClk + cBitsPerPel[1]/8*DClk2 + Video_Bpp*max(DClk,DClk2));
            TotalBandwidth = (unsigned long)(MClk*16*MemEfficiency/1000);
            /*TotalBandwidth is divided by 1000 because MemEfficiency was multiplied by 1000*/
            DBG_DD((" SAMM!!\n"));
            DBG_DD((" Screen[0]= %ld x %ld \n",lpVIAGraphicInfo->Screen[0].dwSaveWidth,lpVIAGraphicInfo->Screen[0].dwSaveHeight));
            DBG_DD((" Screen[1]= %ld x %ld \n",lpVIAGraphicInfo->Screen[1].dwSaveWidth,lpVIAGraphicInfo->Screen[1].dwSaveHeight));
            DBG_DD((" cBitsPerPel[0]= %d \n",cBitsPerPel[0]));
            DBG_DD((" cBitsPerPel[1]= %d \n",cBitsPerPel[1]));
            DBG_DD((" Video_Bpp= %d \n",Video_Bpp));
            DBG_DD((" DClk= %d \n",DClk));
            DBG_DD((" DClk2= %d \n",DClk2));
            DBG_DD((" MClk= %d \n",MClk));
            DBG_DD((" MemEfficiency= %d \n",MemEfficiency));
            DBG_DD((" NeedBandwidth= %ld \n",NeedBandwidth));
            DBG_DD((" TotalBandwidth= %ld \n",TotalBandwidth));
           
	
            if (NeedBandwidth < TotalBandwidth)
                return FALSE;
            else
                return TRUE;
        
        }
    }
    else /*Single*/
    {
        cxScreen[0] = lpVIAGraphicInfo->Screen[0].dwWidth;
        cyScreen[0] = lpVIAGraphicInfo->Screen[0].dwHeight;
        cBitsPerPel[0] = lpVIAGraphicInfo->Screen[0].dwBPP;
        RefreshRate[0] = lpVIAGraphicInfo->Screen[0].dwRefreshRate;

        switch(DRAMType)
        {
            case SDR_100:
                MClk = 50; /*HW base on 128 bit*/
                break;
            case SDR_133:
                MClk = 66.5 ;
                break;
            case DDR_200:
                MClk = 100;
                break;
            case DDR_266:
                MClk = 133;
                break;
            case DDR_333:
                MClk = 166;
                break;
            case DDR_400: /* Added for 3205B0 */
                MClk = 200;
                break;
            default:
                /*Unknow DRAM Type*/
                DBG_DD(("Unknow DRAM Type!\n"));
                MClk = 166;
                break;
        }

        switch(DRAMType)
        {
            case SDR_100:
            case SDR_133:
            case DDR_200:
                MemEfficiency = (int)SINGLE_3205_100;
                break;
            case DDR_266:
            case DDR_333:
                /* For 3205B0 */
                if (lpVideoHWDifference->dw205BFIFO64 == VID_HWDIFF_TRUE )
                {
                    MemEfficiency = (int)SINGLE_3205B_133;
                }
                else  /* for 3205Ax */
                {
                    MemEfficiency = (int)SINGLE_3205_133;
                }
                break;
            case DDR_400:
                /* For 3205B0 */
                if (lpVideoHWDifference->dw205BFIFO64 == VID_HWDIFF_TRUE )
                {
                    MemEfficiency = (int)SINGLE_3205B_133;
                }
                else  /* for 3205Ax */
                {
                    MemEfficiency = (int)SINGLE_3205_133;
                }
                break;
            default:
                /*Unknow DRAM Type .*/
                DBG_DD(("Unknow DRAM Type!\n"));
                MemEfficiency = (int)SINGLE_3205_133;
                break;
        }
        /*MemEfficiency = (int)SINGLE_3205;*/
        /*Find the DClk from table that has the value from VESA Timings spec.*/

        if ( lpVIAGraphicInfo->Screen[0].dwActiveDevice == VIA_DEVICE_LCD )
        {
            cxScreen[0] = lpVIAGraphicInfo->Screen[0].dwPanelWidth;
            cyScreen[0] = lpVIAGraphicInfo->Screen[0].dwPanelHeight;
            /*DCLK of each mode in the VESA specification,
              except when LCD panel is 1400x1050, using 1280x1024 instead.*/
            if ((cxScreen[0] == 1400) && (cyScreen[0] == 1050))
            {
                cxScreen[0] = 1280;
                cyScreen[0] = 1024;
            }
            RefreshRate[0] = 60;
        }

        if (bTV)
        {
            for (i = 0; i < TVsz_dclktable; i ++)
            {
                if((TV_dclktable[i].Width  == cxScreen[0]) &&
                   (TV_dclktable[i].Height == cyScreen[0]))
                {
                    DClk = TV_dclktable[i].DClk;
                    break;
                }
            }
        }
        else
        {
            for (i = 0; i < sz_dclktable; i ++)
            {
                if((dclktable[i].Width  == cxScreen[0]) &&
                   (dclktable[i].Height == cyScreen[0]) &&
                   (dclktable[i].Refresh == RefreshRate[0]))
                   {
                       DClk = dclktable[i].DClk;
                       break;
                   }
            }
        }

        if (DClk)
        {
            NeedBandwidth = (unsigned long)((cBitsPerPel[0]/8 + Video_Bpp)*DClk);
            TotalBandwidth = (unsigned long)(MClk*16*MemEfficiency/1000);
            /*TotalBandwidth is divided by 1000 because MemEfficiency was multiplied by 1000*/
            DBG_DD((" cBitsPerPel= %d : \n",cBitsPerPel[0]));
            DBG_DD((" Video_Bpp= %d : \n",Video_Bpp));
            DBG_DD((" DClk= %d : \n",DClk));
            DBG_DD((" MClk= %d : \n",MClk));
            DBG_DD((" MemEfficiency= %d : \n",MemEfficiency));
            DBG_DD((" NeedBandwidth= %ld : \n",NeedBandwidth));
            DBG_DD((" TotalBandwidth= %ld : \n",TotalBandwidth));
            if (NeedBandwidth < TotalBandwidth)
                return FALSE;
            else
                return TRUE;
        }
    }
    return FALSE;
}
/**************** End of RejectOverlaySupport ***********************/

#ifdef XFree86Server
void SetVideoColor(PVIDDATA pVidData, long Hue, long Saturation, long Brightness, long Contrast)
{
   
    float fA,fB1,fC1,fD,fB2,fC2,fB3,fC3;
    float fPI,fContrast,fSaturation,fHue,fBrightness;
    unsigned long dwA,dwB1,dwC1,dwD,dwB2,dwC2,dwB3,dwC3,dwS,dwColor1,dwColor2;
    unsigned long dwD_Int,dwD_Dec;
    int intD;
    LPVIAGRAPHICINFO lpVIAGraphicInfo = &pVidData->VIAGraphicInfo;
    unsigned char* pjVideo = pVidData->Vhwinfo.pjVideo;

    if (!lpVIAGraphicInfo->IsSupportedColorSpaceChange)
    {
        return;
    }
    
    fPI = (float)(PI/180);

    if((Hue == 0)&&(Saturation == 10000)&&(Brightness == 5000)&&(Contrast == 10000))  /*Default*/
    {
        lpVIAGraphicInfo->ColorSpaceChanged = FALSE;
        switch (lpVIAGraphicInfo->dwProjectIDRevision)
        {
            case UNICHROME_VT3123AX:
                VIDOutD(ColorSpaceReg_1, ColorSpaceValue_1);
                VIDOutD(ColorSpaceReg_2, ColorSpaceValue_2);
                VIDOutD(V3_ColorSpaceReg_1, ColorSpaceValue_1);
                VIDOutD(V3_ColorSpaceReg_2, ColorSpaceValue_2);
                DBG_DD(("00000284 %08x\n",ColorSpaceValue_1));
                DBG_DD(("00000288 %08x\n",ColorSpaceValue_2));
                DBG_DD(("000002C4 %08x\n", ColorSpaceValue_1));
                DBG_DD(("000002C8 %08x\n", ColorSpaceValue_2));
                break;
            case UNICHROME_VT3123CX:
            case UNICHROME_VT3259:
            case UNICHROME_VT3259Lite:
            case UNICHROME_VT3324:
            case UNICHROME_VT3324M:
            case UNICHROME_VT3324M2:
            case UNICHROME_VT3293:    
            case UNICHROME_VT3353:    
                VIDOutD(ColorSpaceReg_1, ColorSpaceValue_1_3123C0);
                VIDOutD(ColorSpaceReg_2, ColorSpaceValue_2_3123C0);
                VIDOutD(V3_ColorSpaceReg_1, ColorSpaceValue_1_3123C0);
                VIDOutD(V3_ColorSpaceReg_2, ColorSpaceValue_2_3123C0);
                DBG_DD(("00000284 %08x\n",ColorSpaceValue_1_3123C0));
                DBG_DD(("00000288 %08x\n",ColorSpaceValue_2_3123C0));
                DBG_DD(("000002C4 %08x\n",ColorSpaceValue_1_3123C0));
                DBG_DD(("000002C8 %08x\n",ColorSpaceValue_2_3123C0));
                break;
            case UNICHROME_VT3327:
            case UNICHROME_VT3336:
            case UNICHROME_VT3364:    
                VIDOutD(ColorSpaceReg_1, ColorSpaceValue_1_3123C0);
                VIDOutD(ColorSpaceReg_2, ColorSpaceValue_2_3123C0);   
                DBG_DD(("00000284 %08x\n",ColorSpaceValue_1_3123C0));
                DBG_DD(("00000288 %08x\n",ColorSpaceValue_2_3123C0));         
                break;
            case UNICHROME_VT3205:
            case UNICHROME_VT3204:
            case UNICHROME_VT3314:
                VIDOutD(V3_ColorSpaceReg_1, ColorSpaceValue_1_32045);
                VIDOutD(V3_ColorSpaceReg_2, ColorSpaceValue_2_32045);
                DBG_DD(("000002C4 %08x\n",ColorSpaceValue_1_32045));
                DBG_DD(("000002C8 %08x\n",ColorSpaceValue_2_32045));

                lpVIAGraphicInfo->ColorSpaceValue1 = ColorSpaceValue_1_32045;
                lpVIAGraphicInfo->ColorSpaceValue2 = ColorSpaceValue_2_32045;
                break;
            default:
                DBG_DD(("overlay.c : SetVideoColor: Unknow Device ID "));
                break;
        }
    }
    else
    {
        lpVIAGraphicInfo->ColorSpaceChanged = TRUE;

        fBrightness = (float)Brightness;
        if (fBrightness == 5000) fBrightness = 5120;
        else if (fBrightness > 5000) fBrightness = 5120*(1+(fBrightness-5000)/4370);
        else fBrightness = 5120*((fBrightness)/5000);
        fBrightness = (float)(-125.9656+251.0721*(fBrightness/10000.));
        fContrast = (float)(Contrast/20000.*1.6645);
        fSaturation = (float)(Saturation/10000.);
        fHue = (float)Hue;

        switch (lpVIAGraphicInfo->dwProjectIDRevision)
        {
            case UNICHROME_VT3123AX:
                fA  = (float)(1.164*fContrast);
                fB1 = (float)(-1.596*fContrast*fSaturation*sin(fHue*fPI));
                fC1 = (float)(1.596*fContrast*fSaturation*cos(fHue*fPI));
                fD  = (float)(1.164*(fBrightness-16));
                fB2 = (float)((0.813*sin(fHue*fPI)-0.391*cos(fHue*fPI))*fContrast*fSaturation);
                fC2 = (float)(-(0.813*cos(fHue*fPI)+0.391*sin(fHue*fPI))*fContrast*fSaturation);
                fB3 = (float)(2.018*fContrast*fSaturation*cos(fHue*fPI));
                fC3 = (float)(2.018*fContrast*fSaturation*sin(fHue*fPI));

                if (fA <= 0)
                    fA = 0;

                if (fA >= 1.9375)
                    fA = 1.9375;

                dwA     = ((unsigned long)(fA*32+1))>>1;
                if (dwA >= 0x1f)
                {
                    dwA = 0x1f;
                }
                else
                {
                    dwA &=0x1f;
                }

                if (fB1 >= 0.75)
                    fB1 = 0.75;

                if (fB1 <= -0.75)
                    fB1 = -0.75;

                if (fB1>=0)
                {
                    dwS = 0;
                }
                else
                {
                    dwS = 1;
                    fB1 = -fB1;
                }

                dwB1    = ((unsigned long)(fB1*8+1))>>1;
                if (dwB1>=0x3)
                {
                    dwB1 = 0x3|(dwS<<2);
                }
                else
                {
                    dwB1 = (dwB1&0x3)|(dwS<<2);
                }

                if (fC1<=1)
                    fC1=1;
                if (fC1>=2.875)
                    fC1=2.875;

                dwC1    = ((unsigned long)(fC1*16+1))>>1;
                if (dwC1>=0x17)
                {
                    dwC1 = 0x17;
                }
                else
                {
                    if (dwC1<=0x8)
                    {
                        dwC1 = 0x8;
                    }
                    else
                    {
                        dwC1 &= 0x1f;
                    }
                }

                if (fD>=127)
                    fD=127;

                if (fD<=-128)
                    fD=-128;

                if (fD>=0)
                {
                    dwS = 0;
                }
                else
                {
                    dwS = 1;
                    fD = fD+128;
                }

                dwD = ((unsigned long)(fD*2+1))>>1;
                if (dwD>=0x7f)
                {
                   dwD = 0x7f|(dwS<<7);
                }
                else
                {
                    dwD = (dwD&0x7f)|(dwS<<7);
                }

                if (fB2>=0)
                    fB2=0;

                if (fB2<=-0.875)
                    fB2=-0.875;

                if (fB2>=0)
                    dwB2 = 0;
                else
                {
                    fB2 = -fB2;
                    dwB2    = ((unsigned long)(fB2*16+1))>>1;
                }

                if (dwB2>=0x7)
                {
                    dwB2 = 0x7;
                }
                else
                {
                    dwB2 &= 0x7;
                }

                if (fC2>=0)
                    fC2=0;

                if (fC2<=-1.875)
                    fC2=-1.875;

                if (fC2>=0)
                    dwC2 = 0;
                else
                {
                    fC2 = -fC2;
                    dwC2    = ((unsigned long)(fC2*16+1))>>1;
                }

                if (dwC2>=0xf)
                {
                    dwC2 = 0xf;
                }
                else
                {
                    dwC2 &= 0xf;
                }

                if (fB3>=3.75)
                    fB3=3.75;

                if (fB3<=0)
                    fB3=0;

                if (fB3<=0)
                    dwB3 = 0;
                else
                {
                    dwB3    = ((unsigned long)(fB3*8+1))>>1;
                }

                if (dwB3>=0xf)
                {
                    dwB3 = 0xf;
                }
                else
                {
                    dwB3 &= 0xf;
                }

                if (fC3>=1.25)
                    fC3=1.25;

                if (fC3<=-1.25)
                    fC3=-1.25;

                if (fC3>=0)
                {
                     dwS = 0;
                 }
                 else
                 {
                     dwS = 1;
                   fC3 = -fC3;
                }

                dwC3    = ((unsigned long)(fC3*8+1))>>1;
                if (dwC3 >= 0x7)
                {
                    dwC3 = 0x7 |(dwS<<3);
                }
                else
                {
                    dwC3 = (dwC3&0x7)|(dwS<<3);
                }

                dwColor1 = (dwA<<24)|(dwB1<<18)|(dwC1<<9)|dwD;
                dwColor2 = (dwB2<<25)|(dwC2<<17)|(dwB3<<10)|(dwC3<<2);

                VIDOutD(ColorSpaceReg_1, dwColor1);
                VIDOutD(ColorSpaceReg_2, dwColor2);
                DBG_DD(("00000284 %08lx\n",dwColor1));
                DBG_DD(("00000288 %08lx\n",dwColor2));

                lpVIAGraphicInfo->ColorSpaceValue1 = dwColor1;
                lpVIAGraphicInfo->ColorSpaceValue2 = dwColor2;
                break;
            case UNICHROME_VT3123CX:
            case UNICHROME_VT3259:
            case UNICHROME_VT3259Lite:
            case UNICHROME_VT3324:
            case UNICHROME_VT3324M:
            case UNICHROME_VT3324M2:
            case UNICHROME_VT3327:
            case UNICHROME_VT3293: 
            case UNICHROME_VT3336:
            case UNICHROME_VT3364:    
            case UNICHROME_VT3353:
                fA  = (float)(1.1875*fContrast);
                fB1 = (float)(-1.625*fContrast*fSaturation*sin(fHue*fPI));
                fC1 = (float)(1.625*fContrast*fSaturation*cos(fHue*fPI));
                fD  = (float)(1.1875*(fBrightness-16));
                fB2 = (float)((0.875*sin(fHue*fPI)-0.375*cos(fHue*fPI))*fContrast*fSaturation);
                fC2 = (float)(-(0.875*cos(fHue*fPI)+0.375*sin(fHue*fPI))*fContrast*fSaturation);
                fB3 = (float)(2.0*fContrast*fSaturation*cos(fHue*fPI));
                fC3 = (float)(2.0*fContrast*fSaturation*sin(fHue*fPI));

                if (fA <= 0)
                    fA = 0;

                if (fA >= 1.9375)
                    fA = 1.9375;

                dwA     = ((unsigned long)(fA*32+1))>>1;
                if (dwA >= 0x1f)
                {
                    dwA = 0x1f;
                }
                else
                {
                    dwA &=0x1f;
                }

                if (fB1 >= 2.125)
                    fB1 = 2.125;

                if (fB1 <= -2.125)
                    fB1 = -2.125;

                if (fB1>=0)
                {
                    dwS = 0;
                }
                else
                {
                    dwS = 1;
                    fB1 = -fB1;
                }

                dwB1    = ((unsigned long)(fB1*16+1))>>1;
                if (dwB1>=0x1f)
                {
                    dwB1 = 0x1f|(dwS<<5);
                }
                else
                {
                    dwB1 = (dwB1&0x1f)|(dwS<<5);
                }

                if (fC1<= -2.125)
                    fC1= -2.125;
                if (fC1>= 2.125)
                    fC1= 2.125;


                if (fC1>=0)
                {
                    dwS = 0;
                }
                else
                {
                    dwS = 1;
                    fC1 = -fC1;
                }

                dwC1    = ((unsigned long)(fC1*16+1))>>1;
                if (dwB1>=0x1f)
                {
                    dwB1 = 0x1f|(dwS<<5);
                }
                else
                {
                    dwB1 = (dwB1&0x1f)|(dwS<<5);
                }

                if (fD>=0)
                {
                    intD=(int)fD;
                    if (intD>127)
                        intD = 127;
                    dwD_Int = ((unsigned long)intD)&0xff;
                    dwD = ((unsigned long)(fD*16+1))>>1;
                    dwD_Dec= dwD&0x7;
                }
                else
                {
                    intD=(int)fD;
                    if (intD< -128)
                      intD = -128;
                    intD = intD+256;
                    dwD_Int = ((unsigned long)intD)&0xff;
                    fD = -fD;
                    dwD = ((unsigned long)(fD*16+1))>>1;
                    dwD_Dec= dwD&0x7;
                }

                if (fB2>= 1.875)
                    fB2= 1.875;

                if (fB2<= -1.875)
                    fB2= -1.875;

                if (fB2>=0)
                {
                    dwS = 0;
                }
                else
                {
                   dwS = 1;
                   fB2 = -fB2;
                }

                dwB2    = ((unsigned long)(fB2*16+1))>>1;

                if (dwB2>=0xf)
                {
                    dwB2 = 0xf|(dwS<<4);
                }
                else
                {
                    dwB2 = (dwB2&0xf)|(dwS<<4);
                }

                if (fC2>= 1.875)
                    fC2= 1.875;

                if (fC2<= -1.875)
                    fC2= -1.875;

                if (fC2>=0)
                {
                    dwS = 0;
                }
                else
                {
                    dwS = 1;
                    fC2 = -fC2;
                }
                dwC2    = ((unsigned long)(fC2*16+1))>>1;

                if (dwC2>=0xf)
                {
                    dwC2 = 0xf|(dwS<<4);
                }
                else
                {
                    dwC2 = (dwC2&0xf)|(dwS<<4);
                }

                if (fB3>= 3.875)
                    fB3= 3.875;

                if (fB3<= -3.875)
                    fB3= -3.875;

                if (fB3>=0)
                {
                    dwS = 0;
                }
                else
                {
                    dwS = 1;
                    fB3 = -fB3;
                }
                dwB3    = ((unsigned long)(fB3*16+1))>>1;

                if (dwB3>=0x1f)
                {
                    dwB3 = 0x1f|(dwS<<5);
                }
                else
                {
                    dwB3 = (dwB3&0x1f)|(dwS<<5);
                }

                if (fC3>= 3.875)
                    fC3= 3.875;

                if (fC3<= -3.875)
                    fC3= -3.875;

                if (fC3>=0)
                {
                    dwS = 0;
                }
                else
                {
                    dwS = 1;
                    fC3 = -fC3;
                }

                dwC3    = ((unsigned long)(fC3*16+1))>>1;
                if (dwC3 >= 0x1f)
                {
                    dwC3 = 0x1f |(dwS<<5);
                }
                else
                {
                    dwC3 = (dwC3&0x1f)|(dwS<<5);
                }

                dwColor1 = (dwA<<24)|(dwB1<<16)|(dwC1<<8)|dwD_Int;
                dwColor2 = (dwD_Dec<<29|dwB2<<24)|(dwC2<<16)|(dwB3<<8)|(dwC3);

                VIDOutD(ColorSpaceReg_1, dwColor1);
                VIDOutD(ColorSpaceReg_2, dwColor2);
                VIDOutD(V3_ColorSpaceReg_1, dwColor1);
                VIDOutD(V3_ColorSpaceReg_2, dwColor2);

                DBG_DD(("00000284 %08lx\n",dwColor1));
                DBG_DD(("00000288 %08lx\n",dwColor2));
                DBG_DD(("000002C4 %08lx\n",dwColor1));
                DBG_DD(("000002C8 %08lx\n",dwColor2));

                lpVIAGraphicInfo->ColorSpaceValue1 = dwColor1;
                lpVIAGraphicInfo->ColorSpaceValue2 = dwColor2;
                break;
            case UNICHROME_VT3205:
            case UNICHROME_VT3204:
            case UNICHROME_VT3314:
                fA  = (float)(1.1875*fContrast);
                fB1 = (float)(-1.625*fContrast*fSaturation*sin(fHue*fPI));
                fC1 = (float)(1.625*fContrast*fSaturation*cos(fHue*fPI));
                fD  = (float)(1.1875*(fBrightness-16));
                fB2 = (float)((0.875*sin(fHue*fPI)-0.375*cos(fHue*fPI))*fContrast*fSaturation);
                fC2 = (float)(-(0.875*cos(fHue*fPI)+0.375*sin(fHue*fPI))*fContrast*fSaturation);
                fB3 = (float)(2.0*fContrast*fSaturation*cos(fHue*fPI));
                fC3 = (float)(2.0*fContrast*fSaturation*sin(fHue*fPI));

                if (fA <= 0)
                fA = 0;

                if (fA >= 1.9375)
                    fA = 1.9375;

                dwA     = ((unsigned long)(fA*32+1))>>1;
                if (dwA >= 0x1f)
                {
                    dwA = 0x1f;
                }
                else
                {
                    dwA &=0x1f;
                }

                if (fB1 >= 2.125)
                   fB1 = 2.125;

                if (fB1 <= -2.125)
                    fB1 = -2.125;

                if (fB1>=0)
                {
                    dwS = 0;
                }
                else
                {
                    dwS = 1;
                    fB1 = -fB1;
                }

                dwB1    = ((unsigned long)(fB1*16+1))>>1;
                if (dwB1>=0x1f)
                {
                    dwB1 = 0x1f|(dwS<<5);
                }
                else
                {
                    dwB1 = (dwB1&0x1f)|(dwS<<5);
                }

                if (fC1<= -2.125)
                    fC1= -2.125;
                if (fC1>= 2.125)
                    fC1= 2.125;

                if (fC1>=0)
                {
                            dwS = 0;
                        }
                        else
                        {
                            dwS = 1;
                    fC1 = -fC1;
                }

                dwC1    = ((unsigned long)(fC1*16+1))>>1;
                if (dwB1>=0x1f)
                {
                    dwB1 = 0x1f|(dwS<<5);
                }
                else
                {
                    dwB1 = (dwB1&0x1f)|(dwS<<5);
                }

                if (fD>=0)
                {
                    intD=(int)fD;
                    if (intD>127)
                        intD = 127;
                    dwD_Int = ((unsigned long)intD)&0xff;
                    dwD = ((unsigned long)(fD*16+1))>>1;
                    dwD_Dec= dwD&0x7;
                }
                else
                {
                    intD=(int)fD;
                    if (intD< -128)
                    intD = -128;
                    intD = intD+256;
                    dwD_Int = ((unsigned long)intD)&0xff;
                    fD = -fD;
                    dwD = ((unsigned long)(fD*16+1))>>1;
                    dwD_Dec= dwD&0x7;
                }

                if (fB2>= 1.875)
                    fB2= 1.875;

                if (fB2<= -1.875)
                   fB2= -1.875;

                if (fB2>=0)
                {
                    dwS = 0;
                }
                else
                {
                    dwS = 1;
                    fB2 = -fB2;
                }

                dwB2    = ((unsigned long)(fB2*16+1))>>1;

                if (dwB2>=0xf)
                {
                    dwB2 = 0xf|(dwS<<4);
                }
                else
                {
                    dwB2 = (dwB2&0xf)|(dwS<<4);
                }

                if (fC2>= 1.875)
                    fC2= 1.875;

                if (fC2<= -1.875)
                    fC2= -1.875;

                if (fC2>=0)
                {
                    dwS = 0;
                }
                else
                {
                    dwS = 1;
                    fC2 = -fC2;
                }
                dwC2    = ((unsigned long)(fC2*16+1))>>1;

                if (dwC2>=0xf)
                {
                    dwC2 = 0xf|(dwS<<4);
                }
                else
                {
                    dwC2 = (dwC2&0xf)|(dwS<<4);
                }

                if (fB3>= 3.875)
                    fB3= 3.875;

                if (fB3<= -3.875)
                    fB3= -3.875;

                if (fB3>=0)
                {
                    dwS = 0;
                }
                else
                {
                    dwS = 1;
                    fB3 = -fB3;
                }
                dwB3    = ((unsigned long)(fB3*16+1))>>1;

                if (dwB3>=0x1f)
                {
                    dwB3 = 0x1f|(dwS<<5);
                }
                else
                {
                    dwB3 = (dwB3&0x1f)|(dwS<<5);
                }

                if (fC3>= 3.875)
                    fC3= 3.875;

                if (fC3<= -3.875)
                    fC3= -3.875;

                if (fC3>=0)
                {
                    dwS = 0;
                }
                else
                {
                    dwS = 1;
                    fC3 = -fC3;
                }

                dwC3    = ((unsigned long)(fC3*16+1))>>1;
                if (dwC3 >= 0x1f)
                {
                    dwC3 = 0x1f |(dwS<<5);
                }
                else
                {
                    dwC3 = (dwC3&0x1f)|(dwS<<5);
                }

                dwColor1 = (dwA<<24)|(dwB1<<16)|(dwC1<<8)|dwD_Int;
                dwColor2 = (dwD_Dec<<29|dwB2<<24)|(dwC2<<16)|(dwB3<<8)|(dwC3);

                VIDOutD(V3_ColorSpaceReg_1, dwColor1);
                VIDOutD(V3_ColorSpaceReg_2, dwColor2);
                DBG_DD(("000002C4 %08lx\n",dwColor1));
                DBG_DD(("000002C8 %08lx\n",dwColor2));

                lpVIAGraphicInfo->ColorSpaceValue1 = dwColor1;
                lpVIAGraphicInfo->ColorSpaceValue2 = dwColor2;
                break;
            default:
                DBG_DD(("overlay.c : SetVideoColor: Unknow Device ID "));
                break;
        }/*switch*/
    }/*if else*/
}/*SetVideoColor*/
#endif

#ifdef XFree86Server
unsigned long vfLoadColorSetting(PVIDDATA pVidData, viaPortPrivPtr pPriv)
{
    FILE         *pOverlaySchemeFile = NULL;
    char         filename[40], *home = NULL;
    Bool         ret = FALSE;
    OVERLAYSCHEMERECORD OverlaySchemeRecord;
    LPOVERLAYSCHEMERECORD lpOverlaySchemeRecord = &OverlaySchemeRecord;
    char         chSchemeFormat[] = "HU:%d SA:%d BR:%d CO:%d";
    int count;

    DBG_DD(("vfLoadColorSetting\n"));
    home="/etc/X11";
    if (home)
    {
        filename[0] = '\0';
        strcpy(filename, home);
        strcat(filename, "/.VIAOVERLAYRC");
        if ((pOverlaySchemeFile = fopen(filename, "r")) == NULL)
        {
            DBG_DD(("No Color setting file exist, Create new ones\n"));
        }
        else
        {
            DBG_DD(("Read Color setting file VIAOVERLAYRC \n"));
            fseek(pOverlaySchemeFile, 4L, SEEK_SET);
            count = fread((void*)&OverlaySchemeRecord, sizeof(OverlaySchemeRecord), 1, pOverlaySchemeFile);
            DBG_DD(("count = %d\n", count));
            sscanf(lpOverlaySchemeRecord->chSchemeRecord, chSchemeFormat,
                   &pPriv->hue, &pPriv->saturation, &pPriv->brightness, &pPriv->contrast);                                          

            DBG_DD(("Hue = %ld\n", pPriv->hue));
            DBG_DD(("Saturation = %ld\n", pPriv->saturation));
            DBG_DD(("Brightness = %ld\n", pPriv->brightness));
            DBG_DD(("Contrast = %ld\n", pPriv->contrast));

            SetVideoColor(pVidData, pPriv->hue, pPriv->saturation, pPriv->brightness, pPriv->contrast);

            fclose(pOverlaySchemeFile);
        }
    }
    return ret;

}/*vfLoadColorSetting*/
#endif


unsigned long Flip_CLE(PVIDDATA pVidData, unsigned long DisplayBufferIndex)
{
    unsigned long handle=pVidData->VMI_Info.handle;
    LPSWDEVICE  lpSWDevice  = &(pVidData->SWDevice[handle]);
    unsigned char* pjVideo = pVidData->Vhwinfo.pjVideo;

    LPDDPIXELFORMAT lpDPFsrc    = &pVidData->DPFsrc[handle];
    unsigned long V3_temp;
    unsigned long dwVideoFlag = pVidData->VidStatus.dwVideoFlagSW[handle];
    unsigned long VidBufIndx;
    LPVIAGRAPHICINFO lpVIAGraphicInfo = &pVidData->VIAGraphicInfo;
    
    DBG_DD(("\n"));
    Macro_VidREGRec(VIDREGREC_RESET_COUNTER, 0,0);
    switch(lpDPFsrc->dwFourCC)
    {
        case FOURCC_UYVY:
        case FOURCC_YUY2:
            DBG_DD(("fourcc_uyvy,yuy2,V1[0:2]=%p,%p,%p\n",VIDInD(V1_STARTADDR0),VIDInD(V1_STARTADDR1),VIDInD(V1_STARTADDR2)));
            DBG_DD(("fourcc_uyvy,yuy2,V3[0:2]=%p,%p,%p\n",VIDInD(V3_STARTADDR_0),VIDInD(V3_STARTADDR_1),VIDInD(V3_STARTADDR_2)));
            if (dwVideoFlag & VIDEO_HQV_INUSE)
            {
                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, HQV_SRC_STARTADDR_Y, 
                                lpSWDevice->dwSWPhysicalAddr[DisplayBufferIndex]);
                if(!lpVIAGraphicInfo->rotate)
                {
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, HQV_CONTROL,
                                ( VIDInD(HQV_CONTROL)&~HQV_FLIP_ODD) |HQV_SW_FLIP|HQV_FLIP_STATUS);
                }
                else
                {
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, HQV_CONTROL,
                                ( VIDInD(HQV_CONTROL)&~HQV_FLIP_ODD) |HQV_SW_FLIP|HQV_FLIP_STATUS|HQV_INT_ENABLE);
                }
                    
                WaitHQVDone(pjVideo);
                if(dwVideoFlag & VIDEO_1_ON_SND)
                {
                    V3_temp = VIDInD(V3_CONTROL)&~V3_FLIP_HW_CAPTURE1;
                    V3_temp = V3_temp|V3_SWAP_SW;
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V3_CONTROL, V3_temp);
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V3_STARTADDR_0|0x80000000, pVidData->Vhwinfo.dwV3SrcAddr[(VIDInD(HQV_CONTROL)>>1)&0x03]);
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V_COMPOSE_MODE, 
                                    VIDInD(V_COMPOSE_MODE)|V3_COMMAND_FIRE);
                }
            }
            else
            {
                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V1_STARTADDR0, 
                                lpSWDevice->dwSWPhysicalAddr[DisplayBufferIndex]);
                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V_COMPOSE_MODE, 
                                VIDInD(V_COMPOSE_MODE)|V1_COMMAND_FIRE);

                if (dwVideoFlag & VIDEO_1_INUSE)
                {
					DBG_DD(("wait v3 command fire\n"));
                    WaitVideoCommandFire(pjVideo);
                }
                else
                {  
					DBG_DD(("wait v3 command fire\n"));
                    WaitVideo3CommandFire(pjVideo);
                }
            }

            Macro_VidREGFlush(pjVideo);

            break;

        case FOURCC_YV12:
        case FOURCC_I420:
        default:
            if (dwVideoFlag & VIDEO_HQV_INUSE)
            {
                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, HQV_SRC_STARTADDR_Y, 
                                lpSWDevice->dwSWPhysicalAddr[DisplayBufferIndex]);
                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, HQV_SRC_STARTADDR_U, 
                                lpSWDevice->dwSWCbPhysicalAddr[DisplayBufferIndex]);
                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, HQV_SRC_STARTADDR_V, 
                                lpSWDevice->dwSWCrPhysicalAddr[DisplayBufferIndex]);
                if(!lpVIAGraphicInfo->rotate)
                {
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, HQV_CONTROL, 
                                (VIDInD(HQV_CONTROL)&~HQV_FLIP_ODD) |HQV_SW_FLIP|HQV_FLIP_STATUS);
                }
                else
                {
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, HQV_CONTROL, 
                                (VIDInD(HQV_CONTROL)&~HQV_FLIP_ODD) |HQV_SW_FLIP|HQV_FLIP_STATUS|HQV_INT_ENABLE);
                }
                WaitHQVDone(pjVideo);
                if(dwVideoFlag & VIDEO_1_ON_SND)
                {
                    V3_temp = VIDInD(V3_CONTROL)&~V3_FLIP_HW_CAPTURE1;
                    V3_temp = V3_temp|V3_SWAP_SW;
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V3_CONTROL,V3_temp);
                    VidBufIndx = (((VIDInD(HQV_CONTROL)>>1)&0x03)+2)%3;
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V3_STARTADDR_0|0x80000000, pVidData->Vhwinfo.dwV3SrcAddr[VidBufIndx]);
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V_COMPOSE_MODE, 
                                                                             VIDInD(V_COMPOSE_MODE)|V3_COMMAND_FIRE);
                }
            }
            else
            {
                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V1_STARTADDR0_U, 
                                lpSWDevice->dwSWCbPhysicalAddr[DisplayBufferIndex]);
                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V1_STARTADDR0_V, 
                                lpSWDevice->dwSWCrPhysicalAddr[DisplayBufferIndex]);
                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V1_STARTADDR0, 
                                lpSWDevice->dwSWPhysicalAddr[DisplayBufferIndex]);
                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V_COMPOSE_MODE, 
                                VIDInD(V_COMPOSE_MODE)|V1_COMMAND_FIRE);

                if (dwVideoFlag & VIDEO_1_INUSE)
                {
                    WaitVideoCommandFire(pjVideo);
                }
                else
                {
                    WaitVideo3CommandFire(pjVideo);
                }
            }

            Macro_VidREGFlush(pjVideo);

            break;
    }

    return VMI_OK;
}

unsigned long SetOverlayToUpdate(PVIDDATA pVidData) 
{
    static int first=0;
    if(first==0) {
        first=1;
        return (unsigned long)(VIDEO_1|VIDEO_3);
    }
    switch ( pVidData->VIAGraphicInfo.dwDeviceID )
        {
        case VIA_DEVICE_VT3204:
        case VIA_DEVICE_VT3205:
        case VIA_DEVICE_VT3314:
            return (unsigned long)(VIDEO_1|VIDEO_3);
            break;
        default:
            return pVidData->VIAGraphicInfo.OverlayToUpdate;
            break;
        }
}

unsigned long SaveV3StartAddr(PVIDDATA pVidData,unsigned long dwData0, unsigned long dwData1,unsigned long dwData2)
{
     Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V3_STARTADDR_0, dwData0);
     Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V3_STARTADDR_1, dwData1);

    if(OverlayToUpdate!=VIDEO_1)
    {
          pVidData->Vhwinfo.dwV3SrcAddr[0]=dwData0;	
          pVidData->Vhwinfo.dwV3SrcAddr[1]=dwData1;	
    }
	 
    if ( pVidData->VideoHWDifference.dwThreeHQVBuffer )    /*CLE_C0*/
    {
        Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V3_STARTADDR_2, dwData2);
        if(OverlayToUpdate!=VIDEO_1)
        {
            pVidData->Vhwinfo.dwV3SrcAddr[2]=dwData2;
        }		
    }
    return VMI_OK;	

}



/****************************************************************************
 *
 * Upd_Video()
 *
 ***************************************************************************/
unsigned long Upd_Video(PVIDDATA pVidData)
{
    unsigned long handle=pVidData->VMI_Info.handle;	    
    unsigned char* pjVideo = pVidData->Vhwinfo.pjVideo;
   
    LPVIDHWDIFFERENCE lpVideoHWDifference = &pVidData->VideoHWDifference;
    LPVIAGRAPHICINFO lpVIAGraphicInfo = &pVidData->VIAGraphicInfo;
    LPVIAVIDCTRL  lpVideoControl = &pVidData->VideoControl;
    LPOVERLAYRECORD lpOverlayRecordSW = pVidData->OverlayRecordSet.OverlayRecordSW;
    unsigned long dwVidCtl=0, dwCompose=(VIDInD(V_COMPOSE_MODE)&~(SELECT_VIDEO_IF_COLOR_KEY|V1_COMMAND_FIRE|V3_COMMAND_FIRE))|V_COMMAND_LOAD_VBI;

    unsigned long dwVideoFlag = pVidData->VidStatus.dwVideoFlagSW[handle];
    unsigned long dwStartAddr = VIDInD(HQV_SRC_STARTADDR_Y);
    OV_RECTL rSrc = lpOverlayRecordSW->dwOverlaySrcRect;
    OV_RECTL rDest = lpOverlayRecordSW->dwOverlayDestRect;
    unsigned long dwSrcPitch = pVidData->SWDevice[handle].dwPitch;
    unsigned long dwOriSrcWidth = lpOverlayRecordSW->dwOriSrcWidth;
    unsigned long dwOriSrcHeight = lpOverlayRecordSW->dwOriSrcHeight;
    LPDDPIXELFORMAT lpDPFsrc = &pVidData->DPFsrc[handle];
    unsigned long dwDeinterlaceMode = lpOverlayRecordSW->dwDeinterlaceMode;
    unsigned long dwColorKeyOn = lpOverlayRecordSW->dwColorKeyOn;
    unsigned long dwChromaKeyOn = lpOverlayRecordSW->dwChromaKeyOn;
    unsigned long dwKeyLow = lpOverlayRecordSW->dwKeyLow;
    unsigned long dwChromaLow = lpOverlayRecordSW->dwChromaLow;
    unsigned long dwChromaHigh = lpOverlayRecordSW->dwChromaHigh;
    int dwChromaIC = lpOverlayRecordSW->dwChromaIC;

    unsigned long srcWidth, srcHeight,dstWidth,dstHeight;
    unsigned long zoomCtl=0, miniCtl=0;
    unsigned long dwHQVCtl=0;
    unsigned long dwHQVfilterCtl=0,dwHQVminiCtl=0;
    unsigned long dwHQVzoomflagH=0,dwHQVzoomflagV=0;
    unsigned long dwHQVsrcWidth=0,dwHQVdstWidth=0;
    unsigned long dwHQVsrcFetch = 0,dwHQVoffset=0;
    unsigned long dwOffset=0,dwFetch=0,dwTmp=0;
    unsigned long dwDisplayCountW=0;

    /* For SAMM expand Mode*/
    unsigned long dstModifiedWidth=0,dstModifiedHeight=0;
    /* For debugging samm expand mode*/
    unsigned long  dwV3hvStartLeft=0,dwV3hvStartTop=0,dwV3EndRight=0,dwV3EndBottom=0;
    int ScrnIndex = 0;

    YCBCRREC YCbCr;
    unsigned long dwVDevFlag=pVidData->VDevInfo.dwVDevFlag;

    DBG_DD((" Upd_Video cle,lpVIAGraphicInfo->duoview=%p dwActiveDevice=%p dwVideoFlag=%p\n",lpVIAGraphicInfo->duoview,lpVIAGraphicInfo->dwActiveDevice,dwVideoFlag));

    OverlayToUpdate = SetOverlayToUpdate(pVidData);

    if ( pVidData->VidStatus.dwVideoFlagSW[handle] & SW_ON_SND )
    {
        if(lpVIAGraphicInfo->SAMM) {
            ScrnIndex = (lpVIAGraphicInfo->Screen_IGA_Map[SCREEN_INDEX_0]==IGA2)? SCREEN_INDEX_0:SCREEN_INDEX_1;
        }
        else if(lpVIAGraphicInfo->duoview&DUOVIEW_ENABLED)
        /*duo view mode check if sw overlay is on snd really*/
        {
            if(!(lpVIAGraphicInfo->duoview&DUOVIEW_V3_ON_IGA2))
            {
                pVidData->VidStatus.dwVideoFlagSW[handle] &= ~SW_ON_SND;
                dwVideoFlag = pVidData->VidStatus.dwVideoFlagSW[handle];
            }
        }
        if (lpVIAGraphicInfo->dwActiveDevice==VIA_DEVICE_LCD)
        {
            ScrnIndex = 0;
        }
    }
    else {
        if(lpVIAGraphicInfo->SAMM) {
            ScrnIndex = (lpVIAGraphicInfo->Screen_IGA_Map[SCREEN_INDEX_0]==IGA2)? SCREEN_INDEX_1:SCREEN_INDEX_0;
        }
    }

    if ( pVidData->dwVideoEngineSwitch == TRUE )
    {
        DBG_DD((" VideoEngineSwitch!! \n"));
        Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V1_CONTROL,VIDInD(V1_CONTROL) & ~V1_ENABLE);
        Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V3_CONTROL,VIDInD(V3_CONTROL) & ~V3_ENABLE);
        Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V_COMPOSE_MODE, (VIDInD(V_COMPOSE_MODE)|V3_COMMAND_FIRE|V1_COMMAND_FIRE));

        if (dwVideoFlag & VIDEO_1_INUSE)/*For cle duoview*/
        {
            WaitVideoCommandFire(pjVideo);
        }
        else
        {
            WaitVideo3CommandFire(pjVideo);
        }
        Macro_VidREGFlush(pjVideo);
        Macro_VidREGRec(VIDREGREC_RESET_COUNTER, 0,0);
        pVidData->dwVideoEngineSwitch = FALSE;
    }

    DBG_DD(("Upd_Video: dwVideoFlag=%p \n",dwVideoFlag));
    DBG_DD(("original rSrc  start (%ld,%ld) (w,h)= (%ld,%ld)\n",
                rSrc.left, rSrc.top,rSrc.right-rSrc.left, rSrc.bottom-rSrc.top));
    DBG_DD(("original rDest start (%ld,%ld) (w,h) (%ld,%ld)\n",
                rDest.left, rDest.top,rDest.right-rDest.left, rDest.bottom-rDest.top));

    /* The register 298 of 3123_C1,
     * the V1 V3 register always load and V1 V3 register load VBI are split.
     * So, in V1 path the Compose flag must be or V_COMMAND_LOAD_VBI,
     * and in V3 path the Compose flag must be or V3_COMMAND_LOAD_VBI.
     */
    switch ( DispatchVGARevisionID(pVidData) )
    {
        case VIA_REVISION_CLEAX :
            dwCompose = dwCompose | V_COMMAND_LOAD_VBI;
            break;
        default :
            if (dwVideoFlag & VIDEO_1_INUSE)
                dwCompose = dwCompose | V_COMMAND_LOAD_VBI;
            else
                dwCompose = dwCompose | V3_COMMAND_LOAD_VBI;
            break;
    }

    if (dwVideoFlag & VIDEO_SHOW)
    {
        lpOverlayRecordSW->dwWidth=dstWidth = rDest.right - rDest.left;
        lpOverlayRecordSW->dwHeight=dstHeight = rDest.bottom - rDest.top;
        srcWidth = (unsigned long)rSrc.right - rSrc.left;
        srcHeight = (unsigned long)rSrc.bottom - rSrc.top;
	DBG_DD(("Show overlay \n"));
	DBG_DD(("ori(%ld,%ld) (w,h) (%ld,%ld)\n",rSrc.left,rSrc.top,srcWidth,srcHeight));
        if((dwVideoFlag& SW_ON_SND) || (dwVideoFlag & VIDEO_1_ON_SND))/*add duo view case*/
            dwVDevFlag|=CTL_UseIGA2Path;

        /*Added for SAMM expand Mode*/
        /*In SAMM expand, we need adjust x and y position, width and height zoom factor.*/
        if  ( ((lpVIAGraphicInfo->Screen[ScrnIndex].dwDVIOn)||(lpVIAGraphicInfo->Screen[ScrnIndex].dwLCDOn))
              &&(lpVIAGraphicInfo->Screen[ScrnIndex].dwExpand) )
        {
            dstModifiedWidth=dstWidth*(lpVIAGraphicInfo->Screen[ScrnIndex].dwPanelWidth)/
                             (lpVIAGraphicInfo->Screen[ScrnIndex].dwWidth)+1;
            dstModifiedHeight=dstHeight*(lpVIAGraphicInfo->Screen[ScrnIndex].dwPanelHeight)/
                              (lpVIAGraphicInfo->Screen[ScrnIndex].dwHeight)+1;

		 if((lpVIAGraphicInfo->dwDeviceID == VIA_DEVICE_CLE1)||
		    (lpVIAGraphicInfo->dwDeviceID == VIA_DEVICE_CLE2))
		 {
		     if(!lpVIAGraphicInfo->SAMM && !(lpVIAGraphicInfo->duoview&DUOVIEW_ENABLED)&&
                        !(lpVIAGraphicInfo->dwActiveDevice==VIA_DEVICE_LCD))
             {
                   	dstModifiedWidth=dstWidth;
                   	dstModifiedHeight=dstHeight+1;
                   	DBG_DD((" VIA_DEVICE_CLE Single/Simultaneious mode,hw horizontal lcd scaling!! \n"));
                   }
		 }
            if((lpVIAGraphicInfo->dwDeviceID == VIA_DEVICE_VT3204)||
              (lpVIAGraphicInfo->dwDeviceID == VIA_DEVICE_VT3314))
	        {
                if((lpVIAGraphicInfo->SAMM)||(lpVIAGraphicInfo->duoview&DUOVIEW_ENABLED)
                    || (lpVIAGraphicInfo->dwActiveDevice==VIA_DEVICE_LCD))
                {
               dstModifiedWidth=dstWidth;
               dstModifiedHeight=dstHeight;
                  DBG_DD((" VIA_DEVICE_VT3204 SAMM mode,hw lcd scaling!! \n"));
	        }
               else
               {
                  dstModifiedWidth=dstWidth;
                  DBG_DD((" VIA_DEVICE_VT3204 Single/Simultaneious mode,hw horizontal lcd scaling!! \n"));
               }
	     }
                              
        }
        else
        {
            dstModifiedWidth=dstWidth;
            dstModifiedHeight=dstHeight;
        }
	DBG_DD(("dst(%ld,%ld) modified(w,h) (%ld,%ld)\n",rDest.left,rDest.top,dstModifiedWidth,dstModifiedHeight));

        if (dwVideoFlag & VIDEO_1_INUSE)
        {
            /* Overlay source format for V1 */
           	DBG_DD(("Overlay source format for V1 \n"));
            /*=* Modify for C1 FIFO *=*/
            switch ( DispatchVGARevisionID(pVidData) )
            {
                case VIA_REVISION_CLEAX :
                    if (pVidData->dwUseExtendedFIFO)
                    {
                        dwVidCtl = (V1_ENABLE|V1_EXPIRE_NUM_A|V1_FIFO_EXTENDED);
                    }
                    else
                    {
                        dwVidCtl = (V1_ENABLE|V1_EXPIRE_NUM);
                    }
                    break;
                default:
                    dwVidCtl = (V1_ENABLE|V1_EXPIRE_NUM_F);
                    break;
            }

            DDOver_GetV1Format(pVidData, dwVideoFlag,lpDPFsrc,&dwVidCtl,&dwHQVCtl);
        }
        else
        {
            DBG_DD(("Overlay source format for V3 \n"));
            switch ( lpVIAGraphicInfo->dwDeviceID )
            {
                case VIA_DEVICE_VT3204:
                case VIA_DEVICE_VT3205:
                case VIA_DEVICE_VT3314:
                    dwVidCtl = (V3_ENABLE|V3_EXPIRE_NUM_3205);
                    break;
                case VIA_DEVICE_CLE1:
                case VIA_DEVICE_CLE2:
                    switch ( DispatchVGARevisionID(pVidData) )
                    {
                        case VIA_REVISION_CLECX :
                            dwVidCtl = (V3_ENABLE|V3_EXPIRE_NUM_F);
                            break;
                        default:
                            /* Overlay source format for V3*/
                            dwVidCtl = (V3_ENABLE|V3_EXPIRE_NUM);
                            break;
                    }
                    break;
                default:
                    DBG_DD(("Unknown DeviceID\n"));
                    break;
            }/* End of DeviceID */

            DDOver_GetV3Format(pVidData, dwVideoFlag,lpDPFsrc,&dwVidCtl,&dwHQVCtl);
        }

        if ( lpVideoHWDifference->dwThreeHQVBuffer )    /* CLE_C0 */
        {
            /* HQV support 3 HQV buffer */
	        DBG_DD((" HQV support 3 HQV buffer  \n"));
            dwHQVCtl &= ~HQV_SW_FLIP;
            dwHQVCtl |= HQV_TRIPLE_BUFF | HQV_FLIP_STATUS;
        }

        /* Starting address of source and Source offset */
        dwOffset = DDOver_GetSrcStartAddress (pVidData, dwVideoFlag,rSrc,rDest,dwSrcPitch,lpDPFsrc,&dwHQVoffset );
        DBG_DD(("Starting address of source and Source offset===dwOffset= 0x%lx \n", dwOffset));

        lpOverlayRecordSW->dwOffset = dwOffset;
        if (lpDPFsrc->dwFourCC == FOURCC_YV12)
        {
            DBG_DD(("lpDPFsrc->dwFourCC == FOURCC_YV12 \n"));
            if (dwVideoFlag & VIDEO_HQV_INUSE)
            {
                dwHQVsrcWidth=(unsigned long)rSrc.right - rSrc.left;
                dwHQVdstWidth=(unsigned long)rDest.right - rDest.left;
                /*if (dwHQVsrcWidth>dwHQVdstWidth)
                {
                    dwOffset = dwOffset * dwHQVdstWidth / dwHQVsrcWidth;
                }*/

                if (dwVideoFlag & VIDEO_1_INUSE)
                {
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V1_STARTADDR0, lpOverlayRecordSW->dwHQVAddr[0]+dwOffset);
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V1_STARTADDR1, lpOverlayRecordSW->dwHQVAddr[1]+dwOffset);
                    if ( lpVideoHWDifference->dwThreeHQVBuffer )    /*CLE_C0*/
                        Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V1_STARTADDR2, lpOverlayRecordSW->dwHQVAddr[2]+dwOffset);
                    if(dwVideoFlag & VIDEO_1_ON_SND)/*For cle duoview*/
                    {
                        SaveV3StartAddr(pVidData,lpOverlayRecordSW->dwHQVAddr[0]+dwOffset,lpOverlayRecordSW->dwHQVAddr[1]+dwOffset,lpOverlayRecordSW->dwHQVAddr[2]+dwOffset);                    
                    }
                }
                else
                {
                    SaveV3StartAddr(pVidData,lpOverlayRecordSW->dwHQVAddr[0]+dwOffset,lpOverlayRecordSW->dwHQVAddr[1]+dwOffset,lpOverlayRecordSW->dwHQVAddr[2]+dwOffset);                
                }
				
                YCbCr = DDOVer_GetYCbCrStartAddress(dwVideoFlag,dwStartAddr,lpOverlayRecordSW->dwOffset,lpOverlayRecordSW->dwUVoffset,dwSrcPitch,dwOriSrcHeight);
                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, HQV_SRC_STARTADDR_Y, YCbCr.dwY);
                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, HQV_SRC_STARTADDR_U, YCbCr.dwCR);
                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, HQV_SRC_STARTADDR_V, YCbCr.dwCB);
            }
            else
            {
                YCbCr = DDOVer_GetYCbCrStartAddress(dwVideoFlag,dwStartAddr,lpOverlayRecordSW->dwOffset,lpOverlayRecordSW->dwUVoffset,dwSrcPitch,dwOriSrcHeight);
                if (dwVideoFlag & VIDEO_1_INUSE)
                {
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V1_STARTADDR0, YCbCr.dwY);
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V1_STARTADDR0_U, YCbCr.dwCR);
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V1_STARTADDR0_V, YCbCr.dwCB);
                }
                else
                {
                    DBG_DD(("Upd_Video() : We do not support YV12 with V3!\n"));
                }
            }
        }
        else
        {
            DBG_DD(("lpDPFsrc->dwFourCC == not FOURCC_YV12 \n"));

            if (dwVideoFlag & VIDEO_HQV_INUSE)
            {
                dwHQVsrcWidth=(unsigned long)rSrc.right - rSrc.left;
                dwHQVdstWidth=(unsigned long)rDest.right - rDest.left;
                if (dwHQVsrcWidth>dwHQVdstWidth)
                {
                    dwOffset = dwOffset * dwHQVdstWidth / dwHQVsrcWidth;
                }

                if (dwVideoFlag & VIDEO_1_INUSE)
                {
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V1_STARTADDR0, lpOverlayRecordSW->dwHQVAddr[0]+dwHQVoffset);
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V1_STARTADDR1, lpOverlayRecordSW->dwHQVAddr[1]+dwHQVoffset);
                    if ( lpVideoHWDifference->dwThreeHQVBuffer )    /*CLE_C0*/
                        Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V1_STARTADDR2, lpOverlayRecordSW->dwHQVAddr[2]+dwHQVoffset);
                }
                else
                {
                    SaveV3StartAddr(pVidData,lpOverlayRecordSW->dwHQVAddr[0]+dwOffset,lpOverlayRecordSW->dwHQVAddr[1]+dwOffset,lpOverlayRecordSW->dwHQVAddr[2]+dwOffset);                
                }
                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, HQV_SRC_STARTADDR_Y, dwStartAddr);
            }
            else
            {
                dwStartAddr += dwOffset;

                if (dwVideoFlag & VIDEO_1_INUSE)
                {
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V1_STARTADDR0, dwStartAddr);	
                }
                else
                {
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V3_STARTADDR_0, dwStartAddr);
                     if(OverlayToUpdate!=VIDEO_1)
                     {
                          DBG_DD(("Upd_Video:OverlayToUpdate=%p\n",OverlayToUpdate));   
                          pVidData->Vhwinfo.dwV3SrcAddr[0]=dwStartAddr;	
                     }					
                }
            }
        }

        /*For SAMM expand Mode*/
        dwFetch = DDOver_GetFetch(dwVideoFlag,lpDPFsrc,srcWidth,dstModifiedWidth,dwOriSrcWidth,&dwHQVsrcFetch);
        DBG_DD(("DDOver_GetFetch===dwFetch= 0x%lx \n", dwFetch));

        if (dwVideoFlag & VIDEO_HQV_INUSE)
        {
            if ( !(dwDeinterlaceMode & DDOVER_INTERLEAVED) && (dwDeinterlaceMode & DDOVER_BOB ) )
            {
                if ( lpVideoHWDifference->dwHQVFetchByteUnit )     /* CLE_C0 */
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, HQV_SRC_FETCH_LINE, (((dwHQVsrcFetch)-1)<<16)|((dwOriSrcHeight<<1)-1));
                else
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, HQV_SRC_FETCH_LINE, (((dwHQVsrcFetch>>3)-1)<<16)|((dwOriSrcHeight<<1)-1));
            }
            else
            {
                if ( lpVideoHWDifference->dwHQVFetchByteUnit )     /* CLE_C0 */
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, HQV_SRC_FETCH_LINE, (((dwHQVsrcFetch)-1)<<16)|(dwOriSrcHeight-1));
                else
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, HQV_SRC_FETCH_LINE, (((dwHQVsrcFetch>>3)-1)<<16)|(dwOriSrcHeight-1));
            }
            if (lpDPFsrc->dwFourCC == FOURCC_YV12)
            {
                if (dwVideoFlag & VIDEO_1_INUSE)
                {
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V1_STRIDE, dwSrcPitch<<1);
                    if (dwVideoFlag & VIDEO_1_ON_SND)/*for DuoView*/
                    {
                        Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V3_STRIDE, dwSrcPitch<<1);
                    }
                }
                else
                {
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V3_STRIDE, dwSrcPitch<<1);
                }
                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, HQV_SRC_STRIDE, ((dwSrcPitch>>1)<<16)|dwSrcPitch);
                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, HQV_DST_STRIDE, (dwSrcPitch<<1));
            }
            else
            {
                if (dwVideoFlag & VIDEO_1_INUSE)
                {
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V1_STRIDE, dwSrcPitch);
                    if (dwVideoFlag & VIDEO_1_ON_SND)/*for DuoView*/
                    {
                        Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V3_STRIDE, dwSrcPitch);
                    }
                }
                else
                {
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V3_STRIDE, dwSrcPitch);
                }
                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, HQV_SRC_STRIDE, dwSrcPitch);
                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, HQV_DST_STRIDE, dwSrcPitch);
            }

        }
        else
        {
            if (dwVideoFlag & VIDEO_1_INUSE)
            {
                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V1_STRIDE, dwSrcPitch );
            }
            else
            {
                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V3_STRIDE, dwSrcPitch | (dwSrcPitch <<15) );
            }
        }
        DBG_DD(("Before set video position\n"));

        /* Destination window key */
        if (dwVideoFlag & VIDEO_1_INUSE)
        {
            /* For SAMM expand Mode */
            if  ( ((lpVIAGraphicInfo->Screen[ScrnIndex].dwDVIOn)||(lpVIAGraphicInfo->Screen[ScrnIndex].dwLCDOn))
                  &&(lpVIAGraphicInfo->Screen[ScrnIndex].dwExpand) )
            {
                if((lpVIAGraphicInfo->dwDeviceID == VIA_DEVICE_CLE1)||
                   (lpVIAGraphicInfo->dwDeviceID == VIA_DEVICE_CLE2))
                {
                    if(lpVIAGraphicInfo->SAMM|| (lpVIAGraphicInfo->duoview &DUOVIEW_ENABLED)
                        ||(lpVIAGraphicInfo->dwActiveDevice==VIA_DEVICE_LCD))
                    {
                        Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V1_WIN_END_Y,
                                        ((rDest.right*(lpVIAGraphicInfo->Screen[ScrnIndex].dwPanelWidth)/
                                          (lpVIAGraphicInfo->Screen[ScrnIndex].dwWidth)) <<16) +
                                        (rDest.bottom*(lpVIAGraphicInfo->Screen[ScrnIndex].dwPanelHeight)/
                                         (lpVIAGraphicInfo->Screen[ScrnIndex].dwHeight)) );
                        if (rDest.top > 0)
                        {
                            Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V1_WIN_START_Y,
                                            ((rDest.left*(lpVIAGraphicInfo->Screen[ScrnIndex].dwPanelWidth)/
                                              (lpVIAGraphicInfo->Screen[ScrnIndex].dwWidth)) <<16 ) +
                                            (rDest.top*(lpVIAGraphicInfo->Screen[ScrnIndex].dwPanelHeight)/
                                             (lpVIAGraphicInfo->Screen[ScrnIndex].dwHeight)) );
                        }
                        else
                        {
                            Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V1_WIN_START_Y,
                                            ((rDest.left*(lpVIAGraphicInfo->Screen[ScrnIndex].dwPanelWidth)/
                                              (lpVIAGraphicInfo->Screen[ScrnIndex].dwWidth)) <<16) );
                        }
                        if(lpVIAGraphicInfo->duoview &DUOVIEW_ENABLED)
                        {
                            Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V3_WIN_END_Y,
                                            (rDest.right << 16) + rDest.bottom );
                            if (rDest.top > 0)
                            {
                                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V3_WIN_START_Y,
                                                (rDest.left<<16) + rDest.top);
                            }
                            else
                            {
                                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V3_WIN_START_Y,
                                                rDest.left << 16);
                            }
                        }
                    }
                    else
                    {
                        Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V1_WIN_END_Y,
                                        ((rDest.right-1)<<16) +(rDest.bottom*(lpVIAGraphicInfo->Screen[ScrnIndex].dwPanelHeight)/
                                                                (lpVIAGraphicInfo->Screen[ScrnIndex].dwHeight)));                         
                        if (rDest.top > 0)
                        {
                            Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V1_WIN_START_Y,
                                            (rDest.left <<16 )+(rDest.top*(lpVIAGraphicInfo->Screen[ScrnIndex].dwPanelHeight)/
                                                                (lpVIAGraphicInfo->Screen[ScrnIndex].dwHeight)) );
                        }
                        else
                        {
                            Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V1_WIN_START_Y,
                                            (rDest.left <<16) );
                        }
                    }
                }
                else
                {
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V1_WIN_END_Y,
                                    (((rDest.right-1)*(lpVIAGraphicInfo->Screen[ScrnIndex].dwPanelWidth)/
                                      (lpVIAGraphicInfo->Screen[ScrnIndex].dwWidth)) <<16) +
                                    (rDest.bottom*(lpVIAGraphicInfo->Screen[ScrnIndex].dwPanelHeight)/
                                     (lpVIAGraphicInfo->Screen[ScrnIndex].dwHeight)) );

                    if (rDest.top > 0)
                    {
                        Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V1_WIN_START_Y,
                                        ((rDest.left*(lpVIAGraphicInfo->Screen[ScrnIndex].dwPanelWidth)/
                                          (lpVIAGraphicInfo->Screen[ScrnIndex].dwWidth)) <<16 ) +
                                        (rDest.top*(lpVIAGraphicInfo->Screen[ScrnIndex].dwPanelHeight)/
                                         (lpVIAGraphicInfo->Screen[ScrnIndex].dwHeight)) );
                    }
                    else
                    {
                        Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V1_WIN_START_Y,
                                        ((rDest.left*(lpVIAGraphicInfo->Screen[ScrnIndex].dwPanelWidth)/
                                          (lpVIAGraphicInfo->Screen[ScrnIndex].dwWidth)) <<16) );
                    }
                }
            }
            else
            {
                 Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V1_WIN_END_Y, ((rDest.right-1)<<16) + (rDest.bottom-1));
                 if (rDest.top > 0)
                 {
                     Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V1_WIN_START_Y, (rDest.left<<16) + rDest.top );
                 }
                 else
                 {
                     Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V1_WIN_START_Y, (rDest.left<<16));
                 }
                 if (dwVideoFlag & VIDEO_1_ON_SND)/*for DuoView*/
                 {
                     Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V3_WIN_END_Y, ((rDest.right-1)<<16) + (rDest.bottom-1));
                     if (rDest.top > 0)
                     {
                         Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V3_WIN_START_Y, (rDest.left<<16) + rDest.top );
                     }  
                     else
                     {
                         Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V3_WIN_START_Y, (rDest.left<<16));
                     }
                }
            }
        }
        else
        {
            /* For SAMM expand Mode */
		    DBG_DD(("V3 path, set video position:\n"));
            if  ( ((lpVIAGraphicInfo->Screen[ScrnIndex].dwDVIOn)||(lpVIAGraphicInfo->Screen[ScrnIndex].dwLCDOn))
                  &&(lpVIAGraphicInfo->Screen[ScrnIndex].dwExpand) )
            {
                dwV3EndRight=((rDest.right-1)*(lpVIAGraphicInfo->Screen[ScrnIndex].dwPanelWidth)/(lpVIAGraphicInfo->Screen[ScrnIndex].dwWidth));
                dwV3EndBottom = (rDest.bottom*(lpVIAGraphicInfo->Screen[ScrnIndex].dwPanelHeight)/(lpVIAGraphicInfo->Screen[ScrnIndex].dwHeight));
	            dwV3hvStartLeft=(rDest.left*(lpVIAGraphicInfo->Screen[ScrnIndex].dwPanelWidth)/(lpVIAGraphicInfo->Screen[ScrnIndex].dwWidth));
		        dwV3hvStartTop=(rDest.top*(lpVIAGraphicInfo->Screen[ScrnIndex].dwPanelHeight)/(lpVIAGraphicInfo->Screen[ScrnIndex].dwHeight));

                DBG_DD(("rExpDest  startX (%ld,%ld) (w,h) (%ld,%ld) \n",
	            dwV3hvStartLeft,dwV3hvStartTop,dwV3EndRight-dwV3hvStartLeft,dwV3EndBottom-dwV3hvStartTop));

                if((lpVIAGraphicInfo->dwDeviceID == VIA_DEVICE_VT3204)||
                  (lpVIAGraphicInfo->dwDeviceID == VIA_DEVICE_VT3314))
                {	
                    if((lpVIAGraphicInfo->SAMM)||(lpVIAGraphicInfo->duoview&DUOVIEW_ENABLED)
                        || (lpVIAGraphicInfo->dwActiveDevice==VIA_DEVICE_LCD))
                    {
                    DBG_DD((" VIA_DEVICE_VT3204 set video position: not expaned!! \n"));
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V3_WIN_END_Y, ((rDest.right-1)<<16) + (rDest.bottom-1));
                    if (rDest.top > 0)
                    {
                        Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V3_WIN_START_Y, (rDest.left<<16) + rDest.top );
                    }
                    else
                    {
                        Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V3_WIN_START_Y, (rDest.left<<16));
	                 }
                    }
                    else
                    {
                       DBG_DD(("vt3204 single/simultainous mode,just enable Horizontal scaling funciton\n"));
                       dwCompose |= BYPASS_LCD_VERTICAL_MAGNIFY; /* enable bypass lcd vertical magnify in single/simultaneous */
                       Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V3_WIN_END_Y,
                                     ((rDest.right-1)<<16) +(/*rDest.bottom*/dwV3EndBottom-1)); 
	                  if (rDest.top > 0)
	                  {
	                      Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V3_WIN_START_Y,(rDest.left<<16 ) +/*rDest.top*/dwV3hvStartTop);
	                  }
	                  else
	                  {
	                      Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V3_WIN_START_Y,(rDest.left <<16) );
	                  }
                    }
                }
                else
                {/*not vt3204 chip*/
                    if((lpVIAGraphicInfo->dwDeviceID == VIA_DEVICE_CLE1)||
                       (lpVIAGraphicInfo->dwDeviceID == VIA_DEVICE_CLE2))
                    {
                        if((lpVIAGraphicInfo->SAMM)||
                            (lpVIAGraphicInfo->dwActiveDevice==VIA_DEVICE_LCD))
                        {
                            Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V3_WIN_END_Y,
                                            ((rDest.right*(lpVIAGraphicInfo->Screen[ScrnIndex].dwPanelWidth)/
                                              (lpVIAGraphicInfo->Screen[ScrnIndex].dwWidth)) <<16) +
                                            dwV3EndBottom );
                            if (rDest.top > 0)
                            {
                                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V3_WIN_START_Y,
                                                (dwV3hvStartLeft <<16 ) +dwV3hvStartTop );
                            }
                            else
                            {
                                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V3_WIN_START_Y,
                                                (dwV3hvStartLeft <<16) );
                            }
                        }
                        else
                        {
                            Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V3_WIN_END_Y,
                                            ((rDest.right-1)<<16) +dwV3EndBottom );                         
                            if (rDest.top > 0)
                            {
                                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V3_WIN_START_Y,
                                                (rDest.left <<16 )+dwV3hvStartTop );
                            }
                            else
                            {
                                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V3_WIN_START_Y,
                                                (rDest.left <<16) );
                            }
                        }
                    }
                    else
				
                    {
                        Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V3_WIN_END_Y,
                                        ((dwV3EndRight-1)<<16) +(dwV3EndBottom-1)); 
                        if (rDest.top > 0)
                        {
                            Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V3_WIN_START_Y,(dwV3hvStartLeft<<16 ) +dwV3hvStartTop );
                        }
                        else
                        {
                            Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V3_WIN_START_Y,(dwV3hvStartLeft <<16) );
                        }
                    }
                }
            }
            else
            {
                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V3_WIN_END_Y, ((rDest.right-1)<<16) + (rDest.bottom-1));
                if (rDest.top > 0)
                {
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V3_WIN_START_Y, (rDest.left<<16) + rDest.top );
                }
                else
                {
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V3_WIN_START_Y, (rDest.left<<16));
                }
            }
        }

        dwCompose |= ALWAYS_SELECT_VIDEO;


        /* Setup X zoom factor*/
        lpOverlayRecordSW->dwFetchAlignment = 0;

        /*For SAMM expand Mode*/
        if ( DDOVER_HQVCalcZoomWidth(pVidData, dwVideoFlag, srcWidth , dstModifiedWidth,
                                  &zoomCtl, &miniCtl, &dwHQVfilterCtl, &dwHQVminiCtl,&dwHQVzoomflagH) == VMI_ERR )
        {
            /* too small to handle */
            dwFetch <<= 20;
            if (dwVideoFlag & VIDEO_1_INUSE)
            {
                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V12_QWORD_PER_LINE, dwFetch);
                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V_COMPOSE_MODE , dwCompose|V1_COMMAND_FIRE );
                if (dwVideoFlag & VIDEO_1_ON_SND)/*for DuoView*/
                {
                    dwFetch |=(VIDInD(V3_ALPHA_FETCH_COUNT)&(~V3_FETCH_COUNT));
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V3_ALPHA_FETCH_COUNT, dwFetch);
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V_COMPOSE_MODE , dwCompose|V3_COMMAND_FIRE );
                }
            }
            else
            {
                dwFetch |=(VIDInD(V3_ALPHA_FETCH_COUNT)&(~V3_FETCH_COUNT));
                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V3_ALPHA_FETCH_COUNT, dwFetch);
                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V_COMPOSE_MODE , dwCompose|V3_COMMAND_FIRE );

            }

            if (dwVideoFlag & VIDEO_1_INUSE)
            {
                WaitVideoCommandFire(pjVideo);
                if (dwVideoFlag & VIDEO_1_ON_SND)/*for DuoView*/
                    WaitVideo3CommandFire(pjVideo);
            }
            else
            {
                WaitVideo3CommandFire(pjVideo);
            }
            Macro_VidREGFlush(pjVideo);
            return VMI_ERR;
        }

        dwFetch <<= 20;
        if (dwVideoFlag & VIDEO_1_INUSE)
        {
            Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V12_QWORD_PER_LINE, dwFetch);
            if (dwVideoFlag & VIDEO_1_ON_SND)
            {/*for DuoView*/
                dwFetch |=(VIDInD(V3_ALPHA_FETCH_COUNT)&(~V3_FETCH_COUNT));
                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V3_ALPHA_FETCH_COUNT, dwFetch);
            }
        }
        else
        {
            dwFetch |=(VIDInD(V3_ALPHA_FETCH_COUNT)&(~V3_FETCH_COUNT));
            Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V3_ALPHA_FETCH_COUNT, dwFetch);
        }

        /* Setup Y zoom factor */
        if ( (dwDeinterlaceMode & DDOVER_INTERLEAVED) && (dwDeinterlaceMode & DDOVER_BOB))
        {
            if (!(dwVideoFlag & VIDEO_HQV_INUSE))
            {
                srcHeight /=2;
                if (dwVideoFlag & VIDEO_1_INUSE)
                {
                    dwVidCtl |= (V1_BOB_ENABLE | V1_FRAME_BASE);
                }
                else
                {
                    dwVidCtl |= (V3_BOB_ENABLE | V3_FRAME_BASE);
                }
            }
            else
            {
                dwHQVCtl |= HQV_FIELD_2_FRAME|HQV_FRAME_2_FIELD|HQV_DEINTERLACE;
            }
        }
        else if (dwDeinterlaceMode & DDOVER_BOB )
        {
            if (dwVideoFlag & VIDEO_HQV_INUSE)
            {
                srcHeight <<=1;
                dwHQVCtl |= HQV_FIELD_2_FRAME|HQV_DEINTERLACE;
            }
            else
            {
                if (dwVideoFlag & VIDEO_1_INUSE)
                {
                    dwVidCtl |= V1_BOB_ENABLE;
                }
                else
                {
                    dwVidCtl |= V3_BOB_ENABLE;
                }
            }
        }

        DDOver_GetDisplayCount(pVidData, dwVideoFlag,lpDPFsrc,srcWidth,&dwDisplayCountW);

        if (dwVideoFlag & VIDEO_1_INUSE)
        {
            Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V1_SOURCE_HEIGHT, (srcHeight<<16)|dwDisplayCountW);
            if (dwVideoFlag & VIDEO_1_ON_SND)
            {/*for DuoView*/
                if (lpVideoHWDifference->dwV3SrcHeightSetting == VID_HWDIFF_TRUE)
                {
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V3_SOURCE_WIDTH, 
                                    (srcHeight<<16)|dwDisplayCountW);
                }
                else
                {
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V3_SOURCE_WIDTH, dwDisplayCountW);
                }
            }
        }
        else
        {
            if (lpVideoHWDifference->dwV3SrcHeightSetting == VID_HWDIFF_TRUE)
            {
                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V3_SOURCE_WIDTH, (srcHeight<<16)|dwDisplayCountW);
            }
            else
            {
                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V3_SOURCE_WIDTH, dwDisplayCountW);
            }
        }

        /*For SAMM expand Mode*/
        if ( DDOVER_HQVCalcZoomHeight(srcHeight,dstModifiedHeight,&zoomCtl,&miniCtl, &dwHQVfilterCtl, &dwHQVminiCtl ,&dwHQVzoomflagV) == VMI_ERR )
        {
            if (dwVideoFlag & VIDEO_1_INUSE)
            {
                if(dwVideoFlag & VIDEO_1_ON_SND)/*for duoview*/
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V_COMPOSE_MODE , 
                                    dwCompose|V1_COMMAND_FIRE|V3_COMMAND_FIRE);
                else
                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V_COMPOSE_MODE , dwCompose|V1_COMMAND_FIRE );
            }
            else
            {
                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V_COMPOSE_MODE , dwCompose|V3_COMMAND_FIRE );
            }

            if (dwVideoFlag & VIDEO_1_INUSE)
            {
                WaitVideoCommandFire(pjVideo);
                if(dwVideoFlag & VIDEO_1_ON_SND)
                    WaitVideo3CommandFire(pjVideo);
            }
            else
            {
                WaitVideo3CommandFire(pjVideo);
            }
            Macro_VidREGFlush(pjVideo);
            return VMI_ERR;
        }

        if (miniCtl & V1_Y_INTERPOLY)
        {
           DBG_DD(("miniCtl & V1_Y_INTERPOLY\n"));
            if (lpDPFsrc->dwFourCC == FOURCC_YV12)
            {
                if (dwVideoFlag & VIDEO_HQV_INUSE)
                {
                    DBG_DD(("FOURCC_YV12\n"));
                    if (dwVideoFlag & VIDEO_1_INUSE)
                    {
                        /*=* Modify for C1 FIFO *=*/
                        switch ( DispatchVGARevisionID(pVidData) )
                        {
                            case VIA_REVISION_CLECX :
                                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V_FIFO_CONTROL,
                                    V1_FIFO_DEPTH64 | V1_FIFO_PRETHRESHOLD56 | V1_FIFO_THRESHOLD56);
                                break;

                            default:
                                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V_FIFO_CONTROL,
                                    V1_FIFO_DEPTH32 | V1_FIFO_PRETHRESHOLD29 | V1_FIFO_THRESHOLD16);
                                break;
                        }
                        if(dwVideoFlag & VIDEO_1_ON_SND)
                        {
                            switch ( lpVIAGraphicInfo->dwDeviceID )
                            {
                            case VIA_DEVICE_CLE1:
                            case VIA_DEVICE_CLE2:
                                switch ( DispatchVGARevisionID(pVidData) )
                                {
                                    case VIA_REVISION_CLEAX :
                                        Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_FIFO_CONTROL,
                                            (VIDInD(ALPHA_V3_FIFO_CONTROL)&ALPHA_FIFO_MASK)|V3_FIFO_DEPTH32 | V3_FIFO_THRESHOLD16);
                                        Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_PREFIFO_CONTROL ,(V3_FIFO_THRESHOLD16>>8) |
                                            ( VIDInD(ALPHA_V3_PREFIFO_CONTROL)& (~V3_FIFO_MASK)) );
                                        break;
                                    default:
                                        Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_FIFO_CONTROL,
                                            (VIDInD(ALPHA_V3_FIFO_CONTROL)&ALPHA_FIFO_MASK)|V3_FIFO_DEPTH64 | V3_FIFO_THRESHOLD56);
                                        Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_PREFIFO_CONTROL ,V3_FIFO_PRETHRESHOLD56 |
                                            ( VIDInD(ALPHA_V3_PREFIFO_CONTROL)& (~V3_FIFO_MASK)) );
                                        break;
                                }
                                break;
                            default:
                                DBG_DD(("Unsupported DeviceID in DuoView mode!\n"));
                                break;
                            }/*End of DeviceID*/
                        }
                    }
                    else
                    {
                        switch ( lpVIAGraphicInfo->dwDeviceID )
                        {
                            case VIA_DEVICE_VT3204:
                                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_FIFO_CONTROL,
                                    (VIDInD(ALPHA_V3_FIFO_CONTROL)&ALPHA_FIFO_MASK)|V3_FIFO_DEPTH100 | V3_FIFO_THRESHOLD89);
                                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_PREFIFO_CONTROL ,V3_FIFO_PRETHRESHOLD89 |
                                    ( VIDInD(ALPHA_V3_PREFIFO_CONTROL)& (~V3_FIFO_MASK)) );
                                break;
                            case VIA_DEVICE_VT3314:
                                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_FIFO_CONTROL,
                                                (VIDInD(ALPHA_V3_FIFO_CONTROL)&
                                                 ALPHA_FIFO_MASK)|V3_FIFO_DEPTH64 | V3_FIFO_THRESHOLD61);
                                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_PREFIFO_CONTROL ,(V3_FIFO_THRESHOLD61>>8)
                                                |(VIDInD(ALPHA_V3_PREFIFO_CONTROL)& (~V3_FIFO_MASK_3314)));
                                break;
                            case VIA_DEVICE_VT3205:
                                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_FIFO_CONTROL,
                                    (VIDInD(ALPHA_V3_FIFO_CONTROL)&ALPHA_FIFO_MASK)|V3_FIFO_DEPTH32 | V3_FIFO_THRESHOLD29);
                                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_PREFIFO_CONTROL ,V3_FIFO_PRETHRESHOLD29 |
                                    ( VIDInD(ALPHA_V3_PREFIFO_CONTROL)& (~V3_FIFO_MASK)) );
                                break;
                            case VIA_DEVICE_CLE1:
                            case VIA_DEVICE_CLE2:
                                switch ( DispatchVGARevisionID(pVidData) )
                                {
                                    case VIA_REVISION_CLEAX :
                                        Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_FIFO_CONTROL,
                                            (VIDInD(ALPHA_V3_FIFO_CONTROL)&ALPHA_FIFO_MASK)|V3_FIFO_DEPTH32 | V3_FIFO_THRESHOLD16);
                                        Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_PREFIFO_CONTROL ,(V3_FIFO_THRESHOLD16>>8) |
                                            ( VIDInD(ALPHA_V3_PREFIFO_CONTROL)& (~V3_FIFO_MASK)) );
                                        break;
                                    default:
                                        Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_FIFO_CONTROL,
                                            (VIDInD(ALPHA_V3_FIFO_CONTROL)&ALPHA_FIFO_MASK)|V3_FIFO_DEPTH64 | V3_FIFO_THRESHOLD56);
                                        Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_PREFIFO_CONTROL ,V3_FIFO_PRETHRESHOLD56 |
                                            ( VIDInD(ALPHA_V3_PREFIFO_CONTROL)& (~V3_FIFO_MASK)) );
                                        break;
                                }
                                break;
                            default:
                                DBG_DD(("Unknown DeviceID\n"));
                                break;
                        }/*End of DeviceID*/
                    }
                }
                else
                {
                    /* Minified Video will be skewed if not work around */
                    if (srcWidth <= 80) /* Fetch count <= 5 */
                    {
                        if (dwVideoFlag & VIDEO_1_INUSE)
                        {
                            Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V_FIFO_CONTROL,V1_FIFO_DEPTH16 );
                        }
                        else
                        {
                            Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_FIFO_CONTROL,
                                (VIDInD(ALPHA_V3_FIFO_CONTROL)&ALPHA_FIFO_MASK)|V3_FIFO_DEPTH16 );
                            Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_PREFIFO_CONTROL ,(V3_FIFO_THRESHOLD16>>8) |
                                ( VIDInD(ALPHA_V3_PREFIFO_CONTROL)& (~V3_FIFO_MASK)) );
                        }
                    }
                    else
                    {
                        if (dwVideoFlag & VIDEO_1_INUSE)
                        {
                            /*=* Modify for C1 FIFO *=*/
                            switch ( DispatchVGARevisionID(pVidData) )
                            {
                                case VIA_REVISION_CLEAX :
                                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V_FIFO_CONTROL,
                                        V1_FIFO_DEPTH16 | V1_FIFO_PRETHRESHOLD12 | V1_FIFO_THRESHOLD8);
                                    break;
                                default:
                                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V_FIFO_CONTROL,
                                        V1_FIFO_DEPTH64 | V1_FIFO_PRETHRESHOLD56 | V1_FIFO_THRESHOLD56);
                                    break;
                            }
                        }
                        else
                        {
                            switch ( lpVIAGraphicInfo->dwDeviceID )
                            {
                                case VIA_DEVICE_VT3204:
                                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_FIFO_CONTROL,
                                        (VIDInD(ALPHA_V3_FIFO_CONTROL)&ALPHA_FIFO_MASK)|V3_FIFO_DEPTH100 | V3_FIFO_THRESHOLD89);
                                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_PREFIFO_CONTROL ,V3_FIFO_PRETHRESHOLD89 |
                                        ( VIDInD(ALPHA_V3_PREFIFO_CONTROL)& (~V3_FIFO_MASK)) );
                                    break;
                                case VIA_DEVICE_VT3314:
                                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_FIFO_CONTROL,
                                                    (VIDInD(ALPHA_V3_FIFO_CONTROL)&
                                                     ALPHA_FIFO_MASK)|V3_FIFO_DEPTH64 | V3_FIFO_THRESHOLD61);
                                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_PREFIFO_CONTROL ,(V3_FIFO_THRESHOLD61>>8)
                                                    |(VIDInD(ALPHA_V3_PREFIFO_CONTROL)& (~V3_FIFO_MASK_3314)));
                                    break;
                                case VIA_DEVICE_VT3205:
                                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_FIFO_CONTROL,
                                        (VIDInD(ALPHA_V3_FIFO_CONTROL)&ALPHA_FIFO_MASK)|V3_FIFO_DEPTH32 | V3_FIFO_THRESHOLD29);
                                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_PREFIFO_CONTROL ,V3_FIFO_PRETHRESHOLD29 |
                                        ( VIDInD(ALPHA_V3_PREFIFO_CONTROL)& (~V3_FIFO_MASK)) );
                                    break;
                                case VIA_DEVICE_CLE1:
                                case VIA_DEVICE_CLE2:
                                    switch ( DispatchVGARevisionID(pVidData) )
                                    {
                                        case VIA_REVISION_CLEAX :
                                            Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_FIFO_CONTROL,
                                                (VIDInD(ALPHA_V3_FIFO_CONTROL)&ALPHA_FIFO_MASK)|V3_FIFO_DEPTH16 | V3_FIFO_THRESHOLD8);
                                            Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_PREFIFO_CONTROL ,(V3_FIFO_THRESHOLD16>>8) |
                                                ( VIDInD(ALPHA_V3_PREFIFO_CONTROL)& (~V3_FIFO_MASK)) );
                                            break;
                                        default:
                                            Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_FIFO_CONTROL,
                                                (VIDInD(ALPHA_V3_FIFO_CONTROL)&ALPHA_FIFO_MASK)|V3_FIFO_DEPTH64 | V3_FIFO_THRESHOLD56);
                                            Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_PREFIFO_CONTROL ,V3_FIFO_PRETHRESHOLD56 |
                                                ( VIDInD(ALPHA_V3_PREFIFO_CONTROL)& (~V3_FIFO_MASK)) );
                                            break;
                                    }
                                    break;
                                default:
                                    DBG_DD(("Unknown DeviceID\n"));
                                    break;
                            }/*End of DeviceID*/
                        }
                    }
                }
            }
            else
            {
                if (dwVideoFlag & VIDEO_1_INUSE)
                {
                    /*=* Modify for C1 FIFO *=*/
                    switch ( DispatchVGARevisionID(pVidData) )
                    {
                        case VIA_REVISION_CLEAX :
                            if (pVidData->dwUseExtendedFIFO)
                            {
                                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V_FIFO_CONTROL,
                                    V1_FIFO_DEPTH48 | V1_FIFO_PRETHRESHOLD40 | V1_FIFO_THRESHOLD40);
                            }
                            else
                            {
                                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V_FIFO_CONTROL,
                                    V1_FIFO_DEPTH32 | V1_FIFO_PRETHRESHOLD29 | V1_FIFO_THRESHOLD16);
                            }
                            break;
                        default:
                            Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V_FIFO_CONTROL,
                                V1_FIFO_DEPTH64 | V1_FIFO_PRETHRESHOLD56 | V1_FIFO_THRESHOLD56);
                            break;
                    }
                    if (dwVideoFlag & VIDEO_1_ON_SND)
                    {
                        if (srcWidth <= 8)
                        {
                            Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_FIFO_CONTROL,
                                            (VIDInD(ALPHA_V3_FIFO_CONTROL)&ALPHA_FIFO_MASK));
                            Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_PREFIFO_CONTROL ,
                                            ( VIDInD(ALPHA_V3_PREFIFO_CONTROL)& (~V3_FIFO_MASK)) );
                        }
                        else
                        {
                            switch ( lpVIAGraphicInfo->dwDeviceID )
                            {
                            case VIA_DEVICE_CLE1:
                            case VIA_DEVICE_CLE2:
                                switch ( DispatchVGARevisionID(pVidData) )
                                {
                                case VIA_REVISION_CLEAX :
                                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_FIFO_CONTROL,
                                                    (VIDInD(ALPHA_V3_FIFO_CONTROL)&ALPHA_FIFO_MASK)|V3_FIFO_DEPTH32 | V3_FIFO_THRESHOLD16);
                                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_PREFIFO_CONTROL ,(V3_FIFO_THRESHOLD16>>8) |
                                                    ( VIDInD(ALPHA_V3_PREFIFO_CONTROL)& (~V3_FIFO_MASK)) );
                                    break;
                                default:
                                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_FIFO_CONTROL,
                                                    (VIDInD(ALPHA_V3_FIFO_CONTROL)&ALPHA_FIFO_MASK)|V3_FIFO_DEPTH64 | V3_FIFO_THRESHOLD56);
                                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_PREFIFO_CONTROL ,V3_FIFO_PRETHRESHOLD56 |
                                                    ( VIDInD(ALPHA_V3_PREFIFO_CONTROL)& (~V3_FIFO_MASK)) );
                                    break;
                                }
                                break;
                            default:
                                DBG_DD(("Unsupported  DeviceID in DuoView mode \n"));
                                break;
                            }/*End of DeviceID*/
                        }
                    }
                }
                else
                {
                    if (srcWidth <= 8)
                    {
                        Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_FIFO_CONTROL,
                            (VIDInD(ALPHA_V3_FIFO_CONTROL)&ALPHA_FIFO_MASK));
                        Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_PREFIFO_CONTROL ,
                            ( VIDInD(ALPHA_V3_PREFIFO_CONTROL)& (~V3_FIFO_MASK)) );
                    }
                    else
                    {
                        switch ( lpVIAGraphicInfo->dwDeviceID )
                        {
                            case VIA_DEVICE_VT3204:
                                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_FIFO_CONTROL,
                                    (VIDInD(ALPHA_V3_FIFO_CONTROL)&ALPHA_FIFO_MASK)|V3_FIFO_DEPTH100 | V3_FIFO_THRESHOLD89);
                                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_PREFIFO_CONTROL ,V3_FIFO_PRETHRESHOLD89 |
                                    ( VIDInD(ALPHA_V3_PREFIFO_CONTROL)& (~V3_FIFO_MASK)) );
                                break;
                            case VIA_DEVICE_VT3314:
                                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_FIFO_CONTROL,
                                                (VIDInD(ALPHA_V3_FIFO_CONTROL)&
                                                 ALPHA_FIFO_MASK)|V3_FIFO_DEPTH64 | V3_FIFO_THRESHOLD61);
                                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_PREFIFO_CONTROL ,(V3_FIFO_THRESHOLD61>>8)
                                                |(VIDInD(ALPHA_V3_PREFIFO_CONTROL)& (~V3_FIFO_MASK_3314)));
                                break;
                            case VIA_DEVICE_VT3205:
                                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_FIFO_CONTROL,
                                    (VIDInD(ALPHA_V3_FIFO_CONTROL)&ALPHA_FIFO_MASK)|V3_FIFO_DEPTH32 | V3_FIFO_THRESHOLD29);
                                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_PREFIFO_CONTROL ,V3_FIFO_PRETHRESHOLD29 |
                                    ( VIDInD(ALPHA_V3_PREFIFO_CONTROL)& (~V3_FIFO_MASK)) );
                                break;
                            case VIA_DEVICE_CLE1:
                            case VIA_DEVICE_CLE2:
                                switch ( DispatchVGARevisionID(pVidData) )
                                {
                                    case VIA_REVISION_CLEAX :
                                        Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_FIFO_CONTROL,
                                            (VIDInD(ALPHA_V3_FIFO_CONTROL)&ALPHA_FIFO_MASK)|V3_FIFO_DEPTH32 | V3_FIFO_THRESHOLD16);
                                        Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_PREFIFO_CONTROL ,(V3_FIFO_THRESHOLD16>>8) |
                                            ( VIDInD(ALPHA_V3_PREFIFO_CONTROL)& (~V3_FIFO_MASK)) );
                                        break;
                                    default:
                                        Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_FIFO_CONTROL,
                                            (VIDInD(ALPHA_V3_FIFO_CONTROL)&ALPHA_FIFO_MASK)|V3_FIFO_DEPTH64 | V3_FIFO_THRESHOLD56);
                                        Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_PREFIFO_CONTROL ,V3_FIFO_PRETHRESHOLD56 |
                                            ( VIDInD(ALPHA_V3_PREFIFO_CONTROL)& (~V3_FIFO_MASK)) );
                                        break;
                                }
                                break;
                            default:
                                DBG_DD(("Unknown DeviceID\n"));
                                break;
                        }/*End of DeviceID*/
                    }
                }
            }
        }
        else
        {
            DBG_DD(("not miniCtl & V1_Y_INTERPOLY"));
            if (lpDPFsrc->dwFourCC == FOURCC_YV12)
            {
                if (dwVideoFlag & VIDEO_HQV_INUSE)
                {
                    if (dwVideoFlag & VIDEO_1_INUSE)
                    {
                        /*=* Modify for C1 FIFO *=*/
                        switch ( DispatchVGARevisionID(pVidData) )
                        {
                            case VIA_REVISION_CLECX :
                                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V_FIFO_CONTROL,
                                    V1_FIFO_DEPTH64 | V1_FIFO_PRETHRESHOLD56 | V1_FIFO_THRESHOLD56);
                                break;

                            default:
                                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V_FIFO_CONTROL,
                                    V1_FIFO_DEPTH32 | V1_FIFO_PRETHRESHOLD29 | V1_FIFO_THRESHOLD16);
                                break;
                        }
                        if (dwVideoFlag & VIDEO_1_ON_SND)
                        {
                            switch ( lpVIAGraphicInfo->dwDeviceID )
                            {
                            case VIA_DEVICE_CLE1:
                            case VIA_DEVICE_CLE2:
                                switch ( DispatchVGARevisionID(pVidData) )
                                {
                                case VIA_REVISION_CLEAX :
                                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_FIFO_CONTROL,
                                                    (VIDInD(ALPHA_V3_FIFO_CONTROL)&ALPHA_FIFO_MASK)|V3_FIFO_DEPTH32 | V3_FIFO_THRESHOLD16);
                                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_PREFIFO_CONTROL ,(V3_FIFO_THRESHOLD16>>8) |
                                                    ( VIDInD(ALPHA_V3_PREFIFO_CONTROL)& (~V3_FIFO_MASK)) );
                                    break;
                                default:
                                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_FIFO_CONTROL,
                                                    (VIDInD(ALPHA_V3_FIFO_CONTROL)&ALPHA_FIFO_MASK)|V3_FIFO_DEPTH64 | V3_FIFO_THRESHOLD56);
                                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_PREFIFO_CONTROL ,V3_FIFO_PRETHRESHOLD56 |
                                                    ( VIDInD(ALPHA_V3_PREFIFO_CONTROL)& (~V3_FIFO_MASK)) );
                                    break;
                                }
                                break;
                            default:
                                DBG_DD(("Unknown DeviceID\n"));
                                break;
                            }/*End of DeviceID*/
                        }
                    }
                    else
                    {
                        switch ( lpVIAGraphicInfo->dwDeviceID )
                        {
                            case VIA_DEVICE_VT3204:
                                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_FIFO_CONTROL,
                                    (VIDInD(ALPHA_V3_FIFO_CONTROL)&ALPHA_FIFO_MASK)|V3_FIFO_DEPTH100 | V3_FIFO_THRESHOLD89);
                                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_PREFIFO_CONTROL ,V3_FIFO_PRETHRESHOLD89 |
                                    ( VIDInD(ALPHA_V3_PREFIFO_CONTROL)& (~V3_FIFO_MASK)) );
                                break;
                            case VIA_DEVICE_VT3314:
                                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_FIFO_CONTROL,
                                                (VIDInD(ALPHA_V3_FIFO_CONTROL)&
                                                 ALPHA_FIFO_MASK)|V3_FIFO_DEPTH64 | V3_FIFO_THRESHOLD61);
                                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_PREFIFO_CONTROL ,(V3_FIFO_THRESHOLD61>>8)
                                                |(VIDInD(ALPHA_V3_PREFIFO_CONTROL)& (~V3_FIFO_MASK_3314)));
                                break;
                            case VIA_DEVICE_VT3205:
                                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_FIFO_CONTROL,
                                    (VIDInD(ALPHA_V3_FIFO_CONTROL)&ALPHA_FIFO_MASK)|V3_FIFO_DEPTH32 | V3_FIFO_THRESHOLD29);
                                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_PREFIFO_CONTROL ,V3_FIFO_PRETHRESHOLD29 |
                                    ( VIDInD(ALPHA_V3_PREFIFO_CONTROL)& (~V3_FIFO_MASK)) );
                                break;
                            case VIA_DEVICE_CLE1:
                            case VIA_DEVICE_CLE2:
                                switch ( DispatchVGARevisionID(pVidData) )
                                {
                                    case VIA_REVISION_CLEAX :
                                        Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_FIFO_CONTROL,
                                            (VIDInD(ALPHA_V3_FIFO_CONTROL)&ALPHA_FIFO_MASK)|V3_FIFO_DEPTH32 | V3_FIFO_THRESHOLD16);
                                        Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_PREFIFO_CONTROL ,(V3_FIFO_THRESHOLD16>>8) |
                                            ( VIDInD(ALPHA_V3_PREFIFO_CONTROL)& (~V3_FIFO_MASK)) );
                                        break;
                                    default:
                                        Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_FIFO_CONTROL,
                                            (VIDInD(ALPHA_V3_FIFO_CONTROL)&ALPHA_FIFO_MASK)|V3_FIFO_DEPTH64 | V3_FIFO_THRESHOLD56);
                                        Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_PREFIFO_CONTROL ,V3_FIFO_PRETHRESHOLD56 |
                                            ( VIDInD(ALPHA_V3_PREFIFO_CONTROL)& (~V3_FIFO_MASK)) );
                                        break;
                                }
                                break;
                            default:
                                DBG_DD(("Unknown DeviceID\n"));
                                break;
                        }/*End of DeviceID*/
                    }
                }
                else
                {
                    /* Minified Video will be skewed if not work around */
                    if (srcWidth <= 80) /* Fetch count <= 5 */
                    {
                        if (dwVideoFlag & VIDEO_1_INUSE)
                        {
                            Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V_FIFO_CONTROL,V1_FIFO_DEPTH16 );
                        }
                        else
                        {
                            Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_FIFO_CONTROL,
                                (VIDInD(ALPHA_V3_FIFO_CONTROL)&ALPHA_FIFO_MASK)|V3_FIFO_DEPTH16 );
                            Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_PREFIFO_CONTROL ,(V3_FIFO_THRESHOLD16>>8) |
                                ( VIDInD(ALPHA_V3_PREFIFO_CONTROL)& (~V3_FIFO_MASK)) );
                        }
                    }
                    else
                    {
                        if (dwVideoFlag & VIDEO_1_INUSE)
                        {
                            /*=* Modify for C1 FIFO *=*/
                            switch ( DispatchVGARevisionID(pVidData) )
                            {
                                case VIA_REVISION_CLEAX :
                                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V_FIFO_CONTROL,
                                        V1_FIFO_DEPTH16 | V1_FIFO_PRETHRESHOLD12 | V1_FIFO_THRESHOLD8);
                                    break;
                                default:
                                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V_FIFO_CONTROL,
                                        V1_FIFO_DEPTH64 | V1_FIFO_PRETHRESHOLD56 | V1_FIFO_THRESHOLD56);
                                    break;
                            }
                        }
                        else
                        {
                            switch ( lpVIAGraphicInfo->dwDeviceID )
                            {
                                case VIA_DEVICE_VT3204:
                                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_FIFO_CONTROL,
                                        (VIDInD(ALPHA_V3_FIFO_CONTROL)&ALPHA_FIFO_MASK)|V3_FIFO_DEPTH100 | V3_FIFO_THRESHOLD89);
                                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_PREFIFO_CONTROL ,V3_FIFO_PRETHRESHOLD89 |
                                        ( VIDInD(ALPHA_V3_PREFIFO_CONTROL)& (~V3_FIFO_MASK)) );
                                    break;
                                case VIA_DEVICE_VT3314:
                                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_FIFO_CONTROL,
                                                    (VIDInD(ALPHA_V3_FIFO_CONTROL)&
                                                     ALPHA_FIFO_MASK)|V3_FIFO_DEPTH64 | V3_FIFO_THRESHOLD61);
                                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_PREFIFO_CONTROL ,(V3_FIFO_THRESHOLD61>>8)
                                                    |(VIDInD(ALPHA_V3_PREFIFO_CONTROL)& (~V3_FIFO_MASK_3314)));
                                    break;
                                case VIA_DEVICE_VT3205:
                                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_FIFO_CONTROL,
                                        (VIDInD(ALPHA_V3_FIFO_CONTROL)&ALPHA_FIFO_MASK)|V3_FIFO_DEPTH32 | V3_FIFO_THRESHOLD29);
                                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_PREFIFO_CONTROL ,V3_FIFO_PRETHRESHOLD29 |
                                        ( VIDInD(ALPHA_V3_PREFIFO_CONTROL)& (~V3_FIFO_MASK)) );
                                    break;
                                case VIA_DEVICE_CLE1:
                                case VIA_DEVICE_CLE2:
                                    switch ( DispatchVGARevisionID(pVidData) )
                                    {
                                        case VIA_REVISION_CLEAX :
                                            Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_FIFO_CONTROL,
                                                (VIDInD(ALPHA_V3_FIFO_CONTROL)&ALPHA_FIFO_MASK)|V3_FIFO_DEPTH16 | V3_FIFO_THRESHOLD8);
                                            Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_PREFIFO_CONTROL ,(V3_FIFO_THRESHOLD16>>8) |
                                                ( VIDInD(ALPHA_V3_PREFIFO_CONTROL)& (~V3_FIFO_MASK)) );
                                            break;
                                        default:
                                            Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_FIFO_CONTROL,
                                                (VIDInD(ALPHA_V3_FIFO_CONTROL)&ALPHA_FIFO_MASK)|V3_FIFO_DEPTH64 | V3_FIFO_THRESHOLD56);
                                            Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_PREFIFO_CONTROL ,V3_FIFO_PRETHRESHOLD56 |
                                                ( VIDInD(ALPHA_V3_PREFIFO_CONTROL)& (~V3_FIFO_MASK)) );
                                            break;
                                    }
                                    break;
                                default:
                                    DBG_DD(("Unknown DeviceID\n"));
                                    break;
                            }/*End of DeviceID*/
                        }
                    }
                }
            }
            else
            {
                if (dwVideoFlag & VIDEO_1_INUSE)
                {
                    /*=* Modify for C1 FIFO *=*/
                    switch ( DispatchVGARevisionID(pVidData) )
                    {
                        case VIA_REVISION_CLEAX :
                            if (pVidData->dwUseExtendedFIFO)
                            {
                                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V_FIFO_CONTROL,
                                    V1_FIFO_DEPTH48 | V1_FIFO_PRETHRESHOLD40 | V1_FIFO_THRESHOLD40);
                            }
                            else
                            {
                                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V_FIFO_CONTROL,
                                    V1_FIFO_DEPTH32 | V1_FIFO_PRETHRESHOLD29 | V1_FIFO_THRESHOLD16);
                            }
                            break;
                        default:
                            Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, V_FIFO_CONTROL,
                                V1_FIFO_DEPTH64 | V1_FIFO_PRETHRESHOLD56 | V1_FIFO_THRESHOLD56);
                            break;
                    }
                    if(dwVideoFlag & VIDEO_1_ON_SND)
                    {
                        if (srcWidth <= 8)
                        {
                            Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_FIFO_CONTROL,
                                            (VIDInD(ALPHA_V3_FIFO_CONTROL)&ALPHA_FIFO_MASK));
                            Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_PREFIFO_CONTROL ,
                                            ( VIDInD(ALPHA_V3_PREFIFO_CONTROL)& (~V3_FIFO_MASK)) );
                        }
                        else
                        {
                            switch ( lpVIAGraphicInfo->dwDeviceID )
                                {
                                case VIA_DEVICE_CLE1:
                                case VIA_DEVICE_CLE2:
                                switch ( DispatchVGARevisionID(pVidData) )
                                {
                                case VIA_REVISION_CLEAX :
                                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_FIFO_CONTROL,
                                                    (VIDInD(ALPHA_V3_FIFO_CONTROL)&ALPHA_FIFO_MASK)|V3_FIFO_DEPTH32 | V3_FIFO_THRESHOLD16);
                                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_PREFIFO_CONTROL ,(V3_FIFO_THRESHOLD16>>8) |
                                            ( VIDInD(ALPHA_V3_PREFIFO_CONTROL)& (~V3_FIFO_MASK)) );
                                    break;
                                default:
                                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_FIFO_CONTROL,
                                                    (VIDInD(ALPHA_V3_FIFO_CONTROL)&ALPHA_FIFO_MASK)|V3_FIFO_DEPTH64 | V3_FIFO_THRESHOLD56);
                                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_PREFIFO_CONTROL ,V3_FIFO_PRETHRESHOLD56 |
                                                    ( VIDInD(ALPHA_V3_PREFIFO_CONTROL)& (~V3_FIFO_MASK)) );
                                    break;
                                }
                                break;
                                default:
                                    DBG_DD(("Unsupported  DeviceID in DuoView mode\n"));
                                    break;
                                }/*End of DeviceID*/
                        }
                    }
                }
                else
                {
                    if (srcWidth <= 8)
                    {
                        Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_FIFO_CONTROL,
                            (VIDInD(ALPHA_V3_FIFO_CONTROL)&ALPHA_FIFO_MASK));
                        Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_PREFIFO_CONTROL ,
                            ( VIDInD(ALPHA_V3_PREFIFO_CONTROL)& (~V3_FIFO_MASK)) );
                    }
                    else
                    {
                        switch ( lpVIAGraphicInfo->dwDeviceID )
                        {
                            case VIA_DEVICE_VT3204:
                                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_FIFO_CONTROL,
                                    (VIDInD(ALPHA_V3_FIFO_CONTROL)&ALPHA_FIFO_MASK)|V3_FIFO_DEPTH100 | V3_FIFO_THRESHOLD89);
                                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_PREFIFO_CONTROL ,V3_FIFO_PRETHRESHOLD89 |
                                    ( VIDInD(ALPHA_V3_PREFIFO_CONTROL)& (~V3_FIFO_MASK)) );
                                break;
                            case VIA_DEVICE_VT3314:
                                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_FIFO_CONTROL,
                                                (VIDInD(ALPHA_V3_FIFO_CONTROL)&
                                                 ALPHA_FIFO_MASK)|V3_FIFO_DEPTH64 | V3_FIFO_THRESHOLD61);
                                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_PREFIFO_CONTROL ,(V3_FIFO_THRESHOLD61>>8) |
                                                (VIDInD(ALPHA_V3_PREFIFO_CONTROL)& (~V3_FIFO_MASK_3314)) );
                            break;
                            case VIA_DEVICE_VT3205:
                                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_FIFO_CONTROL,
                                    (VIDInD(ALPHA_V3_FIFO_CONTROL)&ALPHA_FIFO_MASK)|V3_FIFO_DEPTH32 | V3_FIFO_THRESHOLD29);
                                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_PREFIFO_CONTROL ,V3_FIFO_PRETHRESHOLD29 |
                                    ( VIDInD(ALPHA_V3_PREFIFO_CONTROL)& (~V3_FIFO_MASK)) );
                                break;
                            case VIA_DEVICE_CLE1:
                            case VIA_DEVICE_CLE2:
                                switch ( DispatchVGARevisionID(pVidData) )
                                {
                                    case VIA_REVISION_CLEAX :
                                        Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_FIFO_CONTROL,
                                            (VIDInD(ALPHA_V3_FIFO_CONTROL)&ALPHA_FIFO_MASK)|V3_FIFO_DEPTH32 | V3_FIFO_THRESHOLD16);
                                        Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_PREFIFO_CONTROL ,(V3_FIFO_THRESHOLD16>>8) |
                                            ( VIDInD(ALPHA_V3_PREFIFO_CONTROL)& (~V3_FIFO_MASK)) );
                                        break;
                                    default:
                                        Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_FIFO_CONTROL,
                                            (VIDInD(ALPHA_V3_FIFO_CONTROL)&ALPHA_FIFO_MASK)|V3_FIFO_DEPTH64 | V3_FIFO_THRESHOLD56);
                                        Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, ALPHA_V3_PREFIFO_CONTROL ,V3_FIFO_PRETHRESHOLD56 |
                                            ( VIDInD(ALPHA_V3_PREFIFO_CONTROL)& (~V3_FIFO_MASK)) );
                                        break;
                                }
                                break;
                            default:
                                DBG_DD(("Unknown DeviceID\n"));
                                break;
                        }/*End of DeviceID*/
                    }
                }
            }
        }

        if (dwVideoFlag & VIDEO_HQV_INUSE)
        {
            if((dwVideoFlag & VIDEO_1_ON_SND) && (dstModifiedWidth > dstWidth))
            {/*handle DuoView*/
                unsigned long v3_zoom=0,v3_mini=0,v1_zoom=0,v1_mini=0;
                dwHQVminiCtl = 0;
                if(srcWidth >= dstWidth)/*HQV scales down video, V1 scale up video itself and V3 does nothing*/
                {
                    if(DDOVER_HQVCalcZoomWidth(pVidData, dwVideoFlag, srcWidth , dstWidth,
                                               &dwTmp, &dwTmp, &dwHQVfilterCtl, &dwHQVminiCtl,&dwHQVzoomflagH) == VMI_ERR )
                    {/*caculate HQV paras*/
                        /* too small to handle */
                        return VMI_ERR;
                    }
                    if(DDOVER_HQVCalcZoomWidth(pVidData, dwVideoFlag, dstWidth , dstModifiedWidth,
                                               &v1_zoom, &v1_mini, &dwTmp, &dwTmp, &dwHQVzoomflagH) == VMI_ERR )
                    {/*caculate V1 paras*/
                        /* too small to handle */
                        return VMI_ERR;
                    }
                }
                else/*HQV does not scale video, V1 uses the orginal paras and V3 must recaculate paras*/
                {
                    if(DDOVER_HQVCalcZoomWidth(pVidData, dwVideoFlag, srcWidth , dstWidth,
                                               &v3_zoom, &v3_mini, &dwTmp, &dwTmp,&dwHQVzoomflagH) == VMI_ERR )
                    {
                        /* too small to handle */
                        return VMI_ERR;
                    }
                    v1_zoom |= zoomCtl&0xffff0000;
                    v1_mini |= miniCtl&(V1_X_INTERPOLY);
                }
               if(srcHeight >= dstHeight)/*HQV scales down video, V1 scale up video itself and V3 does nothing*/
                {
                    if(DDOVER_HQVCalcZoomHeight(srcHeight , dstHeight, &dwTmp,
                                                &dwTmp, &dwHQVfilterCtl, &dwHQVminiCtl, &dwHQVzoomflagV) == VMI_ERR )
                    {/*caculate HQV paras*/
                        /* too small to handle */
                        return VMI_ERR;
                    }
                    if(DDOVER_HQVCalcZoomHeight(dstHeight , dstModifiedHeight,&v1_zoom, 
                                                &v1_mini, &dwTmp, &dwTmp, &dwHQVzoomflagV) == VMI_ERR )
                    {/*caculate V1 paras*/
                        /* too small to handle */
                        return VMI_ERR;
                    }
                }
                else/*HQV does not scale video, V1 uses the orginal paras and V3 must recaculate paras*/
                {
                    if(DDOVER_HQVCalcZoomHeight(srcHeight, dstHeight, &v3_zoom, 
                                                &v3_mini, &dwTmp, &dwTmp, &dwHQVzoomflagH) == VMI_ERR )
                    {
                        /* too small to handle */
                        return VMI_ERR;
                    }
                    v1_zoom |= zoomCtl&0x0000ffff;
                    v1_mini |= miniCtl&(V1_Y_INTERPOLY|V1_YCBCR_INTERPOLY);
                }            
                if (dwVideoFlag & VIDEO_1_INUSE)
                {
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V1_MINI_CONTROL, v1_mini);
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V1_ZOOM_CONTROL, v1_zoom);
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V3_MINI_CONTROL, v3_mini);
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V3_ZOOM_CONTROL, v3_zoom);
                }
            }        
            else
            {
            miniCtl=0;
            if (dwHQVzoomflagH||dwHQVzoomflagV)
            {
                dwTmp = 0;
                if (dwHQVzoomflagH)
                {
                    miniCtl = V1_X_INTERPOLY;
                    dwTmp = (zoomCtl&0xffff0000);
                }

                if (dwHQVzoomflagV)
                {
                    if((srcWidth>=800)&&(lpVIAGraphicInfo->dwDeviceID==VIA_DEVICE_VT3314))
                        miniCtl &= ~(V1_Y_INTERPOLY);
                    else
                        miniCtl |= (V1_Y_INTERPOLY | V1_YCBCR_INTERPOLY);

                    dwTmp |= (zoomCtl&0x0000ffff);
                }

                if ((pVidData->dwUseExtendedFIFO))
                {
                    miniCtl &= ~V1_Y_INTERPOLY;
                }

                if (dwVideoFlag & VIDEO_1_INUSE)
                {
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V1_MINI_CONTROL, miniCtl);
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V1_ZOOM_CONTROL, dwTmp);
                    if (dwVideoFlag & VIDEO_1_ON_SND)
                    {
                        Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V3_MINI_CONTROL, miniCtl);
                        Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V3_ZOOM_CONTROL, dwTmp);
                    }
                }
                else
                {
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V3_MINI_CONTROL, miniCtl);
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V3_ZOOM_CONTROL, dwTmp);
                }
            }
            else
            {
                if (dwVideoFlag & VIDEO_1_INUSE)
                {
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V1_MINI_CONTROL, 0);
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V1_ZOOM_CONTROL, 0);
                    if (dwVideoFlag & VIDEO_1_ON_SND)
                    {
                        Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V3_MINI_CONTROL, 0);
                        Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V3_ZOOM_CONTROL, 0);
                    }
                }
                else
                {
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V3_MINI_CONTROL, 0);
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V3_ZOOM_CONTROL, 0);
                }
            }
            }
            if (lpVIAGraphicInfo->HQVFilter_ManualSelect)
            {
                switch(lpVIAGraphicInfo->MpegHQV_VFilterMode)
                {
                    case 0:
                        dwHQVfilterCtl &= 0x0000ffff;
                        dwHQVfilterCtl |= 0x00400000;
                        DBG_DD(("Upd_Video : No MPEG HQV Vertical Filter\n"));
                        break;

                    case 1:
                        dwHQVfilterCtl &= 0x0000ffff;
                        dwHQVfilterCtl |= HQV_V_TAP2_11;
                        DBG_DD(("Upd_Video : MPEG HQV Vertical Filter mode: TAP2_11\n"));
                        break;

                    case 2:
                        dwHQVfilterCtl &= 0x0000ffff;
                        dwHQVfilterCtl |= HQV_V_TAP4_121;
                        DBG_DD(("Upd_Video : MPEG HQV Vertical Filter mode: TAP4_121\n"));
                        break;

                    case 3:
                        dwHQVfilterCtl &= 0x0000ffff;
                        dwHQVfilterCtl |= HQV_V_TAP16_1991;
                        DBG_DD(("Upd_Video : MPEG HQV Vertical Filter mode: TAP16_1991\n"));
                        break;

                    case 4:
                        dwHQVfilterCtl &= 0x0000ffff;
                        dwHQVfilterCtl |= HQV_V_TAP16_141041;
                        DBG_DD(("Upd_Video : MPEG HQV Vertical Filter mode: TAP16_141041\n"));
                        break;

                    case 5:
                        dwHQVfilterCtl &= 0x0000ffff;
                        dwHQVfilterCtl |= HQV_V_TAP32;
                        DBG_DD(("Upd_Video : MPEG HQV Vertical Filter mode: TAP32\n"));
                        break;

                    default:
                        dwHQVfilterCtl &= 0x0000ffff;
                        dwHQVfilterCtl |= HQV_V_FILTER_DEFAULT;
                        DBG_DD(("Upd_Video : MPEG HQV Vertical Filter mode: FILTER_DEFAULT\n"));
                        break;
                }

                switch(lpVIAGraphicInfo->MpegHQV_HFilterMode)
                {
                    case 0:
                        dwHQVfilterCtl &= 0xffff0000;
                        dwHQVfilterCtl |= 0x00000040;
                        DBG_DD(("Upd_Video : No MPEG HQV horizontal Filter\n"));
                        break;

                    case 1:
                        dwHQVfilterCtl &= 0xffff0000;
                        dwHQVfilterCtl |= HQV_H_TAP2_11;
                        DBG_DD(("Upd_Video : MPEG HQV horizontal Filter mode: TAP2_11\n"));
                        break;

                    case 2:
                        dwHQVfilterCtl &= 0xffff0000;
                        dwHQVfilterCtl |= HQV_H_TAP4_121;
                        DBG_DD(("Upd_Video : MPEG HQV horizontal Filter mode: TAP4_121\n"));
                        break;

                    case 3:
                        dwHQVfilterCtl &= 0xffff0000;
                        dwHQVfilterCtl |= HQV_H_TAP16_1991;
                        DBG_DD(("Upd_Video : MPEG HQV horizontal Filter mode: TAP16_1991\n"));
                        break;

                    case 4:
                        dwHQVfilterCtl &= 0xffff0000;
                        dwHQVfilterCtl |= HQV_H_TAP16_141041;
                        DBG_DD(("Upd_Video : MPEG HQV horizontal Filter mode: TAP16_141041\n"));
                        break;

                    case 5:
                        dwHQVfilterCtl &= 0xffff0000;
                        dwHQVfilterCtl |= HQV_H_TAP32;
                        DBG_DD(("Upd_Video : MPEG HQV horizontal Filter mode: TAP32\n"));
                        break;

                    default:
                        dwHQVfilterCtl &= 0xffff0000;
                        dwHQVfilterCtl |= HQV_H_FILTER_DEFAULT;
                        DBG_DD(("Upd_Video : MPEG HQV horizontal Filter mode: FILTER_DEFAULT\n"));
                        break;
                }
            }

            Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,HQV_MINI_CONTROL, dwHQVminiCtl);
            Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,HQV_FILTER_CONTROL, dwHQVfilterCtl);
        }
        else
        {
            if (dwVideoFlag & VIDEO_1_INUSE)
            {
                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V1_MINI_CONTROL, miniCtl);
                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V1_ZOOM_CONTROL, zoomCtl);
            }
            else
            {
                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V3_MINI_CONTROL, miniCtl);
                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V3_ZOOM_CONTROL, zoomCtl);
            }
        }
        if((lpVIAGraphicInfo->Screen_IGA_Map[ScrnIndex]==IGA2)||(dwVDevFlag&CTL_UseIGA2Path))
        {
            if(dwVideoFlag&VIDEO_1_INUSE)
	        dwVidCtl |=V1_ON_SND_DISPLAY;
            else
                dwVidCtl |=V3_ON_SND_DISPLAY_32045;
        }

        /* Colorkey */
        if (dwColorKeyOn)
        {
            DBG_DD(("dwColorKeyOn - Overlay colorkey= low:%08lx \n", dwKeyLow));

            dwKeyLow &= 0x00FFFFFF;

            if (lpVIAGraphicInfo->Screen[ScrnIndex].dwBPP == 8)
            {
                dwKeyLow &= 0xff;
            }

            if (lpVideoHWDifference->dwSupportTwoColorKey == VID_HWDIFF_TRUE)
            {
                dwCompose = (dwCompose & ~0x0f) | SELECT_VIDEO_IF_COLOR_KEY | SELECT_VIDEO3_IF_COLOR_KEY;
            }
            else
            {
                dwCompose = (dwCompose & ~0x0f) | SELECT_VIDEO_IF_COLOR_KEY;
            }

            /*Modified for SAMM Mode*/
            if ( (lpVIAGraphicInfo->Screen_IGA_Map[ScrnIndex]==IGA2) || (dwVDevFlag&CTL_UseIGA2Path))
            {
                DBG_DD(("ScrnIndex == 1||(dwVDevFlag&CTL_UseIGAPath)\n"));
                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,SND_COLOR_KEY, dwKeyLow);
                dwCompose |= SECOND_DISPLAY_COLOR_KEY_ENABLE;
                if ( dwVideoFlag & VIDEO_1_INUSE )
                {
                    if(lpVIAGraphicInfo->VideoUseSecondDisplay)/*use IGA2 as primary*/
                    {
                        if(ScrnIndex == 0)
                    	{
                            dwVidCtl |= V1_ON_SND_DISPLAY;
                    	}
                        else
                            dwVidCtl &= ~V1_ON_SND_DISPLAY;
                    }
                    else /*scrnindex=1*/
                    {
                        if((ScrnIndex == 0) && (!(dwVDevFlag&CTL_UseIGA2Path)))
                            dwVidCtl &= ~V1_ON_SND_DISPLAY;
                        else  
                            dwVidCtl |= V1_ON_SND_DISPLAY;  
                        if(dwVideoFlag & VIDEO_1_ON_SND)
                            Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V3_COLOR_KEY, dwKeyLow);
                    }
                }
                else
                {
                    dwVidCtl |= V3_ON_SND_DISPLAY_32045;
                }
            }
            else
            {
                DBG_DD(("not (ScrnIndex == 1||(dwVDevFlag&CTL_UseIGA2Path))\n"));
                if(lpVIAGraphicInfo->VideoUseSecondDisplay) /*scrnIndex=0 here*/
                  dwVidCtl |= V1_ON_SND_DISPLAY;	
                else
                  dwVidCtl &= ~V1_ON_SND_DISPLAY;  
                if (lpVideoHWDifference->dwSupportTwoColorKey == VID_HWDIFF_TRUE)
                {
                    if (dwVideoFlag & VIDEO_1_INUSE)
                    {
                        Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V_COLOR_KEY, dwKeyLow);
                    }
                    else
                    {
                        Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V3_COLOR_KEY, dwKeyLow);
                    }
                }
                else
                {
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V_COLOR_KEY, dwKeyLow);
                }
            }
        }/* Colorkey*/

        if (dwChromaKeyOn) {
            DBG_DD(("Overlay Chroma key low=%08lx\n", dwChromaLow));
            DBG_DD(("Overlay Chroma key upper=%08lx\n", dwChromaHigh));

            dwChromaLow  &= CHROMA_KEY_LOW;
            dwChromaHigh &= CHROMA_KEY_HIGH;

            dwChromaLow  |= (VIDInD(V_CHROMAKEY_LOW)&(~CHROMA_KEY_LOW));
            dwChromaHigh |= (VIDInD(V_CHROMAKEY_HIGH)&(~CHROMA_KEY_HIGH));

            /* for Chroma Key */
            if (lpDPFsrc->dwFlags & DDPF_FOURCC)
            {
                    switch (lpDPFsrc->dwFourCC) {
                    case FOURCC_YV12:
                            /* to be continued... */
                            break;
                    case FOURCC_YUY2:
                            /* to be continued... */
                            break;
                    default:
                            break;
                    }
            }
            else if (lpDPFsrc->dwFlags & DDPF_RGB)
            {
                    unsigned long dwtmpLowR;
                    unsigned long dwtmpLowG;
                    unsigned long dwtmpLowB;
                    unsigned long dwtmpChromaLow;
                    unsigned long dwtmpHighR;
                    unsigned long dwtmpHighG;
                    unsigned long dwtmpHighB;
                    unsigned long dwtmpChromaHigh;

                    switch (lpDPFsrc->dwRGBBitCount) {
                    case 16:
                            if (lpDPFsrc->dwGBitMask==0x07E0) /* RGB16(5:6:5) */
                            {
                                    dwtmpLowR = (((dwChromaLow >> 11) << 3) | ((dwChromaLow >> 13) & 0x07)) & 0xFF;
                                    dwtmpLowG = (((dwChromaLow >> 5) << 2) | ((dwChromaLow >> 9) & 0x03)) & 0xFF;

                                    dwtmpHighR = (((dwChromaHigh >> 11) << 3) | ((dwChromaHigh >> 13) & 0x07)) & 0xFF;
                                    dwtmpHighG = (((dwChromaHigh >> 5) << 2) | ((dwChromaHigh >> 9) & 0x03)) & 0xFF;
                            }
                            else /* RGB15(5:5:5) */
                            {
                                    dwtmpLowR = (((dwChromaLow >> 10) << 3) | ((dwChromaLow >> 12) & 0x07)) & 0xFF;
                                    dwtmpLowG = (((dwChromaLow >> 5) << 3) | ((dwChromaLow >> 7) & 0x07)) & 0xFF;

                                    dwtmpHighR = (((dwChromaHigh >> 10) << 3) | ((dwChromaHigh >> 12) & 0x07)) & 0xFF;
                                    dwtmpHighG = (((dwChromaHigh >> 5) << 3) | ((dwChromaHigh >> 7) & 0x07)) & 0xFF;
                            }
                            dwtmpLowB = (((dwChromaLow << 3) | (dwChromaLow >> 2)) & 0x07) & 0xFF;
                            dwtmpChromaLow = (dwtmpLowG << 16) | (dwtmpLowR << 8) | dwtmpLowB;
                            dwChromaLow = ((dwChromaLow >> 24) << 24) | dwtmpChromaLow;

                            dwtmpHighB = (((dwChromaHigh << 3) | (dwChromaHigh >> 2)) & 0x07) & 0xFF;
                            dwtmpChromaHigh = (dwtmpHighG << 16) | (dwtmpHighR << 8) | dwtmpHighB;
                            dwChromaHigh = ((dwChromaHigh >> 24) << 24) | dwtmpChromaHigh;
                            break;

                    case 32: /* 32 bit RGB */
                            dwtmpLowR = (dwChromaLow >> 16) & 0xFF;
                            dwtmpLowG = (dwChromaLow >> 8) & 0xFF;
                            dwtmpLowB = dwChromaLow & 0xFF;
                            dwtmpChromaLow = (dwtmpLowG << 16) | (dwtmpLowR << 8) | dwtmpLowB;
                            dwChromaLow = ((dwChromaLow >> 24) << 24) | dwtmpChromaLow;

                            dwtmpHighR = (dwChromaHigh >> 16) & 0xFF;
                            dwtmpHighG = (dwChromaHigh >> 8) & 0xFF;
                            dwtmpHighB = dwChromaHigh & 0xFF;
                            dwtmpChromaHigh = (dwtmpHighG << 16) | (dwtmpHighR << 8) | dwtmpHighB;
                            dwChromaHigh = ((dwChromaHigh >> 24) << 24) | dwtmpChromaHigh;
                            break;

                    default:
                            break;
                    }
            } /* End of DDPF_FOURCC */

            Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V_CHROMAKEY_HIGH,dwChromaHigh);
            if (dwVideoFlag & VIDEO_1_INUSE)
            {
                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V_CHROMAKEY_LOW, dwChromaLow);
                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V1_MINI_CONTROL, miniCtl & 0xFFFFFFF8);
                if (dwVideoFlag & VIDEO_1_ON_SND)/*for DuoView mode*/
                {
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V_CHROMAKEY_LOW, dwChromaLow|V_CHROMAKEY_V3);
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V3_MINI_CONTROL, miniCtl & 0xFFFFFFF8);
                }
            }
            else
            {
                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V_CHROMAKEY_LOW, dwChromaLow|V_CHROMAKEY_V3);
                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V3_MINI_CONTROL, miniCtl & 0xFFFFFFF8);
            }

            /* Modified for select video if (color key & chroma key) */
            if (dwCompose & SELECT_VIDEO_IF_COLOR_KEY)
               dwCompose |= SELECT_VIDEO_IF_CHROMA_KEY;
            else
               dwCompose &= ~SELECT_VIDEO_IF_CHROMA_KEY;

            /* add Chroma key inverse control */
            if (dwChromaIC)
                dwCompose |= CHROMA_KEY_INVERSE_CONTROL;
            else
                dwCompose &= ~CHROMA_KEY_INVERSE_CONTROL;
        }

        DBG_DD(("determine which video stream is on top- lpVideoControl->dwCompose 0x%lx\n", lpVideoControl->dwCompose));

        if (lpVideoControl->dwCompose & (VW_TV1_TOP | VW_TV_TOP) )
        {
            if ( pVidData->VidStatus.dwVideoFlagMPEG[0] & VIDEO_1_INUSE )
            {
                dwCompose &= ~COMPOSE_V3_TOP;
            }
            else
            {
                dwCompose |= COMPOSE_V3_TOP;
            }
        }
        else if (lpVideoControl->dwCompose & (VW_TV0_TOP | VW_DVD_TOP) )
        {
            if ( pVidData->VidStatus.dwVideoFlagMPEG[0] & VIDEO_1_INUSE )
            {
                dwCompose |= COMPOSE_V3_TOP;
            }
            else
            {
                dwCompose &= ~COMPOSE_V3_TOP;
            }
        }

        if ( ( !(pVidData->VidStatus.dwVideoFlagSW[0]&SW_ON_SND)&&
               ((pVidData->VidStatus.dwVideoFlagMPEG[0]&MPEG_ON_SND)||
                (pVidData->VidStatus.dwVideoFlagTV0&CAP0_ON_SND)||
                (pVidData->VidStatus.dwVideoFlagTV1&CAP1_ON_SND)) ) ||
             ( (pVidData->VidStatus.dwVideoFlagSW[0]&SW_ON_SND)&&
               (!(pVidData->VidStatus.dwVideoFlagMPEG[0]&MPEG_ON_SND)||
                !(pVidData->VidStatus.dwVideoFlagTV0&CAP0_ON_SND)||
                !(pVidData->VidStatus.dwVideoFlagTV1&CAP1_ON_SND)) ) )
        {
            dwCompose = dwCompose | COMPOSE_V3_TOP;
        }

        DBG_DD(("for 3123.C0 H/W bug  - dwCompose 0x%8lx\n", dwCompose));

        if (!lpVIAGraphicInfo->IsHWCursorEnabled)
        {
            if ( dwVideoFlag & SW_ON_SND )
            {
                VIDOutD(HI_CONTROL, VIDInD(HI_CONTROL)|V4_DISPLAY_PATH); 
            }
            else
            {
                VIDOutD(HI_CONTROL, VIDInD(HI_CONTROL)&(~V4_DISPLAY_PATH));
            }
        }
	
        /* Setup video control*/
        if (dwVideoFlag & VIDEO_HQV_INUSE)
        {
            if (!(dwVideoFlag & VIDEO_ON))
            {
                DBG_DD((" Setup video contro -  First HQV\n"));
                if (dwVideoFlag & VIDEO_1_INUSE)
                {
                    DBG_DD((" wait v1 command fire\n"));
		            WaitVideoCommandFire(pjVideo);
                    if (dwVideoFlag & VIDEO_1_ON_SND)/*for DuoView mode*/
                        WaitVideo3CommandFire(pjVideo);
                }
                else
                {
                    DBG_DD((" wait v3 command fire\n"));
                    WaitVideo3CommandFire(pjVideo);
                }
#ifdef CS_FLAG_HWROTATION
                if (  lpVIAGraphicInfo->rotate )
                {
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,HQV_MINI_CONTROL, 0);
                }
#endif /* CS_FLAG_HWROTATION */
                Macro_VidREGFlush(pjVideo);

                if ( lpVideoHWDifference->dwHQVInitPatch )
                {
                    WaitHQVFlipClear(pjVideo, ((dwHQVCtl&~HQV_SW_FLIP)|HQV_FLIP_STATUS)&~HQV_ENABLE);
                    VIDOutD( HQV_CONTROL, dwHQVCtl);
                    UniversalWait(pjVideo, HQV_CONTROL, HQV_FLIP_STATUS, 1);
                    WaitHQVFlipClear(pjVideo, ((dwHQVCtl&~HQV_SW_FLIP)|HQV_FLIP_STATUS)&~HQV_ENABLE);
                    VIDOutD( HQV_CONTROL, dwHQVCtl);
                    UniversalWait(pjVideo, HQV_CONTROL, HQV_FLIP_STATUS, 1);
                }
                else    /* CLE_C0 */
                {
                    VIDOutD( HQV_CONTROL, dwHQVCtl & ~HQV_SW_FLIP);
                    VIDOutD( HQV_CONTROL, dwHQVCtl | HQV_SW_FLIP);
                    UniversalWait(pjVideo, HQV_CONTROL, HQV_FLIP_STATUS, 1);
                }

                if (dwVideoFlag & VIDEO_1_INUSE)
                {
                    VIDOutD(V1_CONTROL, dwVidCtl);
                    VIDOutD(V_COMPOSE_MODE, (dwCompose|V1_COMMAND_FIRE ));
                    if (dwVideoFlag & VIDEO_1_ON_SND)/*for DuoView mode*/
                    {
                        if ((lpVideoHWDifference->dwNeedV3Prefetch==TRUE) && (lpVideoHWDifference->dwNeedV4Prefetch==TRUE))
                        {
                            /* Gamma */
                            VIDOutD(V3_CONTROL, (VIDInD(V3_CONTROL)&V3_GAMMA_ENABLE_3204)|dwVidCtl );
                            /* V4 Pre-Fetch Enable */
                            VIDOutD(HI_CONTROL, VIDInD(HI_CONTROL) |V4_PREFETCH_ON_3204 );
                            VIDOutD(V_COMPOSE_MODE, (dwCompose|V3_COMMAND_FIRE|COMPOSE_V3_TOP|V3_COMMAND_LOAD_VBI));
                        }
                        else
                        {
                            VIDOutD(V3_CONTROL, dwVidCtl);
                            VIDOutD(V_COMPOSE_MODE, (dwCompose|V3_COMMAND_FIRE|COMPOSE_V3_TOP|V3_COMMAND_LOAD_VBI));
                        }
                    }
                    if (pVidData->dwUseExtendedFIFO)
                    {
                        /* Set Display FIFO */
                        UniversalWait(pjVideo, V_FLAGS, VBI_STATUS, 0);
                
                    #ifdef XFree86Server
                        OUTB(0x17, 0x3C4); OUTB(0x2f, 0x3C5);
                        OUTB(0x16, 0x3C4); OUTB((pVidData->Save_3C4_16&0xf0)|0x14, 0x3C5);
                        OUTB(0x18, 0x3C4); OUTB(0x56, 0x3C5);                			   
			        #endif
                    }
                }
                else
                {
                    if ((lpVideoHWDifference->dwNeedV3Prefetch==TRUE) && (lpVideoHWDifference->dwNeedV4Prefetch==TRUE))
                    {
                        /* Gamma */
                        VIDOutD(V3_CONTROL, (VIDInD(V3_CONTROL)&V3_GAMMA_ENABLE_3204)|dwVidCtl );
                        /* V4 Pre-Fetch Enable */
                        VIDOutD(HI_CONTROL, VIDInD(HI_CONTROL) |V4_PREFETCH_ON_3204 );
                        VIDOutD(V_COMPOSE_MODE, (dwCompose|V3_COMMAND_FIRE ));
                    }
                    else
                    {
                        VIDOutD(V3_CONTROL, dwVidCtl);
                        VIDOutD(V_COMPOSE_MODE, (dwCompose|V3_COMMAND_FIRE ));
                    }
                }
            }
            else
            {
                DBG_DD(("    Normal called\n"));
                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER, HQV_CONTROL, dwHQVCtl|HQV_FLIP_STATUS);

                if (dwVideoFlag & VIDEO_1_INUSE)
                {
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V1_CONTROL, dwVidCtl);
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V_COMPOSE_MODE, (dwCompose|V1_COMMAND_FIRE ));
                    if(dwVideoFlag & VIDEO_1_ON_SND)/*for DuoView mode*/
                    {
                        if ((lpVideoHWDifference->dwNeedV3Prefetch == TRUE)&&(lpVideoHWDifference->dwNeedV4Prefetch == TRUE))
                        {
                            /* V3 Pre-Fetch Enable */
                            Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V3_CONTROL, (VIDInD(V3_CONTROL)&V3_GAMMA_ENABLE_3204)|dwVidCtl);
                            /* V4 Pre-Fetch Enable */
                            Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,HI_CONTROL, VIDInD(HI_CONTROL) |V4_PREFETCH_ON_3204 );
                            Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V_COMPOSE_MODE, (dwCompose|V3_COMMAND_FIRE|V3_COMMAND_LOAD_VBI|COMPOSE_V3_TOP));
                        }
                        else
                        {
                            Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V3_CONTROL, dwVidCtl);
                            Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V_COMPOSE_MODE, (dwCompose|V3_COMMAND_FIRE|V3_COMMAND_LOAD_VBI|COMPOSE_V3_TOP));
                        }
                    }
                }
                else
                {
                    if ((lpVideoHWDifference->dwNeedV3Prefetch == TRUE)&&(lpVideoHWDifference->dwNeedV4Prefetch == TRUE))
                    {
                        /* V3 Pre-Fetch Enable */
                        Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V3_CONTROL, (VIDInD(V3_CONTROL)&V3_GAMMA_ENABLE_3204)|dwVidCtl);
                        /* V4 Pre-Fetch Enable */
                        Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,HI_CONTROL, VIDInD(HI_CONTROL) |V4_PREFETCH_ON_3204 );
                        Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V_COMPOSE_MODE, (dwCompose|V3_COMMAND_FIRE ));
                    }
                    else
                    {
                        Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V3_CONTROL, dwVidCtl);
                        Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V_COMPOSE_MODE, (dwCompose|V3_COMMAND_FIRE ));
                    }
                }

                /* WaitHQVDone(pjVideo); */
                if (dwVideoFlag & VIDEO_1_INUSE)
                {
                    WaitVideoCommandFire(pjVideo);
                    if(dwVideoFlag & VIDEO_1_ON_SND)/*for DuoView mode*/
                        WaitVideo3CommandFire(pjVideo);
                }
                else
                {
                    WaitVideo3CommandFire(pjVideo);
                }
#ifdef CS_FLAG_HWROTATION
                if (lpVIAGraphicInfo->rotate ) 
                {
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,HQV_MINI_CONTROL, 0);
                }
#endif /* CS_FLAG_HWROTATION */
                Macro_VidREGFlush(pjVideo);
            }
        }
        else
        {
            if (dwVideoFlag & VIDEO_1_INUSE)
            {
                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V1_CONTROL, dwVidCtl);
                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V_COMPOSE_MODE, (dwCompose|V1_COMMAND_FIRE ));
            }
            else
            {
                if ((lpVideoHWDifference->dwNeedV3Prefetch == TRUE)&&(lpVideoHWDifference->dwNeedV4Prefetch == TRUE))
                {
                    /* V3 Pre-Fetch Enable */
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V3_CONTROL, (VIDInD(V3_CONTROL)&V3_GAMMA_ENABLE_3204)|dwVidCtl);
                    /* V4 Pre-Fetch Enable */
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,HI_CONTROL, VIDInD(HI_CONTROL) |V4_PREFETCH_ON_3204 );
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V_COMPOSE_MODE, (dwCompose|V3_COMMAND_FIRE ));
                }
                else
                {
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V3_CONTROL, dwVidCtl);
                    Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V_COMPOSE_MODE, (dwCompose|V3_COMMAND_FIRE ));
                }
            }
            if (dwVideoFlag & VIDEO_1_INUSE)
            {
                WaitVideoCommandFire(pjVideo);
            }
            else
            {
                WaitVideo3CommandFire(pjVideo);
            }
            Macro_VidREGFlush(pjVideo);
        }
    }
    else
    {
        /* Hide overlay */
        DBG_DD(("Hide overlay \n"));

        if ( lpVideoHWDifference->dwHQVDisablePatch )     /* CLE_C0 */
        {
         #ifdef XFree86Server
            OUTB(0x2E, 0x3C4);
            OUTB(0xEF, 0x3C5);
	     #endif
        }

        if (dwVideoFlag & VIDEO_1_INUSE)
        {
        Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V_FIFO_CONTROL,V1_FIFO_PRETHRESHOLD12 |
             V1_FIFO_THRESHOLD8 |V1_FIFO_DEPTH16);
        }
        else
        {
        Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,ALPHA_V3_FIFO_CONTROL, ALPHA_FIFO_THRESHOLD4
             | ALPHA_FIFO_DEPTH8  | V3_FIFO_THRESHOLD24 | V3_FIFO_DEPTH32 );
        }

        if (dwVideoFlag&VIDEO_HQV_INUSE)
        {
            Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,HQV_CONTROL, (VIDInD(HQV_CONTROL) & (~HQV_ENABLE)));
        }

        if (dwVideoFlag&VIDEO_1_INUSE)
        {
            Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V1_CONTROL, (VIDInD(V1_CONTROL) & (~V1_ENABLE)));
            Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V_COMPOSE_MODE, (VIDInD(V_COMPOSE_MODE)|V1_COMMAND_FIRE));
            if(dwVideoFlag&VIDEO_1_ON_SND)
            {
                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V3_CONTROL, (VIDInD(V3_CONTROL) & (~V3_ENABLE)));
                Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V_COMPOSE_MODE, (VIDInD(V_COMPOSE_MODE)|V3_COMMAND_FIRE));
            }
        }
        else
        {
            Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V3_CONTROL, (VIDInD(V3_CONTROL) & (~V3_ENABLE)));
            Macro_VidREGRec(VIDREGREC_SAVE_REGISTER,V_COMPOSE_MODE, (VIDInD(V_COMPOSE_MODE)|V3_COMMAND_FIRE));
        }

        if (dwVideoFlag & VIDEO_1_INUSE)
        {
            WaitVideoCommandFire(pjVideo);
            if (dwVideoFlag & VIDEO_1_ON_SND)/*for DuoView*/
                WaitVideo3CommandFire(pjVideo);
        }
        else
        {
            WaitVideo3CommandFire(pjVideo);
        }
        Macro_VidREGFlush(pjVideo);

        if ( lpVideoHWDifference->dwHQVDisablePatch )     /* CLE_C0 */
        {
         #ifdef XFree86Server
            OUTB(0x2E, 0x3C4);
            OUTB(0xFF, 0x3C5);   		 
	     #endif
        }
    }

    return VMI_OK;
}

/* Create overlay surface depend on FOURCC */
unsigned long CreateSurface(PVIDDATA pVidData, LPDDSURFACEDESC lpDDSurfaceDesc)
{
    unsigned long handle=pVidData->VMI_Info.handle;
    unsigned long   dwWidth, dwHeight, dwPitch=0;
    unsigned long   dwBackBufferCount;
    unsigned long   dwRet=VMI_OK;
    unsigned long   dwAddr = 0;
    unsigned long   MPEGFBSIZE  = 0;
    unsigned long   SUBPFBSIZE  = 0;
    unsigned long   HQVFBSIZE   = 0;
    unsigned long   TVFBSIZE    = 0;
    unsigned long   ALPHAFBSIZE = 0;
    unsigned long   SWFBSIZE    = 0;
    int     i, iCount;       /* iCount for clean HQV FB use */
    unsigned char  * lpTmpAddr;    /* for clean HQV FB use */

    LPVIDHWDIFFERENCE lpVideoHWDifference = &pVidData->VideoHWDifference;
    LPMPGDEVICE lpMPGDevice = &(pVidData->MPGDevice[0]);
    LPSUBDEVICE lpSubDevice = &(pVidData->ddLock.SubDev[0]);
    LPSWDEVICE  lpSWDevice  = &(pVidData->SWDevice[0]);
    LPALPHADEVICE lpAlpDevice = &(pVidData->ddLock.ALPDevice);
    LPVIDEOMEMORY lpVidMemory = &VidMemory;
    LPVIAGRAPHICINFO lpVIAGraphicInfo = &pVidData->VIAGraphicInfo;
    LPVIAVIDCTRL  lpVideoControl = &pVidData->VideoControl;
    LPDDPIXELFORMAT lpDPFsrc = &pVidData->DPFsrc[0];
    LPOVERLAYRECORD lpOverlayRecordMPEG = &pVidData->OverlayRecordSet.OverlayRecordMPEG[0];
    LPOVERLAYRECORD lpOverlayRecordSW = &pVidData->OverlayRecordSet.OverlayRecordSW[0];

    unsigned char* pjVideo = pVidData->Vhwinfo.pjVideo;
    unsigned char* pjScreen = pVidData->Vhwinfo.pjScreen;

    int ScrnIndex = lpVIAGraphicInfo->scrnIndex;

    EngMgr_RegClient_Para regclient_para;       /* for EngMgr_RegClient() */
    EngMgr_AdjustClientPriority_Para AdjustClientPriority_Para; /* for EngMgr_AdjustClientPriority */

    DBG_DD(("\n"));

    if ( lpDDSurfaceDesc == NULL )
        return VMI_ERR;

    /*ENGINEMANAGER*/
    if (!uAreadyInit)
    {
        InitEngineManager();
    }

    switch (lpDDSurfaceDesc->dwFourCC)
    {
        case FOURCC_NV12 :

            DBG_DD(("    Create MPEG Surface: \n"));
            lpDPFsrc->dwFlags = DDPF_FOURCC;
            lpDPFsrc->dwFourCC = FOURCC_NV12;

            /* init Video status flag */
            /*ENGINEMANAGER*/
            if( VidClient[CLIENT_MPEG].uClientHandle == CLIENT_INVALID_HANDLE )
	        {
	            /*register the client*/
		        regclient_para.func = EngCallback_MPEG;
		        regclient_para.uPriority = NORMAL_PRIORITY;
		        regclient_para.pClient_handle = &VidClient[CLIENT_MPEG].uClientHandle;
		        EngMgr_RegClient(&regclient_para);
            }

            switch ( lpVIAGraphicInfo->dwDeviceID )
            {
                case VIA_DEVICE_VT3204:
                case VIA_DEVICE_VT3205:
                case VIA_DEVICE_VT3314:
                    if ( VIA_Engine_Allocate(pVidData, CLIENT_MPEG, VID_ENGINE, 1) ) /*V3*/
                    {
                        pVidData->VidStatus.dwVideoFlagMPEG[0] = VIDEO_ACTIVE | VIDEO_SHOW | VIDEO_MPEG_INUSE | VIDEO_3_INUSE | MPEG_USE_V3;
                        DBG_DD(("    Create MPEG Surface: VIDEO_3_INUSE !!!!\n"));
                    }
                    else
                    {
                        DBG_DD((" No more Video Engine(V3) for MPEG \n"));
                        return VMI_ERR;
                    }
                    break;
                case VIA_DEVICE_CLE1:
                case VIA_DEVICE_CLE2:
                case VIA_DEVICE_VT3259:
                default:
                    if ( VIA_Engine_Allocate(pVidData, CLIENT_MPEG, VID_ENGINE, 0) ) /*V1*/
                    {
                        pVidData->VidStatus.dwVideoFlagMPEG[0] = VIDEO_ACTIVE | VIDEO_SHOW | VIDEO_MPEG_INUSE | VIDEO_1_INUSE | MPEG_USE_V1;
                        DBG_DD(("    Create MPEG Surface: VIDEO_1_INUSE !!!!\n"));
                        if((lpVIAGraphicInfo->duoview & DUOVIEW_ENABLED))
                        {
                            pVidData->VidStatus.dwVideoFlagMPEG[0] = VIDEO_ACTIVE | VIDEO_SHOW | VIDEO_MPEG_INUSE | VIDEO_1_INUSE | MPEG_USE_V1;
                            if ( VIA_Engine_Allocate(pVidData, CLIENT_MPEG,VID_ENGINE,1) ) /*V3*/
                            {
                                pVidData->VidStatus.dwVideoFlagMPEG[0] |=  VIDEO_3_INUSE|VIDEO_1_ON_SND;
                                DBG_DD(("    Create MPEG2 Surface: DuoView VIDEO_3_INUSE !!!!\n"));                                  
                            }
                        }
                    }
                    else if ( VIA_Engine_Allocate(pVidData, CLIENT_MPEG, VID_ENGINE, 1) ) /*V3*/
                    {
                        pVidData->VidStatus.dwVideoFlagMPEG[0] = VIDEO_ACTIVE | VIDEO_SHOW | VIDEO_MPEG_INUSE | VIDEO_3_INUSE | MPEG_USE_V3;
                        DBG_DD(("    Create MPEG Surface: VIDEO_3_INUSE !!!!\n"));
                    }
                    else
                    {
                        DBG_DD((" No more Video Engine for MPEG!! \n"));
                        return VMI_ERR;
                    }
                    break;
            } /*End of DeviceID*/

            if ( !(pVidData->VidStatus.dwVideoFlagMPEG[0] & VIDEO_1_INUSE)
                 && (pVidData->VidStatus.dwVideoFlagTV1 & VIDEO_1_INUSE) )
            {
                if ( VIA_Engine_Swap(pVidData, CLIENT_MPEG,VID_ENGINE,1,0) ) /*Swap V3 to V1*/
                {
                    pVidData->VidStatus.dwVideoFlagMPEG[0] &= ~VIDEO_3_INUSE;
                    pVidData->VidStatus.dwVideoFlagMPEG[0] |= VIDEO_1_INUSE;
                    DBG_DD((" Switch V3 to V1 !! \n"));
                }
            }

            AdjustClientPriority_Para.uClient_handle = VidClient[CLIENT_MPEG].uClientHandle;
            AdjustClientPriority_Para.uPriority = URGENT_PRIORITY ;
            EngMgr_AdjustClientPriority(&AdjustClientPriority_Para);

            if ( VIA_Engine_Allocate(pVidData, CLIENT_MPEG,HQV_ENGINE,0) ) /*HQV0*/
            {
                pVidData->VidStatus.dwVideoFlagMPEG[0] |= VIDEO_HQV_INUSE | MPEG_USE_HQV;
                lpVideoControl->dwHighQVDO = VW_HIGHQVDO_MPEG;
                DBG_DD(("    Create MPEG Surface: HQV_INUSE !!!!\n"));
            }
            else
            {
                DBG_DD(("    Create MPEG Surface: no HQV_INUSE !!!!\n"));
            }

            AdjustClientPriority_Para.uClient_handle = VidClient[CLIENT_MPEG].uClientHandle;
            AdjustClientPriority_Para.uPriority = NORMAL_PRIORITY ;
            EngMgr_AdjustClientPriority(&AdjustClientPriority_Para);

            if (pVidData->VidStatus.dwVideoFlagMPEG[0] != 0)
            {
                switch ( lpVIAGraphicInfo->dwDeviceID )
                {
                    case VIA_DEVICE_CLE1:
                    case VIA_DEVICE_CLE2:
                        switch ( DispatchVGARevisionID(pVidData) )
                        {
                            case VIA_REVISION_CLEAX :
                            case VIA_REVISION_CLECX :
                                pVidData->VidStatus.dwVideoFlagMPEG[0] |= MPEG_USE_SW_FLIP;
                                DBG_DD(("    Create MPEG Surface: MPEG_USE_SW_FLIP !!!!\n"));
                                break;

                            default :
                                pVidData->VidStatus.dwVideoFlagMPEG[0] |= MPEG_USE_HW_FLIP;
                                DBG_DD(("    Create MPEG Surface: MPEG_USE_HW_FLIP !!!!\n"));
                                break;
                        }
                        break;

                    case VIA_DEVICE_VT3259:
                    default:
                        pVidData->VidStatus.dwVideoFlagMPEG[0] |= MPEG_USE_SW_FLIP;
                        break;
                }
            }
            

            /* Color Space Conversion */
            if (!lpVIAGraphicInfo->ColorSpaceChanged)
            {
                switch ( lpVIAGraphicInfo->dwDeviceID )
                {
                    case VIA_DEVICE_VT3204:
                    case VIA_DEVICE_VT3205:
                    case VIA_DEVICE_VT3314:
                        VIDOutD(V3_ColorSpaceReg_1, ColorSpaceValue_1_32045);
                        VIDOutD(V3_ColorSpaceReg_2, ColorSpaceValue_2_32045);
                        DBG_DD(("000002C4 %08x\n",ColorSpaceValue_1_32045));
                        DBG_DD(("000002C8 %08x\n",ColorSpaceValue_2_32045));
                        break;

                    case VIA_DEVICE_CLE1:
                    case VIA_DEVICE_CLE2:
                        switch ( DispatchVGARevisionID(pVidData) )
                        {
                            case VIA_REVISION_CLEAX :
                                VIDOutD(ColorSpaceReg_1, ColorSpaceValue_1);
                                VIDOutD(ColorSpaceReg_2, ColorSpaceValue_2);
                                VIDOutD(V3_ColorSpaceReg_1, ColorSpaceValue_1);
                                VIDOutD(V3_ColorSpaceReg_2, ColorSpaceValue_2);
                                DBG_DD(("00000284 %08x\n",ColorSpaceValue_1));
                                DBG_DD(("00000288 %08x\n",ColorSpaceValue_2));
                                DBG_DD(("000002C4 %08x\n", ColorSpaceValue_1));
                                DBG_DD(("000002C8 %08x\n", ColorSpaceValue_2));
                                break;

                            case VIA_REVISION_CLECX :
                            default :
                                VIDOutD(ColorSpaceReg_1, ColorSpaceValue_1_3123C0);
                                VIDOutD(ColorSpaceReg_2, ColorSpaceValue_2_3123C0);
                                VIDOutD(V3_ColorSpaceReg_1, ColorSpaceValue_1_3123C0);
                                VIDOutD(V3_ColorSpaceReg_2, ColorSpaceValue_2_3123C0);
                                DBG_DD(("00000284 %08x\n",ColorSpaceValue_1_3123C0));
                                DBG_DD(("00000288 %08x\n",ColorSpaceValue_2_3123C0));
                                DBG_DD(("000002C4 %08x\n",ColorSpaceValue_1_3123C0));
                                DBG_DD(("000002C8 %08x\n",ColorSpaceValue_2_3123C0));
                                break;
                        }
                        break;

                    case VIA_DEVICE_VT3259:
                    default:
                        VIDOutD(ColorSpaceReg_1, ColorSpaceValue_1_3123C0);
                        VIDOutD(ColorSpaceReg_2, ColorSpaceValue_2_3123C0);
                        VIDOutD(V3_ColorSpaceReg_1, ColorSpaceValue_1_3123C0);
                        VIDOutD(V3_ColorSpaceReg_2, ColorSpaceValue_2_3123C0);
                        DBG_DD(("00000284 %08x\n",ColorSpaceValue_1_3123C0));
                        DBG_DD(("00000288 %08x\n",ColorSpaceValue_2_3123C0));
                        DBG_DD(("000002C4 %08x\n",ColorSpaceValue_1_3123C0));
                        DBG_DD(("000002C8 %08x\n",ColorSpaceValue_2_3123C0));
                        break;
                }
            }
            else
            {
                VIDOutD(ColorSpaceReg_1, lpVIAGraphicInfo->ColorSpaceValue1);
                VIDOutD(ColorSpaceReg_2, lpVIAGraphicInfo->ColorSpaceValue2);
                VIDOutD(V3_ColorSpaceReg_1, lpVIAGraphicInfo->ColorSpaceValue1);
                VIDOutD(V3_ColorSpaceReg_2, lpVIAGraphicInfo->ColorSpaceValue2);
            }

            if (!lpVIAGraphicInfo->SetMpeg_FBNumber)
            {
                dwBackBufferCount = lpDDSurfaceDesc->dwBackBufferCount;
                DBG_DD(("MPEG decode frame buffer number = %ld\n",dwBackBufferCount));
            }
            else
            {
                dwBackBufferCount = 7;
                DBG_DD(("MPEG decode frame buffer number = %ld\n",dwBackBufferCount));
            }

            dwWidth  = lpDDSurfaceDesc->dwWidth;
            dwHeight = lpDDSurfaceDesc->dwHeight;
            dwPitch  = ALIGN_TO_32_BYTE(dwWidth);

            MPEGFBSIZE = (((dwPitch * 3)/2)*dwHeight);       /* 1.5 bytes per pixel */

            if (!lpVIAGraphicInfo->DRMEnabled)
            {
                lpVidMemory->MPEGMemRequest[0].checkID = VIA_MID;
                lpVidMemory->MPEGMemRequest[0].size    = MPEGFBSIZE*(dwBackBufferCount+1) + dwPitch*4;
                lpVidMemory->MPEGMemRequest[0].type    = VIDMGR_TYPE_VIDEO;
                if ( MEM_OK != viaMgr_vid_ioctl(pVidData, VIAMGR_ALLOC_VIDEO, (char *)(&lpVidMemory->MPEGMemRequest[0]) ) )
                {
                    DBG_DD((" Memory Allocation fail for MPEG Video.\n"));
                   return VMI_ERR_CANNOT_CREATE_SURFACE;
                }
                dwAddr = ALIGN_TO_32_BYTE(lpVidMemory->MPEGMemRequest[0].offset);
            }

            lpMPGDevice->dwMPEGPhysicalAddr[0]   = dwAddr;
            lpMPGDevice->lpMPEGOverlaySurface[0] = pjScreen+dwAddr;

            lpMPGDevice->dwMPEGYPhysicalAddr[0]  = dwAddr;
            lpMPGDevice->dwMPEGCbPhysicalAddr[0] = lpMPGDevice->dwMPEGYPhysicalAddr[0]
                                                   + (dwPitch*dwHeight);
            lpMPGDevice->dwMPEGCrPhysicalAddr[0] = lpMPGDevice->dwMPEGCbPhysicalAddr[0]
                                                   + ((dwPitch>>1)*(dwHeight>>1));

            DBG_DD(("//    :Buffer[0]: 0x%x\n", lpMPGDevice->dwMPEGPhysicalAddr[0]));
            for ( i=1; i<= dwBackBufferCount; i++)
            {
                lpMPGDevice->dwMPEGPhysicalAddr[i]=lpMPGDevice->dwMPEGPhysicalAddr[i-1]+MPEGFBSIZE;
                lpMPGDevice->lpMPEGOverlaySurface[i]=lpMPGDevice->lpMPEGOverlaySurface[i-1]+MPEGFBSIZE;

                lpMPGDevice->dwMPEGYPhysicalAddr[i]  = lpMPGDevice->dwMPEGYPhysicalAddr[i-1]
                                                       + MPEGFBSIZE;
                lpMPGDevice->dwMPEGCbPhysicalAddr[i] = lpMPGDevice->dwMPEGYPhysicalAddr[i]
                                                       + (dwPitch*dwHeight);
                lpMPGDevice->dwMPEGCrPhysicalAddr[i] = lpMPGDevice->dwMPEGCbPhysicalAddr[i]
                                                       + ((dwPitch>>1)*(dwHeight>>1));
                DBG_DD(("//    :Buffer[%d]: 0x%x\n", i, lpMPGDevice->dwMPEGPhysicalAddr[i]));
            }
            /* fill in the MPEG buffer with Y:0x00, U,V:0x80 (YV12-black color) to clear MPEG buffers */
            for ( i=0; i<= dwBackBufferCount; i++)
            {
                lpTmpAddr = pjScreen + lpMPGDevice->dwMPEGYPhysicalAddr[i];
                for(iCount=0;iCount<(dwPitch * dwHeight);iCount++)
                    *lpTmpAddr++=0x00;
                for(iCount=0;iCount<(dwPitch * (dwHeight>>1));iCount++)
                    *lpTmpAddr++=0x80;
            }

            if (lpMPGDevice->dwMPEGPhysicalAddr[dwBackBufferCount] + MPEGFBSIZE >=
                        (unsigned long)lpVIAGraphicInfo->Screen[ScrnIndex].VideoHeapEnd)
            {
                DBG_DD(("//    :Memory not enough for MPEG\n"));
                return VMI_ERR_CANNOT_CREATE_SURFACE;
            }

            lpMPGDevice->dwWidth  = dwWidth;           /* not changed later */
            lpMPGDevice->dwHeight = dwHeight;          /* not changed later */
            lpMPGDevice->dwPitch  = dwPitch;           /* not changed later */
            lpMPGDevice->gdwMPGSrcWidth  = dwWidth;    /* changed if window is out of screen */
            lpMPGDevice->gdwMPGSrcHeight = dwHeight;   /* changed if window is out of screen */
            lpMPGDevice->byDeviceType = DEV_MPEG;

            /* Fill image data in overlay record */
            lpOverlayRecordMPEG->dwOriSrcWidth  = lpMPGDevice->dwWidth;
            lpOverlayRecordMPEG->dwOriSrcHeight = lpMPGDevice->dwHeight;
            lpOverlayRecordMPEG->dwOriSrcPitch  = lpMPGDevice->dwPitch;


            if (!lpVIAGraphicInfo->DRMEnabled)
            {
                    lpVidMemory->MPEGDecodeCountMemRequest.checkID = VIA_MID;
                    lpVidMemory->MPEGDecodeCountMemRequest.size    = 16;
                    lpVidMemory->MPEGDecodeCountMemRequest.type    = VIDMGR_TYPE_VIDEO;
                    if ( MEM_OK != viaMgr_vid_ioctl(pVidData, VIAMGR_ALLOC_VIDEO, (char *)(&lpVidMemory->MPEGDecodeCountMemRequest) ) )
                    {
                        DBG_DD((" Memory Allocation fail for MPEG Video.\n"));
                       return VMI_ERR_CANNOT_CREATE_SURFACE;
                    }

                    lpMPGDevice->dwDecodeEndCountPhysicalAddr
                            = ALIGN_TO_32_BYTE(lpVidMemory->MPEGDecodeCountMemRequest.offset);
            }
            lpTmpAddr = pjScreen + lpMPGDevice->dwDecodeEndCountPhysicalAddr;
            for (i=0; i<16; i++)  *lpTmpAddr++=0x00;


            /* If M/B uses newer chips than CLE A2, we enable errorconcealment function */
            switch ( lpVIAGraphicInfo->dwDeviceID )
            {
                case VIA_DEVICE_CLE1:
                case VIA_DEVICE_CLE2:
                    switch ( DispatchVGARevisionID(pVidData) )
                    {
                        case VIA_REVISION_CLEAX :
                            lpMPGDevice->dwEnableErrorConcealment = 0;
                            break;

                        case VIA_REVISION_CLECX :
                        default :
                            lpMPGDevice->dwEnableErrorConcealment = 1;
                            break;
                    }
                    break;

                case VIA_DEVICE_VT3259:
                default:
                    lpMPGDevice->dwEnableErrorConcealment = 1;
                    break;
            }

            /*  In some case , we don't need to create MPEG HQV surface */
            if (!(pVidData->VidStatus.dwVideoFlagMPEG[0] & MPEG_USE_HQV))
            {
            break;
            }
            else
            {
                if (!lpVIAGraphicInfo->DRMEnabled)
                {
                    if (lpVidMemory->HQVMemRequest.offset != 0)
                    {
                        break;
                    }
                }
                else
                {
                    if (lpVidMemory->HQVfbRequest.offset != 0)
                    {
                        break;
                    }
                }
            }

        case FOURCC_HQVMPEG :
            DBG_DD(("    Create MPEG HQV Surface: \n"));
            dwWidth  = lpMPGDevice->dwWidth;
            dwHeight = lpMPGDevice->dwHeight;
            dwPitch  = lpMPGDevice->dwPitch;

            HQVFBSIZE = dwPitch * dwHeight * 2;

            if (!lpVIAGraphicInfo->DRMEnabled)
            {
                lpVidMemory->HQVMemRequest.checkID = VIA_MID;

                if ( lpVideoHWDifference->dwThreeHQVBuffer )    /* CLE_C0 */
                    lpVidMemory->HQVMemRequest.size    = (HQVFBSIZE * 3);
                else
                    lpVidMemory->HQVMemRequest.size    = (HQVFBSIZE*2);

                lpVidMemory->HQVMemRequest.type    = VIDMGR_TYPE_VIDEO;
                if ( MEM_OK != viaMgr_vid_ioctl(pVidData, VIAMGR_ALLOC_VIDEO, (char *)(&lpVidMemory->HQVMemRequest) ) )
                {
                    DBG_DD((" Memory Allocation fail for MPEG HQV Video.\n"));
                    return VMI_ERR_CANNOT_CREATE_SURFACE;
                }
                dwAddr = ALIGN_TO_32_BYTE(lpVidMemory->HQVMemRequest.offset);
            }

            lpMPGDevice->dwHQVAddr[0] = dwAddr;
            lpMPGDevice->dwHQVAddr[1] = dwAddr + HQVFBSIZE;
            lpOverlayRecordMPEG->dwHQVAddr[0] = dwAddr;
            lpOverlayRecordMPEG->dwHQVAddr[1] = dwAddr + HQVFBSIZE;

            DBG_DD(("lpMPGDevice->dwHQVAddr[0]= %08lx\n",lpMPGDevice->dwHQVAddr[0]));
            DBG_DD(("lpMPGDevice->dwHQVAddr[1]= %08lx\n",lpMPGDevice->dwHQVAddr[1]));

            if ( lpVideoHWDifference->dwThreeHQVBuffer )    /* CLE_C0 */
            {
                lpMPGDevice->dwHQVAddr[2] = lpMPGDevice->dwHQVAddr[1] + HQVFBSIZE;
                lpOverlayRecordMPEG->dwHQVAddr[2] = lpOverlayRecordMPEG->dwHQVAddr[1] + HQVFBSIZE;
                DBG_DD(("lpMPGDevice->dwHQVAddr[2]= %08lx\n",lpMPGDevice->dwHQVAddr[2]));

                if (lpMPGDevice->dwHQVAddr[2] + HQVFBSIZE >=
                                    (unsigned long)lpVIAGraphicInfo->Screen[ScrnIndex].VideoHeapEnd)
            {
                DBG_DD(("//    :Memory not enough for MPEG with HQV\n"));
                return VMI_ERR_CANNOT_CREATE_SURFACE;
            }
            }
            else
            {
                if (lpMPGDevice->dwHQVAddr[1] + HQVFBSIZE >=
                                        (unsigned long)lpVIAGraphicInfo->Screen[ScrnIndex].VideoHeapEnd)
                {
                    DBG_DD(("//    :Memory not enough for MPEG with HQV\n"));
                    return VMI_ERR_CANNOT_CREATE_SURFACE;
                }
            }

            /* fill in the HQV buffer with 0x8000 (YUV-black color) to clear HQV buffers */
            for(iCount=0;iCount<HQVFBSIZE*2;iCount++)
            {
                lpTmpAddr = pjScreen + dwAddr + iCount;
                if((iCount%2) == 0)
                    *(lpTmpAddr)=0x00;
                else
                    *(lpTmpAddr)=0x80;
            }

            /* dwPitch -1 to hide the noise dot in the left down corner */
            /* dwHeight -1 to hide green line in the bottom */

            if ( lpVideoHWDifference->dwHQVFetchByteUnit )     /* CLE_C0 */
                VIDOutD(HQV_SRC_FETCH_LINE, (((dwPitch)-1)<<16)|(dwHeight-1));
            else
                VIDOutD(HQV_SRC_FETCH_LINE, (((dwPitch>>3)-1)<<16)|(dwHeight-1));

            VIDOutD(HQV_SRC_STRIDE, ((dwPitch>>1)<<16)|dwPitch);
            VIDOutD(HQV_DST_STRIDE, (dwPitch<<1));

            VIDOutD(HQV_DST_STARTADDR0,lpMPGDevice->dwHQVAddr[0]);
            VIDOutD(HQV_DST_STARTADDR1,lpMPGDevice->dwHQVAddr[1]);
            DBG_DD(("000003EC %08lx\n",VIDInD(HQV_DST_STARTADDR0) ));
            DBG_DD(("000003F0 %08lx\n",VIDInD(HQV_DST_STARTADDR1) ));
            if ( lpVideoHWDifference->dwThreeHQVBuffer )    /*CLE_C0*/
            {
                VIDOutD(HQV_DST_STARTADDR2,lpMPGDevice->dwHQVAddr[2]);
                DBG_DD(("000003FC %08lx\n",VIDInD(HQV_DST_STARTADDR2) ));
            }

            DBG_DD(("000003E0 %08lx\n",VIDInD(HQV_SRC_FETCH_LINE) ));
            DBG_DD(("000003F8 %08lx\n",VIDInD(HQV_SRC_STRIDE) ));
            DBG_DD(("000003F4 %08lx\n",VIDInD(HQV_DST_STRIDE) ));
            break;

        case FOURCC_SUBP :
            DBG_DD(("    Create subpicture Surface: \n"));

            dwBackBufferCount = lpDDSurfaceDesc-> dwBackBufferCount;
            dwWidth  = lpDDSurfaceDesc->dwWidth;
            dwHeight = lpDDSurfaceDesc->dwHeight;
            dwPitch  = ALIGN_TO_32_BYTE(dwWidth);

            SUBPFBSIZE = dwPitch*dwHeight;    /* 4 bits per pixel,AI44 mode */


            if (!lpVIAGraphicInfo->DRMEnabled)
            {
                lpVidMemory->SUBPMemRequest[0].checkID = VIA_MID;
                lpVidMemory->SUBPMemRequest[0].size    = SUBPFBSIZE*2;
                lpVidMemory->SUBPMemRequest[0].type    = VIDMGR_TYPE_VIDEO;
                if ( MEM_OK != viaMgr_vid_ioctl(pVidData, VIAMGR_ALLOC_VIDEO, (char *)(&lpVidMemory->SUBPMemRequest[0]) ) )
                {
                    DBG_DD((" Memory Allocation fail for SubPicture.\n"));
                    return VMI_ERR_CANNOT_CREATE_SURFACE;
                }
                dwAddr = ALIGN_TO_32_BYTE(lpVidMemory->SUBPMemRequest[0].offset);
            }

            lpSubDevice->dwSUBPhysicalAddr[0] = dwAddr;

            lpSubDevice->dwSUBPhysicalAddr[1] = lpSubDevice->dwSUBPhysicalAddr[0] + SUBPFBSIZE;
            DBG_DD(("SP_Buffer[0]: 0x%x\n", lpSubDevice->dwSUBPhysicalAddr[0]));
            DBG_DD(("SP_Buffer[1]: 0x%x\n", lpSubDevice->dwSUBPhysicalAddr[1]));

            if (lpSubDevice->dwSUBPhysicalAddr[1] + SUBPFBSIZE  >=
                                              (unsigned long)lpVIAGraphicInfo->Screen[ScrnIndex].VideoHeapEnd)
            {
                DBG_DD(("Memory not enough for SubPicture\n"));
                return VMI_ERR_CANNOT_CREATE_SURFACE;
            }
	    /* add the subdevice's overalay surface	*/
            lpSubDevice->lpSUBOverlaySurface[0] = pjScreen + lpSubDevice->dwSUBPhysicalAddr[0];
	     lpSubDevice->lpSUBOverlaySurface[1]	= pjScreen + lpSubDevice->dwSUBPhysicalAddr[1];
		 
            lpSubDevice->gdwSUBSrcWidth = dwWidth;
            lpSubDevice->gdwSUBSrcHeight= dwHeight;
            lpSubDevice->dwPitch        = dwPitch;
            lpOverlayRecordMPEG->dwSPPitch  = dwPitch;

            /* Subpicture Frame Buffer Stride */
            VIDOutD(SUBP_CONTROL_STRIDE, SUBP_AI44|dwPitch);
            DBG_DD(("000003C0 %08lx\n",VIDInD(SUBP_CONTROL_STRIDE) ));

            lpMPGDevice->dwSUBPPitch  = dwPitch;

            lpMPGDevice->byDeviceType = DEV_SUBP;
            lpMPGDevice->gdwSUBP_NotVisible = 1;

            break;

        case FOURCC_ALPHA :
            DBG_DD(("    Create ALPHA Surface: \n"));
            dwWidth  = lpDDSurfaceDesc->dwWidth;
            dwHeight = lpDDSurfaceDesc->dwHeight;
            if(lpVIAGraphicInfo->dwDeviceID == VIA_DEVICE_VT3314)
                dwPitch  = ALIGN_TO_16_BYTE(dwWidth);
            else 
                dwPitch  = ALIGN_TO_16_BYTE(dwWidth/2);

            ALPHAFBSIZE = dwPitch*dwHeight;

            if (!lpVIAGraphicInfo->DRMEnabled)
            {
                lpVidMemory->ALPHAMemRequest.checkID = VIA_MID;
                lpVidMemory->ALPHAMemRequest.size    = ALPHAFBSIZE;
                lpVidMemory->ALPHAMemRequest.type    = VIDMGR_TYPE_VIDEO;
                if ( MEM_OK != viaMgr_vid_ioctl(pVidData, VIAMGR_ALLOC_VIDEO, (char *)(&lpVidMemory->ALPHAMemRequest) ) )
                {
                   DBG_DD((" Memory Allocation fail for Alpha Surface.\n"));
                   return VMI_ERR_CANNOT_CREATE_SURFACE;
                }
                dwAddr = ALIGN_TO_32_BYTE(lpVidMemory->ALPHAMemRequest.offset);
            }

            lpAlpDevice->dwALPPhysicalAddr   = dwAddr;
            lpAlpDevice->lpALPOverlaySurface = pjScreen+dwAddr;
            lpAlpDevice->dwPitch = dwPitch;
            lpAlpDevice->gdwALPDstHeight = dwHeight;
            lpAlpDevice->gdwALPDstWidth  = dwWidth ;
            lpAlpDevice->gdwALPDstLeft   = 0;
            lpAlpDevice->gdwALPDstTop    = 0;
            if (lpAlpDevice->dwALPPhysicalAddr+ALPHAFBSIZE >=
                                    (unsigned long)lpVIAGraphicInfo->Screen[ScrnIndex].VideoHeapEnd)
            {
                DBG_DD(("Memory not enough\n"));

                return VMI_ERR_CANNOT_CREATE_SURFACE;
            }
				
            break;

       case FOURCC_YUY2 :
#ifdef CS_FLAG_HWROTATION
            if(lpVIAGraphicInfo->IsSurfaceUsedForRotate)
            {
                DBG_DD(("    Create SW YUY2 Surface: for rotate. \n"));
                lpDPFsrc = &pVidData->DPFsrc_rot;
                lpSWDevice = &(pVidData->SWDevice_rot[handle]);                
            }
            else
#endif /* CS_FLAG_HWROTATION */
            {
                DBG_DD(("    Create SW YUY2 Surface: \n"));
            }
            
            lpDPFsrc->dwFlags = DDPF_FOURCC;
            lpDPFsrc->dwFourCC = FOURCC_YUY2;

            if (!lpVIAGraphicInfo->IsSurfaceUsedForRotate)
            {
                /* init Video status flag*/
                /*ENGINEMANAGER*/
                if( VidClient[CLIENT_SW].uClientHandle == CLIENT_INVALID_HANDLE )
    	        {
    	            /*register the client*/
    		        regclient_para.func = EngCallback_SW;
    		        regclient_para.uPriority = NORMAL_PRIORITY;
    		        regclient_para.pClient_handle = &VidClient[CLIENT_SW].uClientHandle;
    		        EngMgr_RegClient(&regclient_para);
    		        DBG_DD((" *regclient_para.pClient_handle = %ld \n",*regclient_para.pClient_handle));
    		        DBG_DD((" VidClient[CLIENT_SW].uClientHandle= %ld \n",VidClient[CLIENT_SW].uClientHandle));
    		        /*VidClient[CLIENT_MPEG].pPossessed_Eng =NULL;*/
                }

                switch ( lpVIAGraphicInfo->dwDeviceID )
                {
                    case VIA_DEVICE_VT3204:
                    case VIA_DEVICE_VT3205:
                    case VIA_DEVICE_VT3314:
                        if ( VIA_Engine_Allocate(pVidData, CLIENT_SW,VID_ENGINE,1) ) /*V3*/
                        {
                        pVidData->VidStatus.dwVideoFlagSW[0] = VIDEO_HQV_INUSE | SW_USE_HQV | VIDEO_3_INUSE;
                        }
                        else
                        {
                            DBG_DD((" No more Video Engine (V3) for YUY2!! \n"));
                            return VMI_ERR;
                            /*pVidData->VidStatus.dwVideoFlagSW[0] = 0;*/
                        }break;
                    case VIA_DEVICE_CLE1:
                    case VIA_DEVICE_CLE2:
                    case VIA_DEVICE_VT3259:
                    default:
                        if ( VIA_Engine_Allocate(pVidData, CLIENT_SW,VID_ENGINE,0) ) /*V1*/
                        {
                            pVidData->VidStatus.dwVideoFlagSW[0] = VIDEO_ACTIVE | VIDEO_SHOW | VIDEO_1_INUSE;
                            DBG_DD(("    Create YUY2 Surface: VIDEO_1_INUSE !!!!\n"));
                            if((lpVIAGraphicInfo->duoview & 0x01))
                                {
                                    if ( VIA_Engine_Allocate(pVidData, CLIENT_SW,VID_ENGINE,1) ) /*V3*/
                                        {
                                            pVidData->VidStatus.dwVideoFlagSW[0] |=  VIDEO_3_INUSE|VIDEO_1_ON_SND;
                                            DBG_DD(("    Create YUY2 Surface: DuoView VIDEO_3_INUSE !!!!\n"));
                                            
                                        }

                                }
                        }
                        else if ( VIA_Engine_Allocate(pVidData, CLIENT_SW,VID_ENGINE,1) ) /*V3*/
                        {
                            pVidData->VidStatus.dwVideoFlagSW[0] = VIDEO_ACTIVE | VIDEO_SHOW | VIDEO_3_INUSE;
                            DBG_DD(("    Create YUY2 Surface: VIDEO_3_INUSE !!!!\n"));
                        }
                        else
                        {
                            DBG_DD((" No more Video Engine for YUY2!! \n"));
                            return VMI_ERR;
                        }
                        break;
                } /*End of DeviceID*/
                if ( VIA_Engine_Allocate(pVidData, CLIENT_SW,HQV_ENGINE,0) ) /*HQV*/
                {
                    pVidData->VidStatus.dwVideoFlagSW[0] |= VIDEO_HQV_INUSE | SW_USE_HQV;
                    lpVideoControl->dwHighQVDO = VW_HIGHQVDO_SW;
                    DBG_DD(("    Create YUY2 Surface: HQV_INUSE !!!!\n"));
                }
                else
                {
                    DBG_DD(("    Create YUY2 Surface: no HQV_INUSE !!!!\n"));
                }
            }

            /*write Color Space Conversion param.*/
            if (!lpVIAGraphicInfo->ColorSpaceChanged)
            {
                switch ( lpVIAGraphicInfo->dwDeviceID )
                {
                    case VIA_DEVICE_VT3204:
                    case VIA_DEVICE_VT3205:
                    case VIA_DEVICE_VT3314:
                        VIDOutD(V3_ColorSpaceReg_1, ColorSpaceValue_1_32045);
                        VIDOutD(V3_ColorSpaceReg_2, ColorSpaceValue_2_32045);
                        DBG_DD(("000002C4 %08x\n",ColorSpaceValue_1_32045));
                        DBG_DD(("000002C8 %08x\n",ColorSpaceValue_2_32045));
                        break;

                    case VIA_DEVICE_CLE1:
                    case VIA_DEVICE_CLE2:
                        switch ( DispatchVGARevisionID(pVidData) )
                        {
                            case VIA_REVISION_CLEAX :
                                VIDOutD(ColorSpaceReg_1, ColorSpaceValue_1);
                                VIDOutD(ColorSpaceReg_2, ColorSpaceValue_2);
                                VIDOutD(V3_ColorSpaceReg_1, ColorSpaceValue_1);
                                VIDOutD(V3_ColorSpaceReg_2, ColorSpaceValue_2);
                                DBG_DD(("00000284 %08x\n",ColorSpaceValue_1));
                                DBG_DD(("00000288 %08x\n",ColorSpaceValue_2));
                                DBG_DD(("000002C4 %08x\n", ColorSpaceValue_1));
                                DBG_DD(("000002C8 %08x\n", ColorSpaceValue_2));
                                break;
                            case VIA_REVISION_CLECX :
                            default : /* For future chip */
                                VIDOutD(ColorSpaceReg_1, ColorSpaceValue_1_3123C0);
                                VIDOutD(ColorSpaceReg_2, ColorSpaceValue_2_3123C0);
                                VIDOutD(V3_ColorSpaceReg_1, ColorSpaceValue_1_3123C0);
                                VIDOutD(V3_ColorSpaceReg_2, ColorSpaceValue_2_3123C0);
                                DBG_DD(("00000284 %08x\n",ColorSpaceValue_1_3123C0));
                                DBG_DD(("00000288 %08x\n",ColorSpaceValue_2_3123C0));
                                DBG_DD(("000002C4 %08x\n",ColorSpaceValue_1_3123C0));
                                DBG_DD(("000002C8 %08x\n",ColorSpaceValue_2_3123C0));
                                break;
                        }
                        break;

                    case VIA_DEVICE_VT3259:
                    default:
                        VIDOutD(ColorSpaceReg_1, ColorSpaceValue_1_3123C0);
                        VIDOutD(ColorSpaceReg_2, ColorSpaceValue_2_3123C0);
                        VIDOutD(V3_ColorSpaceReg_1, ColorSpaceValue_1_3123C0);
                        VIDOutD(V3_ColorSpaceReg_2, ColorSpaceValue_2_3123C0);
                        DBG_DD(("00000284 %08x\n",ColorSpaceValue_1_3123C0));
                        DBG_DD(("00000288 %08x\n",ColorSpaceValue_2_3123C0));
                        DBG_DD(("000002C4 %08x\n",ColorSpaceValue_1_3123C0));
                        DBG_DD(("000002C8 %08x\n",ColorSpaceValue_2_3123C0));
                        break;
                }/*End of DeviceID   */
            }
            else
            {
                VIDOutD(ColorSpaceReg_1, lpVIAGraphicInfo->ColorSpaceValue1);
                VIDOutD(ColorSpaceReg_2, lpVIAGraphicInfo->ColorSpaceValue2);
                VIDOutD(V3_ColorSpaceReg_1, lpVIAGraphicInfo->ColorSpaceValue1);
                VIDOutD(V3_ColorSpaceReg_2, lpVIAGraphicInfo->ColorSpaceValue2);
            }

            dwBackBufferCount = lpDDSurfaceDesc->dwBackBufferCount;
            dwWidth  = lpDDSurfaceDesc->dwWidth;
            dwHeight = lpDDSurfaceDesc->dwHeight;
            dwPitch  = ALIGN_TO_32_BYTE(dwWidth)*2;
            DBG_DD(("    srcWidth= %ld \n", dwWidth));
            DBG_DD(("    srcHeight= %ld \n", dwHeight));

            SWFBSIZE = dwPitch*dwHeight;    /* YUYV */

            if (!lpVIAGraphicInfo->DRMEnabled)
            {
                
                dwAddr = VIAAllocFB(pVidData, &lpVidMemory->SWMemRequest[handle], SWFBSIZE*2);
                
            }
            else
            {
                lpVidMemory->SWfbRequest.context=1;
                lpVidMemory->SWfbRequest.size=SWFBSIZE*2;
                lpVidMemory->SWfbRequest.type= VIDEO;
            }

            if (dwAddr ==0)
            {
                DBG_DD((" Memory Allocation fail for SW YUY2 Video.\n"));
                return VMI_ERR_CANNOT_CREATE_SURFACE;
            }

            /* fill in the SW buffer with 0x8000 (YUY2-black color) to clear FB buffer */
            lpTmpAddr = pjScreen + dwAddr;
            for(iCount=0;iCount<(SWFBSIZE*2);iCount++)
            {
                if((iCount%2) == 0)
                    *lpTmpAddr++=0x00;
                else
                {
#ifdef CS_FLAG_HWROTATION 
		            if(lpVIAGraphicInfo->rotate)
                    {      
            		    *lpTmpAddr++=0x00; /* fill the buffer with rgb blcak */
                    }
                    else
#endif /* CS_FLAG_HWROTATION */		
                        *lpTmpAddr++=0x80;
                    
                }
            }

            lpSWDevice->dwSWPhysicalAddr[0]   = dwAddr;
            lpSWDevice->lpSWOverlaySurface[0] = pjScreen+dwAddr;

            lpSWDevice->dwSWPhysicalAddr[1] = lpSWDevice->dwSWPhysicalAddr[0] + SWFBSIZE;
            lpSWDevice->lpSWOverlaySurface[1] = lpSWDevice->lpSWOverlaySurface[0] + SWFBSIZE;

            DBG_DD(("lpSWDevice->dwSWPhysicalAddr[0] 0x%08lx\n", dwAddr));
            DBG_DD(("lpSWDevice->dwSWPhysicalAddr[1] 0x%08lx\n", dwAddr + SWFBSIZE));

            if (lpVIAGraphicInfo->IsSurfaceUsedForRotate)
            {
                break;
            }

            lpSWDevice->gdwSWSrcWidth = dwWidth;
            lpSWDevice->gdwSWSrcHeight= dwHeight;
            lpSWDevice->dwPitch = dwPitch;

            /* Fill image data in overlay record */
            lpOverlayRecordSW->dwOriSrcWidth  = dwWidth;
            lpOverlayRecordSW->dwOriSrcHeight = dwHeight;
            lpOverlayRecordSW->dwOriSrcPitch  = dwPitch;

            if (!(pVidData->VidStatus.dwVideoFlagSW[0] & SW_USE_HQV))
            {
                break;
            }

        case FOURCC_HQVYUY2 :
            DBG_DD(("Create HQV_YUY2 Surface\n"));
            dwWidth  = lpSWDevice->gdwSWSrcWidth;
            dwHeight = lpSWDevice->gdwSWSrcHeight;
            dwPitch  = lpSWDevice->dwPitch;

            HQVFBSIZE = dwPitch * dwHeight;

            if (!lpVIAGraphicInfo->DRMEnabled)
            {
                if ( lpVideoHWDifference->dwThreeHQVBuffer )  /* CLE_C0 */
                {
                    
                    dwAddr = VIAAllocFB(pVidData, &lpVidMemory->HQVMemRequest, HQVFBSIZE*3);
                    
                   
                }
                else
                {
                    
                    dwAddr = VIAAllocFB(pVidData, &lpVidMemory->HQVMemRequest, HQVFBSIZE*2);
                    
                }
            }
            else
            {
                lpVidMemory->HQVfbRequest.context=1;
                if ( lpVideoHWDifference->dwThreeHQVBuffer )
                {
                    lpVidMemory->HQVfbRequest.size=HQVFBSIZE*3;
                }
                else
                {
                    lpVidMemory->HQVfbRequest.size=HQVFBSIZE*2;
                }
                lpVidMemory->HQVfbRequest.type= VIDEO;
            }

            if (dwAddr ==0)
            {
                DBG_DD((" Memory Allocation fail for SW YUY2 HQV Video.\n"));
                return VMI_ERR_CANNOT_CREATE_SURFACE;
            }

            lpOverlayRecordSW->dwHQVAddr[0] = dwAddr;
            lpOverlayRecordSW->dwHQVAddr[1] = dwAddr + HQVFBSIZE;

            if ( lpVideoHWDifference->dwThreeHQVBuffer )    /*CLE_C0*/
            {
                lpOverlayRecordSW->dwHQVAddr[2] = lpOverlayRecordSW->dwHQVAddr[1] + HQVFBSIZE;

                if (lpOverlayRecordSW->dwHQVAddr[2] + HQVFBSIZE >=
                                        (unsigned long)lpVIAGraphicInfo->Screen[ScrnIndex].VideoHeapEnd)
                {
                    DBG_DD(("Memory not enough for MPEG with HQV\n"));
                    return VMI_ERR_CANNOT_CREATE_SURFACE;
                }
            }
            else
            {
                if (lpOverlayRecordSW->dwHQVAddr[1] + HQVFBSIZE >=
                                        (unsigned long)lpVIAGraphicInfo->Screen[ScrnIndex].VideoHeapEnd)
                {
                    DBG_DD(("Memory not enough for MPEG with HQV\n"));
                    return VMI_ERR_CANNOT_CREATE_SURFACE;
                }
            }

            /*fill in the HQV buffer with 0x8000 (YUY2-black color) to clear HQV buffers*/
            for(iCount=0;iCount<HQVFBSIZE*2;iCount++)
            {
                lpTmpAddr = pjScreen + dwAddr + iCount;
                if((iCount%2) == 0)
                    *(lpTmpAddr)=0x00;
                else
                    *(lpTmpAddr)=0x80;
            }

            VIDOutD(HQV_DST_STARTADDR0,lpOverlayRecordSW->dwHQVAddr[0]);
            VIDOutD(HQV_DST_STARTADDR1,lpOverlayRecordSW->dwHQVAddr[1]);
            DBG_DD(("000003EC %08lx\n",VIDInD(HQV_DST_STARTADDR0) ));
            DBG_DD(("000003F0 %08lx\n",VIDInD(HQV_DST_STARTADDR1) ));
            if ( lpVideoHWDifference->dwThreeHQVBuffer )    /*CLE_C0*/
            {
                VIDOutD(HQV_DST_STARTADDR2,lpOverlayRecordSW->dwHQVAddr[2]);
                DBG_DD(("000003FC %08lx\n",VIDInD(HQV_DST_STARTADDR2) ));
            }

            break;

       case FOURCC_YV12 :
            DBG_DD(("    Create SW YV12 Surface: \n"));
            lpDPFsrc->dwFlags = DDPF_FOURCC;
            lpDPFsrc->dwFourCC = FOURCC_YV12;

            /* init Video status flag*/
            /*ENGINEMANAGER*/
            if( VidClient[CLIENT_SW].uClientHandle == CLIENT_INVALID_HANDLE )
	        {
	            /*register the client*/
		        regclient_para.func = EngCallback_SW;
		        regclient_para.uPriority = NORMAL_PRIORITY;
		        regclient_para.pClient_handle = &VidClient[CLIENT_SW].uClientHandle;
		        EngMgr_RegClient(&regclient_para);
            }

            switch ( lpVIAGraphicInfo->dwDeviceID )
            {
                case VIA_DEVICE_VT3204:
                case VIA_DEVICE_VT3205:
                case VIA_DEVICE_VT3314:
                    if ( VIA_Engine_Allocate(pVidData, CLIENT_SW,VID_ENGINE,1) ) /*V3*/
                    {
                    pVidData->VidStatus.dwVideoFlagSW[0] = VIDEO_HQV_INUSE | SW_USE_HQV | VIDEO_3_INUSE;
                    }
                    else
                    {
                        DBG_DD((" No more Video Engine (V3) for YV12!! \n"));
                        return VMI_ERR;
                    }
                    break;
                case VIA_DEVICE_CLE1:
                case VIA_DEVICE_CLE2:
                case VIA_DEVICE_VT3259:
                default:
                    if ( VIA_Engine_Allocate(pVidData, CLIENT_SW,VID_ENGINE,0) ) /*V1*/
                    {
                        pVidData->VidStatus.dwVideoFlagSW[0] = VIDEO_ACTIVE | VIDEO_SHOW | VIDEO_1_INUSE;
                        DBG_DD(("    Create YV12 Surface: VIDEO_1_INUSE !!!!\n"));
                    }
                    else if ( VIA_Engine_Allocate(pVidData, CLIENT_SW,VID_ENGINE,1) ) /*V3*/
                    {
                        pVidData->VidStatus.dwVideoFlagSW[0] = VIDEO_ACTIVE | VIDEO_SHOW | VIDEO_3_INUSE;
                        DBG_DD(("    Create YV12 Surface: VIDEO_3_INUSE !!!!\n"));
                    }
                    else
                    {
                        DBG_DD((" No more Video Engine for YV12!! \n"));
                        return VMI_ERR;
                    }
                    break;
            } /*End of DeviceID*/

            if ( VIA_Engine_Allocate(pVidData, CLIENT_SW,HQV_ENGINE,0) ) /*HQV*/
            {
                pVidData->VidStatus.dwVideoFlagSW[0] |= VIDEO_HQV_INUSE | SW_USE_HQV;
                lpVideoControl->dwHighQVDO = VW_HIGHQVDO_SW;
                DBG_DD(("    Create YV12 Surface: HQV_INUSE !!!!\n"));
            }
            else
            {
                DBG_DD(("    Create YV12 Surface: no HQV_INUSE !!!!\n"));
            }


            /* write Color Space Conversion param. */
            if (!lpVIAGraphicInfo->ColorSpaceChanged)
            {
                switch ( lpVIAGraphicInfo->dwDeviceID )
                {
                    case VIA_DEVICE_VT3204:
                    case VIA_DEVICE_VT3205:
                    case VIA_DEVICE_VT3314:
                        VIDOutD(V3_ColorSpaceReg_1, ColorSpaceValue_1_32045);
                        VIDOutD(V3_ColorSpaceReg_2, ColorSpaceValue_2_32045);
                        DBG_DD(("000002C4 %08x\n",ColorSpaceValue_1_32045));
                        DBG_DD(("000002C8 %08x\n",ColorSpaceValue_2_32045));
                        break;

                    case VIA_DEVICE_CLE1:
                    case VIA_DEVICE_CLE2:
                        switch ( DispatchVGARevisionID(pVidData) )
                        {
                            case VIA_REVISION_CLEAX :
                                VIDOutD(ColorSpaceReg_1, ColorSpaceValue_1);
                                VIDOutD(ColorSpaceReg_2, ColorSpaceValue_2);
                                VIDOutD(V3_ColorSpaceReg_1, ColorSpaceValue_1);
                                VIDOutD(V3_ColorSpaceReg_2, ColorSpaceValue_2);
                                DBG_DD(("00000284 %08x\n",ColorSpaceValue_1));
                                DBG_DD(("00000288 %08x\n",ColorSpaceValue_2));
                                DBG_DD(("000002C4 %08x\n", ColorSpaceValue_1));
                                DBG_DD(("000002C8 %08x\n", ColorSpaceValue_2));
                                break;
                            case VIA_REVISION_CLECX :
                            default : /* For future chip */
                                VIDOutD(ColorSpaceReg_1, ColorSpaceValue_1_3123C0);
                                VIDOutD(ColorSpaceReg_2, ColorSpaceValue_2_3123C0);
                                VIDOutD(V3_ColorSpaceReg_1, ColorSpaceValue_1_3123C0);
                                VIDOutD(V3_ColorSpaceReg_2, ColorSpaceValue_2_3123C0);
                                DBG_DD(("00000284 %08x\n",ColorSpaceValue_1_3123C0));
                                DBG_DD(("00000288 %08x\n",ColorSpaceValue_2_3123C0));
                                DBG_DD(("000002C4 %08x\n",ColorSpaceValue_1_3123C0));
                                DBG_DD(("000002C8 %08x\n",ColorSpaceValue_2_3123C0));
                                break;
                        }
                        break;

                    case VIA_DEVICE_VT3259:
                    default:
                        VIDOutD(ColorSpaceReg_1, ColorSpaceValue_1_3123C0);
                        VIDOutD(ColorSpaceReg_2, ColorSpaceValue_2_3123C0);
                        VIDOutD(V3_ColorSpaceReg_1, ColorSpaceValue_1_3123C0);
                        VIDOutD(V3_ColorSpaceReg_2, ColorSpaceValue_2_3123C0);
                        DBG_DD(("00000284 %08x\n",ColorSpaceValue_1_3123C0));
                        DBG_DD(("00000288 %08x\n",ColorSpaceValue_2_3123C0));
                        DBG_DD(("000002C4 %08x\n",ColorSpaceValue_1_3123C0));
                        DBG_DD(("000002C8 %08x\n",ColorSpaceValue_2_3123C0));
                        break;
                }/*End of DeviceID   */
            }
            else
            {
                VIDOutD(ColorSpaceReg_1, lpVIAGraphicInfo->ColorSpaceValue1);
                VIDOutD(ColorSpaceReg_2, lpVIAGraphicInfo->ColorSpaceValue2);
                VIDOutD(V3_ColorSpaceReg_1, lpVIAGraphicInfo->ColorSpaceValue1);
                VIDOutD(V3_ColorSpaceReg_2, lpVIAGraphicInfo->ColorSpaceValue2);
            }

            dwBackBufferCount = lpDDSurfaceDesc->dwBackBufferCount;
            dwWidth  = lpDDSurfaceDesc->dwWidth;
            dwHeight = lpDDSurfaceDesc->dwHeight;
            dwPitch  = ALIGN_TO_32_BYTE(dwWidth);
            DBG_DD(("    srcWidth= %ld \n", dwWidth));
            DBG_DD(("    srcHeight= %ld \n", dwHeight));
            
            SWFBSIZE = (((dwPitch * 3)/2)* dwHeight);    /* 1.5 bytes per pixel */

            if (!lpVIAGraphicInfo->DRMEnabled)
            {
                
                dwAddr = VIAAllocFB(pVidData, &lpVidMemory->SWMemRequest[handle], SWFBSIZE*2);
               
            }
            else
            {
                lpVidMemory->SWfbRequest.context=1;
                lpVidMemory->SWfbRequest.size=SWFBSIZE*2;
                lpVidMemory->SWfbRequest.type= VIDEO;

            }

            if (dwAddr ==0)
            {
                DBG_DD((" Memory Allocation fail for SW YV12 Video.\n"));
                return VMI_ERR_CANNOT_CREATE_SURFACE;
            }

            /* fill in the SW buffer with 0x8000 (YUY2-black color) to clear FB buffer */
            lpTmpAddr = pjScreen + dwAddr;
            for(iCount=0;iCount<(SWFBSIZE*2);iCount++)
            {
                if((iCount%2) == 0)
                    *lpTmpAddr++=0x00;
                else
                    *lpTmpAddr++=0x80;
            }

            lpSWDevice->dwSWPhysicalAddr[0]   = dwAddr;
            lpSWDevice->dwSWCrPhysicalAddr[0] = lpSWDevice->dwSWPhysicalAddr[0]
                                             + (dwPitch*dwHeight);
            lpSWDevice->dwSWCbPhysicalAddr[0] = lpSWDevice->dwSWCrPhysicalAddr[0]
                                             + ((dwPitch>>1)*(dwHeight>>1));
            lpSWDevice->lpSWOverlaySurface[0] = pjScreen+dwAddr;

            lpSWDevice->dwSWPhysicalAddr[1] = dwAddr + SWFBSIZE;
            lpSWDevice->dwSWCrPhysicalAddr[1] = lpSWDevice->dwSWPhysicalAddr[1]
                                             + (dwPitch*dwHeight);
            lpSWDevice->dwSWCbPhysicalAddr[1] = lpSWDevice->dwSWCrPhysicalAddr[1]
                                             + ((dwPitch>>1)*(dwHeight>>1));
            lpSWDevice->lpSWOverlaySurface[1] = lpSWDevice->lpSWOverlaySurface[0] + SWFBSIZE;

            DBG_DD(("lpSWDevice->dwSWPhysicalAddr[0] 0x%08lx\n", dwAddr));
            DBG_DD(("lpSWDevice->dwSWPhysicalAddr[1] 0x%08lx\n", dwAddr + SWFBSIZE));

            lpSWDevice->gdwSWSrcWidth = dwWidth;
            lpSWDevice->gdwSWSrcHeight= dwHeight;
            lpSWDevice->dwPitch = dwPitch;

            /* Fill image data in overlay record */
            lpOverlayRecordSW->dwOriSrcWidth  = dwWidth;
            lpOverlayRecordSW->dwOriSrcHeight = dwHeight;
            lpOverlayRecordSW->dwOriSrcPitch  = dwPitch;

            if (!(pVidData->VidStatus.dwVideoFlagSW[0] & SW_USE_HQV))
            {
                break;
            }

        case FOURCC_HQVYV12 :
            /*
             *  if sw video use HQV dwpitch should changed
             */
            DBG_DD(("Create HQV_YV12 Surface\n"));
            dwWidth  = lpSWDevice->gdwSWSrcWidth;
            dwHeight = lpSWDevice->gdwSWSrcHeight;
            dwPitch  = lpSWDevice->dwPitch * 2;
            lpSWDevice->dwWidth = dwWidth;
            lpSWDevice->dwHeight = dwHeight;
            
            HQVFBSIZE = dwPitch * dwHeight;

            if (!lpVIAGraphicInfo->DRMEnabled)
            {
                if ( lpVideoHWDifference->dwThreeHQVBuffer )  /* CLE_C0 */
                {
                    
                    dwAddr = VIAAllocFB(pVidData, &lpVidMemory->HQVMemRequest, HQVFBSIZE*3);
                    
                }
                else
                {
                    
                    dwAddr = VIAAllocFB(pVidData, &lpVidMemory->HQVMemRequest, HQVFBSIZE*2);
                    
                }
            }
            else
            {
                lpVidMemory->HQVfbRequest.context=1;
                if ( lpVideoHWDifference->dwThreeHQVBuffer )
                {
                    lpVidMemory->HQVfbRequest.size=HQVFBSIZE*3;
                }
                else
                {
                    lpVidMemory->HQVfbRequest.size=HQVFBSIZE*2;
                }
                lpVidMemory->HQVfbRequest.type= VIDEO;

            }

            if (dwAddr ==0)
            {
                DBG_DD((" Memory Allocation fail for SW YV12 HQV Video.\n"));
                return VMI_ERR_CANNOT_CREATE_SURFACE;
            }

            lpOverlayRecordSW->dwHQVAddr[0] = dwAddr;
            lpOverlayRecordSW->dwHQVAddr[1] = dwAddr + HQVFBSIZE;
            lpSWDevice->dwHQVAddr[0] = dwAddr;
            lpSWDevice->dwHQVAddr[1] = dwAddr + HQVFBSIZE;

            if ( lpVideoHWDifference->dwThreeHQVBuffer )    /*CLE_C0*/
            {
                lpOverlayRecordSW->dwHQVAddr[2] = lpOverlayRecordSW->dwHQVAddr[1] + HQVFBSIZE;
                lpSWDevice->dwHQVAddr[2] = lpSWDevice->dwHQVAddr[1] + HQVFBSIZE;
                
                if (lpOverlayRecordSW->dwHQVAddr[2] + HQVFBSIZE >=
                            (unsigned long)lpVIAGraphicInfo->Screen[ScrnIndex].VideoHeapEnd)
                {
                    DBG_DD(("Memory not enough for MPEG with HQV\n"));
                    return VMI_ERR_CANNOT_CREATE_SURFACE;
                }
            }
            else
            {
                if (lpOverlayRecordSW->dwHQVAddr[1] + HQVFBSIZE >=
                            (unsigned long)lpVIAGraphicInfo->Screen[ScrnIndex].VideoHeapEnd)
                {
                    DBG_DD(("Memory not enough for MPEG with HQV\n"));
                    return VMI_ERR_CANNOT_CREATE_SURFACE;
                }
            }

            /*fill in the HQV buffer with 0x8000 (YUY2-black color) to clear HQV buffers*/
            for(iCount=0;iCount<HQVFBSIZE*2;iCount++)
            {
                lpTmpAddr = pjScreen + dwAddr + iCount;
                if((iCount%2) == 0)
                    *(lpTmpAddr)=0x00;
                else
                    *(lpTmpAddr)=0x80;
            }

            VIDOutD(HQV_DST_STARTADDR0,lpOverlayRecordSW->dwHQVAddr[0]);
            VIDOutD(HQV_DST_STARTADDR1,lpOverlayRecordSW->dwHQVAddr[1]);
            DBG_DD(("000003EC %08lx\n",VIDInD(HQV_DST_STARTADDR0) ));
            DBG_DD(("000003F0 %08lx\n",VIDInD(HQV_DST_STARTADDR1) ));
            if ( lpVideoHWDifference->dwThreeHQVBuffer )    /*CLE_C0*/
            {
                VIDOutD(HQV_DST_STARTADDR2,lpOverlayRecordSW->dwHQVAddr[2]);
                DBG_DD(("000003FC %08lx\n",VIDInD(HQV_DST_STARTADDR2) ));
            }

            break;



        default :
            break;
    }

    return dwRet;

} /* CreateSurface */





/* Destroy Surface */
unsigned long DestroySurface(PVIDDATA pVidData, LPDDSURFACEDESC lpDDSurfaceDesc)
{
    unsigned long handle=pVidData->VMI_Info.handle;
    LPVIDEOMEMORY lpVidMemory = &VidMemory;
    LPVIAGRAPHICINFO lpVIAGraphicInfo = &pVidData->VIAGraphicInfo;
    LPDDPIXELFORMAT lpDPFsrc = &pVidData->DPFsrc[handle];
    EngMgr_FreeEngine_Para Free_Para;
    unsigned long PortId;

    DBG_DD(("\n"));

    if(lpVIAGraphicInfo->duoview)  
    {
        switch (lpDDSurfaceDesc->dwFourCC)
        {
        case FOURCC_NV12 :
        case FOURCC_YUY2:
        case FOURCC_YV12:
            lpVIAGraphicInfo->bVideoEngineUsedOut=0;
			break;
        default:
            break;
        }
    }

    switch (lpDDSurfaceDesc->dwFourCC)
    {
        case FOURCC_HQVMPEG :
            DBG_DD((" Destroy HQVMPEG Surface! \n"));
            lpVidMemory->HQVMemRequest.checkID = VIA_MID;
            lpVidMemory->HQVMemRequest.type    = VIDMGR_TYPE_VIDEO;

            if (!lpVIAGraphicInfo->DRMEnabled)
            {
                if (lpVidMemory->HQVMemRequest.offset != 0)
                {
                    if ( MEM_OK != viaMgr_vid_ioctl(pVidData, VIAMGR_FREE_VIDEO, (char *)(&lpVidMemory->HQVMemRequest) )  )
                    {
                       DBG_DD((" Memory Free fail for MPEG HQV Video \n"));
                    }
                    lpVidMemory->HQVMemRequest.offset = 0;
                    DBG_DD((" Memory Free for MPEG HQV Video \n"));
                }
            }

            PortId = CLIENT_MPEG;
            Free_Para.uClient_handle = VidClient[PortId].uClientHandle;
            Free_Para.uEng_type = HQV_ENGINE;
            Free_Para.uEng_id = 0;
            EngMgr_FreeEngine(pVidData, &Free_Para);

            break;

        case FOURCC_SUBP :
            DBG_DD((" Destroy SUBP Surface! \n"));
            lpVidMemory->SUBPMemRequest[0].checkID = VIA_MID;
            lpVidMemory->SUBPMemRequest[0].type    = VIDMGR_TYPE_VIDEO;

            if (!lpVIAGraphicInfo->DRMEnabled)
            {
                if (lpVidMemory->SUBPMemRequest[0].offset != 0)
                {
                    if ( MEM_OK != viaMgr_vid_ioctl(pVidData, VIAMGR_FREE_VIDEO, (char *)(&lpVidMemory->SUBPMemRequest[0]) ) )
                    {
                       DBG_DD((" Memory Free fail for SubPicture \n"));
                    }
                    lpVidMemory->SUBPMemRequest[0].offset = 0;
                    DBG_DD((" Memory Free for SubPicture \n"));
                }
            }

            break;

        case FOURCC_ALPHA :
            DBG_DD((" Destroy ALPHA Surface! \n"));
            lpVidMemory->ALPHAMemRequest.checkID = VIA_MID;
            lpVidMemory->ALPHAMemRequest.type    = VIDMGR_TYPE_VIDEO;

            if (!lpVIAGraphicInfo->DRMEnabled)
            {
                if (lpVidMemory->ALPHAMemRequest.offset != 0)
                {
                    if ( MEM_OK != viaMgr_vid_ioctl(pVidData, VIAMGR_FREE_VIDEO, (char *)(&lpVidMemory->ALPHAMemRequest) ) )
                    {
                        DBG_DD((" Memory Free fail for ALPHA \n"));
                    }
                    lpVidMemory->ALPHAMemRequest.offset = 0;
                    DBG_DD((" Memory Free for ALPHA \n"));
                }
            }

        case FOURCC_YUY2 :
#ifdef CS_FLAG_HWROTATION /* For video rotation */
            if (lpVIAGraphicInfo->IsSurfaceUsedForRotate)
            {
                lpDPFsrc = &pVidData->DPFsrc_rot;
                DBG_DD((" Destroy YUY2 Surface: for rotate! \n"));
            }
            else
#endif /* CS_FLAG_HWROTATION */
            {
                DBG_DD((" Destroy YUY2 Surface! \n"));
            }

            lpDPFsrc->dwFlags = 0;
            lpDPFsrc->dwFourCC = 0;
            lpVidMemory->SWMemRequest[handle].checkID = VIA_MID;
            lpVidMemory->SWMemRequest[handle].type    = VIDMGR_TYPE_VIDEO;

            if (!lpVIAGraphicInfo->DRMEnabled)
            {
                if (lpVidMemory->SWMemRequest[handle].offset != 0)
                {
                    if ( MEM_OK != viaMgr_vid_ioctl(pVidData, VIAMGR_FREE_VIDEO, (char *)(&lpVidMemory->SWMemRequest[handle]) ) )
                    {
                       DBG_DD((" Memory Free fail for SW YUY2 Surface \n"));
                    }
                    lpVidMemory->SWMemRequest[handle].offset = 0;
                    DBG_DD((" Memory Free for SW YUY2 surface \n"));
                }
            }
            else
            {
                DBG_DD((" Memory Free for SW YUY2 Surface \n"));
            }

            pVidData->VidStatus.dwVideoFlagSW[0] &= ~VIDEO_ACTIVE;

            if (!(pVidData->VidStatus.dwVideoFlagSW[0] & SW_USE_HQV))
            {
                VIA_Engine_Free_All(pVidData, CLIENT_SW); /*ENGINEMANAGER*/
                pVidData->VidStatus.dwVideoFlagSW[0] = 0;
            }
            else
            {
                /* Free SW-HQV surface */
                if (!lpVIAGraphicInfo->DRMEnabled)
                {
                    if (lpVidMemory->HQVMemRequest.offset != 0)
                    {
                        if ( MEM_OK != viaMgr_vid_ioctl(pVidData, VIAMGR_FREE_VIDEO, (char *)(&lpVidMemory->HQVMemRequest) )  )
                        {
                           DBG_DD((" Memory Free fail for SW HQV Video \n"));
                        }
                        lpVidMemory->HQVMemRequest.offset = 0;
                        DBG_DD((" Memory Free for SW HQV Video \n"));
                    }
   
                }
		        else
		        {
		  	        DBG_DD((" Memory Free for SW HQV Surface \n"));

                #if (defined(XF86DRI))
                    if ( ioctl( drm_fd, DRM_IOCTL_VIA_FREEMEM, &lpVidMemory->HQVfbRequest ) )
                    {
                         DBG_DD(("\nCannot free SWOV-HQV memory use DRM module!\n"));
                         return -errno;
                     }
			        else
                         lpVidMemory->HQVfbRequest.offset = 0;
              
                     drmClose(drm_fd);
		         #endif
		        }	
		
                /* unregister client */
                VIA_Engine_Free_All(pVidData, CLIENT_SW);/*ENGINEMANAGER*/

                /* Clear VideoFlag */
                pVidData->VidStatus.dwVideoFlagSW[0] = 0;
            }
            break;

        case FOURCC_HQVYUY2 :
            DBG_DD((" Destroy YUY2 HQV Surface! \n"));
            lpVidMemory->HQVMemRequest.checkID = VIA_MID;
            lpVidMemory->HQVMemRequest.type    = VIDMGR_TYPE_VIDEO;

            if (!lpVIAGraphicInfo->DRMEnabled)
            {
                if (lpVidMemory->HQVMemRequest.offset != 0)
                {
                    if ( MEM_OK != viaMgr_vid_ioctl(pVidData, VIAMGR_FREE_VIDEO, (char *)(&lpVidMemory->HQVMemRequest) ) )
                    {
                       DBG_DD((" Memory Free fail for SW YUY2 HQV Surface \n"));
                    }
                    lpVidMemory->HQVMemRequest.offset = 0;
                    DBG_DD((" Memory Free for SW YUY2 HQV Surface \n"));
                }
            }
            else
            {
            #if (defined(XF86DRI))
                if ( ioctl( drm_fd, DRM_IOCTL_VIA_FREEMEM, &lpVidMemory->HQVfbRequest ) )
                {
                    DBG_DD(("\nCannot free HQV memory use DRM module!\n"));
                    return -errno;
                }
		        else
		  	        lpVidMemory->HQVfbRequest.offset = 0;		
                drmClose(drm_fd);
            #endif
            }

            PortId = CLIENT_SW;
            Free_Para.uClient_handle = VidClient[PortId].uClientHandle;
            Free_Para.uEng_type = HQV_ENGINE;
            Free_Para.uEng_id = 0;
            EngMgr_FreeEngine(pVidData, &Free_Para);
            break;

        case FOURCC_YV12 :
            DBG_DD((" Destroy YV12 Surface! \n"));
            lpDPFsrc->dwFlags = 0;
            lpDPFsrc->dwFourCC = 0;
            lpVidMemory->SWMemRequest[handle].checkID = VIA_MID;
            lpVidMemory->SWMemRequest[handle].type    = VIDMGR_TYPE_VIDEO;

            if (!lpVIAGraphicInfo->DRMEnabled)
            {
                if (lpVidMemory->SWMemRequest[handle].offset != 0)
                {
                    if ( MEM_OK != viaMgr_vid_ioctl(pVidData, VIAMGR_FREE_VIDEO, (char *)(&lpVidMemory->SWMemRequest[handle]) ) )
                    {
                       DBG_DD((" Memory Free fail for SW YV12 Surface \n"));
                    }
                    lpVidMemory->SWMemRequest[handle].offset = 0;
                    DBG_DD((" Memory Free for SW YV12 surface \n"));
                }
            }
            else
            {
                DBG_DD((" Memory Free for SW YV12 Surface \n"));

            #if (defined(XF86DRI))
                if ( ioctl( drm_fd, DRM_IOCTL_VIA_FREEMEM, &lpVidMemory->SWfbRequest ) )
                {
                    DBG_DD(("\nCannot free SWOV memory use DRM module!\n"));
                    return -errno;
                }
		        else
                   lpVidMemory->SWfbRequest.offset = 0;
                if (!(pVidData->VidStatus.dwVideoFlagSW[0] & SW_USE_HQV))
                {
                    drmClose(drm_fd);
                }
            #endif
            }

            pVidData->VidStatus.dwVideoFlagSW[0] &= ~VIDEO_ACTIVE;

            if (!(pVidData->VidStatus.dwVideoFlagSW[0] & SW_USE_HQV))
            {
                VIA_Engine_Free_All(pVidData, CLIENT_SW); /*ENGINEMANAGER*/
                pVidData->VidStatus.dwVideoFlagSW[0] = 0;
            }
            else
            {
                /* Free SW-HQV surface */
                if (!lpVIAGraphicInfo->DRMEnabled)
                {
                    if (lpVidMemory->HQVMemRequest.offset != 0)
                    {
                        if ( MEM_OK != viaMgr_vid_ioctl(pVidData, VIAMGR_FREE_VIDEO, (char *)(&lpVidMemory->HQVMemRequest) )  )
                        {
                           DBG_DD((" Memory Free fail for SW HQV Video \n"));
                        }
                        lpVidMemory->HQVMemRequest.offset = 0;
                        DBG_DD((" Memory Free for SW HQV Video \n"));
                    }
                }
		        else
		        {
		  	        DBG_DD((" Memory Free for SW HQV Surface \n"));

                #if  (defined(XF86DRI))
                    if ( ioctl( drm_fd, DRM_IOCTL_VIA_FREEMEM, &lpVidMemory->HQVfbRequest ) )
                    {
                        DBG_DD(("\nCannot free SWOV-HQV memory use DRM module!\n"));
                        return -errno;
                    }
			        else
                        lpVidMemory->HQVfbRequest.offset = 0;
                         
                    drmClose(drm_fd);
		        #endif
                }	
				
                /* unregister client */
                VIA_Engine_Free_All(pVidData, CLIENT_SW);/*ENGINEMANAGER*/

                /* Clear VideoFlag */
                pVidData->VidStatus.dwVideoFlagSW[0] = 0;
            }
            break;

        case FOURCC_HQVYV12 :
            DBG_DD((" Destroy YV12 HQV Surface! \n"));
            lpVidMemory->HQVMemRequest.checkID = VIA_MID;
            lpVidMemory->HQVMemRequest.type    = VIDMGR_TYPE_VIDEO;

            if (!lpVIAGraphicInfo->DRMEnabled)
            {
                if (lpVidMemory->HQVMemRequest.offset != 0)
                {
                    if ( MEM_OK != viaMgr_vid_ioctl(pVidData, VIAMGR_FREE_VIDEO, (char *)(&lpVidMemory->HQVMemRequest) ) )
                    {
                       DBG_DD((" Memory Free fail for SW YV12 HQV Surface \n"));
                    }
                    lpVidMemory->HQVMemRequest.offset = 0;
                    DBG_DD((" Memory Free for SW YV12 HQV Surface \n"));
                }
            }
            else
            {
            #if (defined(XF86DRI))
                if ( ioctl( drm_fd, DRM_IOCTL_VIA_FREEMEM, &lpVidMemory->HQVfbRequest ) )
                {
                    DBG_DD(("\nCannot free HQV memory use DRM module!\n"));
                    return -errno;
                }
		        else
                    lpVidMemory->HQVfbRequest.offset = 0;
			
                drmClose(drm_fd);
            #endif
            }

            PortId = CLIENT_SW;
            Free_Para.uClient_handle = VidClient[PortId].uClientHandle;
            Free_Para.uEng_type = HQV_ENGINE;
            Free_Para.uEng_id = 0;
            EngMgr_FreeEngine(pVidData, &Free_Para);
            break;

        default :
            break;
    }

    /* memset(&MPGDevice, 0, sizeof(MPGDevice)); */
    return VMI_OK;
} /* DestroySurface */



